Package: distantia
Type: Package
Title: Advanced Toolset for Efficient Time Series Dissimilarity
        Analysis
Version: 2.0.3
Authors@R: 
    person(given = "Blas M.", 
          family = "Benito", , 
          email = "blasbenito@gmail.com", 
          role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-5105-7232"))
URL: https://blasbenito.github.io/distantia/
BugReports: https://github.com/BlasBenito/distantia/issues
Description: Fast C++ implementation of Dynamic Time Warping for time series dissimilarity analysis, with applications in environmental monitoring and sensor data analysis, climate science, signal processing and pattern recognition, and financial data analysis. Built upon the ideas presented in Benito and Birks (2020) <doi:10.1111/ecog.04895>, provides tools for analyzing time series of varying lengths and structures, including irregular multivariate time series. Key features include individual variable contribution analysis, restricted permutation tests for statistical significance, and imputation of missing data via GAMs. Additionally, the package provides an ample set of tools to prepare and manage time series data.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Depends: R (>= 4.1), doFuture
Imports: Rcpp, zoo, foreach, future.apply, lubridate, progressr
Suggests: roxyglobals, spelling, sf, lwgeom, testthat (>= 3.0.0)
Config/roxyglobals/filename: globals.R
Config/roxyglobals/unique: FALSE
Config/testthat/edition: 3
LinkingTo: Rcpp
Language: en-US
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-13 13:16:25 UTC; blas
Author: Blas M. Benito [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-5105-7232>)
Maintainer: Blas M. Benito <blasbenito@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-13 13:40:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-13 14:44:55 UTC; unix
Archs: distantia.so.dSYM

# Generated by roxygen2: do not edit by hand

export(auto_distance_cpp)
export(auto_sum_cpp)
export(auto_sum_full_cpp)
export(auto_sum_path_cpp)
export(color_continuous)
export(color_discrete)
export(cost_matrix_diagonal_cpp)
export(cost_matrix_diagonal_weighted_cpp)
export(cost_matrix_orthogonal_cpp)
export(cost_path_cpp)
export(cost_path_diagonal_bandwidth_cpp)
export(cost_path_diagonal_cpp)
export(cost_path_orthogonal_bandwidth_cpp)
export(cost_path_orthogonal_cpp)
export(cost_path_slotting_cpp)
export(cost_path_sum_cpp)
export(cost_path_trim_cpp)
export(distance)
export(distance_bray_curtis_cpp)
export(distance_canberra_cpp)
export(distance_chebyshev_cpp)
export(distance_chi_cpp)
export(distance_cosine_cpp)
export(distance_euclidean_cpp)
export(distance_hamming_cpp)
export(distance_hellinger_cpp)
export(distance_jaccard_cpp)
export(distance_ls_cpp)
export(distance_manhattan_cpp)
export(distance_matrix)
export(distance_matrix_cpp)
export(distance_russelrao_cpp)
export(distance_sorensen_cpp)
export(distantia)
export(distantia_aggregate)
export(distantia_boxplot)
export(distantia_cluster_hclust)
export(distantia_cluster_kmeans)
export(distantia_dtw)
export(distantia_dtw_plot)
export(distantia_ls)
export(distantia_matrix)
export(distantia_model_frame)
export(distantia_spatial)
export(distantia_stats)
export(distantia_time_delay)
export(f_binary)
export(f_clr)
export(f_detrend_difference)
export(f_detrend_linear)
export(f_detrend_poly)
export(f_hellinger)
export(f_list)
export(f_log)
export(f_percent)
export(f_proportion)
export(f_proportion_sqrt)
export(f_rescale_global)
export(f_rescale_local)
export(f_scale_global)
export(f_scale_local)
export(f_trend_linear)
export(f_trend_poly)
export(importance_dtw_cpp)
export(importance_dtw_legacy_cpp)
export(importance_ls_cpp)
export(momentum)
export(momentum_aggregate)
export(momentum_boxplot)
export(momentum_dtw)
export(momentum_ls)
export(momentum_model_frame)
export(momentum_spatial)
export(momentum_stats)
export(momentum_to_wide)
export(permute_free_by_row_cpp)
export(permute_free_cpp)
export(permute_restricted_by_row_cpp)
export(permute_restricted_cpp)
export(psi_auto_distance)
export(psi_auto_sum)
export(psi_cost_matrix)
export(psi_cost_path)
export(psi_cost_path_sum)
export(psi_distance_lock_step)
export(psi_distance_matrix)
export(psi_dtw_cpp)
export(psi_equation)
export(psi_equation_cpp)
export(psi_ls_cpp)
export(psi_null_dtw_cpp)
export(psi_null_ls_cpp)
export(subset_matrix_by_rows_cpp)
export(tsl_Inf_to_NA)
export(tsl_NaN_to_NA)
export(tsl_aggregate)
export(tsl_burst)
export(tsl_colnames_clean)
export(tsl_colnames_get)
export(tsl_colnames_prefix)
export(tsl_colnames_set)
export(tsl_colnames_suffix)
export(tsl_count_NA)
export(tsl_diagnose)
export(tsl_handle_NA)
export(tsl_init)
export(tsl_initialize)
export(tsl_join)
export(tsl_names_clean)
export(tsl_names_get)
export(tsl_names_set)
export(tsl_names_test)
export(tsl_ncol)
export(tsl_nrow)
export(tsl_plot)
export(tsl_repair)
export(tsl_resample)
export(tsl_simulate)
export(tsl_smooth)
export(tsl_stats)
export(tsl_subset)
export(tsl_time)
export(tsl_time_summary)
export(tsl_to_df)
export(tsl_transform)
export(utils_as_time)
export(utils_block_size)
export(utils_boxplot_common)
export(utils_check_args_distantia)
export(utils_check_args_matrix)
export(utils_check_args_momentum)
export(utils_check_args_path)
export(utils_check_args_tsl)
export(utils_check_args_zoo)
export(utils_check_distance_args)
export(utils_check_list_class)
export(utils_clean_names)
export(utils_cluster_hclust_optimizer)
export(utils_cluster_kmeans_optimizer)
export(utils_cluster_silhouette)
export(utils_coerce_time_class)
export(utils_color_breaks)
export(utils_digits)
export(utils_distantia_df_split)
export(utils_drop_geometry)
export(utils_global_scaling_params)
export(utils_is_time)
export(utils_line_color)
export(utils_line_guide)
export(utils_matrix_guide)
export(utils_matrix_plot)
export(utils_new_time)
export(utils_new_time_type)
export(utils_optimize_loess)
export(utils_optimize_spline)
export(utils_prepare_df)
export(utils_prepare_matrix)
export(utils_prepare_matrix_list)
export(utils_prepare_time)
export(utils_prepare_vector_list)
export(utils_prepare_zoo_list)
export(utils_rescale_vector)
export(utils_time_keywords)
export(utils_time_keywords_dictionary)
export(utils_time_keywords_translate)
export(utils_time_units)
export(utils_tsl_pairs)
export(zoo_aggregate)
export(zoo_name_clean)
export(zoo_name_get)
export(zoo_name_set)
export(zoo_permute)
export(zoo_plot)
export(zoo_resample)
export(zoo_simulate)
export(zoo_smooth_exponential)
export(zoo_smooth_window)
export(zoo_time)
export(zoo_to_tsl)
export(zoo_vector_to_matrix)
importFrom(Rcpp,evalCpp)
importFrom(doFuture,"%dofuture%")
useDynLib(distantia)
useDynLib(distantia, .registration = TRUE)

# `distantia` Time Series Dissimilarity <a href="https://github.com/BlasBenito/distantia"><img src="man/figures/logo.png" align="right" height="138" /></a>

<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.3269235.svg)](https://doi.org/10.5281/zenodo.3269235)
[![CRAN_Release_Badge](http://www.r-pkg.org/badges/version/distantia)](https://CRAN.R-project.org/package=distantia)
[![CRAN_Download_Badge](http://cranlogs.r-pkg.org/badges/grand-total/distantia)](https://CRAN.R-project.org/package=distantia)
[![R-CMD-check](https://github.com/BlasBenito/distantia/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/BlasBenito/distantia/actions/workflows/R-CMD-check.yaml)

<!-- badges: end -->

## Summary

The R package **`distantia`** offers an efficient, feature-rich toolkit
for managing, comparing, and analyzing time series data. It is designed
to handle a wide range of scenarios, including:

- Multivariate and univariate time series.
- Regular and irregular sampling.
- Time series of different lengths.

### Key Features

#### Comprehensive Analytical Tools

- 10 distance metrics: see `distantia::distances`.
- The normalized dissimilarity metric `psi`.
- Free and Restricted Dynamic Time Warping (DTW) for shape-based
  comparison.
- A Lock-Step method for sample-to-sample comparison.
- Restricted permutation tests for robust inferential support.
- Analysis of contribution to dissimilarity of individual variables in
  multivariate time series.
- Hierarchical and K-means clustering of time series based on
  dissimilarity matrices.

#### Computational Efficiency

- A **C++ back-end** powered by [Rcpp](https://www.rcpp.org/).
- **Parallel processing** managed through the
  [future](https://future.futureverse.org/) package.
- **Efficient data handling** via
  [zoo](https://CRAN.R-project.org/package=zoo).

#### Time Series Management Tools

- Introduces **time series lists** (TSL), a versatile format for
  handling collections of time series stored as lists of `zoo` objects.
- Includes a suite of `tsl_...()` functions for generating, resampling,
  transforming, analyzing, and visualizing univariate and multivariate
  time series.

## Citation

If you find this package useful, please cite it as:

*Blas M. Benito, H. John B. Birks (2020). distantia: an open-source
toolset to quantify dissimilarity between multivariate ecological
time-series. Ecography, 43(5), 660-667. doi:
[10.1111/ecog.04895](https://nsojournals.onlinelibrary.wiley.com/doi/10.1111/ecog.04895).*

*Blas M. Benito (2024). distantia: A Toolset for Time Series
Dissimilarity Analysis. R package version 2.0. url:
<https://blasbenito.github.io/distantia/>.*

## Install

Version 2.0.0 of `distantia` can be installed from CRAN and GitHub.

``` r
#CRAN
install.packages("distantia")

#GitHub
remotes::install_github(
  repo = "blasbenito/distantia", 
  ref = "main"
  )
```

## Getting Started

This section showcases several features of the package `distantia`.
Please check the **Articles** section for further details.

### Setup

All heavy-duty functions in `distantia` support parallelization via the
[future](https://future.futureverse.org/) package. However, due to the
high efficiency of the C++ backend of `distantia`, parallel execution is
only worth it for very large datasets and restricted permutation
analyses.

Progress bars provided by the
[progressr](https://progressr.futureverse.org/) package are also
available. Unfortunately, the latter does not work in R Markdown
documents like this one.

``` r
library(distantia, quietly = TRUE)
library(future)
# library(progressr)

#parallelization setup
# future::plan(future::multisession)

#progress bar (does not work in Rmarkdown)
#progressr::handlers(global = TRUE)
```

### Example Data

The `albatross` data frame contains daily GPS data of 4 individuals of
Waved Albatross in the Pacific captured during the summer of 2008. Below
are the first 6 rows of this data frame:

    #>   name       time         x         y     speed temperature  heading
    #> 1 X132 2008-05-31 -89.62097 -1.389512 0.1473333    29.06667 212.0307
    #> 2 X132 2008-06-01 -89.62101 -1.389508 0.2156250    28.25000 184.0337
    #> 3 X132 2008-06-02 -89.62101 -1.389503 0.2143750    27.68750 123.1269
    #> 4 X132 2008-06-03 -89.62099 -1.389508 0.2018750    27.81250 183.4600
    #> 5 X132 2008-06-04 -89.62098 -1.389507 0.2256250    27.68750 114.8931
    #> 6 X132 2008-06-05 -89.62925 -1.425734 1.3706667    25.73333 245.8033

The code below transforms the data to a *Time Series List* with
`tsl_initialize()` and applies global scaling and centering with
`tsl_transform()` and `f_scale_global` to facilitate time series
comparisons.

``` r
tsl <- tsl_initialize(
  x = distantia::albatross,
  name_column = "name",
  time_column = "time",
  lock_step = TRUE
) |> 
  tsl_transform(
    f = f_scale_global
  )

tsl_plot(
  tsl = tsl,
  ylim = "relative"
)
```

<img src="man/figures/README-unnamed-chunk-4-1.png" alt="" width="100%" />

### Dissimilarity Analysis

#### Lock-Step Analysis

Lock-step analysis performs direct comparisons between samples observed
at the same time without any time distortion. It requires time series of
the same length, preferably observed at the same times.

``` r
df_ls <- distantia(
  tsl = tsl,
  lock_step = TRUE
)

df_ls[, c("x", "y", "psi")]
#>      x    y      psi
#> 1 X132 X134 1.888451
#> 3 X132 X153 2.128340
#> 5 X134 X153 2.187862
#> 4 X134 X136 2.270977
#> 2 X132 X136 2.427479
#> 6 X136 X153 2.666099
```

The “psi” column shows normalized dissimilarity values and is used to
sort the data frame from lowest to highest dissimilarity. Hence, the
first row shows the most similar pair of time series.

The function `distantia_boxplot()` enables a quick identification of the
time series that are either more dissimilar (top) or similar (bottom) to
others.

``` r
distantia_boxplot(df = df_ls, text_cex = 0.8)
```

![](man/figures/README-unnamed-chunk-6-1.png)<!-- -->

#### Dynamic Time Warping

By default, `distantia()` computes unrestricted dynamic time warping
with orthogonal and diagonal least-cost paths.

``` r
df_dtw <- distantia(
  tsl = tsl
)

df_dtw[, c("x", "y", "psi")]
#>      x    y      psi
#> 1 X132 X134 1.163499
#> 5 X134 X153 2.115504
#> 4 X134 X136 2.206298
#> 3 X132 X153 2.306471
#> 2 X132 X136 2.307844
#> 6 X136 X153 2.686551
```

The function `distantia_dtw_plot()` provides detailed insights into the
alignment between a pair of time series resulting from DTW.

``` r
distantia_dtw_plot(
  tsl = tsl[c("X132", "X153")]
)
```

<img src="man/figures/README-unnamed-chunk-8-1.png" alt="" width="100%" />

Deviations from the perfect diagonal in the least-cost path reveal
adjustments made by DTW to align time series by shape rather than time.

The article [Dynamic Time Warping vs
Lock-Step](https://blasbenito.github.io/distantia/articles/dynamic_time_warping_and_lock_step.html)
provides further insights into the advantages and disadvantages of each
method in different scenarios.

### Permutation Test

The function `distantia()` implements restricted permutation tests to
assess the significance of dissimilarity scores. It provides several
setups to support different assumptions.

For example, the configuration below rearranges complete rows within
7-day blocks, assuming strong dependencies within rows and between
observations that are close in time.

``` r
future::plan(future::multisession)

df_dtw <- distantia(
  tsl = tsl,
  repetitions = 1000,
  permutation = "restricted_by_row",
  block_size = 7 #one week
)

future::plan(future::sequential)

df_dtw[, c("x", "y", "psi", "p_value")]
#>      x    y      psi p_value
#> 1 X132 X134 1.163499   0.001
#> 5 X134 X153 2.115504   0.380
#> 4 X134 X136 2.206298   0.134
#> 3 X132 X153 2.306471   0.376
#> 2 X132 X136 2.307844   0.002
#> 6 X136 X153 2.686551   0.005
```

The “p_value” column represents the fraction of permutations yielding a
psi score lower than the observed value, and indicates the strength of
similarity between two time series. A significance threshold (e.g.,
0.05, depending on the number of repetitions) helps identify pairs of
time series with a robust similarity.

### Variable Importance

When comparing multivariate time series, certain variables contribute
more to similarity or dissimilarity. The `momentum()` function uses a
leave-one-out algorithm to quantify each variable’s contribution to the
overall dissimilarity between two time series.

``` r
df_importance <- momentum(
  tsl = tsl
)

df_importance[, c("x", "y", "variable", "importance", "effect")]
#>       x    y    variable  importance               effect
#> 1  X132 X134           x   89.721187 decreases similarity
#> 2  X132 X134           y  101.305396 decreases similarity
#> 3  X132 X134       speed  -28.286386 increases similarity
#> 4  X132 X134 temperature   78.130508 decreases similarity
#> 5  X132 X134     heading  -43.644053 increases similarity
#> 6  X132 X136           x   15.687570 decreases similarity
#> 7  X132 X136           y   82.867368 decreases similarity
#> 8  X132 X136       speed  -67.196851 increases similarity
#> 9  X132 X136 temperature  382.039900 decreases similarity
#> 10 X132 X136     heading -104.245839 increases similarity
#> 11 X132 X153           x  467.261463 decreases similarity
#> 12 X132 X153           y  159.727491 decreases similarity
#> 13 X132 X153       speed  -44.549191 increases similarity
#> 14 X132 X153 temperature   -4.016121 increases similarity
#> 15 X132 X153     heading  -88.852346 increases similarity
#> 16 X134 X136           x   36.205194 decreases similarity
#> 17 X134 X136           y   90.757712 decreases similarity
#> 18 X134 X136       speed  -61.923595 increases similarity
#> 19 X134 X136 temperature  348.244258 decreases similarity
#> 20 X134 X136     heading  -96.737145 increases similarity
#> 21 X134 X153           x  761.132445 decreases similarity
#> 22 X134 X153           y   26.329542 decreases similarity
#> 23 X134 X153       speed  -62.801312 increases similarity
#> 24 X134 X153 temperature  -23.264433 increases similarity
#> 25 X134 X153     heading  -76.874072 increases similarity
#> 26 X136 X153           x  530.402462 decreases similarity
#> 27 X136 X153           y   24.217594 decreases similarity
#> 28 X136 X153       speed  -67.659029 increases similarity
#> 29 X136 X153 temperature  255.711643 decreases similarity
#> 30 X136 X153     heading -119.877355 increases similarity
```

Positive “importance” values indicate variables contributing to
dissimilarity, while negative values indicate contributions to
similarity. The function documentation provides more details on how
importance scores are computed.

The `momentum_boxplot()` function provides a quick insight into which
variables contribute the most to similarity or dissimilarity across all
pairs of time series.

``` r
momentum_boxplot(
  df = df_importance
)
```

<img src="man/figures/README-unnamed-chunk-11-1.png" alt="" width="75%" />

### Clustering

The package `distantia` provides tools to group time series by
dissimilarity using hierarchical or K-means clustering. The example
below applies the former to the `albatross` dataset to identify groups
of individuals with the most similar movement time series.

``` r
dtw_hclust <- distantia_cluster_hclust(
  df = df_dtw,
  clusters = NULL, #automatic mode
  method = NULL    #automatic mode
  )

#cluster object
dtw_hclust$cluster_object
#> 
#> Call:
#> stats::hclust(d = d_dist, method = method)
#> 
#> Cluster method   : ward.D 
#> Number of objects: 4

#number of clusters
dtw_hclust$clusters
#> [1] 3

#clustering data frame
#group label in column "cluster"
# negatives in column "silhouette_width" highlight anomalous cluster assignments
dtw_hclust$df
#>   name cluster silhouette_width
#> 1 X132       1        0.4955501
#> 2 X134       1        0.4500134
#> 3 X136       2        0.0000000
#> 4 X153       3        0.0000000

#tree plot
par(mar=c(3,1,1,3))

plot(
  x = stats::as.dendrogram(
    dtw_hclust$cluster_object
    ),
  horiz = TRUE
)
```

<img src="man/figures/README-unnamed-chunk-12-1.png" alt="" width="75%" />

This is just a summary of the features implemented in the package.
Please visit the **Articles** section to find out more about
`distantia`.

## Getting help

If you encounter bugs or issues with the documentation, please [file an
issue on GitHub](https://github.com/BlasBenito/distantia/issues).

#' Aggregate `distantia()` Data Frames Across Parameter Combinations
#'
#' @description
#'
#' The function [distantia()] allows dissimilarity assessments based on several combinations of arguments at once. For example, when the argument `distance` is set to `c("euclidean", "Manhattan")`, the output data frame will show two dissimilarity scores for each pair of compared time series, one based on Euclidean distances, and another based on Manhattan distances.
#'
#' This function computes dissimilarity stats across combinations of parameters.
#'
#' If psi scores smaller than zero occur in the aggregated output, then the smaller psi value is added to the column `psi` to start dissimilarity scores at zero.
#'
#' If there is only one combination of arguments in the input data frame, no aggregation occurs and all parameter columns are removed.
#'
#' @param df (required, data frame) Output of [distantia()], [distantia_ls()], [distantia_dtw()], or [distantia_time_delay()]. Default: NULL
#' @param f (optional, function) Function to summarize psi scores (for example, `mean`) when there are several combinations of parameters in `df`. Ignored when there is a single combination of arguments in the input. Default: `mean`
#' @param ... (optional, arguments of `f`) Further arguments to pass to the function `f`.
#'
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#' #three time series
#' #climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl,
#'     guide_columns = 3
#'     )
#' }
#'
#' #distantia with multiple parameter combinations
#' #-------------------------------------
#' df <- distantia(
#'   tsl = tsl,
#'   distance = c("euclidean", "manhattan"),
#'   lock_step = TRUE
#' )
#'
#' df[, c(
#'   "x",
#'   "y",
#'   "distance",
#'   "psi"
#' )]
#'
#' #aggregation using means
#' df <- distantia_aggregate(
#'   df = df,
#'   f = mean
#' )
#'
#' df
#' @family distantia_support
#' 
distantia_aggregate <- function(
    df = NULL,
    f = mean,
    ...
){

  df_type <- attributes(df)$type

  if(df_type != "distantia_df"){
    stop("distantia::distantia_aggregate(): argument 'df' must be the output of 'distantia()'.", call. = FALSE)
  }

  if(!is.function(f)){
    stop("distantia::distantia_aggregate(): argument 'f' must be a function such as mean, median, mean, max, and the likes.", call. = FALSE)
  }

  # split ----
  df_list <- utils_distantia_df_split(
    df = df
  )

  #no aggregation needed
  if(length(df_list) == 1){
    df$distance <- NULL
    df$diagonal <- NULL
    df$bandwidth <- NULL
    df$repetitions <- NULL
    df$seed <- NULL
    df$block_size <- NULL
    df$permutation <- NULL
    df$lock_step <- NULL
    return(df)
  }

  # aggregate ----
    df <- stats::aggregate(
      psi ~ x + y,
      data = df,
      FUN = f,
      ...
    )

  #start aggregated distances at zero
  if(min(df$psi) < 0){
    df$psi <- df$psi + abs(min(df$psi))
  }

  #add type
  attr(
    x = df,
    which = "type"
  ) <- df_type

  df

}

test_that("`distantia_stats()` returns one row per time series with expected columns", {
  tsl <- tsl_simulate(n = 4, irregular = FALSE)

  df <- distantia(tsl = tsl, lock_step = TRUE)

  out <- distantia_stats(df = df)

  expect_equal(class(out), "data.frame")
  expect_equal(nrow(out), length(tsl))
  expect_true(all(c("name", "mean", "min", "q1", "median", "q3", "max", "sd", "range") %in% colnames(out)))
  expect_true(all(is.numeric(out$mean)))
  expect_true(all(out$range >= 0))
})

test_that("`distantia_stats()` warns on NA psi and still returns stats", {
  tsl <- tsl_simulate(n = 3, irregular = FALSE)

  df <- distantia(tsl = tsl, lock_step = TRUE)

  # inject an NA psi to simulate a flat/constant series edge case
  df$psi[1] <- NA

  expect_warning(
    out <- distantia_stats(df = df),
    regexp = "NA psi"
  )

  expect_equal(class(out), "data.frame")
  expect_true(all(!is.na(out$mean)))
})

test_that("`distantia_aggregate()` works with multiple parameter combinations", {
  tsl <- tsl_simulate(n = 3, irregular = FALSE)

  # lock_step = c(TRUE, FALSE) ensures 'diagonal' column is present, triggering
  # the multi-combo aggregate path (the original bug: aggregate(x=df, by=formula))
  df <- distantia(tsl = tsl, distance = c("euclidean", "manhattan"), lock_step = c(TRUE, FALSE))

  out <- distantia_aggregate(df = df, f = mean)

  expect_equal(class(out), "data.frame")
  expect_true("psi" %in% colnames(out))
  expect_true("x" %in% colnames(out))
  expect_true("y" %in% colnames(out))
  # one row per pair (3 pairs for 3 time series, all combos collapsed)
  expect_equal(nrow(out), 3L)
})

#' Distance Matrix
#'
#' @description
#' Demonstration function to compute the distance matrix between two univariate or multivariate time series.
#'
#' This function does not accept NA data in the matrices `x` and `y`.
#'
#' @param x (required, zoo object or numeric matrix) a time series with no NAs. Default: NULL
#' @param y (zoo object or numeric matrix) a time series with the same columns as `x` and no NAs. Default: NULL
#' @inheritParams distantia
#' @examples
#' #distance metric
#' d <- "euclidean"
#'
#' #simulate two irregular time series
#' x <- zoo_simulate(
#'   name = "x",
#'   rows = 100,
#'   seasons = 2,
#'   seed = 1
#' )
#'
#' y <- zoo_simulate(
#'   name = "y",
#'   rows = 80,
#'   seasons = 2,
#'   seed = 2
#' )
#'
#' if(interactive()){
#'   zoo_plot(x = x)
#'   zoo_plot(x = y)
#' }
#'
#' #distance matrix
#' dist_matrix <- psi_distance_matrix(
#'   x = x,
#'   y = y,
#'   distance = d
#' )
#'
#' if(interactive()){
#'   utils_matrix_plot(
#'     m = dist_matrix
#'     )
#' }
#' @return numeric matrix
#' @export
#' @autoglobal
#' @family psi_demo
psi_distance_matrix <- function(
    x = NULL,
    y = NULL,
    distance = "euclidean"
){

  distance <- utils_check_distance_args(
    distance = distance
  )[1]

  if(zoo::is.zoo(x)){
    x_index <- zoo::index(x)
  } else {
    x_index <- 1:nrow(x)
  }

  if(zoo::is.zoo(y)){
    y_index <- zoo::index(y)
  } else {
    y_index <- 1:nrow(y)
  }

  #computing distance matrix
  m <- distance_matrix_cpp(
    x = as.matrix(x),
    y = as.matrix(y),
    distance = distance
    )

  #adding names
  dimnames(m) <- list(
    as.character(y_index),
    as.character(x_index)
  )

  x_name <- attributes(x)$name
  if(is.null(x_name)){
    x_name <- "x"
  }

  y_name <- attributes(y)$name
  if(is.null(y_name)){
    y_name <- "y"
  }

  #adding attributes
  attr(x = m, which = "x_time") <- x_index
  attr(x = m, which = "y_time") <- y_index
  attr(x = m, which = "x_name") <- x_name
  attr(x = m, which = "y_name") <- y_name
  attr(x = m, which = "type") <- "distance"
  attr(x = m, which = "distance") <- distance

  m

}

test_that("psi_equation returns NA when b is zero", {
  expect_true(is.na(psi_equation(a = 0, b = 0, diagonal = TRUE)))
  expect_true(is.na(psi_equation(a = 1, b = 0, diagonal = TRUE)))
})

test_that("`psi_distance_lock_step()` errors on unequal-length series", {
  x <- zoo_simulate(name = "x", rows = 10, seed = 1)
  y <- zoo_simulate(name = "y", rows = 20, seed = 2)
  expect_error(
    psi_distance_lock_step(x = x, y = y, distance = "euclidean"),
    "same number of rows"
  )
})

test_that("`psi_...()` works", {

  d <- "euclidean"

  diagonal <- TRUE

  x <- zoo_simulate(
    name = "x",
    rows = 20,
    seasons = 1,
    seed = 1
    )

  expect_true(zoo::is.zoo(x))
  expect_true(nrow(zoo::coredata(x)) == 20)

  y <- zoo_simulate(
    name = "y",
    rows = 30,
    seasons = 1,
    seed = 2
    )

  expect_true(zoo::is.zoo(y))
  expect_true(nrow(zoo::coredata(y)) == 30)

  dist_matrix <- psi_distance_matrix(
    x = x,
    y = y,
    distance = d
    )

  expect_true(is.matrix(dist_matrix))
  expect_true(ncol(dist_matrix) == 20)
  expect_true(nrow(dist_matrix) == 30)


  cost_matrix <- psi_cost_matrix(
    dist_matrix = dist_matrix,
    diagonal = diagonal
    )

  expect_true(is.matrix(cost_matrix))
  expect_equal(dim(dist_matrix), dim(cost_matrix))


  cost_path <- psi_cost_path(
    dist_matrix = dist_matrix,
    cost_matrix = cost_matrix,
    diagonal = diagonal
  )

  expect_true(is.data.frame(cost_path))
  expect_equal(colnames(cost_path), c("x", "y", "dist", "cost"))

  a <- psi_cost_path_sum(path = cost_path)

  expect_true(is.numeric(a))

  b <- psi_auto_sum(
    x = x,
    y = y,
    distance = d
    )

  expect_true(is.numeric(b))

  psi <- psi_equation(
    a = a,
    b = b,
    diagonal = diagonal
    )

  expect_true(is.numeric(psi))

  #compare with distantia
  expect_equal(
    distantia(
      tsl = list(x = x, y = y),
      distance = d
      )$psi,
    psi
    )

})

#' Two-Way Dissimilarity Plots of Time Series Lists
#'
#' @description
#'
#' Plots two sequences, their distance or cost matrix, their least cost path, and all relevant values used to compute dissimilarity.
#'
#' Unlike [distantia()], this function does not accept vectors as inputs for the arguments to compute dissimilarity (`distance`, `diagonal`,  and `weighted`), and only plots a pair of sequences at once.
#'
#' The argument `lock_step` is not available because this plot does not make sense in such a case.
#'
#' @inheritParams distantia
#' @param matrix_type (optional, character string): one of "cost" or "distance" (the abbreviation "dist" is accepted as well). Default: "cost".
#' @param matrix_color (optional, character vector) Vector of colors for the distance or cost matrix. If NULL, uses the palette "Zissou 1" provided by [grDevices::hcl.colors()]. Default: NULL
#' @param path_width (optional, numeric) width of the least cost path. Default: 1
#' @param path_color (optional, character string) color of the least-cost path. Default: "black"
#' @param diagonal_width (optional, numeric) width of the diagonal. Set to 0 to remove the diagonal line. Default: 0.5
#' @param diagonal_color (optional, character string) color of the diagonal. Default: "white"
#' @param line_color (optional, character vector) Vector of colors for the time series plot. If NULL, defaults to a subset of `matrix_color`.
#' @param line_width (optional, numeric vector) Width of the time series plot. Default: 1
#' @param text_cex (optional, numeric) Multiplier of the text size. Default: 1
#'
#' @examples
#' #three time series
#' #climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
#' #convert to time series list
#' #scale and center to neutralize effect of different scales in temperature, rainfall, and ndvi
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global #see help(f_scale_global)
#'   )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl,
#'     guide_columns = 3
#'     )
#' }
#'
#' #visualize dynamic time warping
#' if(interactive()){
#'
#'   #plot pair with cost matrix (default)
#'   distantia_dtw_plot(
#'     tsl = tsl[c("Spain", "Sweden")] #only two time series!
#'   )
#'
#'   #plot pair with distance matrix
#'   distantia_dtw_plot(
#'     tsl = tsl[c("Spain", "Sweden")],
#'     matrix_type = "distance"
#'   )
#'
#'   #plot pair with different distance
#'   distantia_dtw_plot(
#'     tsl = tsl[c("Spain", "Sweden")],
#'     distance = "manhattan", #sed data(distances)
#'     matrix_type = "distance"
#'   )
#'
#'
#'   #with different colors
#'   distantia_dtw_plot(
#'     tsl = tsl[c("Spain", "Sweden")],
#'     matrix_type = "distance",
#'     matrix_color = grDevices::hcl.colors(
#'       n = 100,
#'       palette = "Inferno"
#'     ),
#'     path_color = "white",
#'     path_width = 2,
#'     line_color = grDevices::hcl.colors(
#'       n = 3, #same as variables in tsl
#'       palette = "Inferno"
#'     )
#'   )
#'
#' }
#' @return multipanel plot
#' @autoglobal
#' @export
#' @family distantia
distantia_dtw_plot <- function(
    tsl = NULL,
    distance = "euclidean",
    diagonal = TRUE,
    bandwidth = 1,
    matrix_type = "cost",
    matrix_color = NULL,
    path_width = 1,
    path_color = "black",
    diagonal_width = 1,
    diagonal_color = "white",
    line_color = NULL,
    line_width = 1,
    text_cex = 1
){

  #length > 1
  if(length(tsl) < 2){
    stop("distantia::distantia_dtw_plot(): argument 'tsl' must be a time series list of length 2 or higher.", call. = FALSE)
  }

  #only two elements
  if(length(tsl) > 2){

    message("distantia::distantia_dtw_plot(): Argument 'tsl' has more than two time series. Using ", paste0(names(tsl)[1:2], collapse = " and "), " to build the plot. Please use tsl_subset() or the notation txl[c('a', 'b')] to select a different pair of time series.")

    tsl <- tsl_subset(
      tsl = tsl,
      names = c(1, 2)
    )

  }

  #count and handle NA
  na_count <- tsl_count_NA(
    tsl = tsl
  ) |>
    unlist()

  if(sum(na_count) > 0){
    message(
      "distantia::distantia_dtw_plot(): ",
      sum(na_count),
      " NA cases were imputed with distantia::tsl_handle_NA() in these time series: '",
      paste(names(na_count[na_count > 0]), collapse = "', '"),
      "'."
    )
  }

  tsl <- tsl_handle_NA(
    tsl = tsl
  )

  #subset numeric and shared columns
  tsl <- tsl_subset(
    tsl = tsl,
    numeric_cols = TRUE,
    shared_cols = TRUE
  )

  # Preserve user's config
  old.par <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old.par))

  # Check arguments ----

  #check matrix argument
  matrix_type <- match.arg(
    arg = matrix_type,
    choices =  c("cost", "distance", "dist"),
    several.ok = FALSE
  )

  if(is.null(matrix_color)){

    matrix_color = color_continuous(
      n = 100
    )

    if(is.null(line_color)){
      line_color <- matrix_color
    }

  }

  # Psi computation ----
  xy_psi <- distantia(
    tsl = tsl,
    distance = distance[1],
    diagonal = diagonal[1],
    bandwidth = bandwidth[1]
  )$psi

  xy_psi <- round(xy_psi, 3)

  # Distance and cost matrices ----
  dist_m <- psi_distance_matrix(
    x = tsl[[1]],
    y = tsl[[2]],
    distance = distance
  )

  cost_m <- psi_cost_matrix(
    dist_matrix = dist_m,
    diagonal = diagonal
  )

  # Cost path ----
  path <- psi_cost_path(
    dist_matrix = dist_m,
    cost_matrix = cost_m,
    diagonal = diagonal,
    bandwidth = bandwidth
  )

  # select matrix to plot
  if(matrix_type %in% c("distance", "dist")){
    m <- dist_m
    rm(cost_m)
  } else {
    m <- cost_m
    rm(dist_m)
  }

  #path sum
  path_sum <- psi_cost_path_sum(
    path = path
  ) |>
    round(2)

  plt_all <- graphics::par()$plt

  # Plotting areas ----
  plt_y <- c(0.2, 0.35, 0.25, 0.8)
  plt_x <- c(0.35, 0.8, 0.12, 0.25)
  plt_m <- c(0.35, 0.8, 0.25, 0.8)
  plt_matrix_guide <- c(0.82, 0.84, 0.25, 0.8)
  plt_line_guide <- c(0.78, 1, 0.05, 0.25)

  # Plot matrix ----
  graphics::par(
    oma = c(1, 1, 1, 1)
  )

  graphics::par(
    plt = plt_m
    )

  utils_matrix_plot(
    m = m,
    matrix_color = matrix_color,
    subtitle = paste0(
      "Psi = ",
      xy_psi
    ),
    text_cex = text_cex,
    path = path,
    path_width = path_width,
    path_color = path_color,
    diagonal_width = diagonal_width,
    diagonal_color = diagonal_color,
    guide = FALSE,
    subpanel = TRUE
  )

  graphics::mtext(
    text = paste(
      "Cost sum = ",
      path_sum
    ),
    line = -1.3,
    cex = 0.9 * text_cex,
    adj = 0.05,
    col = path_color
  )

  # Plot time series y ----
  graphics::par(
    plt = plt_y,
    new = TRUE,
    mgp = c(0, 0.5, 0),
    bty = "n"
    )

  zoo_plot(
    x = tsl[[2]],
    line_color = line_color,
    line_width = line_width,
    text_cex = text_cex,
    guide = FALSE,
    vertical = TRUE,
    subpanel = TRUE
  )

  y_sum <- psi_auto_distance(
    x = tsl[[2]],
    distance = distance
  ) |>
    round(2)

  graphics::mtext(
    text = paste(
      "Auto sum = ",
      y_sum
    ),
    side = 2,
    line = -1,
    cex = 0.9 * text_cex,
    adj = 0.05
  )

  # Plot time series x ----
  graphics::par(
    plt = plt_x,
    new = TRUE,
    mgp = c(0, 0.5, 0),
    bty = "n"
    )

  zoo_plot(
    x = tsl[[1]],
    line_color = line_color,
    line_width = line_width,
    text_cex = text_cex,
    guide = FALSE,
    vertical = FALSE,
    subpanel = TRUE
  )

  x_sum <- psi_auto_distance(
    x = tsl[[1]],
    distance = distance
  ) |>
    round(2)

  graphics::mtext(
    text = paste(
      "Auto sum = ",
      x_sum
    ),
    side = 3,
    line = -1,
    cex = 0.9 * text_cex,
    adj = 0.05
  )

  # Plot matrix guide ----
  graphics::par(
    plt = plt_matrix_guide,
    new = TRUE,
    mgp = c(3, 0.5, 0)
  )

  utils_matrix_guide(
    m = m,
    matrix_color = matrix_color,
    text_cex = text_cex
  )

  # Plot line guide ----
  graphics::par(
    plt = plt_line_guide,
    new = TRUE
  )

  utils_line_guide(
    x = tsl[[2]],
    position = "topleft",
    line_color = line_color,
    text_cex = text_cex * 0.6,
    line_width = line_width,
    subpanel = TRUE
  )

  #reset to main plotting area
  # graphics::par(plt = plt_all)

  invisible()

}

#' Aggregate Time Series List Over Time Periods
#'
#' @description
#'
#' Time series aggregation involves grouping observations and summarizing group values with a statistical function. This operation is useful to:
#' \itemize{
#'   \item Decrease (downsampling) the temporal resolution of a time series.
#'   \item Highlight particular states of a time series over time. For example, a daily temperature series can be aggregated by month using `max` to represent the highest temperatures each month.
#'   \item Transform irregular time series into regular.
#' }
#'
#' This function aggregates time series lists \strong{with overlapping times}. Please check such overlap by assessing the columns "begin" and "end " of the data frame resulting from `df <- tsl_time(tsl = tsl)`. Aggregation will be limited by the shortest time series in your time series list. To aggregate non-overlapping time series, please subset the individual components of `tsl` one by one either using [tsl_subset()] or the syntax `tsl = my_tsl[[i]]`.
#'
#' \strong{Methods}
#'
#' Any function returning a single number from a numeric vector can be used to aggregate a time series list. Quoted and unquoted function names can be used. Additional arguments to these functions can be passed via the argument `...`. Typical examples are:
#'
#' \itemize{
#'   \item `mean` or `"mean"`: see [mean()].
#'   \item `median` or `"median"`: see [stats::median()].
#'   \item `quantile` or "quantile": see [stats::quantile()].
#'   \item `min` or `"min"`: see [min()].
#'   \item `max` or `"max"`: see [max()].
#'   \item `sd` or `"sd"`: to compute standard deviation, see [stats::sd()].
#'   \item `var` or `"var"`: to compute the group variance, see [stats::var()].
#'   \item `length` or `"length"`: to compute group length.
#'   \item `sum` or `"sum"`: see [sum()].
#' }
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param new_time (required, numeric, numeric vector, Date vector, POSIXct vector, or keyword) Definition of the aggregation pattern. The available options are:
#' \itemize{
#'   \item numeric vector: only for the "numeric" time class, defines the breakpoints for time series aggregation.
#'   \item "Date" or "POSIXct" vector: as above, but for the time classes "Date" and "POSIXct." In any case, the input vector is coerced to the time class of the `tsl` argument.
#'   \item numeric: defines fixed time intervals in the units of `tsl` for time series aggregation. Used as is when the time class is "numeric", and coerced to integer and interpreted as days for the time classes "Date" and "POSIXct".
#'   \item keyword (see [utils_time_units()]): the common options for the time classes "Date" and "POSIXct" are: "millennia", "centuries", "decades", "years", "quarters", "months", and "weeks". Exclusive keywords for the "POSIXct" time class are: "days", "hours", "minutes", and "seconds". The time class "numeric" accepts keywords coded as scientific numbers, from "1e8" to "1e-8".
#' }
#' @param f (required, function name) Name of function taking a vector as input and returning a single value as output. Typical examples are `mean`, `max`,`min`, `median`, and `quantile`. Default: `mean`.
#' @param ... (optional) further arguments for `f`.
#'
#' @return time series list
#' @export
#' @seealso [zoo_aggregate()]
#' @examples
#'
#' # yearly aggregation
#' #----------------------------------
#' #long-term monthly temperature of 20 cities
#' tsl <- tsl_initialize(
#'   x = cities_temperature,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #plot time series
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl[1:4],
#'     guide_columns = 4
#'   )
#' }
#'
#' #check time features
#' tsl_time(tsl)[, c("name", "resolution", "units")]
#'
#' #aggregation: mean yearly values
#' tsl_year <- tsl_aggregate(
#'   tsl = tsl,
#'   new_time = "year",
#'   f = mean
#' )
#'
#' #check time features
#' tsl_time(tsl_year)[, c("name", "resolution", "units")]
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl_year[1:4],
#'     guide_columns = 4
#'   )
#' }
#'
#'
#' # other supported keywords
#' #----------------------------------
#'
#' #simulate full range of calendar dates
#' tsl <- tsl_simulate(
#'   n = 2,
#'   rows = 1000,
#'   time_range = c(
#'     "0000-01-01",
#'     as.character(Sys.Date())
#'   )
#' )
#'
#' #mean value by millennia (extreme case!!!)
#' tsl_millennia <- tsl_aggregate(
#'   tsl = tsl,
#'   new_time = "millennia",
#'   f = mean
#' )
#'
#' if(interactive()){
#'   tsl_plot(tsl_millennia)
#' }
#'
#' #max value by centuries
#' tsl_century <- tsl_aggregate(
#'   tsl = tsl,
#'   new_time = "century",
#'   f = max
#' )
#'
#' if(interactive()){
#'   tsl_plot(tsl_century)
#' }
#'
#' #quantile 0.75 value by centuries
#' tsl_centuries <- tsl_aggregate(
#'   tsl = tsl,
#'   new_time = "centuries",
#'   f = stats::quantile,
#'   probs = 0.75 #argument of stats::quantile()
#' )
#'
#' @family tsl_processing
tsl_aggregate <- function(
    tsl = NULL,
    new_time = NULL,
    f = mean,
    ...
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  new_time <- utils_new_time(
    tsl = tsl,
    new_time = new_time,
    keywords = "aggregate"
  )

  #testing version
  # tsl <- future.apply::future_lapply(
  #   X = tsl,
  #   FUN = function(x){
  #
  #     x <- zoo_aggregate(
  #       x = x,
  #       new_time = new_time,
  #       f = f
  #     )
  #
  #   }
  # )

  #progress bar
  p <- progressr::progressor(along = tsl)

  #parallelized loop
  tsl <- future.apply::future_lapply(
    X = tsl,
    FUN = function(x, ...){

      p()

      x <- zoo_aggregate(
        x = x,
        new_time = new_time,
        f = f,
        ... = ...
      )

    },
    ... = ...
  )

  tsl <- tsl_names_set(
    tsl = tsl
  )

  tsl

}

test_that("`tsl_aggregate()` works", {

  tsl <- tsl_initialize(
    x = cities_temperature,
    name_column = "name",
    time_column = "time"
  )

  tsl_year <- tsl_aggregate(tsl = tsl, new_time = "year", f = mean)

  expect_true(tsl_time(tsl)$resolution[1] < tsl_time(tsl_year)$resolution[1])

  expect_true(tsl_nrow(tsl)[[1]] > tsl_nrow(tsl_year)[[1]])

})

test_that("`utils_cluster_silhouette()` handles k = 1 (single cluster)", {
  m <- matrix(
    c(0, 1, 2, 1, 0, 1, 2, 1, 0),
    nrow = 3,
    dimnames = list(c("a", "b", "c"), c("a", "b", "c"))
  )
  labels <- c(a = 1L, b = 1L, c = 1L)

  out_df <- utils_cluster_silhouette(labels = labels, d = m, mean = FALSE)
  expect_equal(colnames(out_df), c("name", "cluster", "silhouette_width"))
  expect_true(all(is.na(out_df$silhouette_width)))

  out_mean <- utils_cluster_silhouette(labels = labels, d = m, mean = TRUE)
  expect_true(is.na(out_mean))
})

test_that("`utils_cluster_silhouette()`, `utils_cluster_hclust_optimizer()` and `utils_cluster_kmeans_optimizer()` works", {

  tsl <- tsl_aggregate(
    tsl_subset(tsl_initialize(
      x = covid_prevalence,
      name_column = "name", time_column = "time"
    ), names = 1:10),
    new_time = "months", fun = max
  )

  psi_matrix <- distantia_matrix(distantia(tsl = tsl, lock_step = TRUE))
  hclust_optimization <- utils_cluster_hclust_optimizer(d = psi_matrix)

  expect_equal(class(hclust_optimization), "data.frame")
  expect_equal(colnames(hclust_optimization), c("clusters", "method", "silhouette_mean"))

  psi_hclust <- stats::hclust(d = as.dist(psi_matrix[[1]]))
  psi_hclust_labels <- stats::cutree(tree = psi_hclust, k = 3, )
  expect_equal(colnames(utils_cluster_silhouette(labels = psi_hclust_labels, d = psi_matrix)), c("name", "cluster", "silhouette_width"))

  kmeans_optimization <- utils_cluster_kmeans_optimizer(d = psi_matrix)

  expect_equal(class(kmeans_optimization), "data.frame")
  expect_equal(colnames(kmeans_optimization), c("clusters", "silhouette_mean"))

  psi_kmeans <- stats::kmeans(x = as.dist(psi_matrix[[1]]), centers = 3)
  expect_equal(colnames(utils_cluster_silhouette(labels = psi_kmeans$cluster, d = psi_matrix)), c("name", "cluster", "silhouette_width"))

})

#' Importance Model Frame
#'
#' @description
#' This function generates a model frame for statistical or machine learning analysis from these objects:
#'
#' \itemize{
#'   \item An importance data frame generated by [momentum()], [momentum_ls()], or [momentum_dtw()]. The output model frame will have as many rows as this data frame.
#'   \item A data frame with static descriptors of the time series. These descriptors are converted to distances between pairs of time series via [distance_matrix()].
#'   \item A list defining composite predictors. This feature allows grouping together predictors that have a common meaning. For example, `composite_predictors = list(temperature = c("temperature_mean", "temperature_min", "temperature_max"))` generates a new predictor named "temperature", which results from computing the multivariate distances between the vectors of temperature variables of each pair of time series. Predictors in one such group will be scaled before distance computation if their maximum standard deviations differ by a factor of 10 or more.
#' }
#'
#'
#' The resulting data frame contains the following columns:
#' \itemize{
#'   \item `x` and `y`: names of the pair of time series represented in the row.
#'   \item response columns.
#'   \item predictors columns: representing the distance between the values of the given static predictor between `x` and `y`.
#'   \item (optional) `geographic_distance`: If `predictors_df` is an `sf` data frame, then geographic distances are computed via [sf::st_distance()].
#' }
#'
#' This function supports a parallelization setup via [future::plan()].
#'
#' @param response_df (required, data frame) Output of [momentum()], [momentum_ls()], or [momentum_dtw()]. Default: NULL
#' @param predictors_df (required, data frame or sf data frame) Data frame with numeric predictors for the model frame. Must have a column with the time series names in `response_df$x` and `response_df$y`. If an `sf` data frame, the column "geographic_distance" with distances between pairs of time series is added to the model frame. Default: NULL
#' @param composite_predictors (optional, list) List defining composite predictors. For example, `composite_predictors = list(a = c("b", "c"))` uses the columns `"b"` and `"c"` from `predictors_df` to generate the predictor `a` as the multivariate distance between `"b"` and `"c"` for each pair of time series in `response_df`. Default: NULL
#' @param scale (optional, logical) If TRUE, all predictors are scaled and centered with [scale()]. Default: TRUE
#' @param distance (optional, string) Method to compute the distance between predictor values for all pairs of time series in `response_df`. Default: "euclidean".
#'
#' @return A data frame with the attribute "predictors".
#' @export
#' @autoglobal
#' @examples
#'
#' #Fagus sylvatica dynamics in Europe
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #importance analysis
#' df <- momentum_ls(tsl = tsl)
#'
#' #generate model frame
#' model_frame <- momentum_model_frame(
#'   response_df = df,
#'   predictors_df = fagus_coordinates,
#'   scale = TRUE
#' )
#'
#' head(model_frame)
#'
#' #names of response and predictors
#' #and an additive formula
#' #are stored as attributes
#' attributes(model_frame)$predictors
#'
#' @family momentum_support
momentum_model_frame <- distantia_model_frame

#' Aggregate Cases in Zoo Time Series
#'
#' @param x (required, zoo object) Time series to aggregate. Default: NULL
#' @param new_time (optional, zoo object, keyword, or time vector) New time to aggregate `x` to. The available options are:
#' \itemize{
#'   \item NULL: the highest resolution keyword returned by `zoo_time(x)$keywords` is used to generate a new time vector to aggregate `x`.
#'   \item zoo object: the index of the given zoo object is used as template to aggregate `x`.
#'   \item time vector: a vector with new times to resample `x` to. If time in `x` is of class "numeric", this vector must be numeric as well. Otherwise, vectors of classes "Date" and "POSIXct" can be used indistinctly.
#'   \item keyword: a valid keyword returned by `zoo_time(x)$keywords`, used to generate a time vector with the relevant units.
#'   \item numeric of length 1: interpreted as new time interval, in the highest resolution units returned by `zoo_time(x)$units`.
#' }
#' @param f (optional, quoted or unquoted function name) Name of a function to aggregate numeric vectors. Typical examples are `mean`, `max`,`min`, `median`, and `quantile`. Default: `mean`.
#' @param ... (optional, additional arguments) additional arguments to `f`.
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' #full range of calendar dates
#' x <- zoo_simulate(
#'   rows = 1000,
#'   time_range = c(
#'     "0000-01-01",
#'     as.character(Sys.Date())
#'     )
#' )
#'
#' #plot time series
#' if(interactive()){
#'   zoo_plot(x)
#' }
#'
#'
#' #find valid aggregation keywords
#' x_time <- zoo_time(x)
#' x_time$keywords
#'
#' #mean value by millennia (extreme case!!!)
#' x_millennia <- zoo_aggregate(
#'   x = x,
#'   new_time = "millennia",
#'   f = mean
#' )
#'
#' if(interactive()){
#'   zoo_plot(x_millennia)
#' }
#'
#' #max value by centuries
#' x_centuries <- zoo_aggregate(
#'   x = x,
#'   new_time = "centuries",
#'   f = max
#' )
#'
#' if(interactive()){
#'   zoo_plot(x_centuries)
#' }
#'
#' #quantile 0.75 value by centuries
#' x_centuries <- zoo_aggregate(
#'   x = x,
#'   new_time = "centuries",
#'   f = stats::quantile,
#'   probs = 0.75 #argument of stats::quantile()
#' )
#'
#' if(interactive()){
#'   zoo_plot(x_centuries)
#' }
#' @family zoo_functions
zoo_aggregate <- function(
    x = NULL,
    new_time = NULL,
    f = mean,
    ...
){

  if(zoo::is.zoo(x) == FALSE){
    stop("distantia::zoo_aggregate(): argument 'x' must be a zoo object.", call. = FALSE)
  }

  f <- tryCatch(
    match.fun(f),
    error = function(e){
      stop("distantia::zoo_aggregate(): Argument 'f' must be a function name. Examples of valid options are: 'mean', 'median', 'max', 'min', and 'sd'.", call. = FALSE)
    }
  )

  #new_time from keyword
  if(is.null(new_time)){

    old_time <- zoo_time(
      x = x,
      keywords = "aggregate"
    )

    new_time <- utils::tail(
      x = unlist(old_time$keywords),
      n = 1
      )

    message("distantia::zoo_aggregate(): aggregating 'x' with keyword '", new_time, "'.")

  }

  #process new_time
  new_time <- utils_new_time(
    tsl = zoo_to_tsl(x = x),
    new_time = new_time,
    keywords = "aggregate"
  )

  #new_time to aggregation intervals
  new_time_groups <- cut(
    x = zoo::index(x),
    breaks = new_time,
    include.lowest = TRUE,
    right = TRUE
  ) |>
    as.numeric()

  #aggregate x as y
  y <- stats::aggregate(
    x = x,
    by = new_time_groups,
    FUN = f,
    ... = ...
  )

  colnames(y) <- colnames(x)

  y <- zoo::zoo(
    x = y,
    order.by = new_time
  )

  #reset name
  zoo_name_set(
    x = y,
    name = attr(x, "name", exact = TRUE)
  )

}

test_that("`zoo_aggregate()` works", {

  x <- zoo_simulate(rows = 1000, time_range = c("0000-01-01", as.character(Sys.Date())), seed = 1)

  x_millennia <- zoo_aggregate(x = x, new_time = "millennia", f = mean)

  expect_equal(nrow(x_millennia), 4)

  x_centuries <- zoo_aggregate(x = x, new_time = "centuries", f = max)

  expect_equal(nrow(x_centuries), 22)

})

#' Lock-Step Dissimilarity Analysis of Time Series Lists
#'
#' @description
#'
#' A minimalistic but slightly faster version of [distantia()] that computes lock-step dissimilarity scores.
#'
#' @inheritParams distantia
#'
#' @return data frame:
#' \itemize{
#'   \item `x`: time series name.
#'   \item `y`: time series name.
#'   \item `distance`: name of the distance metric.
#'   \item `lock_step`: always `TRUE`; flags this result as lock-step.
#'   \item `psi`: psi dissimilarity of the sequences `x` and `y`.
#' }
#' @export
#' @autoglobal
#' @examples
#'
#'
#' #load fagus_dynamics as tsl
#' #global centering and scaling
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl,
#'     guide_columns = 3
#'     )
#' }
#'
#' #lock-step dissimilarity analysis
#' df_ls <- distantia_ls(
#'   tsl = tsl,
#'   distance = "euclidean"
#' )
#'
#' df_ls
#'
#' @family distantia
distantia_ls <- function(
    tsl = NULL,
    distance = "euclidean"
){

  #check input arguments
  args <- utils_check_args_distantia(
    tsl = tsl,
    distance = distance
  )

  tsl <- args$tsl
  distance <- args$distance[1]

  #count rows in time series
  row_counts <- tsl |>
    tsl_nrow() |>
    unlist() |>
    unique()

  if(length(row_counts) > 1){

    stop(
      "distantia::distantia_ls(): time series in 'tsl' do not have the same number of rows, cannot perform a lock-step analysis.",
      call. = FALSE
    )

  }

  df <- utils_tsl_pairs(
    tsl = tsl
  )

  #iterate over pairs of time series
  df$psi <- foreach::foreach(
    i = seq_len(nrow(df)),
    .combine = "c",
    .errorhandling = "pass",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

    psi_ls_cpp(
      x = tsl[[df[i, "x"]]],
      y = tsl[[df[i, "y"]]],
      distance = distance
    )

  }

  df$distance <- distance
  df$lock_step <- TRUE
  df <- df[, c("x", "y", "distance", "lock_step", "psi")]

  df <- df[order(df$psi), ]

  #add type
  attr(
    x = df,
    which = "type"
  ) <- "distantia_df"

  df

}

test_that("Functions `distantia()`, `distantia_ls()`, and `distantia_dtw()` work.", {

  #short multivariate tsl for testing
  tsl <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  ) |>
    tsl_subset(
      time = c("2010-01-01", "2011-01-01")
    ) |>
    tsl_transform(
      f = f_scale_global
      )

  #params
  distance = c("euclidean", "manhattan")
  permutation = c(
    "restricted_by_row",
    "restricted",
    "free_by_row",
    "free"
    )
  diagonal <- c(TRUE, FALSE)
  bandwidth <- c(1, 0.5)
  lock_step <- c(TRUE, FALSE)
  robust <- c(TRUE, FALSE)

  #distantia output only
  #message about robust having two values
  testthat::expect_message(
    df <- momentum(
      tsl = tsl,
      distance = distance,
      diagonal = diagonal,
      bandwidth = bandwidth,
      lock_step = lock_step,
      robust = robust
    )
  )

  expect_true(all(c("x", "y", "psi", "variable", "importance", "effect", "psi_difference", "psi_without", "psi_only_with", "distance", "diagonal", "bandwidth", "lock_step", "robust") %in% colnames(df)))

  #check for invalid combinations with lock_step
  expect_true(
    nrow(
      df[df$lock_step == TRUE & !is.na(df$diagonal),]
    ) == 0
  )

  expect_true(
    nrow(
      df[df$lock_step == TRUE & !is.na(df$bandwidth),]
    ) == 0
  )

  expect_true(all(names(tsl) %in% unique(c(df$x, df$y))))

  expect_true(
    all(df[df$x == "Germany" & df$y == "Sweden", "psi"] < df[df$x == "Spain" & df$y == "Sweden", "psi"])
  )

 #compare with momentum ls
  df_ls <- momentum_ls(
    tsl = tsl
  )

  df_ls <- df_ls[, c("x", "y", "variable", "importance")]

  rownames(df_ls) <- NULL

  df_test <- df[df$lock_step == TRUE & df$distance == "euclidean", c("x", "y", "variable", "importance")]

  rownames(df_test) <- NULL

  expect_true(all.equal(df_ls, df_test))

  #compare with distantia_dtw
  df_dtw <- momentum_dtw(
    tsl = tsl
  )

  df_dtw <- df_dtw[, c("x", "y", "variable", "importance")]

  rownames(df_dtw) <- NULL

  df_test <- df[df$lock_step == FALSE & df$diagonal == TRUE & df$bandwidth == 1 &  df$distance == "euclidean", c("x", "y", "variable", "importance")]

  rownames(df_test) <- NULL

  expect_true(all.equal(df_dtw, df_test))

  #test distantia aggregate
  df_agg <- momentum_aggregate(
    df = df
  )

  expect_true(
    nrow(df_agg) == 9
  )

  expect_true(all(c("x", "y", "variable", "psi", "importance", "psi_without", "psi_only_with") %in% colnames(df_agg)))

  #test distantia stats
  df_stats <- momentum_stats(
    df = df
  )

  expect_true(
    nrow(df_stats) == 3
  )

  expect_true(all(c("variable", "mean", "min", "q1", "median", "q3", "max", "sd", "range") %in% colnames(df_stats)))

  #momentum to wide
  df_wide <- momentum_to_wide(
    df = df
  )

  expect_true(all(c("importance__evi", "importance__rainfall", "importance__temperature") %in% colnames(df_wide)))

})

test_that("`distantia_spatial()` produces a valid sf network", {

  tsl <- distantia::tsl_initialize(
    x = distantia::covid_prevalence,
    name_column = "name", time_column = "time"
  )

  df_psi <- distantia::distantia_ls(tsl = tsl)

  sf_psi <- distantia::distantia_spatial(
    df = df_psi, sf = distantia::covid_counties,
    network = TRUE
  )

  # correct class
  expect_true(inherits(sf_psi, "sf"))

  # row count preserved
  expect_equal(nrow(df_psi), nrow(sf_psi))

  # required columns exist
  expect_true("edge_name" %in% colnames(sf_psi))
  expect_true("length"    %in% colnames(sf_psi))

  # intermediate id columns must not leak into output
  expect_false(any(c("id", "id_x", "id_y", "id.x", "id.y") %in% colnames(sf_psi)))

  # all linestring lengths are positive (confirms ids were resolved correctly)
  expect_true(all(sf_psi$length > 0))

  # geometry type is LINESTRING
  expect_equal(
    unique(as.character(sf::st_geometry_type(sf_psi))),
    "LINESTRING"
  )

})

test_that("`distantia_spatial()` warns on identical coordinates (zero-length linestring)", {

  tsl2 <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  )[c("Germany", "Spain")]

  df <- distantia_ls(tsl = tsl2)

  sf_dup <- sf::st_as_sf(
    data.frame(name = c("Germany", "Spain"), x = c(10.0, 10.0), y = c(51.0, 51.0)),
    coords = c("x", "y"),
    crs = 4326
  )

  expect_warning(
    distantia_spatial(df = df, sf = sf_dup),
    regexp = "zero-length",
    fixed = FALSE
  )

})

test_that("`distantia_spatial()` gives informative error when sf has no matching name column", {

  tsl2 <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  )[c("Germany", "Spain")]

  df <- distantia_ls(tsl = tsl2)

  # sf with wrong column name — no column contains "Germany" and "Spain"
  sf_bad <- sf::st_as_sf(
    data.frame(site = c("A", "B"), x = c(10.0, 11.0), y = c(51.0, 52.0)),
    coords = c("x", "y"),
    crs = 4326
  )

  expect_error(
    distantia_spatial(df = df, sf = sf_bad),
    regexp = "no column in 'sf' contains all time series names"
  )

})

test_that("`momentum_spatial()` is an alias and produces the same output", {

  tsl <- distantia::tsl_initialize(
    x = distantia::covid_prevalence,
    name_column = "name", time_column = "time"
  )

  df_psi <- distantia::distantia_ls(tsl = tsl)

  sf_via_distantia <- distantia::distantia_spatial(
    df = df_psi, sf = distantia::covid_counties,
    network = TRUE
  )

  sf_via_momentum <- distantia::momentum_spatial(
    df = df_psi, sf = distantia::covid_counties,
    network = TRUE
  )

  expect_equal(sf_via_distantia, sf_via_momentum)

})

test_that("`distantia_model_frame()` works", {

  tsl <- tsl_initialize(
    x = covid_prevalence, name_column = "name",
    time_column = "time"
  )

  df <- distantia_ls(tsl = tsl)

  composite_predictors <- list(economy = c(
    "poverty_percentage",
    "median_income", "domestic_product"
  ))

  model_frame <- distantia_model_frame(
    response_df = df, predictors_df = covid_counties,
    composite_predictors = composite_predictors, scale = TRUE
  )

  expect_equal(class(model_frame), "data.frame")

  expect_true("economy" %in% colnames(model_frame))

  expect_true("geographic_distance" %in% colnames(model_frame))

  expect_equal(attributes(model_frame)$response, "psi")

  model <- lm(formula = attributes(model_frame)$formula, data = model_frame)

  expect_equal(class(model), "lm")

})

#' Dissimilarity Analysis of Time Series Lists
#'
#' @description
#'
#' This function combines *dynamic time warping* or *lock-step comparison* with the *psi dissimilarity score* and *permutation methods* to assess dissimilarity between pairs of time series or any other sort of data composed of events ordered across a relevant dimension.
#'
#' **Dynamic Time Warping** (DTW) finds the optimal alignment between two time series by minimizing the cumulative distance between their samples. It applies dynamic programming to identify the least-cost path through a distance matrix between all pairs of samples. The resulting sum of distances along the least cost path is a metric of time series similarity. DTW disregards the exact timing of samples and focuses on their order and pattern similarity between time series, making it suitable for comparing both *regular and irregular time series of the same or different lengths*, such as phenological data from different latitudes or elevations, time series from various years or periods, and movement trajectories like migration paths. Additionally, `distantia()` implements constrained DTW via Sakoe-Chiba bands with the `bandwidth` argument, which defines a region around the distance matrix diagonal to restrict the spread of the least cost path.
#'
#' **Lock-step** (LS) sums pairwise distances between samples in *regular or irregular time series of the same length*, preferably captured at the same times. This method is an alternative to dynamic time warping when the goal is to assess the synchronicity of two time series.
#'
#' The **psi score** normalizes the cumulative sum of distances between two time series by the cumulative sum of distances between their consecutive samples to generate a comparable dissimilarity score. If for two time series \eqn{x} and \eqn{y} \eqn{D_xy} represents the cumulative sum of distances between them, either resulting from dynamic time warping or the lock-step method, and \eqn{S_xy} represents the cumulative sum of distances of their consecutive samples, then the psi score can be computed in two ways depending on the scenario:
#'
#' **Equation 1:** \eqn{\psi = \frac{D_{xy} - S_{xy}}{S_{xy}}}
#'
#' **Equation 2:** \eqn{\psi = \frac{D_{xy} - S_{xy}}{S_{xy}} + 1}
#'
#' When $D_xy$ is computed via dynamic time warping **ignoring the distance matrix diagonals** (`diagonal = FALSE`), then *Equation 1* is used. On the other hand, if $D_xy$ results from the lock-step method (`lock_step = TRUE`), or from dynamic time warping considering diagonals (`diagonal = TRUE`), then *Equation 2* is used instead:
#'
#' In both equations, a psi score of zero indicates maximum similarity.
#'
#' **Permutation methods** are provided here to help assess the robustness of observed psi scores by direct comparison with a null distribution of psi scores resulting from randomized versions of the compared time series. The fraction of null scores smaller than the observed score is returned as a *p_value* in the function output and interpreted as "the probability of finding a lower psi score (higher similarity) by chance".
#'
#'  In essence, restricted permutation is useful to answer the question "how robust is the similarity between two time series?"
#'
#' Four different permutation methods are available:
#' \itemize{
#'   \item **"restricted"**: Separates the data into blocks of contiguous rows, and re-shuffles data points randomly within these blocks, independently by row and column. Applied when the data is structured in blocks that should be preserved during permutations (e.g., "seasons", "years", "decades", etc) and the columns represent independent variables.
#'   \item **"restricted_by_row"**: Separates the data into blocks of contiguous rows, and re-shuffles complete rows within these blocks. This method is suitable for cases where the data is organized into blocks as described above, but columns represent interdependent data (e.g., rows represent percentages or proportions), and maintaining the relationships between data within each row is important.
#'   \item **"free"**: Randomly re-shuffles data points across the entire time series, independently by row and column. This method is useful for loosely structured time series where data independence is assumed. When the data exhibits a strong temporal structure, this approach may lead to an overestimation of the robustness of dissimilarity scores.
#'   \item **"free_by_row"**: Randomly re-shuffles complete rows across the entire time series. This method is useful for loosely structured time series where dependency between columns is assumed (e.g., rows represent percentages or proportions). This method has the same drawbacks as the "free" method, when the data exhibits a strong temporal structure.
#' }
#'
#' This function allows computing dissimilarity between pairs of time series using different combinations of arguments at once. For example, when the argument `distance` is set to `c("euclidean", "manhattan")`, the output data frame will show two dissimilarity scores for each pair of time series, one based on euclidean distances, and another based on manhattan distances. The same happens for most other parameters.
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr). However, due to the high performance of the C++ backend, parallelization might only result in efficiency gains when running permutation tests with large number of iterations, or working with very long time series.
#'
#' @param tsl (required, time series list) list of zoo time series. Default: NULL
#' @param distance (optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset [distances]. Default: "euclidean".
#' @param diagonal (optional, logical vector). If TRUE, diagonals are included in the dynamic time warping computation. Default: TRUE
#' @param bandwidth (optional, numeric) Proportion of space at each side of the cost matrix diagonal (aka *Sakoe-Chiba band*) defining a valid region for dynamic time warping, used to control the flexibility of the warping path. This method prevents degenerate alignments due to differences in magnitude between time series when the data is not properly scaled. If `1` (default), DTW is unconstrained. If `0`, DTW is fully constrained and the warping path follows the matrix diagonal. Recommended values may vary depending on the nature of the data. Ignored if `lock_step = TRUE`. Default: 1.
#' @param lock_step (optional, logical vector) If TRUE, time series captured at the same times are compared sample wise (with no dynamic time warping). Requires time series in argument `tsl` to be fully aligned, or it will return an error. Default: FALSE.
#' @param permutation (optional, character vector) permutation method, only relevant when `repetitions` is higher than zero. Valid values are: "restricted_by_row", "restricted", "free_by_row", and "free". Default: "restricted_by_row".
#' @param block_size (optional, integer) Size of the row blocks for the restricted permutation test. Only relevant when permutation methods are "restricted" or "restricted_by_row" and `repetitions` is higher than zero. A block of size `n` indicates that a row can only be permuted within a block of `n` adjacent rows. If NULL, defaults to the rounded one tenth of the shortest time series in `tsl`. Default: NULL.
#' @param repetitions (optional, integer vector) number of permutations to compute the p-value. If 0, p-values are not computed. Otherwise, the minimum is 2. The resolution of the p-values and the overall computation time depends on the number of permutations. Default: 0
#' @param seed (optional, integer) initial random seed to use for replicability when computing p-values. Default: 1
#'
#' @return data frame with columns:
#' \itemize{
#'   \item `x`: time series name.
#'   \item `y`: time series name.
#'   \item `distance`: name of the distance metric.
#'   \item `diagonal`: value of the argument `diagonal`.
#'   \item `lock_step`: value of the argument `lock_step`.
#'   \item `repetitions` (only if `repetitions > 0`): value of the argument `repetitions`.
#'   \item `permutation` (only if `repetitions > 0`): name of the permutation method used to compute p-values.
#'   \item `seed` (only if `repetitions > 0`): random seed used in the permutations.
#'   \item `psi`: psi dissimilarity of the sequences `x` and `y`.
#'   \item `null_mean` (only if `repetitions > 0`): mean of the null distribution of psi scores.
#'   \item `null_sd` (only if `repetitions > 0`): standard deviation of the null distribution of psi values.
#'   \item `p_value`  (only if `repetitions > 0`): proportion of scores smaller or equal than `psi` in the null distribution.
#' }
#' @export
#' @autoglobal
#' @examples
#'
#' #parallelization setup
#' #not worth it for this data size
#' # future::plan(
#' #   strategy = future::multisession,
#' #   workers = 2
#' # )
#'
#' #progress bar (does not work in R examples)
#' # progressr::handlers(global = TRUE)
#'
#'
#' #load fagus_dynamics as tsl
#' #global centering and scaling
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl,
#'     guide_columns = 3
#'     )
#' }
#'
#' #dynamic time warping dissimilarity analysis
#' #-------------------------------------------
#' #permutation restricted by row to preserve dependency of ndvi on temperature and rainfall
#' #block size is 3 months to permute within same season
#' df_dtw <- distantia(
#'   tsl = tsl,
#'   distance = "euclidean",
#'   permutation = "restricted_by_row",
#'   block_size = 3, #months
#'   repetitions = 10, #increase to 100 or more
#'   seed = 1
#' )
#'
#' #focus on the important details
#' df_dtw[, c("x", "y", "psi", "p_value", "null_mean", "null_sd")]
#' #higher psi values indicate higher dissimilarity
#' #p-values indicate chance of finding a random permutation with a psi smaller than the observed
#'
#' #visualize dynamic time warping
#' if(interactive()){
#'
#'   distantia_dtw_plot(
#'     tsl = tsl[c("Spain", "Sweden")],
#'     distance = "euclidean"
#'   )
#'
#' }
#'
#' #recreating the null distribution
#' #direct call to C++ function
#' #use same args as distantia() call
#' psi_null <- psi_null_dtw_cpp(
#'   x = tsl[["Spain"]],
#'   y = tsl[["Sweden"]],
#'   distance = "euclidean",
#'   repetitions = 10, #increase to 100 or more
#'
#'   permutation = "restricted_by_row",
#'   block_size = 3,
#'   seed = 1
#' )
#'
#' #compare null mean with output of distantia()
#' mean(psi_null)
#' df_dtw$null_mean[3]
#'
#' @family distantia
#' @importFrom doFuture "%dofuture%"
distantia <- function(
    tsl = NULL,
    distance = "euclidean",
    diagonal = TRUE,
    bandwidth = 1,
    lock_step = FALSE,
    permutation = "restricted_by_row",
    block_size = NULL,
    repetitions = 0,
    seed = 1
){


  #check input arguments
  args <- utils_check_args_distantia(
    tsl = tsl,
    distance = distance,
    diagonal = diagonal,
    bandwidth = bandwidth,
    lock_step = lock_step,
    repetitions = repetitions,
    permutation = permutation,
    block_size = block_size,
    seed = seed
  )

  tsl <- args$tsl
  distance <- args$distance
  diagonal <- args$diagonal
  bandwidth <- args$bandwidth
  lock_step <- args$lock_step
  repetitions <- args$repetitions
  permutation <- args$permutation
  block_size <- args$block_size
  seed <- args$seed

  #lock-step check
  if(any(lock_step == TRUE)){

    #count rows in time series
    row_counts <- tsl |>
      tsl_nrow() |>
      unlist() |>
      unique()

    if(length(row_counts) > 1){

          message(
            "distantia::distantia(): argument 'lock_step' is TRUE, but time series in 'tsl' do not have the same number of rows. Setting 'lock_step' to FALSE."
          )

      lock_step <- FALSE

    }

  }


  #iterations data
  if(repetitions == 0){

    df <- utils_tsl_pairs(
      tsl = tsl,
      args_list = list(
        distance = distance,
        diagonal = diagonal,
        bandwidth = bandwidth,
        lock_step = lock_step
      )
    )

    df$psi <- NA

  } else {

    df <- utils_tsl_pairs(
      tsl = tsl,
      args_list = list(
        distance = distance,
        diagonal = diagonal,
        bandwidth = bandwidth,
        lock_step = lock_step,
        repetitions = repetitions,
        permutation = permutation,
        block_size = block_size,
        seed = seed
      )
    )

    df$psi <- NA
    df$p_value <- NA
    df$null_mean <- NA
    df$null_sd <- NA

  }

  iterations <- seq_len(nrow(df))

  p <- progressr::progressor(along = iterations)

  #iterate over pairs of time series
  df_distantia <- foreach::foreach(
    i = iterations,
    .combine = "rbind",
    .errorhandling = "pass"
  ) %dofuture% {

    # p()

    df.i <- df[i, ]

    x <- tsl[[df.i$x]]
    y <- tsl[[df.i$y]]

    if(df.i$lock_step == TRUE){

      df.i$psi <- psi_ls_cpp(
        x = x,
        y = y,
        distance = df.i$distance
      )

      if(repetitions > 0){

        psi_null <- psi_null_ls_cpp(
          x = x,
          y = y,
          distance = df.i$distance,
          repetitions = df.i$repetitions,
          permutation = df.i$permutation,
          block_size = df.i$block_size,
          seed = df.i$seed
        )

        df.i$p_value <- sum(psi_null <= df.i$psi) / repetitions
        df.i$null_mean <- mean(psi_null)
        df.i$null_sd <- stats::sd(psi_null)

      }

    } else {

      df.i$psi <- psi_dtw_cpp(
        x = x,
        y = y,
        distance = df.i$distance,
        diagonal = df.i$diagonal,
        weighted = TRUE,
        ignore_blocks = FALSE,
        bandwidth = df.i$bandwidth
      )

      if(repetitions > 0){

        psi_null <- psi_null_dtw_cpp(
          x = x,
          y = y,
          distance = df.i$distance,
          diagonal = df.i$diagonal,
          weighted = TRUE,
          ignore_blocks = FALSE,
          bandwidth = df.i$bandwidth,
          repetitions = df.i$repetitions,
          permutation = df.i$permutation,
          block_size = df.i$block_size,
          seed = df.i$seed
        )

        df.i$p_value <- sum(psi_null <= df.i$psi) / repetitions
        df.i$null_mean <- mean(psi_null)
        df.i$null_sd <- stats::sd(psi_null)

      }

    }

    return(df.i)

  } |>
    suppressWarnings()

  df_distantia <- df_distantia[order(df_distantia$psi), ]

  #remove dtw arguments if only lock-step was used
  if(
    "lock_step" %in% colnames(df_distantia) &&
    sum(df_distantia[["lock_step"]]) == nrow(df_distantia)
    ){
    df_distantia$diagonal <- NULL
    df_distantia$bandwidth <- NULL
  }

  #add type
  attr(
    x = df_distantia,
    which = "type"
  ) <- "distantia_df"

  df_distantia

}

test_that("`distantia_cluster_kmeans()` works", {

  tsl <- tsl_initialize(
    x = covid_prevalence,
    name_column = "name",
    time_column = "time"
  )

  tsl <- tsl_subset(tsl = tsl, names = 1:10)

  distantia_df <- distantia(tsl = tsl, lock_step = TRUE)

  distantia_kmeans <- distantia_cluster_kmeans(
    df = distantia_df,
    clusters = NULL
  )

  expect_equal(names(distantia_kmeans), c("cluster_object", "clusters", "silhouette_width", "df", "d", "optimization"))

  expect_true(class(distantia_kmeans$cluster_object) == "kmeans")

  expect_true("matrix" %in% class(distantia_kmeans$d))

  expect_true(class(distantia_kmeans$df) == "data.frame")

  expect_true(is.numeric(distantia_kmeans$silhouette_width))

})

test_that("`distantia_cluster_hclust()` works", {

   tsl <- tsl_initialize(
    x = covid_prevalence,
    name_column = "name",
    time_column = "time"
  )

  tsl <- tsl_subset(tsl = tsl, names = 1:10)

  distantia_df <- distantia(
    tsl = tsl,
    lock_step = TRUE
    )

  distantia_clust <- distantia_cluster_hclust(
    df = distantia_df,
    clusters = NULL,
    method = NULL
  )

  expect_equal(names(distantia_clust), c("cluster_object", "clusters", "silhouette_width", "df", "d", "optimization"))

  expect_true(class(distantia_clust$cluster_object) == "hclust")

  expect_true(class(distantia_clust$d) == "dist")

  expect_true(class(distantia_clust$df) == "data.frame")

  expect_true(is.numeric(distantia_clust$silhouette_width))

})

test_that("`utils_distantia_df_split()` works", {

  tsl <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  )

  df <- distantia(
    tsl = tsl, distance = c("euclidean", "manhattan"),
    lock_step = c(TRUE, FALSE)
  )

  df_split <- utils_distantia_df_split(df = df)

  expect_equal(length(df_split), 4)

})

#' Transform Time Series List to Data Frame
#'
#' @param tsl (required, list) Time series list. Default: NULL
#'
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#'
#' tsl <- tsl_simulate(
#'   n = 3,
#'   rows = 10,
#'   time_range = c(
#'     "2010-01-01",
#'     "2020-01-01"
#'   ),
#'   irregular = FALSE
#' )
#'
#' df <- tsl_to_df(
#'   tsl = tsl
#' )
#'
#' names(df)
#' nrow(df)
#' head(df)
#' @family tsl_management
tsl_to_df <- function(
    tsl = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  df_list <- lapply(
    X = tsl,
    FUN = function(x){

      data.frame(
        name = attributes(x)$name,
        time = zoo::index(x),
        as.data.frame(x)
      )

    }

  )

  df <- do.call(
    what = "rbind",
    args = df_list
  )

  rownames(df) <- NULL

  df

}

test_that("`tsl_transform()` works", {

  tsl <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  ) |>
    tsl_subset(
      time = c(
        "2010-01-01",
        "2011-01-01"
        )
    )

  # f_binary ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_binary
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #output is binary
  expect_equal(
    lapply(tsl_test, range) |>
      unlist() |>
      unique(),
    c(0, 1)
  )

  #f_rescale_global ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_rescale_global
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #check values
  tsl_test_df <- tsl_to_df(tsl = tsl_test)
  expect_equal(
    min(tsl_test_df$temperature),
    0
  )
  expect_equal(
    max(tsl_test_df$temperature),
    1
  )
  expect_equal(
    min(tsl_test_df$evi),
    0
  )
  expect_equal(
    max(tsl_test_df$evi),
    1
  )
  expect_equal(
    min(tsl_test_df$rainfall),
    0
  )
  expect_equal(
    max(tsl_test_df$rainfall),
    1
  )

  #f_rescale_local ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_rescale_local
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #check values
  tsl_test_stats <- tsl_stats(tsl = tsl_test)
  expect_equal(
    unique(tsl_test_stats$min),
    0
  )
  expect_equal(
    unique(tsl_test_stats$max),
    1
  )

  #f_scale_global ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_scale_global
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  tsl_test_df <- tsl_to_df(tsl = tsl_test)
  expect_equal(
    mean(tsl_test_df$temperature),
    0
  )
  expect_equal(
    sd(tsl_test_df$temperature),
    1
  )
  expect_equal(
    mean(tsl_test_df$evi),
    0
  )
  expect_equal(
    sd(tsl_test_df$evi),
    1
  )
  expect_equal(
    mean(tsl_test_df$rainfall),
    0
  )
  expect_equal(
    sd(tsl_test_df$rainfall),
    1
  )


  #f_scale_local ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_scale_local
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #check values
  tsl_test_stats <- tsl_stats(tsl = tsl_test)
  expect_true(
    all.equal.numeric(
      target = rep(0, nrow(tsl_test_stats)),
      current = tsl_test_stats$mean
    )
  )
  expect_true(
    all.equal.numeric(
      target = rep(1, nrow(tsl_test_stats)),
      current = tsl_test_stats$sd
    )
  )

  #f_detrend_difference ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_detrend_difference
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #f_detrend_linear ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_detrend_linear
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #f_detrend_poly ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_detrend_poly
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #f_trend_linear ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_trend_linear
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #f_trend_poly ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_trend_poly
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  # Warning message:
  #   distantia::utils_prepare_time():  duplicated time indices in 'Krumbach_I':
  #   - value 6.8 replaced with 6.825
    tsl <- tsl_initialize(
      x = eemian_pollen,
      name_column = "name",
      time_column = "time"
    ) |>
      tsl_subset(
        names = 1:3
      ) |>
      suppressMessages()


  #f_clr ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_clr
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #f_hellinger ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_hellinger
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #f_log ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_log
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #f_percent ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_percent
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #f_proportion ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_proportion
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

  #f_proportion_sqrt ----
  tsl_test <- tsl_transform(
    tsl = tsl,
    f = f_proportion_sqrt
  )

  #output has same dimensions
  expect_equal(
    lapply(tsl_test, dim) |>
      unlist() |>
      unique(),
    lapply(tsl, dim) |>
      unlist() |>
      unique()
  )

  #output has no NA
  expect_equal(
    tsl_count_NA(tsl = tsl_test) |>
      unlist() |>
      unique(),
    0
  )

})

test_that("`tsl_to_df()` works", {
  tsl <- tsl_simulate(n = 3, rows = 10, time_range = c(
    "2010-01-01",
    "2020-01-01"
  ), irregular = FALSE,
  seed = 1)
  df <- tsl_to_df(tsl = tsl)
  expect_equal(class(df), "data.frame")
  expect_true(all(unique(unlist(tsl_colnames_get(tsl))) %in% colnames(df)))
  expect_equal(nrow(df), sum(unlist(tsl_nrow(tsl = tsl))))

})

#' Join Time Series Lists
#'
#' @description
#' Joins an arbitrary number of time series lists by name and time. Pairs of zoo objects are joined with [zoo::merge.zoo()]. Names that are not shared across all input TSLs are ignored, and observations with no matching time are filled with NA and then managed via [tsl_handle_NA()] depending on the value of the argument `na_action`.
#'
#' @param ... (required, time series lists) names of the time series lists to merge.
#' @inheritParams tsl_handle_NA
#'
#' @return time series list
#' @export
#' @autoglobal
#' @examples
#'
#' #generate two time series list to join
#' tsl_a <- tsl_simulate(
#'   n = 2,
#'   cols = 2,
#'   irregular = TRUE,
#'   seed = 1
#' )
#'
#' #needs renaming
#' tsl_b <- tsl_simulate(
#'   n = 3,
#'   cols = 2,
#'   irregular = TRUE,
#'   seed = 2
#' ) |>
#'   tsl_colnames_set(
#'     names = c("c", "d")
#'   )
#'
#' #join
#' tsl <- tsl_join(
#'   tsl_a,
#'   tsl_b
#' )
#'
#' #plot result
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl
#'   )
#' }
#'
#' @family tsl_management
tsl_join <- function(
    ...,
    na_action = "impute"
    ){

  #gather lists
  tsls <- list(...)

  if (length(tsls) <= 1) {
    stop(
      "distantia::tsl_join(): more than TSL is required.",
      call. = FALSE
      )
  }

  #get shared list names
  shared_names <- Reduce(
    f = intersect,
    x = lapply(
      X = tsls,
      FUN = tsl_names_get
      )
    )

  if(length(shared_names) == 0){
    stop(
      "distantia::tsl_join(): input time series lists must have shared names.",
      call. = FALSE
    )
  }

  #subset common elements in all lists
  tsls <- future.apply::future_lapply(
    X = tsls,
    FUN = function(x){

      tsl_subset(
        tsl = x,
        names = shared_names
      )

    })

  tsl <- future.apply::future_lapply(
    X = shared_names,
    FUN = function(name) {

      do.call(
        what = zoo::merge.zoo,
        args = c(
          lapply(
            X = tsls,
            FUN = `[[`,
            name
          ),
          list(
            all = TRUE,
            fill = NA
          )
        )
      )
    }
  )

  tsl <- tsl_names_set(
    tsl = tsl,
    names = shared_names
  )

  #handle NA cases
  tsl_na_count <- tsl_count_NA(
    tsl = tsl
  ) |>
    unlist() |>
    sum()

  if(tsl_na_count > 0){
    message(
      "distantia::tsl_join(): ", tsl_na_count, " NA cases generated by the join due to non-matching times were handled via distantia::tsl_handle_NA() with 'na_action = '",
      na_action, "''."
    )
  }

  tsl <- tsl_handle_NA(
    tsl = tsl,
    na_action = na_action
  )

  tsl

}

#' New Time for Time Series Aggregation
#'
#' @description
#' Internal function called by [tsl_aggregate()] and [tsl_resample()] to help transform the input argument `new_time` into the proper format for time series aggregation or resampling.
#'
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param new_time (required, zoo object, numeric, numeric vector, Date vector, POSIXct vector, or keyword) breakpoints defining aggregation groups. Options are:
#' \itemize{
#'   \item numeric vector: only for the "numeric" time class, defines the breakpoints for time series aggregation.
#'   \item "Date" or "POSIXct" vector: as above, but for the time classes "Date" and "POSIXct." In any case, the input vector is coerced to the time class of the `tsl` argument.
#'   \item numeric: defines fixed with time intervals for time series aggregation. Used as is when the time class is "numeric", and coerced to integer and interpreted as days for the time classes "Date" and "POSIXct".
#'   \item keyword (see [utils_time_units()] and [tsl_time_summary()]): the common options for the time classes "Date" and "POSIXct" are: "millennia", "centuries", "decades", "years", "quarters", "months", and "weeks". Exclusive keywords for the "POSIXct" time class are: "days", "hours", "minutes", and "seconds". The time class "numeric" accepts keywords coded as scientific numbers, from "1e8" to "1e-8".
#' }
#' @param keywords (optional, character string or vector) Defines what keywords are returned. If "aggregate", returns valid keywords for [zoo_aggregate()]. If "resample", returns valid keywords for [zoo_resample()]. Default: "aggregate".
#'
#' @return Vector of class numeric, Date, or POSIXct
#' @export
#' @autoglobal
#' @examples
#' #three time series
#' #climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' # new time for aggregation using keywords
#' #-----------------------------------
#'
#' #get valid keywords for aggregation
#' tsl_time_summary(
#'   tsl = tsl,
#'   keywords = "aggregate"
#' )$keywords
#'
#' #if no keyword is used, for aggregation the highest resolution keyword is selected automatically
#' new_time <- utils_new_time(
#'   tsl = tsl,
#'   new_time = NULL,
#'   keywords = "aggregate"
#' )
#'
#' new_time
#'
#' #if no keyword is used
#' #for resampling a regular version
#' #of the original time based on the
#' #average resolution is used instead
#' new_time <- utils_new_time(
#'   tsl = tsl,
#'   new_time = NULL,
#'   keywords = "resample"
#' )
#'
#' new_time
#'
#' #aggregation time vector form keyword "years"
#' new_time <- utils_new_time(
#'   tsl = tsl,
#'   new_time = "years",
#'   keywords = "aggregate"
#' )
#'
#' new_time
#'
#' #same from shortened keyword
#' #see utils_time_keywords_dictionary()
#' utils_new_time(
#'   tsl = tsl,
#'   new_time = "year",
#'   keywords = "aggregate"
#' )
#'
#' #same for abbreviated keyword
#' utils_new_time(
#'   tsl = tsl,
#'   new_time = "y",
#'   keywords = "aggregate"
#' )
#'
#' #from a integer defining a time interval in days
#' utils_new_time(
#'   tsl = tsl,
#'   new_time = 365,
#'   keywords = "aggregate"
#' )
#'
#' #using this vector as input for aggregation
#' tsl_aggregated <- tsl_aggregate(
#'   tsl = tsl,
#'   new_time = new_time
#' )
#' @family internal_time_handling
utils_new_time <- function(
    tsl = NULL,
    new_time = NULL,
    keywords = "aggregate"
){

  keywords <- match.arg(
    arg = keywords,
    choices = c(
      "aggregate",
      "resample"
    ),
    several.ok = FALSE
  )

  old_time <- tsl_time_summary(
    tsl = tsl,
    keywords = keywords
  )

  #default values
  if(is.null(new_time)){

    if(keywords == "aggregate"){

      new_time <- utils::tail(
        x = unlist(old_time$keywords),
        n = 1
      )

      message("distantia::utils_new_time(): aggregating 'tsl' with keyword '", new_time, "'.")

    } else {

      old_time_df <- tsl_time(
        tsl = tsl,
        keywords = "resample"
      )

      new_time <- seq(
        from = max(old_time_df$begin),
        to = min(old_time_df$end),
        length.out = floor(mean(old_time_df$rows))
      )

      message("distantia::utils_new_time(): resampling 'tsl' to its average resolution.")

    }

  }

  # new_time is zoo ----
  if(zoo::is.zoo(new_time)){
    new_time <- zoo::index(new_time)
  }

  # new_time is tsl ----
  if(
    is.list(new_time) &&
    zoo::is.zoo(new_time[[1]])
  ){

    new_time <- tsl_time(
      tsl = new_time
    )

    new_time <- seq(
      from = max(new_time$begin),
      to = min(new_time$end),
      by = mean(new_time$resolution)
    )

  }

  # new_time type ----
  new_time <- utils_new_time_type(
    tsl = tsl,
    new_time = new_time
  )

  new_time_type <- attributes(new_time)$new_time_type


  # generate new_time ----

  ## standard_keyword ----
  if(new_time_type == "standard_keyword"){

    new_time <- seq(
      from = lubridate::floor_date(
        x = old_time$begin,
        unit = new_time
        ),
      to = lubridate::ceiling_date(
        x = old_time$end,
        unit = new_time
        ),
      by = new_time
    )

  } #end of standard keyword


  ## non_standard_keyword ----
  if(new_time_type == "non_standard_keyword"){

    # time units
    time_units <- old_time$units_df[
      old_time$units_df$units == new_time,
    ]

    #non-standard are always in "years"
    unit <- paste0(time_units$factor, " years")

    new_time <- seq(
      from = lubridate::floor_date(
        x = old_time$begin,
        unit = unit
      ),
      to = lubridate::ceiling_date(
        x = old_time$end,
        unit = unit
      ),
      by = unit
    )

  }


  ## numeric_interval ----
  if(new_time_type == "numeric_interval"){

    if(old_time$class == "numeric"){

      new_time <- seq(
        from = old_time$begin - new_time,
        to = old_time$end + new_time,
        by = new_time
      )

    } else {

      #non-standard are always in "days"
      new_time <- as.integer(new_time)
      unit <- paste0(new_time, " days")

      new_time <- seq.Date(
        from = old_time$begin,
        to = old_time$end,
        by = unit
      )

    }

  } #end of numeric_interval


  ## numeric_vector ----
  if(new_time_type == "numeric_vector"){

    begin <- min(new_time) - min(diff(new_time))

    if(min(new_time) > old_time$begin){
      new_time <- c(
        min(new_time) - min(diff(new_time)),
        new_time
      )
    }

    if(max(new_time) < old_time$end){
      new_time <- c(
        new_time,
        max(new_time) + min(diff(new_time))
      )
    }

  }

  #coerce time class
  new_time <- utils_coerce_time_class(
    x = new_time,
    to = ifelse(
      test = "POSIXct" %in% old_time$class,
      yes = "POSIXct",
      no = old_time$class
    )
  )

  new_time

}


#' @export
#' @rdname utils_new_time
#' @autoglobal
utils_new_time_type <- function(
    tsl = NULL,
    new_time = NULL,
    keywords = c(
      "resample",
      "aggregate"
    )
){

  time_summary <- tsl_time_summary(
    tsl = tsl,
    keywords = keywords
  )

  time_summary_min_threshold <- min(time_summary$units_df$threshold)

  # ERROR: too many time classes ----
  if(length(time_summary$class) > 1){

    stop(
      "distantia::utils_new_time_type(): the time class of all zoo objects in 'tsl' must be the same, but they are: '",
      paste(time_summary$class, collapse = "', '"),
      "'.",
      call. = FALSE
    )
  }

  # intervals ----
  if(length(new_time) == 1){

    # numeric interval ----
    if(
      is.numeric(new_time) &&
      new_time >= time_summary_min_threshold
    ){

      attr(
        x = new_time,
        which = "new_time_type"
      ) <- "numeric_interval"

      return(new_time)

    }


    ## keyword ----
    new_time <- utils_time_keywords_translate(
      keyword = new_time
    )

    ### keyword for numerics ----
    if(new_time %in% time_summary$keywords){

      if(time_summary$class %in% c("numeric", "integer")){

        new_time <- as.numeric(new_time)

        attr(
          x = new_time,
          which = "new_time_type"
        ) <- "numeric_interval"

        return(new_time)

      }

      ### keyword for Date or POSIXct ----
      standard_keyword <- time_summary$units_df[
        time_summary$units_df$units == new_time,
        "keyword"
      ]

      #### standard keyword ----
      if(standard_keyword == TRUE){

        attr(
          x = new_time,
          which = "new_time_type"
        ) <- "standard_keyword"

        return(new_time)

      }

      #### non-standard keyword ----
      attr(
        x = new_time,
        which = "new_time_type"
      ) <- "non_standard_keyword"

      return(new_time)

    }

    stop(
      "distantia::utils_new_time_type(): argument 'new_time' of length 1 must be:\n",
      "  - one of these keywords: '",
      paste0(time_summary$keywords, collapse = "', '"),
      ".\n",
      "  - a number higher than ",
      time_summary_min_threshold,
      ".",
      call. = FALSE
    )

  }

  # break points ----

  ## numeric_vector ----
  if(
    is.numeric(new_time) &&
    time_summary$class %in% c("numeric", "integer")
  ){

    attr(
      x = new_time,
      which = "new_time_type"
    ) <- "numeric_vector"

    return(new_time)

  }

  ## Date or POSIXct vector ----
  new_time <- utils_as_time(
    x = new_time,
    to_class = time_summary$class
  )

  if(
    "POSIXct" %in% class(new_time)
  ){

    attr(
      x = new_time,
      which = "new_time_type"
    ) <- "POSIXct_vector"

    return(new_time)

  }


  if(
    "Date" %in% class(new_time)
  ){

    attr(
      x = new_time,
      which = "new_time_type"
    ) <- "Date_vector"

    return(new_time)

  }

  stop(
    "distantia::utils_new_time_type(): argument 'new_time' of length higher than one must be a vector of class ",
    time_summary$class,
    " with values between ",
    time_summary$begin, " and ",
    time_summary$end,
    ".",
    call. = FALSE
  )

}

test_that("`utils_new_time()` works", {

  tsl <- tsl_initialize(
    x = fagus_dynamics, name_column = "name",
    time_column = "time"
  )

  expect_equal(tsl_time_summary(tsl = tsl, keywords = "aggregate")$keywords, c("decades", "years", "quarters"))

  expect_message(
    new_time <- utils_new_time(tsl = tsl, new_time = NULL, keywords = "aggregate")
  )

  expect_equal(
    unique(
      as.numeric(
        diff(new_time)
        )
      ),
    c(90, 91, 92)
    )

})

#' Test if Vector is a Valid Time Class
#'
#' @param x (required, vector) Vector to test. If the class of the vector elements is 'numeric', 'POSIXct', or 'Date', the function returns TRUE. Default: NULL.
#'
#' @return logical
#' @export
#' @autoglobal
#' @examples
#'
#' utils_is_time(
#'   x = c("2024-01-01", "2024-02-01")
#' )
#'
#' utils_is_time(
#'   x = utils_as_time(
#'     x = c("2024-01-01", "2024-02-01")
#'     )
#' )
#' @family internal_time_handling
utils_is_time <- function(
    x = NULL
    ){

  x_class <- class(x)
  if("POSIXct" %in% x_class){
    x_class <- "POSIXct"
  }

  #x is valid time class
  if(x_class %in% c(
    "integer",
    "numeric",
    "Date",
    "POSIXct"
  )){
    return(TRUE)
  }

  FALSE

}

test_that("`utils_is_time()` works", {

  expect_equal(utils_is_time(x = c("2024-01-01", "2024-02-01")), FALSE)

  expect_equal(utils_is_time(x = utils_as_time(x = c("2024-01-01", "2024-02-01"))), TRUE)

})

#' Default Discrete Color Palettes
#'
#' @description
#' Uses the function [grDevices::palette.colors()] to generate discrete color palettes using the following rules:
#' \itemize{
#'   \item `n <= 9`: "Okabe-Ito".
#'   \item `n == 10`: "Tableau 10"
#'   \item `n > 10 && n <= 12`: "Paired"
#'   \item `n > 12 && n <= 26`: "Alphabet"
#'   \item `n > 26 && n <= 36`: "Polychrome 36"
#' }
#'
#'
#' @inheritParams color_continuous
#'
#' @return color vector
#' @export
#' @autoglobal
#' @examples
#'
#' color_discrete(n = 9)
#' @family internal_plotting
color_discrete <- function(
    n = NULL,
    rev = FALSE
    ){

  if(n == 1){
    return("#C8443F")
  }
  if(n <= 9 && n > 1){
    pal <- "Okabe-Ito"
  }
  if(n == 10){
    pal <- "Tableau 10"
  }
  if(n > 10 && n <= 12){
    pal <- "Paired"
  }
  if(n > 12 && n <= 26){
    pal <- "Alphabet"
  }
  if(n > 26 && n <= 36){
    pal <- "Polychrome 36"
  }
  if(n > 36){
    return(
      color_continuous(
        n = n,
        rev = rev
        )
    )
  }

  out <- grDevices::palette.colors(
    n = n,
    palette = pal
  )

  if(rev){
    out <- rev(out)
  }

  out

}

#' Coerces Vector to a Given Time Class
#'
#' @param x (required, vector of class Date or POSIXct) time vector to convert. Default: NULL
#' @param to (required, class name) class to coerce `x` to. Either "Date", "POSIXct", "integer" or "numeric". Default: "POSIXct"
#'
#' @return time vector
#' @export
#' @autoglobal
#' @examples
#' x <- utils_coerce_time_class(
#'   x = c("2024-01-01", "2024-02-01"),
#'   to = "Date"
#' )
#'
#' x
#' class(x)
#'
#' x <- utils_coerce_time_class(
#'   x = c("2024-01-01", "2024-02-01"),
#'   to = "POSIXct"
#' )
#'
#' x
#' class(x)
#'
#' x <- utils_coerce_time_class(
#'   x = c("2024-01-01", "2024-02-01"),
#'   to = "numeric"
#' )
#'
#' x
#' class(x)
#' @family internal_time_handling
utils_coerce_time_class <- function(
    x = NULL,
    to = "POSIXct"
){

  classes <- c(
    "POSIXct",
    "Date",
    "numeric",
    "integer"
    )

  to <- match.arg(
    arg = to,
    choices = classes,
    several.ok = FALSE
  )

  if(is.character(x)){
    x <- x |>
      utils_as_time() |>
      suppressWarnings()
  }

  if(any(class(x) %in% classes) == FALSE){
    warning(
      "Class of argument 'x' must be one of '",
      paste(classes, collapse = "', '"),
    "'."
    )
    return(x)
  }

  if(to == "POSIXct"){

    if("Date" %in% class(x)){

      x <- as.POSIXct(x)

    }

  }

  if(to == "Date"){

    if("POSIXct" %in% class(x)){

      x <- format(
        x = x,
        format = "%Y-%m-%d"
      )

      x <- as.Date(x)

    }

  }

  if(to == "numeric"){

    x <- as.numeric(x)

  }

  if(to == "integer"){

    x <- as.integer(x)

  }

  x

}

test_that("`utils_coerce_time_class()` works", {
  x <- utils_coerce_time_class(
    x = c("2024-01-01", "2024-02-01"),
    to = "Date"
  )
  expect_equal(class(x), "Date")

  x <- utils_coerce_time_class(
    x = c("2024-01-01", "2024-02-01"),
    to = "POSIXct"
  )
  expect_true("POSIXct" %in% class(x))


  x <- utils_coerce_time_class(
    x = c("2024-01-01", "2024-02-01"),
    to = "numeric"
  )
  expect_equal(class(x), "numeric")

})

#' Dynamic Time Warping Variable Importance Analysis of Multivariate Time Series Lists
#'
#' @description
#' A minimalistic but slightly faster version of [momentum()] that computes dynamic time warping importance analysis with the "robust" setup in multivariate time series lists.
#'
#' @inheritParams momentum
#'
#' @return data frame:
#' \itemize{
#'   \item `x`: name of the time series `x`.
#'   \item `y`: name of the time series `y`.
#'   \item `psi`: psi score of `x` and `y`.
#'   \item `variable`: name of the individual variable.
#'   \item `importance`: importance score of the variable.
#'   \item `effect`: interpretation of the "importance" column, with the values "increases similarity" and "decreases similarity".
#' }
#' @export
#' @autoglobal
#' @examples
#'
#' tsl <- tsl_initialize(
#'   x = distantia::albatross,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' df <- momentum_dtw(
#'   tsl = tsl,
#'   distance = "euclidean"
#'   )
#'
#' #focus on important columns
#' df[, c(
#'   "x",
#'   "y",
#'   "variable",
#'   "importance",
#'   "effect"
#'   )]
#'
#' @family momentum
momentum_dtw <- function(
    tsl = NULL,
    distance = "euclidean"
){

  #check input arguments
  args <- utils_check_args_momentum(
    tsl = tsl,
    distance = distance
  )

  tsl <- args$tsl
  distance <- args$distance

  if(length(distance) > 1){

    distance <- distance[1]

    message(
      "distantia::momentum_dtw(): Using first value of the 'distance' argument."
    )

  }

  #stop if tsl is univariate
  tsl_ncol <- tsl_ncol(tsl = tsl) |>
    unlist() |>
    unique()

  if(1 %in% tsl_ncol){
    stop(
      "distantia::momentum_dtw(): variable contribution analysis requires multivariate time series, but 'tsl' contains univariate time series.",
      call. = FALSE
    )
  }

  #tsl pairs
  df <- utils_tsl_pairs(
    tsl = tsl,
    args_list = list(
      distance = distance
    )
  )

  iterations <- seq_len(nrow(df))

  #iterate over pairs of time series
  df <- foreach::foreach(
    i = iterations,
    .combine = "rbind",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

    df.i <- df[i, ]

    importance.i <- importance_dtw_cpp(
      x = tsl[[df.i[["x"]]]],
      y = tsl[[df.i[["y"]]]],
      distance = df.i[["distance"]],
      diagonal = TRUE,
      bandwidth = 1
    )

    #set NaN to zero for constant pairs of sequences
    importance.i[is.na(importance.i)] <- 0

    importance.i <- merge(
      x = df.i,
      y = importance.i
    )

    return(importance.i)

  } #end of loop

  #interpretation
  df$effect <- ifelse(
    test = df$importance > 0,
    yes = "decreases similarity",
    no = "increases similarity"
  )

  df <- df[
    , c(
      "x",
      "y",
      "psi",
      "variable",
      "importance",
      "effect"
    )
  ]

  attr(
    x = df,
    which = "type"
  ) <- "momentum_df"

  df

}

#' Momentum Boxplot
#'
#' @description
#' Creates a boxplot of a data frame returned by [momentum()], summarizing the contribution to similarity (negative) and/or dissimilarity (positive) of individual variables across all time series.
#'
#' @inheritParams momentum_aggregate
#' @inheritParams distantia_boxplot
#'
#' @return boxplot
#' @export
#' @autoglobal
#' @examples
#' tsl <- tsl_initialize(
#'   x = distantia::albatross,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' df <- momentum(
#'   tsl = tsl,
#'   lock_step = TRUE
#'   )
#'
#' momentum_boxplot(
#'   df = df
#'   )
#' @family momentum_support
momentum_boxplot <- function(
    df = NULL,
    fill_color = NULL,
    f = median,
    text_cex = 1
){

  old_par <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old_par))

  if(attributes(df)$type != "momentum_df"){
    stop(
      "distantia::momentum_boxplot(): argument 'df' must be the output of distantia::momentum()",
      call. = FALSE
    )
  }

  #aggregate df if needed
  df <- momentum_aggregate(
    df = df,
    f = f
  )

  #boxplot
  utils_boxplot_common(
    variable = df$variable,
    value = df$importance,
    fill_color = fill_color,
    f = f,
    main = "Contribution to \n Similarity (<0) / Dissimilarity (>0)",
    xlab = "Momentum (psi %)",
    ylab = "",
    text_cex = text_cex
  )

  #zero line
  graphics::abline(
    v = 0,
    col = "gray50",
    lwd = 2
  )

}

#' Summary Statistics of Time Series Lists
#'
#' @description
#' This function computes a variety of summary statistics for each time series and numeric column within a time series list. The statistics include common metrics  such as minimum, maximum, quartiles, mean, standard deviation, range, interquartile range, skewness, kurtosis, and autocorrelation for specified lags.
#'
#' For irregular time series, autocorrelation computation is performed after regularizing the time series via interpolation with [zoo_resample()]. This regularization does not affect the computation of all other stats.
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param lags (optional, integer) An integer specifying the number of autocorrelation lags to compute. If NULL, autocorrelation computation is disabled. Default: 1.
#'
#' @return data frame:
#' \itemize{
#' \item name: name of the zoo object.
#' \item rows: rows of the zoo object.
#' \item columns: columns of the zoo object.
#' \item time_units: time units of the zoo time series (see [zoo_time()]).
#' \item time_begin: beginning time of the time series.
#' \item time_end: end time of the time series.
#' \item time_length: total length of the time series, expressed in time units.
#' \item time_resolution: average distance between consecutive observations
#' \item variable: name of the variable, a column of the zoo object.
#' \item min: minimum value of the zoo column.
#' \item q1: first quartile (25th percentile).
#' \item median: 50th percentile.
#' \item q3: third quartile (75th percentile).
#' \item max: maximum value.
#' \item mean: average value.
#' \item sd: standard deviation.
#' \item range: range of the variable, computed as max - min.
#' \item iq_range: interquartile range of the variable, computed as q3 - q1.
#' \item skewness: asymmetry of the variable distribution.
#' \item kurtosis: "tailedness" of the variable distribution.
#' \item ac_lag_1, ac_lag_2, ...: autocorrelation values for the specified lags.
#' }
#'
#' @examples
#'
#'
#' #three time series
#' #climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#'
#' #stats computation
#' df <- tsl_stats(
#'   tsl = tsl,
#'   lags = 3
#'   )
#'
#' df
#' @export
#' @autoglobal
#' 
#' @family tsl_processing
tsl_stats <- function(
    tsl = NULL,
    lags = 1L
){

  #subset numeric columns
  tsl <- tsl_subset(
    tsl = tsl,
    numeric_cols = TRUE,
    shared_cols = FALSE
  )

  #manage lags
  if(!is.null(lags)){
    lags <- max(1L, as.integer(lags))
  }

  #progress bar and iterator
  p <- progressr::progressor(along = tsl)



  #computing stats
  df <- foreach::foreach(
    i = seq_len(length(tsl)),
    .combine = "rbind",
    .errorhandling = "pass"
  ) %dofuture% {

    p()

    #general details of the zoo object i
    zoo.i <- tsl[[i]]

    zoo.i.data <- zoo::coredata(zoo.i)
    zoo.i.index <- zoo::index(zoo.i)
    zoo.i.name <- zoo_name_get(x = zoo.i)

    zoo.i.details <- data.frame(
      name = zoo.i.name,
      variable = colnames(zoo.i.data),
      NA_count = apply(
        X = zoo.i.data,
        MARGIN = 2,
        FUN = function(x) sum(is.na(x))
        )
    )

    zoo.i.details.names <- names(zoo.i.details)

    #column stats
    zoo.i.stats <- apply(
      X = zoo.i,
      FUN = function(x){

        x <- zoo::zoo(
          x = x,
          order.by = zoo.i.index
        )

        x <- zoo_vector_to_matrix(
          x = x,
          name = zoo.i.name
        )

        n <- nrow(x)

        #column stats
        x.stats <- data.frame(
          min = min(x, na.rm = TRUE),
          q1 = stats::quantile(x, probs = 0.25, na.rm = TRUE),
          median = stats::median(x = x, na.rm = TRUE),
          q3 = stats::quantile(x, probs = 0.75, na.rm = TRUE),
          max = max(x, na.rm = TRUE),
          mean = mean(x, na.rm = TRUE),
          sd = stats::sd(x, na.rm = TRUE)
        )

        #ranges
        x.stats$range <- x.stats$max - x.stats$min
        x.stats$iq_range <- x.stats$q3 - x.stats$q1

        #skewness
        x.stats$skewness <- sum((x - x.stats$mean)^3, na.rm = TRUE) / (n * x.stats$sd^3)

        #kurtosis
        x.stats$kurtosis <- sum((x - x.stats$mean)^4, na.rm = TRUE) / (n * x.stats$sd^4) - 3 * (n - 1)^2 / ((n - 2) * (n - 3))

        #autocorrelation
        if(!is.null(lags)){

          #fill NA
          if(sum(is.na(x)) > 0){

            x <- zoo::na.spline(
              object = x
            )

          }

          #if irregular, resample
          x.regular <- zoo::is.regular(
            x = x,
            strict = TRUE
          )

          if(x.regular == FALSE){
            x <- zoo_resample(
              x = x
            )
          }

          x.ac <- stats::acf(
            x = as.numeric(x),
            lag.max = lags,
            na.action = na.omit,
            plot = FALSE
          )$acf |>
            as.vector() |>
            utils::tail(n = lags) |>
            t() |>
            as.data.frame()

          names(x.ac) <- paste0(
            "ac_lag_",
            seq_len(lags)
          )

          x.stats <- cbind(
            x.stats,
            x.ac
          )

        }

        return(x.stats)

      },
      MARGIN = 2
    )

    zoo.i.stats <- do.call(
      what = "rbind",
      args = zoo.i.stats
    )

    zoo.i.stats.names <- names(zoo.i.stats)

    zoo.i.stats$variable <- row.names(zoo.i.stats)

    rownames(zoo.i.stats) <- NULL

    #merging general details and column stats
    zoo.i.df <- merge(
      zoo.i.details,
      zoo.i.stats
    )

    #reorder columns after merge
    zoo.i.df <- zoo.i.df[,
      c(
        zoo.i.details.names,
        zoo.i.stats.names
        )
      ]

    return(zoo.i.df)

  }

  #sort by variable
  df <- df[order(df$variable), ]

  rownames(df) <- NULL

  df

}

test_that("`tsl_smooth()` works", {

  tsl <- tsl_simulate(n = 2,
                      seed = 1)

  tsl_sd <- tsl_stats(tsl = tsl)$sd

  tsl_smooth <- tsl_smooth(tsl = tsl, window = 5, f = mean)

  tsl_smooth_sd <- tsl_stats(tsl = tsl_smooth)$sd

  expect_true(
    all(
      tsl_sd > tsl_smooth_sd
    )
  )

})

test_that("`tsl_stats()` works", {

  tsl <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  )

  df <- tsl_stats(tsl = tsl, lags = 3)

  expect_true(
    nrow(df) == 9
  )


  expect_true(
    ncol(df) == 17
  )

  expect_true(
    sum(is.na(df)) == 0
  )

  expect_true(
    all(
      colnames(df) %in% c("name", "variable", "NA_count", "min", "q1", "median", "q3", "max", "mean", "sd", "range", "iq_range", "skewness", "kurtosis", "ac_lag_1", "ac_lag_2", "ac_lag_3")
    )
  )

})

#' Exponential Smoothing of Zoo Time Series
#'
#' @description
#' Applies exponential smoothing to a zoo time series object, where each value is a weighted average of the current value and past smoothed values. This method is useful for reducing noise in time series data while preserving the general trend.
#'
#' @param x (required, zoo object) Time series to smooth. Default: NULL
#' @param alpha (required, numeric) Smoothing factor in the range (0, 1]. Determines the weight of the current value relative to past values. A higher value gives more weight to recent observations, while a lower value gives more weight to past observations. Default: 0.2
#'
#' @return zoo object
#' @autoglobal
#' @export
#' @family zoo_functions
#' @examples
#' x <- zoo_simulate()
#'
#' x_smooth <- zoo_smooth_exponential(
#'   x = x,
#'   alpha = 0.2
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(x_smooth)
#' }
zoo_smooth_exponential <- function(
    x = NULL,
    alpha = 0.2
){

  if(zoo::is.zoo(x) == FALSE){
    stop("distantia::zoo_smooth_exponential(): argument 'x' must be a zoo object.", call. = FALSE)
  }

  if (alpha <= 0 || alpha > 1) stop("distantia::zoo_smooth_exponential(): 'alpha' must be in the range (0, 1].")

  # recursive exponential smoothing
  y <- apply(
    X = zoo::coredata(x),
    MARGIN = 2,
    FUN = function(col) {
      Reduce(
        f = function(prev, curr) alpha * curr + (1 - alpha) * prev,
        x = col,
        accumulate = TRUE
      )
    }
  )

  y <- zoo::zoo(
    x = y,
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )

  y

}

test_that("`zoo_smooth()` works", {

  x <- zoo_simulate(cols = 1, seed = 1)

  x_smooth <- zoo_smooth_exponential(x = x, alpha = 0.2)

  expect_true(sd(x[, 1]) > sd(x_smooth[, 1]))

  x_smooth <- zoo_smooth_window(x = x)

  expect_true(sd(x[, 1]) > sd(x_smooth[, 1]))

})

#' Get Time Features from Zoo Objects
#'
#' @description
#' This function generates a data frame summarizing the time features (class, length, resolution, and others) of zoo time series.
#'
#'
#' @param x (required, zoo object) Zoo time series to analyze. Default: NULL.
#' @param keywords (optional, character string or vector) Defines what keywords are returned. If "aggregate", returns valid keywords for [zoo_aggregate()]. If "resample", returns valid keywords for [zoo_resample()]. If both, returns all valid keywords. Default: c("aggregate", "resample").
#'
#' @return Data frame with the following columns:
#'   \itemize{
#'     \item `name` (string): time series name.
#'     \item `rows` (integer): number of observations.
#'     \item `class` (string): time class, one of "Date", "POSIXct", or "numeric."
#'     \item `units` (string): units of the time series.
#'     \item `length` (numeric): total length of the time series expressed in `units`.
#'     \item `resolution` (numeric): average interval between observations expressed in `units`.
#'     \item `begin` (date or numeric): begin time of the time series.
#'     \item `end` (date or numeric): end time of the time series.
#'     \item `keywords` (character vector): valid keywords for [tsl_aggregate()] or [tsl_resample()], depending on the value of the argument `keywords`.
#'   }
#' @export
#' @autoglobal
#' @examples
#' #simulate a zoo time series
#' x <- zoo_simulate(
#'   rows = 150,
#'   time_range = c(
#'     Sys.Date() - 365,
#'     Sys.Date()
#'   ),
#'   irregular = TRUE
#' )
#'
#' #time data frame
#' zoo_time(
#'   x = x
#' )
#' @family zoo_functions
zoo_time <- function(
    x = NULL,
    keywords = c(
      "resample",
      "aggregate"
    )
){

  if(zoo::is.zoo(x) == FALSE){
    stop("distantia::zoo_time(): argument 'x' must be a zoo time series.", call. = FALSE)
  }

  keywords <- match.arg(
    arg = keywords,
    choices = c(
      "resample",
      "aggregate"
    ),
    several.ok = TRUE
  )

  if("name" %in% names(attributes(x))){
    x_name <- attributes(x)$name
  } else {
    x_name <- ""
  }

  x_time <- zoo::index(x)

  #class
  x_class <- class(x_time)
  if("POSIXct" %in% x_class){
    x_class <- "POSIXct"
  }

  #n
  x_rows <- length(x_time)

  #range
  x_range <- range(x_time)

  #length
  x_length <- diff(x_range)
  x_length_units <- attributes(x_length)$units
  if(is.null(x_length_units)){
    x_length_units <- x_class
  }
  x_length <- as.numeric(x_length)

  #resolution
  x_time_diff <- diff(x_time)
  if(!(x_class %in% c("numeric", "integer"))){
    units(x_time_diff) <- x_length_units
  }

  x_resolution <- as.numeric(mean(x_time_diff))

  #output data frame
  df <- data.frame(
    name = x_name,
    rows = x_rows,
    class = x_class,
    units = x_length_units,
    length = x_length,
    resolution = x_resolution,
    begin = min(x_range),
    end = max(x_range)
  )

  #units
  df_units <- utils_time_units(
    all_columns = TRUE,
    class = x_class
  )

  #subset by x_length_units
  df_units <- df_units[
    seq(
      from = min(
        which(
          df_units$base_units == x_length_units
          )
        ),
      to = nrow(df_units),
      by = 1
    ),
  ]


  #exception for very short time units
  if(!(x_length_units %in% c("hours", "mins", "secs"))){

    df_units <- df_units[
      df_units$threshold <= (x_length + x_resolution) &
        df_units$threshold >= x_resolution / 10,
    ]

  }

  #identify aggregation and resampling keywords
  df_units$aggregate <- FALSE
  df_units$resample <- FALSE

  #aggregation with an average of two samples
  df_units$aggregate[
    df_units$threshold > x_resolution * 2
    ] <- TRUE

  #resampling one order of magnitude above and below resolution
  df_units$resample[
    df_units$threshold >= x_resolution / 10 &
      df_units$threshold <= x_resolution * 10
    ] <- TRUE

  if(length(keywords) == 1){

    if(keywords == "aggregate"){
      df_units <- df_units[df_units$aggregate == TRUE, ]
    }

    if(keywords == "resample"){
      df_units <- df_units[df_units$resample == TRUE, ]
    }

  }

  if(nrow(df_units) > 0){
    df$keywords <- I(list(df_units$units))
  } else {
    if(x_class != "numeric"){
      df$keywords <- x_length_units
    }
  }

  df

}

test_that("`zoo_time()` works", {
  x <- zoo_simulate(rows = 150, time_range = c(
    Sys.Date() - 365,
    Sys.Date()
  ), irregular = TRUE, seed = 1)

  df <- zoo_time(x = x)

  expect_equal(class(df), "data.frame")
  expect_true(all(colnames(df) %in% c("name", "rows", "class", "units", "length", "resolution", "begin", "end", "keywords")))
})

#' Transform Values in Time Series Lists
#'
#' @description
#'
#' Function for time series transformations without changes in data dimensions. Generally, functions introduced via the argument `f` should not change the dimensions of the output time series list. See [tsl_resample()] and [tsl_aggregate()] for transformations requiring changes in time series dimensions.
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param f (required, transformation function) name of a function taking a matrix as input. Currently, the following options are implemented, but any other function taking a matrix as input (for example, [scale()]) should work as well:
#' \itemize{
#'   \item f_proportion: proportion computed by row.
#'   \item f_percent: percentage computed by row.
#'   \item f_hellinger: Hellinger transformation computed by row
#'   \item f_scale_local: Local centering and/or scaling based on the variable mean and standard deviation in each time series within `tsl`.
#'   \item f_scale_global: Global centering and/or scaling based on the variable mean and standard deviation across all time series within `tsl`.
#'   \item f_smooth: Time series smoothing with a user defined rolling window.
#'   \item f_detrend_difference: Differencing detrending of time series via [diff()].
#'   \item f_detrend_linear: Detrending of seasonal time series via linear modeling.
#'   \item f_detrend_gam: Detrending of seasonal time series via Generalized Additive Models.
#' }
#' @param ... (optional, additional arguments of `f`) Optional arguments for the transformation function.
#'
#' @return time series list
#' @export
#' @autoglobal
#' @examples
#'
#' #two time series
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_subset(
#'     names = c("Spain", "Sweden"),
#'     colnames = c("rainfall", "temperature")
#'   )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl
#'   )
#' }
#'
#' #centering and scaling
#' #-----------------------------------------
#' #same mean and standard deviation are used to scale each variable across all time series
#' tsl_scale <- tsl_transform(
#'   tsl = tsl,
#'   f = f_scale_local
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl_scale,
#'     guide_columns = 3
#'   )
#' }
#'
#'
#' #rescaling to a new range
#' #-----------------------------------------
#'
#' #rescale between -100 and 100
#' tsl_rescaled <- tsl_transform(
#'   tsl = tsl,
#'   f = f_rescale_local,
#'   new_min = -100,
#'   new_max = 100
#' )
#'
#' #old range
#' sapply(X = tsl, FUN = range)
#'
#' #new range
#' sapply(X = tsl_rescaled, FUN = range)
#'
#'
#'
#' #numeric transformations
#' #-----------------------------------------
#' #eemian pollen counts
#' tsl <- tsl_initialize(
#'   x = distantia::eemian_pollen,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl
#'   )
#' }
#'
#' #percentages
#' tsl_percentage <- tsl_transform(
#'   tsl = tsl,
#'   f = f_percent
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl_percentage
#'   )
#' }
#'
#' #hellinger transformation
#' tsl_hellinger <- tsl_transform(
#'   tsl = tsl,
#'   f = f_hellinger
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl_hellinger
#'   )
#' }
#'
#' @family tsl_processing
tsl_transform <- function(
    tsl = NULL,
    f = NULL,
    ...
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  if(is.function(f) == FALSE){

    stop(
      "distantia::tsl_transform(): Argument 'f' must be a function name with no quotes. Valid options are: ",
      paste(
        f_list(),
        collapse = ", "
      ),
      ".",
      call. = FALSE
    )

  }

  #progress bar
  p <- progressr::progressor(along = tsl)

  #apply centering and or scaling
  #if f has the formals "center" and "scale"
  if(".global" %in% names(formals(f))){

    #remove exclusive columns
    tsl <- tsl_subset(
      tsl = tsl,
      numeric_cols = TRUE,
      shared_cols = TRUE
    )

    #compute global scaling parameters
    scaling_params <- utils_global_scaling_params(
      tsl = tsl,
      f = f
    )

    # scaling or rescaling with scaling params
    tsl <- future.apply::future_lapply(
      X = tsl,
      FUN = function(x){

        p()

        f(
          x = x,
          center = scaling_params$center,
          scale = scaling_params$scale,
          old_min = scaling_params$old_min,
          old_max = scaling_params$old_max,
        )

      }
    )


  } else {

    #apply other transformation function
    tsl <- future.apply::future_lapply(
      X = tsl,
      FUN = function(x, ...){

        p()

        f(
          x = x,
          ... = ...
        )

      },
      ... = ...
    )


  }


  #reset names of zoo objects
  tsl <- tsl_names_set(
    tsl = tsl
  )

  tsl

}

test_that("`distantia_time_delay()` works", {

  tsl <- tsl_transform(
    tsl_transform(
      tsl_init(x = cities_temperature[cities_temperature$name %in%
      c("London", "Kinshasa"), ], name = "name", time = "time"),
    f = f_scale_local
  ),
  f = f_detrend_poly,
  degree = 35
  )

  df_shift <- distantia_time_delay(
    tsl = tsl,
    directional = TRUE
    )

  expect_equal(nrow(df_shift), 2)

  df_shift <- distantia_time_delay(
    tsl = tsl,
    directional = FALSE
  )

  expect_equal(nrow(df_shift), 1)

  expect_true(all(colnames(df_shift) %in% c("x", "y", "distance", "units", "mean", "min", "q1", "median", "q3", "max", "sd", "range", "modal")))

})

#' Remove Geometry Column from SF Data Frames
#'
#' Replicates the functionality of `sf::st_drop_geometry()` without depending on the `sf` package.
#'
#' @param df (required, data frame) Input data frame. Default: NULL.
#'
#' @return A data frame.
#' @autoglobal
#' @export
#' @family tsl_processing_internal
utils_drop_geometry <- function(
    df = NULL
){

  #remove geometry column from df
  sf.column <- attributes(df)$sf_column

  if(!is.null(sf.column)){

    df <- as.data.frame(df)
    df[[sf.column]] <- NULL
    attr(df, "sf_column") <- NULL

  }

  df

}

#' Convert List of Matrices to List of Data Frames
#'
#' @param x (required, list of matrices). Default: NULL.
#'
#' @return List of data frames.
#' @export
#' @autoglobal
#' @family internal
utils_prepare_matrix_list <- function(
    x = NULL
){

  if(
    utils_check_list_class(
      x = x,
      expected_class = "matrix"
      ) == FALSE
    ){

    return(x)

  }

  x <- lapply(
    X = x,
    FUN = function(x){
      if(is.null(colnames(x))){
        colnames(x) <- paste0("x", seq_len(ncol(x)))
      }
      x <- as.data.frame(x)

      return(x)
    }
  )

  x

}

#' Optimize the Silhouette Width of K-Means Clustering Solutions
#'
#' @description
#' Generates k-means solutions from 2 to `nrow(d) - 1` number of clusters and returns the number of clusters with the highest silhouette width median. See [utils_cluster_silhouette()] for more details.
#'
#' This function supports parallelization via [future::plan()] and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @param d (required, matrix) distance matrix typically resulting from [distantia_matrix()], but any other square matrix should work. Default: NULL
#' @param seed (optional, integer) Random seed to be used during the K-means computation. Default: 1
#'
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#'
#' #weekly covid prevalence
#' #in 10 California counties
#' #aggregated by month
#' tsl <- tsl_initialize(
#'   x = covid_prevalence,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_subset(
#'     names = 1:10
#'   ) |>
#'   tsl_aggregate(
#'     new_time = "months",
#'     fun = max
#'   )
#'
#' if(interactive()){
#'   #plotting first three time series
#'   tsl_plot(
#'     tsl = tsl_subset(
#'       tsl = tsl,
#'       names = 1:3
#'     ),
#'     guide_columns = 3
#'   )
#' }
#'
#' #compute dissimilarity matrix
#' psi_matrix <- distantia(
#'   tsl = tsl,
#'   lock_step = TRUE
#' ) |>
#'   distantia_matrix()
#'
#' #optimize hierarchical clustering
#' kmeans_optimization <- utils_cluster_kmeans_optimizer(
#'   d = psi_matrix
#' )
#'
#' #best solution in first row
#' head(kmeans_optimization)
#' 
#' @family distantia_support
utils_cluster_kmeans_optimizer <- function(
    d = NULL,
    seed = 1
    ){

  if(is.list(d)){
    d <- d[[1]]
  }

  if(!is.matrix(d)){
    stop("distantia::utils_cluster_kmeans_optimizer(): argument 'd' must be a matrix.", call. = FALSE)
  }

  if(nrow(d) != ncol(d)){
    stop("distantia::utils_cluster_kmeans_optimizer(): argument 'd' must be a square distance matrix.", call. = FALSE)
  }

  clusters_vector <- seq(
    from = 2,
    to = nrow(d) - 1,
    by = 1
  )

  p <- progressr::progressor(along = clusters_vector)



  sil <- foreach::foreach(
    i = clusters_vector,
    .combine = "c",
    .errorhandling = "pass",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

    p()

    set.seed(seed)

    k <- stats::kmeans(
      x = d,
      centers = i,
      algorithm = "Hartigan-Wong",
      nstart = nrow(d)
    )

    utils_cluster_silhouette(
      labels = k$cluster,
      d = d,
      mean = TRUE
    )

  }

  if(is.vector(sil) == FALSE){

    stop("distantia::utils_cluster_kmeans_optimizer(): Clustering optimization failed. Error message was: '", paste0(sil$message), "'.")

  }

  optimization_df <- data.frame(
    clusters = clusters_vector,
    silhouette_mean = sil
  )

  optimization_df <- optimization_df[order(-optimization_df$silhouette_mean), ]

  rownames(optimization_df) <- NULL

  optimization_df

}

#' Coerce Coredata of Univariate Zoo Time Series to Matrix
#'
#' @description
#' Transforms vector coredata of univariate zoo time series to class matrix. If the input zoo time series has the attribute "name", the output inherits the value of such attribute.
#'
#' Multivariate zoo objects (those with multiple columns) are returned unchanged.
#'
#' @param x (required, zoo object) zoo time series. Default: NULL
#' @param name (required, character string) name of the matrix column. Default: NULL
#'
#' @return zoo time series
#' @export
#' @autoglobal
#' @examples
#' #create zoo object from vector
#' x <- zoo::zoo(
#'   x = runif(100)
#' )
#'
#' #coredata is not a matrix
#' is.matrix(zoo::coredata(x))
#'
#' #convert to matrix
#' y <- zoo_vector_to_matrix(
#'   x = x
#' )
#'
#' #coredata is now a matrix
#' is.matrix(zoo::coredata(y))
#' @family zoo_functions
zoo_vector_to_matrix <- function(
    x = NULL,
    name = NULL
){

  #check x
  x <- utils_check_args_zoo(x = x)

  #get coredata
  x_data <- zoo::coredata(x)

  #return if matrix
  if(is.matrix(x_data)){
    return(x)
  }

  #get name
  x_name <- attr(x, which = "name", exact = TRUE)

  #convert to matrix
  y_data <- as.matrix(x_data)

  #set colnames
  colnames(y_data) <- name

  #convert to zoo
  y <- zoo::zoo(
    x = y_data,
    order.by = zoo::index(x)
  )

  #set name
  if(!is.null(x_name)){
    y <- zoo_name_set(
      x = y,
      name = x_name
    )
  }

  y

}

#' Plot Distance or Cost Matrix and Least Cost Path
#'
#' @description
#' This function is a simplified version of [fields::imagePlot()], by [Douglas Nychka](https://dnychka.github.io/). The original version is recommended in case more customization than the provided here is needed.
#'
#'
#' @param m (required, numeric matrix) distance or cost matrix generated by [psi_distance_matrix()] or [psi_cost_matrix()], but any numeric matrix will work. Default: NULL
#' @param matrix_color (optional, character vector) vector of colors. Uses the palette "Zissou 1" by default. Default: NULL
#' @param title (optional, character string) plot title. By default, names of the sequences used to compute the matrix `m`. Default: NULL
#' @param subtitle (optional, character string) plot subtitle. Default: NULL
#' @param xlab (optional, character string) title of the x axis (matrix columns). By default, the name of one of the sequences used to compute the matrix `m`. Default: NULL
#' @param ylab (optional, character string) title of the y axis (matrix rows). By default, the name of one of the sequences used to compute the matrix `m`. Default: NULL
#' @param text_cex (optional, numeric) multiplier of the text size for the plot labels and titles. Default: 1
#' @param path (optional, data frame) least cost path generated with [psi_cost_path()]. This data frame must have the attribute `type == "cost_path`, and must have been computed from the same sequences used to compute the matrix `m`. Default: NULL.
#' @inheritParams distantia_dtw_plot
#' @param guide (optional, logical) if TRUE, a color guide for the matrix `m` is added by [utils_matrix_guide()].
#' @param subpanel (optional, logical) internal argument used when generating the multi-panel plot produced by [distantia_dtw_plot()].
#'
#' @return plot
#' @examples
#' #prepare time series list
#' tsl <- tsl_simulate(
#'   n = 2,
#'   independent = TRUE
#' )
#'
#' #distance matrix between time series
#' dm <- psi_distance_matrix(
#'   x = tsl[[1]],
#'   y = tsl[[2]]
#' )
#'
#' #cost matrix
#' cm <- psi_cost_matrix(
#'   dist_matrix = dm
#'   )
#'
#' #least cost path
#' cp <- psi_cost_path(
#'   dist_matrix = dm,
#'   cost_matrix = cm
#' )
#'
#' #plot cost matrix and least cost path
#' if(interactive()){
#'   utils_matrix_plot(
#'     m = cm,
#'     path = cp,
#'     guide = TRUE
#'   )
#' }
#' @export
#' @autoglobal
#' @family internal_plotting
utils_matrix_plot <- function(
    m = NULL,
    matrix_color = NULL,
    title = NULL,
    subtitle = NULL,
    xlab = NULL,
    ylab = NULL,
    text_cex = 1,
    path = NULL,
    path_width = 1,
    path_color = "black",
    diagonal_width = 1,
    diagonal_color = "white",
    guide = TRUE,
    subpanel = FALSE
){

  #First to upper
  firstup <- function(x) {
    substr(x, 1, 1) <- toupper(substr(x, 1, 1))
    x
  }

  # Preserve user's config
  if(subpanel == FALSE){
    old.par <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(old.par))
  }

  #check m
  m <- utils_check_args_matrix(m = m)

  #specific behaviours by matrix type
  m_type <- attributes(m)$type

  #generic matrix
  if(is.null(m_type)){

    m_type <- "generic"
    attr(x = m, which = "x_name") <- "x"
    attr(x = m, which = "y_name") <- "y"
    attr(x = m, which = "x_time") <- seq_len(ncol(m))
    attr(x = m, which = "y_time") <- seq_len(nrow(m))
    guide_title <- "Legend"

  } else {

    if(m_type == "distance"){
      guide_title <- paste0(
        attributes(m)$distance,
        "\ndistance"
      )
    }

    if(m_type == "cost"){
      guide_title <- paste0(
        attributes(m)$distance,
        "\ncumulative\ncost"
      )
    }

  }

  x_name <- attributes(m)$x_name
  y_name <- attributes(m)$y_name


  if(is.null(title)){
    title <- paste0(
      attributes(m)$y_name,
      " vs. ",
      attributes(m)$x_name
    )
  }

  if(is.null(xlab)){
    xlab <- x_name
  }

  if(is.null(ylab)){
    ylab <- y_name
  }

  #get axes time
  axis_x_labels <- attributes(m)$x_time
  axis_y_labels <- attributes(m)$y_time

  #to pretty
  axis_x_labels_pretty <- pretty(x = axis_x_labels)
  axis_y_labels_pretty <- pretty(x = axis_y_labels)

  #axis positions closest to pretty version
  axis_x_at <- lapply(
    X = axis_x_labels_pretty,
    FUN = function(x){
      which.min(abs(as.numeric(x - axis_x_labels)))
    }
  ) |>
    unlist()

  axis_y_at <- lapply(
    X = axis_y_labels_pretty,
    FUN = function(x){
      which.min(abs(as.numeric(x - axis_y_labels)))
    }
  ) |>
    unlist()

  axis_x_labels <- attributes(axis_x_labels_pretty)$labels
  axis_y_labels <- attributes(axis_y_labels_pretty)$labels

  #psi matrix
  if(m_type == "distantia_matrix"){

    if(is.null(title)){
      title <- "Dissimilarity Matrix"
    }

    guide_title <- "Psi \ndistance"

    if(is.null(xlab)){
      xlab <- ""
    }

    if(is.null(ylab)){
      ylab <- ""
    }

    axis_x_at <- seq_len(ncol(m))
    axis_y_at <- seq_len(nrow(m))

    axis_x_labels <- axis_y_labels <- dimnames(m)[[2]]

  }

  guide_title <- firstup(x = guide_title)

  #axis title
  #leave more space if axis needs dates
  if(
    class(attributes(m)$x_time) %in% c("Date", "POSIXct") &&
    subpanel == FALSE
  ){

    axis_title_distance <- 2.5
    axis_labels_cex <- 0.6 * text_cex

  } else {

    axis_title_distance <- 2.2
    axis_labels_cex <- 0.8 * text_cex

  }

  #title cex
  axis_title_cex <- 0.9 * text_cex

  #plotting areas
  plt_all <- graphics::par()$plt
  plt_m <- plt_all

  if(guide == TRUE){

    plt_m <- c(
      plt_m[1],
      plt_m[2] - 0.1,
      plt_m[3],
      plt_m[4]
    )

    plt_guide <- c(
      plt_m[2] + 0.02,
      plt_all[2] - 0.05,
      plt_m[3],
      plt_m[4]
    )

  }

  #title
  main_title_distance <- ifelse(
    test = is.null(subtitle),
    yes = 1.2,
    no = 2
  )
  main_title_cex <- 1.2 * text_cex

  #subtitle
  subtitle_distance <- 0.5
  subtitle_cex <- 1 * text_cex

  if(is.null(matrix_color)){
    matrix_color = color_continuous(
      n = 100
    )
  }

  breaks <- utils_color_breaks(
    m = m,
    n = length(matrix_color)
  )

  graphics::par(
    plt = plt_m
  )

  #plot matrix
  graphics::image(
    x = seq_len(ncol(m)),
    y = seq_len(nrow(m)),
    z = t(m),
    breaks = breaks,
    col = matrix_color,
    xlab = "",
    ylab = "",
    axes = FALSE,
    useRaster = FALSE,
    add = FALSE
  )

  if(subpanel == FALSE){

    graphics::title(
      xlab = xlab,
      line = axis_title_distance,
      cex.lab = axis_title_cex
    )

    graphics::title(
      ylab = ylab,
      line = axis_title_distance,
      cex.lab = axis_title_cex
    )


    graphics::axis(
      side = 1,
      at = axis_x_at,
      labels = axis_x_labels,
      cex.axis = axis_labels_cex,
      las = 2
    )

    graphics::axis(
      side = 2,
      at = axis_y_at,
      labels = axis_y_labels,
      cex.axis = axis_labels_cex,
      las = 2
    )

  }

  graphics::title(
    main = title,
    cex.main = main_title_cex,
    line = main_title_distance
  )

  # matrix subtitle ----
  if(!is.null(subtitle)){

    graphics::mtext(
      side = 3,
      line = subtitle_distance,
      at = NA,
      adj = NA,
      padj = NA,
      outer = FALSE,
      cex = subtitle_cex,
      subtitle
    )

  }

  #plot diagonal
  graphics::lines(
    x = c(0, ncol(m) + 0.5),
    y = c(0, nrow(m) + 0.5),
    lwd = diagonal_width,
    col = diagonal_color
  )

  # least cost path ----
  if(!is.null(path)){

    path <- utils_check_args_path(
      path = path
    )

    if(is.null(attributes(path)$x_name)){
      attr(x = path, which = "x_name") <- "x"
    }

    if(is.null(attributes(path)$y_name)){
      attr(x = path, which = "y_name") <- "y"
    }

    if(
      attributes(path)$y_name != y_name ||
      attributes(path)$x_name != x_name
    ){
      stop("distantia::utils_matrix_plot(): time series names in arguments 'm' and 'path' do not match.", call. = FALSE)
    }

    #rename path columns
    colnames(path)[colnames(path) == "y"] <- y_name
    colnames(path)[colnames(path) == "x"] <- x_name

    #Sakoe-Chiba band
    if("bandwidth" %in% colnames(path)){

      bandwidth <- path$bandwidth[1]

      graphics::lines(
        x = c(ncol(m) * bandwidth, ncol(m)),
        y = c(0, nrow(m) - (nrow(m) * bandwidth)),
        lwd = diagonal_width,
        lty = "dotted",
        col = diagonal_color
      )

      graphics::lines(
        x = c(0, ncol(m) - (ncol(m) * bandwidth)),
        y = c(nrow(m) * bandwidth, nrow(m)),
        lwd = diagonal_width,
        lty = "dotted",
        col = diagonal_color
      )

    }

    #plot cost path
    graphics::lines(
      x = path[[attributes(m)$x_name]],
      y = path[[attributes(m)$y_name]],
      lwd = path_width,
      col = path_color
    )

  }

  #guide
  if(guide == TRUE){

    graphics::par(
      plt = plt_guide,
      new = TRUE
    )

    utils_matrix_guide(
      m = m,
      matrix_color = matrix_color,
      breaks = breaks,
      title = guide_title,
      text_cex = text_cex
    )

  }

  invisible()

}

#' Get Number of Rows in Time Series Lists
#'
#' @param tsl (required, list) Time series list. Default: NULL
#'
#' @return list
#' @export
#' @examples
#' #Simulate zoo time series
#' tsl <- tsl_simulate(
#'   rows = 150
#'   )
#'
#' #Count rows
#' tsl_nrow(
#'   tsl = tsl
#' )
#' @autoglobal
#' @family tsl_management
tsl_nrow <- function(
    tsl = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  lapply(
    X = tsl,
    FUN = function(x){
      nrow(x)
    }
  )

}

test_that("`tsl_ncol()` and `tsl_nrow()` works", {

   tsl <- tsl_simulate(
    n = 2,
    cols = 6,
    irregular = FALSE,
    rows = 100,
    seed = 1
    )

  expect_equal(tsl_ncol(tsl = tsl)$A, 6)
  expect_equal(tsl_ncol(tsl = tsl)$B, 6)
  expect_equal(tsl_nrow(tsl = tsl)$A, 100)
  expect_equal(tsl_nrow(tsl = tsl)$B, 100)

})

test_that("`tsl_handle_NA()` works", {

  tsl <- tsl_simulate(na_fraction = 0.25,
                      seed = 1)

  tsl_nrow <- tsl_nrow(tsl = tsl) |>
    unlist() |>
    sum()

  tsl_no_na <- tsl_handle_NA(tsl = tsl, na_action = "omit")

  expect_true(
    tsl_nrow > tsl_nrow(tsl = tsl_no_na) |>
      unlist() |>
      sum()
  )

  expect_equal(sum(unlist(tsl_count_NA(tsl = tsl_no_na))), 0)

  tsl_no_na <- tsl_handle_NA(tsl = tsl, na_action = "impute")

  expect_true(
    tsl_nrow == tsl_nrow(tsl = tsl_no_na) |>
      unlist() |>
      sum()
  )

  expect_equal(sum(unlist(tsl_count_NA(tsl = tsl_no_na))), 0)

})

#' Global Centering and Scaling Parameters of Time Series Lists
#'
#' @description
#' Internal function to compute global scaling parameters (mean and standard deviation) for time series lists. Used within [tsl_transform()] when the scaling function [f_scale_global()] is used as input for the argument `f`.
#'
#' This function removes exclusive columns from the data. See function [tsl_subset()].
#'
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param f (required, function) function [f_scale_global()]. Default: NULL
#' @param ... (optional, arguments of `f`) Optional arguments for the transformation function.
#'
#' @return list
#' @export
#' @autoglobal
#' @family tsl_processing_internal
utils_global_scaling_params <- function(
    tsl = NULL,
    f = NULL,
    ...
){

  #user values for center and scale
  args <- list(...)

  if("center" %in% names(args)){
    #get user input
    center <- args$center
  } else if ("center" %in% names(formals(f))){
    #get function default
    center <- formals(f)$center
  } else {
    center <- "none"
  }

  if("scale" %in% names(args)){
    scale <- args$scale
  } else if ("scale" %in% names(formals(f))){
    scale <- formals(f)$scale
  } else {
    scale <- "none"
  }

  if("old_min" %in% names(args)){
    old_min <- args$old_min
  } else if ("old_min" %in% names(formals(f))){
    old_min <- formals(f)$old_min
  } else {
    old_min <- "none"
  }

  if("old_max" %in% names(args)){
    old_max <- args$old_max
  } else if ("old_max" %in% names(formals(f))){
    old_max <- formals(f)$old_max
  } else {
    old_max <- "none"
  }

  #joining data for mean and sd computation
  tsl.matrix <- do.call(
    what = "rbind",
    args = lapply(
      X = tsl,
      FUN = zoo::coredata
    )
  )

  #output list
  out_list <- list()

  #computing mean
  if(center == TRUE){
    out_list$center <- apply(
      X = tsl.matrix,
      MARGIN = 2,
      FUN = mean,
      na.rm = TRUE
    )
  }

  #computing sd
  if(scale == TRUE){
    out_list$scale <- apply(
      X = tsl.matrix,
      MARGIN = 2,
      FUN = stats::sd,
      na.rm = TRUE
    )
  }

  #computing old_min
  if(is.null(old_min)){
    out_list$old_min <- apply(
      X = tsl.matrix,
      MARGIN = 2,
      FUN = min,
      na.rm = TRUE
    )
  }

  if(is.null(old_max)){
    out_list$old_max <- apply(
      X = tsl.matrix,
      MARGIN = 2,
      FUN = max,
      na.rm = TRUE
    )
  }

  out_list

}

#' Checks Classes of List Elements Against Expectation
#'
#' @param x (required, list). Default: NULL.
#' @param expected_class (required, class name) One of "data.frame", "matrix", or "vector". Default: "data.frame".
#'
#' @return Logical value indicating whether the elements match the expected class.
#' @autoglobal
#' @export
#' @family internal
utils_check_list_class <- function(
    x = NULL,
    expected_class = "data.frame"
){

  if(inherits(x = x, what = "list") == FALSE){
    stop("distantia::utils_check_list_class(): argument 'x' is not a list.", call. = FALSE)
  }

  expected_class <- match.arg(
    arg = expected_class,
    choices = c(
      "data.frame",
      "matrix",
      "vector"
    ),
    several.ok = FALSE
  )


  if(expected_class != "vector"){

    #check element class
    x.class <- lapply(
      X = x,
      FUN = class
    ) |>
      unlist() |>
      as.data.frame()

  } else {

    #check element class
    x.class <- lapply(
      X = x,
      FUN = function(x)
        if(is.vector(x)){
          return("vector")
        } else {
          return(class(x)[1])
        }
    ) |>
      unlist() |>
      as.data.frame()

  }

  colnames(x.class) <- "class"

  if(any(x.class$class %in% expected_class)){

    return(TRUE)

  } else {

    return(FALSE)

  }

}

#' Diagnose Issues in Time Series Lists
#'
#' @description
#' A Time Series List (`tsl` for short) is a named list of zoo time series. This type of object, not defined as a class, is used throughout the `distantia` package to contain time series data ready for processing and analysis.
#'
#' The structure and values of a `tsl` must fulfill several general conditions:
#'
#' Structure:
#' \itemize{
#'   \item List names match the attributes "name" of the zoo time series.
#'   \item Zoo time series must have at least one shared column name.
#'   \item The index (as extracted by [zoo::index()]) of all zoo objects must be of the same class (either "Date", "POSIXct", "numeric", or "integer").
#'   \item The "core data" (as extracted by [zoo::coredata()]) of univariate zoo time series must be of class "matrix".
#' }
#'
#' Values (optional, when `full = TRUE`):
#' \itemize{
#'   \item All time series have at least one shared numeric column.
#'   \item There are no NA, Inf, or NaN values in the time series.
#' }
#'
#' This function analyzes a `tsl` without modifying it and returns messages describing what conditions are not met, and provides hints on how to fix most issues.
#'
#' @param tsl (required, list of zoo time series) Time series list to diagnose. Default: NULL
#' @param full (optional, logical) If TRUE, a full diagnostic is triggered. Otherwise, only the data structure is tested. Default: TRUE
#'
#' @return invisible
#' @export
#' @autoglobal
#' @examples
#' #creating three zoo time series
#'
#' #one with NA values
#' x <- zoo_simulate(
#'   name = "x",
#'   cols = 1,
#'   na_fraction = 0.1
#'   )
#'
#' #with different number of columns
#' #with repeated name
#' y <- zoo_simulate(
#'   name = "x",
#'   cols = 2
#'   )
#'
#' #with different time class
#' z <- zoo_simulate(
#'   name = "z",
#'   cols = 1,
#'   time_range = c(1, 100)
#'   )
#'
#' #adding a few structural issues
#'
#' #changing the column name of x
#' colnames(x) <- c("b")
#'
#' #converting z to vector
#' z <- zoo::zoo(
#'   x = runif(nrow(z)),
#'   order.by = zoo::index(z)
#' )
#'
#' #storing zoo objects in a list
#' #with mismatched names
#' tsl <- list(
#'   a = x,
#'   b = y,
#'   c = z
#' )
#'
#' #running full diagnose
#' tsl_diagnose(
#'   tsl = tsl,
#'   full = TRUE
#'   )
#' @family tsl_management
tsl_diagnose <- function(
    tsl = NULL,
    full = TRUE
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 2
  )

  if(!is.logical(full)){
    stop("distantia::tsl_diagnose(): argument 'full' must be TRUE to run a full diagnostic, and FALSE otherwise.", call. = FALSE)
  }

  if(is.null(tsl)){
    stop("distantia::tsl_diagnose(): argument 'tsl' must not be NULL.", call. = FALSE)
  }

  if(!is.list(tsl)){
    stop("distantia::tsl_diagnose(): argument 'tsl' must be a list.", call. = FALSE)
  }


  #all possible issues
  all_issues <- list(
    tsl_objects_zoo = " - time series in 'tsl' must be 'zoo' objects: coerce all at once with tsl <- lapply(tsl, zoo::zoo) or individually with zoo::zoo().",
    tsl_names_issues =  "  - list and time series names must match and be unique: reset names with distantia::tsl_names_set().",
    zoo_no_colnames = "  - missing column names in zoo time series: use distantia::tsl_colnames_set() to rename columns as needed.",
    zoo_duplicated_colnames = "  - zoo time series have duplicated column names: use distantia::tsl_colnames_clean() or distantia::tsl_colnames_set() to deduplicate.",
    zoo_no_shared_columns = "  - no shared column names across time series: use distantia::tsl_colnames_get() and distantia::tsl_colnames_set() to identify and rename columns as needed.",
    zoo_time_class = "  - time in all time series must be of the same class: use lapply(tsl, function(x) class(zoo::index(x))) to identify and remove or modify the objects with a mismatching class.",
    zoo_non_numeric_columns = "  - columns shared across all time series must be numeric: the function distantia::tsl_subset() may help fix this issue.",
    zoo_NA_cases = "  - there are NA, Inf, -Inf, or NaN cases in the time series: interpolate or remove them with distantia::tsl_handle_NA().",
    zoo_no_matrix = "  - core data of univariate zoo time series must be of class 'matrix': use lapply(tsl, distantia::zoo_vector_to_matrix) to fix this issue."
  )

  # initialize feedback
  issues_structure <- vector()
  issues_values <- vector()

  #zoo objects

  # elements in tsl are not zoo objects
  if(
    !all(
      sapply(
        X = tsl,
        FUN = zoo::is.zoo
      )
    )
  ){

    issues_structure <- c(
      issues_structure,
      all_issues[["tsl_objects_zoo"]]
    )

  } else {

    #all zoo objects are matrices
    zoo.is.not.matrix <- sapply(
      X = tsl,
      FUN = function(x) !is.matrix(zoo::coredata(x))
    ) |>
      any()

    if(zoo.is.not.matrix){

      issues_structure <- c(
        issues_structure,
        all_issues[["zoo_no_matrix"]]
      )

    }

  }

  #naming issues
  if(
    tsl_names_test(tsl = tsl) == FALSE
    ){

    issues_structure <- c(
      issues_structure,
      all_issues[["tsl_names_issues"]]
    )

  }

  #zoo colnames

  # zoo objects have column names
  zoo_colnames_all <- tsl_colnames_get(
    tsl = tsl,
    names = "all"
  )

  if(any(unlist(zoo_colnames_all) == "unnamed")){

    issues_structure <- c(
      issues_structure,
      all_issues[["zoo_no_colnames"]]
    )

  }


  #test for duplicated colnames
  zoo_colnames_duplicated <- lapply(
    X = zoo_colnames_all,
    FUN = duplicated
  ) |>
    unlist() |>
    any()

  #duplicated colnames in tsl
  if(zoo_colnames_duplicated){

    issues_structure <- c(
      issues_structure,
      all_issues[["zoo_duplicated_colnames"]]
    )

  }


  # zoo objects have shared colnames
  zoo_colnames_shared <- tsl_colnames_get(
    tsl = tsl,
    names = "shared"
  )

  #there are no shared columns
  if(all(is.na(zoo_colnames_shared))){

    issues_structure <- c(
      issues_structure,
      all_issues[["zoo_no_shared_columns"]]
    )

  }

  #zoo time class
  zoo.time.classes <- unique(tsl_time(
    tsl = tsl
  )$class)

  if(length(zoo.time.classes) != 1){

    issues_structure <- c(
      issues_structure,
      all_issues[["zoo_time_class"]]
    )

  }

  #run full test
  if(full == TRUE){

    #all columns in zoo objects are numeric
    zoo.columns.numeric <- sapply(
      X = tsl,
      FUN = function(x){

        #vector
        if(is.null(dim(x))){

          out <- is.numeric(x)

        } else {
          #matrix

          out <- apply(
            X = zoo::coredata(x),
            FUN = is.numeric,
            MARGIN = 2
          ) |>
            all()

        }

        out

      }

    ) |>
      all()

    if(
      zoo.columns.numeric == FALSE
      ){

      issues_values <- c(
        issues_values,
        all_issues[["zoo_non_numeric_columns"]]
      )

    }

    # NA values
    na.count <- tsl_count_NA(
      tsl = tsl
    ) |>
      unlist() |>
      sum()

    if(na.count > 0){

      issues_values <- c(
        issues_values,
        all_issues[["zoo_NA_cases"]]
      )

    }

  } #end of full == TRUE

  if(length(issues_structure) > 0){

    message(
      "distantia::tsl_diagnose(): issues in TSL structure:\n",
      "---------------------------------------------------\n\n",
      paste(
        issues_structure,
        collapse = "\n\n"
      )
    )

  }

  if(length(issues_values) > 0){

    message(
      "\ndistantia::tsl_diagnose(): issues in TSL values:\n",
      "--------------------------------------------------\n\n",
      paste(
        issues_values,
        collapse = "\n\n"
      )
    )

  }

  invisible()

}

test_that("`tsl_repair()` works", {

  x <- zoo_simulate(name = "x", cols = 1, na_fraction = 0.1, seed = 1)
  y <- zoo_simulate(name = "x", cols = 2, seed = 2)
  z <- zoo_simulate(name = "z", cols = 1, time_range = c(1, 100), seed = 3)
  colnames(x) <- c("b")
  z <- zoo::zoo(x = runif(nrow(z)), order.by = zoo::index(z))
  tsl <- list(a = x, b = y, c = z)

  expect_message(
    tsl_diagnose(tsl = tsl, full = TRUE)
    ) |>
    suppressMessages()

  expect_message(
    tsl <- tsl_repair(tsl)
  ) |>
    suppressMessages()

})

test_that("`tsl_diagnose()` works", {

  x <- zoo_simulate(name = "x", cols = 1, na_fraction = 0.1, seed = 1)
  y <- zoo_simulate(name = "x", cols = 2, seed = 2)
  z <- zoo_simulate(name = "z", cols = 1, time_range = c(1, 100), seed = 3)
  colnames(x) <- c("b")

  z <- zoo::zoo(x = runif(nrow(z)), order.by = zoo::index(z))

  tsl <- list(a = x, b = y, c = z)

  expect_message(tsl_diagnose(tsl = tsl)) |>
    suppressMessages()

})

#' Plot Time Series List
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param columns (optional, integer) Number of columns of the multipanel plot. Default: 1
#' @param xlim (optional, numeric vector) Numeric vector with the limits of the x axis. Applies to all sequences. Default: NULL
#' @param ylim (optional, numeric vector or character string) Numeric vector of length two with the limits of the vertical axis or a keyword. Accepted keywords are:
#' \itemize{
#'   \item "absolute" (default): all time series are plotted using the overall data range. When this option is used, horizontal lines indicating the overall mean, minimum, and maximum are shown as reference.
#'   \item "relative": each time series is plotted using its own range. Equivalent result can be achieved using `ylim = NULL`.
#' }
#' @param line_color (optional, character vector) vector of colors for the distance or cost matrix. If NULL, uses an appropriate palette generated with [grDevices::palette.colors()]. Default: NULL
#' @param line_width (optional, numeric vector) Width of the time series plot. Default: 1
#' @param text_cex (optional, numeric) Multiplicator of the text size. Default: 1
#' @param guide (optional, logical) If TRUE, plots a legend. Default: TRUE
#' @param guide_columns (optional, integer) Columns of the line guide. Default: 1
#' @param guide_cex (optional, numeric) Size of the guide's text and separation between the guide's rows. Default: 0.7.
#'
#' @return plot
#' @export
#' @autoglobal
#' @examples
#' #simulate zoo time series
#' tsl <- tsl_simulate(
#'   cols = 3
#'   )
#'
#' if(interactive()){
#'
#'   #default plot
#'   tsl_plot(
#'     tsl = tsl
#'     )
#'
#'   #relative vertical limits
#'   tsl_plot(
#'     tsl = tsl,
#'     ylim = "relative"
#'   )
#'
#'   #changing layout
#'   tsl_plot(
#'     tsl = tsl,
#'     columns = 2,
#'     guide_columns = 2
#'   )
#'
#'   #no legend
#'   tsl_plot(
#'     tsl = tsl,
#'     guide = FALSE
#'   )
#'
#'   #changing color
#'   tsl_plot(
#'     tsl = tsl,
#'     line_color = c("red", "green", "blue"))
#'
#' }
#' @family tsl_visualization
tsl_plot <- function(
    tsl = NULL,
    columns = 1,
    xlim = NULL,
    ylim = "absolute",
    line_color = NULL,
    line_width = 1,
    text_cex = 1,
    guide = TRUE,
    guide_columns = 1,
    guide_cex = 0.8
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #set NaN and Inf to NA
  #avoids issues when computing axes limits
  tsl <- tsl_Inf_to_NA(
    tsl = tsl
  )

  tsl <- tsl_NaN_to_NA(
    tsl = tsl
  )

  axis_title_distance <- 2.5

  # preserve user's config
  old.par <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old.par))

  #manage ylim
  if(all(ylim == "absolute") == TRUE){

    tsl_vars <- tsl |>
      tsl_colnames_get() |>
      unlist() |>
      unique()

    df <- tsl_to_df(tsl = tsl)
    m <- as.matrix(df[, tsl_vars])

    ylim_min <- min(x = m, na.rm = TRUE)
    ylim_max <- max(x = m, na.rm = TRUE)
    ylim_mean <- mean(x = m, na.rm = TRUE)

    ylim <- c(ylim_min, ylim_max)

    rm(df, m)

  }

  if(all(ylim == "relative") == TRUE){
    ylim <- NULL
  }

  #generate colors
  line_color <- utils_line_color(
    x = tsl,
    line_color = line_color
  )

  #define new par
  if(guide == TRUE){
    n_panels <- length(tsl) + 1
  } else {
    n_panels <- length(tsl)
  }

  rows <- ceiling(n_panels/columns)

  graphics::par(
    mfrow = c(rows, columns),
    oma = c(2, 1, 1, 1),
    mar = c(1, 3.5, 1, 0.5)
  )

  #plot panels
  for(i in seq_len(n_panels)){

    if(i <= length(tsl)){

      zoo_plot(
        x = tsl[[i]],
        line_color = line_color,
        line_width = line_width,
        xlim = xlim,
        ylim = ylim,
        title = "",
        xlab = NULL,
        ylab = "",
        text_cex = text_cex,
        guide = FALSE,
        guide_cex = guide_cex,
        subpanel = FALSE
      )

      if(exists("ylim_min")){
        graphics::abline(
          h = ylim_min,
          col = "gray50",
          lty = 3
          )
      }

      if(exists("ylim_max")){
        graphics::abline(
          h = ylim_max,
          col = "gray50",
          lty = 3
          )
      }

      if(exists("ylim_mean")){
        graphics::abline(
          h = ylim_mean,
          col = "gray50",
          lty = 3
          )
      }

      graphics::title(
        ylab = attributes(tsl[[i]])$name,
        line = axis_title_distance,
        cex.lab = text_cex
      )

    } else {

      if(guide == TRUE){

        utils_line_guide(
          x = tsl,
          position = "center",
          line_color = line_color,
          line_width = line_width,
          text_cex = guide_cex,
          guide_columns = guide_columns,
          subpanel = TRUE
        )

      }

    }

  }

  invisible()

}

#' Multivariate TSL to Univariate TSL
#'
#' @description
#' Takes a time series list with multivariate zoo objects to generate a new one with one univariate zoo object per variable. A time series list with the zoo objects "A" and "B", each with the columns "a", "b", and "c", becomes a time series list with the zoo objects "A__a", "A__b", "A__c", "B__a", "B__b", and "B__c". The only column of each new zoo object is named "x".
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param sep (required, character string) separator between the time series name and the column name. Default: "__"
#'
#' @return time series list: list of univariate zoo objects with a column named "x".
#' @export
#' @autoglobal
#' @examples
#'
#' tsl <- tsl_simulate(
#'   n = 2,
#'   time_range = c(
#'     "2010-01-01",
#'     "2024-12-31"
#'   ),
#'   cols = 3
#' )
#'
#' tsl_names_get(tsl)
#' tsl_colnames_get(tsl)
#'
#' if(interactive()){
#'   tsl_plot(tsl)
#' }
#'
#' tsl <- tsl_burst(tsl)
#'
#' tsl_names_get(tsl)
#' tsl_colnames_get(tsl)
#'
#' if(interactive()){
#'   tsl_plot(tsl)
#' }
#' @family tsl_management
tsl_burst <- function(
    tsl = NULL,
    sep = "__"
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #return tsl if univariate already
  tsl_columns <- tsl_ncol(tsl) |>
    unlist() |>
    unique()

  if(all(tsl_columns == 1)){
    return(tsl)
  }

  iterations_df <- expand.grid(
    tsl_name = names(tsl),
    time_series_name = lapply(
      X = tsl,
      FUN = colnames
    ) |>
      unlist(),
    stringsAsFactors = FALSE
  )

  iterations_df <- iterations_df[order(iterations_df$tsl_name), ]

  iterations_df$new_name <- paste(
    iterations_df$tsl_name,
    iterations_df$time_series_name,
    sep = sep
  )

  tsl_new <- list()

  for(i in seq_len(nrow(iterations_df))){

    zoo_i <- tsl[[iterations_df$tsl_name[i]]][, iterations_df$time_series_name[i], drop = FALSE]

    colnames(zoo_i) <- "x"

    tsl_new[[iterations_df$new_name[i]]] <- zoo_i

  }

  tsl_new <- tsl_names_set(tsl = tsl_new)

  tsl_new

}

test_that("`tsl_burst()` works", {

  n = 2
  cols = 3

  tsl <- tsl_simulate(
    n = n,
    time_range = c("2010-01-01", "2024-12-31"),
    cols = cols,
    seed = 1
  )


  tsl_ <- tsl_burst(tsl)

  expect_true(
    length(tsl_) == n * cols
  )

  expect_true(
    unique(unlist(tsl_colnames_get(tsl = tsl_))) == "x"
  )

})

#' Time Features of Time Series Lists
#'
#' @description
#' The functions [tsl_time()] and [tsl_time_summary()] summarize the time features of a time series list.
#' \itemize{
#'   \item [tsl_time()] returns a data frame with one row per time series in the argument 'tsl'
#'   \item [tsl_time_summary()] returns a list with the features captured by [tsl_time()], but aggregated across time series.
#' }
#'
#' Both functions return keywords useful for the functions [tsl_aggregate()] and [tsl_resample()], depending on the value of the argument `keywords`.
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param keywords (optional, character string or vector) Defines what keywords are returned. If "aggregate", returns valid keywords for [zoo_aggregate()]. If "resample", returns valid keywords for [zoo_resample()]. If both, returns all valid keywords. Default: c("aggregate", "resample").
#' @return
#' \itemize{
#'   \item [tsl_time()]: data frame with the following columns:
#'   \itemize{
#'     \item `name` (string): time series name.
#'     \item `rows` (integer): number of observations.
#'     \item `class` (string): time class, one of "Date", "POSIXct", or "numeric"
#'     \item `units` (string): units of the time series.
#'     \item `length` (numeric): total length of the time series expressed in `units`.
#'     \item `resolution` (numeric): average interval between observations expressed in `units`.
#'     \item `begin` (date or numeric): begin time of the time series.
#'     \item `end` (date or numeric): end time of the time series.
#'     \item `keywords` (character vector): valid keywords for [tsl_aggregate()] or [tsl_resample()], depending on the value of the argument `keywords`.
#'   }
#'   \item [tsl_time_summary()]: list with the following objects:
#'   \itemize{
#'     \item `class` (string): time class, one of "Date", "POSIXct", or "numeric"
#'     \item `units` (string): units of the time series.
#'     \item `begin` (date or numeric): begin time of the time series.
#'     \item `end` (date or numeric): end time of the time series.
#'     \item `resolution_max` (numeric): longer time interval between consecutive samples expressed in `units`.
#'     \item `resolution_min` (numeric): shorter time interval between consecutive samples expressed in `units`.
#'     \item `keywords` (character vector): valid keywords for [tsl_aggregate()] or [tsl_resample()], depending on the value of the argument `keywords`.
#'     \item `units_df` (data frame) data frame for internal use within [tsl_aggregate()] and [tsl_resample()].
#'   }
#' }
#' @export
#' @autoglobal
#' @examples
#' #simulate a time series list
#' tsl <- tsl_simulate(
#'   n = 3,
#'   rows = 150,
#'   time_range = c(
#'     Sys.Date() - 365,
#'     Sys.Date()
#'   ),
#'   irregular = TRUE
#' )
#'
#' #time data frame
#' tsl_time(
#'   tsl = tsl
#' )
#'
#' #time summary
#' tsl_time_summary(
#'   tsl = tsl
#' )
#' @family tsl_management
tsl_time <- function(
    tsl = NULL,
    keywords = c(
      "resample",
      "aggregate"
    )
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  time_list <- lapply(
    X = tsl,
    FUN = zoo_time,
    keywords = keywords
  )

  out <- do.call(
    what = "rbind",
    args = time_list
  )

  rownames(out) <- NULL

  out

}


#' @rdname tsl_time
#' @export
#' @autoglobal
tsl_time_summary <- function(
    tsl = NULL,
    keywords = c(
      "resample",
      "aggregate"
    )
){

  df <- tsl_time(
    tsl = tsl,
    keywords = keywords
    )

  keywords <- unique(unlist(df$keywords))

  begin <- min(df$begin)
  if(is.numeric(begin)){
    if(!is.integer(begin)){
      begin <- round(x = begin, digits = 4)
      end <- round(x = begin, digits = 4)
    }
  }

  tsl_units <- utils_time_units(
    all_columns = TRUE
  )

  units_df <- tsl_units[
    tsl_units$units %in% keywords,
    c("factor", "base_units", "units", "threshold", "keyword")
  ]

  resolution_range <- lapply(
    X = tsl,
    FUN = function(x){
      range(diff(zoo::index(x)))
    }
  ) |>
    unlist()

  list(
    class = unique(df$class),
    units = unique(df$units),
    begin = min(df$begin),
    end = max(df$end),
    resolution_max = min(resolution_range),
    resolution_min = max(resolution_range),
    keywords = keywords,
    units_df = units_df
  )

}

test_that("`tsl_resample()` works", {

  tsl <- tsl_simulate(
    n = 2,
    rows = 100,
    irregular = TRUE,
    seed = 1
    )

  expect_true(tsl_time_summary(tsl)$resolution_min != tsl_time_summary(tsl)$resolution_max)

  expect_message(
    tsl <- tsl_resample(tsl = tsl)
  ) |>
    suppressMessages()

  expect_equal(tsl_time_summary(tsl)$resolution_min, tsl_time_summary(tsl)$resolution_max)

})

test_that("`tsl_time()` works", {

  tsl <- tsl_simulate(
    n = 3,
    rows = 150,
    time_range = c(Sys.Date() -
    365, Sys.Date()),
    irregular = TRUE
    )

  df <- tsl_time(tsl = tsl)

  expect_equal(class(df), "data.frame")
  expect_equal(nrow(df), length(tsl))
  expect_true(all(df$name ==  tsl_names_get(tsl)))

})

test_that("`tsl_subset()` works", {

  tsl <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  )

  tsl_new <- tsl_subset(
    tsl = tsl, names = c("Sweden", "Germany"),
    colnames = c("rainfall", "temperature"),
    time = c(
      "2010-01-01",
      "2015-01-01"
    )
  )

  expect_true(
    all(tsl_names_get(tsl = tsl_new) == c("Sweden", "Germany"))
    )

  expect_equal(tsl_colnames_get(tsl = tsl_new)$Sweden, c("rainfall", "temperature"))

  expect_equal(tsl_colnames_get(tsl = tsl_new)$Germany, c("rainfall", "temperature"))

  expect_equal(unique(tsl_time(tsl = tsl_new)$begin), as.Date("2010-01-01"))

  expect_equal(unique(tsl_time(tsl = tsl_new)$end), as.Date("2015-01-01"))

})

#' Valid Aggregation Keywords
#'
#' @description
#' Internal function to obtain valid aggregation keywords from a zoo object or a time series list.
#'
#'
#' @param tsl (required, list) Time series list. Default: NULL
#'
#' @return Character string representing a valid aggregation keyword, or "none".
#' @export
#' @autoglobal
#' @examples
#' #one minute time series
#' #-----------------------------------
#' tsl <- tsl_simulate(
#'   time_range = c(
#'     Sys.time() - 60,
#'     Sys.time()
#'   )
#' )
#'
#' #valid keywords for aggregation and/or resampling
#' utils_time_keywords(
#'   tsl = tsl
#' )
#'
#' #10 minutes time series
#' #-----------------------------------
#' tsl <- tsl_simulate(
#'   time_range = c(
#'     Sys.time() - 600,
#'     Sys.time()
#'   )
#' )
#'
#' utils_time_keywords(
#'   tsl = tsl
#' )
#'
#' #10 hours time series
#' #-----------------------------------
#' tsl <- tsl_simulate(
#'   time_range = c(
#'     Sys.time() - 6000,
#'     Sys.time()
#'   )
#' )
#'
#' utils_time_keywords(
#'   tsl = tsl
#' )
#'
#' #10 days time series
#' #-----------------------------------
#' tsl <- tsl_simulate(
#'   time_range = c(
#'     Sys.Date() - 10,
#'     Sys.Date()
#'   )
#' )
#'
#' utils_time_keywords(
#'   tsl = tsl
#' )
#'
#' #10 years time series
#' #-----------------------------------
#' tsl <- tsl_simulate(
#'   time_range = c(
#'     Sys.Date() - 3650,
#'     Sys.Date()
#'   )
#' )
#'
#' utils_time_keywords(
#'   tsl = tsl
#' )
#' @family internal_time_handling
utils_time_keywords <- function(
    tsl = NULL
){

  time_df <- tsl_time(
    tsl = tsl
  )

  keywords <- time_df$keywords |>
    unlist() |>
    unique()

  if(length(keywords) == 0){
    keywords <- NA
  }

  keywords

}

test_that("`utils_time_keywords()` works", {

  tsl <- tsl_simulate(time_range = c(Sys.time() - 60, Sys.time()),
                      seed = 1)

  expect_equal(utils_time_keywords(tsl = tsl), "seconds")

  tsl <- tsl_simulate(time_range = c(Sys.time() - 600, Sys.time()),
                      seed = 1)
  expect_equal(utils_time_keywords(tsl = tsl), c("minutes", "seconds"))

  tsl <- tsl_simulate(time_range = c(Sys.time() - 6000, Sys.time()),
                      seed = 1)
  expect_equal(utils_time_keywords(tsl = tsl), c("hours", "minutes", "seconds"))

  tsl <- tsl_simulate(time_range = c(Sys.Date() - 10, Sys.Date()),
                      seed = 1)
  expect_equal(utils_time_keywords(tsl = tsl), c("weeks", "days"))

  tsl <- tsl_simulate(time_range = c(Sys.Date() - 3650, Sys.Date()),
                      seed = 1)
  expect_true(all(utils_time_keywords(tsl = tsl) %in% c("years", "quarters", "months", "weeks", "decades")))
})

#' Spatial Representation of `momentum()` Data Frames
#'
#' @description
#' Given an sf data frame with geometry types POLYGON, MULTIPOLYGON, or POINT representing time series locations, this function transforms the output of [momentum()], [momentum_ls()], or [momentum_dtw()] to an sf data frame.
#'
#' If `network = TRUE`, the sf data frame is of type LINESTRING, with edges connecting time series locations. This output is helpful for building many-to-many importance maps (see examples).
#'
#' If `network = FALSE`, the sf data frame contains the geometry from the input `sf` argument. This output helps build one-to-many importance maps.
#'
#' @inheritParams momentum_to_wide
#' @inheritParams distantia_spatial
#' @examples
#' tsl <- distantia::tsl_initialize(
#'   x = distantia::eemian_pollen,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#' #reduce size to speed-up example runtime
#' distantia::tsl_subset(
#'   names = 1:3
#'   )
#'
#' df_momentum <- distantia::momentum(
#'   tsl = tsl
#' )
#'
#' #network many to many
#' sf_momentum <- distantia::momentum_spatial(
#'   df = df_momentum,
#'   sf = distantia::eemian_coordinates,
#'   network = TRUE
#' )
#'
#' #network map
#' # mapview::mapview(
#' #   sf_momentum,
#' #   layer.name = "Importance - Abies",
#' #   label = "edge_name",
#' #   zcol = "importance__Abies",
#' #   lwd = 3
#' # ) |>
#' #   suppressWarnings()
#'
#' @return sf data frame (LINESTRING geometry)
#' @export
#' @autoglobal
#' @family momentum_support
momentum_spatial <- distantia_spatial

#' Cumulative Sum of Distances Between Consecutive Cases in a Time Series
#'
#' @description
#' Demonstration function to compute the sum of distances between consecutive cases in a time series
#'
#' @param x (required, zoo object or matrix) univariate or multivariate time series with no NAs. Default: NULL
#' @inheritParams distantia
#'
#' @return numeric value
#' @export
#' @autoglobal
#' @examples
#' #distance metric
#' d <- "euclidean"
#'
#' #simulate zoo time series
#' x <- zoo_simulate(
#'   name = "x",
#'   rows = 100,
#'   seasons = 2,
#'   seed = 1
#' )
#'
#' #sum distance between consecutive samples
#' psi_auto_distance(
#'   x = x,
#'   distance = d
#' )
#' @family psi_demo
psi_auto_distance <- function(
    x = NULL,
    distance = "euclidean"
    ){

  #check x
  x <- utils_check_args_zoo(
    x = x,
    arg_name = "x"
  )

  #check distance
  distance <- utils_check_distance_args(
    distance = distance
  )[1]

  #auto sum
  x.auto_distance <- auto_distance_cpp(
    x = x,
    distance = distance
  )

  x.auto_distance

}

#' Common Boxplot Component of `distantia_boxplot()` and `momentum_boxplot()`
#'
#' @param variable (required, character vector) vector of variable or time series names. Default: NULL
#' @param value (required, numeric vector) numeric values to compute the boxplot. Must have the same length as `variable`. Default: NULL
#' @inheritParams distantia_boxplot
#' @param main (optional, string) boxplot title. Default: NULL
#' @param xlab (optional, string) x axis label. Default: NULL
#' @param ylab (optional, string) y axis label. Default: NULL
#'
#' @return boxplot
#' @export
#' @autoglobal
#' @examples
#' utils_boxplot_common(
#'   variable = rep(x = c("a", "b"), times = 50),
#'   value = stats::runif(100)
#' )
#' @family internal
utils_boxplot_common <- function(
    variable = NULL,
    value = NULL,
    fill_color = NULL,
    f = median,
    main = NULL,
    xlab = NULL,
    ylab = NULL,
    text_cex = 1
    ){

  if(is.null(variable) || is.null(value)){
    stop("distantia::utils_boxplot_common(): arguments 'variable' and 'value' cannot be NULL.", call. = FALSE)
  }

  if(length(variable) != length(value)){
    stop("distantia::utils_boxplot_common(): arguments 'variable' and 'value' must have the same length.", call. = FALSE)
  }

  if(!is.numeric(value)){
    stop("distantia::utils_boxplot_common(): arguments 'value' must be numeric.", call. = FALSE)
  }

  if(!is.character(variable)){
    variable <- as.character(variable)
  }

  if(is.null(fill_color)){

    fill_color <- color_continuous(
      n = length(unique(variable))
    )

  }

  df_plot <- data.frame(
    variable = variable,
    value = value
  )

  #order by function
  variable_means <- tapply(
    X = df_plot$value,
    INDEX = df_plot$variable,
    FUN = f
  )

  variable_order <- sort(
    variable_means,
    decreasing = FALSE
  ) |>
    names()

  df_plot$variable <- factor(
    x = df_plot$variable,
    levels = variable_order
  )

  #notch or not
  notch <- {
    if(min(table(df_plot$variable)) > 30){
      TRUE
    } else {
      FALSE
    }
  }

  #adjusts margin to longer variable name
  left_margin <- (max(nchar(unique(variable))) * text_cex) / 3

  graphics::par(
    mar = c(5, max(left_margin, 5), 4, 2),
    las = 1,
    cex.main = 0.7 * text_cex,
    cex.lab = 0.7 * text_cex,
    cex.axis = 0.6 * text_cex
  )

  graphics::boxplot(
    formula = value ~ variable,
    data = df_plot,
    xlab = xlab,
    ylab = ylab,
    horizontal = TRUE,
    notch = notch,
    col = fill_color,
    main = main
  )



}

#' Convert Dissimilarity Analysis Data Frame to Distance Matrix
#'
#' @description
#' Transforms a data frame resulting from [distantia()] into a dissimilarity matrix.
#'
#'
#' @inheritParams distantia_aggregate
#'
#' @return numeric matrix
#' @export
#' @autoglobal
#' @examples
#' #weekly covid prevalence in three California counties
#' #load as tsl
#' #subset 5 counties
#' #sum by month
#' tsl <- tsl_initialize(
#'   x = covid_prevalence,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_subset(
#'     names = 1:5
#'   ) |>
#'   tsl_aggregate(
#'     new_time = "months",
#'     method = sum
#'   )
#'
#' if(interactive()){
#'
#'   #plotting first three time series
#'   tsl_plot(
#'     tsl = tsl,
#'     guide_columns = 3
#'     )
#'
#'   dev.off()
#'
#' }
#'
#' #dissimilarity analysis
#' #two combinations of arguments
#' distantia_df <- distantia(
#'   tsl = tsl,
#'   lock_step = c(TRUE, FALSE)
#' )
#'
#' #to dissimilarity matrix
#' distantia_matrix <- distantia_matrix(
#'   df = distantia_df
#' )
#'
#' #returns a list of matrices
#' lapply(
#'   X = distantia_matrix,
#'   FUN = class
#'   )
#'
#' #these matrices have attributes tracing how they were generated
#' lapply(
#'   X = distantia_matrix,
#'   FUN = \(x) attributes(x)$distantia_args
#' )
#'
#' #plot matrix
#' if(interactive()){
#'
#'   #plot first matrix (default behavior of utils_matrix_plot())
#'   utils_matrix_plot(
#'     m = distantia_matrix
#'   )
#'
#'   #plot second matrix
#'   utils_matrix_plot(
#'     m = distantia_matrix[[2]]
#'   )
#'
#' }
#' @family distantia_support
distantia_matrix <- function(
    df = NULL
){

  df_type <- attributes(df)$type

  types <- c(
    "distantia_df"
  )

  if(!(df_type %in% types)){
    stop("distantia::distantia_matrix(): argument 'df' must be the output of distantia(), distantia_ls() or distantia_time_warping().", call. = FALSE)
  }

  utils_distantia_df_to_matrix <- function(
    df = NULL,
    x = "x",
    y = "y",
    value = "psi"
  ){

    #subset df
    df <- df_ <- df[, c(x, y, value)]

    #add mirrored pairs
    df_$x <- df$y
    df_$y <- df$x

    #merge everything
    df <- rbind(df, df_)

    #rows and col names
    xy_names <- unique(c(df[[x]], df[[y]]))

    #empty square matrix
    m <- matrix(
      data = NA,
      nrow = length(xy_names),
      ncol = length(xy_names)
    )

    #named vector to map row/column names to indices
    index_map <- stats::setNames(
      object = seq_along(xy_names),
      nm = xy_names
    )

    #vectorized indexing to fill in the matrix
    m[
      cbind(
        index_map[df[[y]]],
        index_map[df[[x]]]
      )
    ] <- df[[value]]

    #dim names
    rownames(m) <- xy_names
    colnames(m) <- xy_names

    #to dist
    m <- m |>
      stats::as.dist() |>
      as.matrix()

    m

  }

  df_list <- utils_distantia_df_split(
    df = df
  )

  m_list <- lapply(
    X = df_list,
    FUN = utils_distantia_df_to_matrix
  )

  #set attributes
  for(i in seq_len(length(m_list))){

    df_i <- df_list[[i]]
    distantia_args <- unique(df_i[, !(colnames(df_i) %in% c("x", "y", "psi"))])

    attr(
      x = m_list[[i]],
      which = "distantia_args"
    ) <- distantia_args

    attr(
      x = m_list[[i]],
      which = "type"
    ) <- "distantia_matrix"

    attr(
      x = m_list[[i]],
      which = "distance"
    ) <- "psi"

  }

  m_list

}

#' Normalized Dissimilarity Score
#'
#' @description
#'
#' Demonstration function to compute the `psi` dissimilarity score (Birks and Gordon 1985). Psi is computed as \eqn{\psi = (2a / b) - 1}, where \eqn{a} is the sum of distances between the relevant samples of two time series, and \eqn{b} is the cumulative sum of distances between consecutive samples in the two time series.
#'
#' When `diagonal = TRUE`, one is added to the result of the equation above.
#'
#' @param a (required, numeric) Result of [psi_cost_path_sum()], the sum of distances of the least cost path between two time series. Default: NULL
#' @param b (required, numeric) Result of [psi_auto_sum()], the cumulative sum of the consecutive cases of two time series. Default: NULL
#' @param diagonal (optional, logical) Used to correct `psi` when diagonals are used during the computation of the least cost path. If the cost matrix and least cost path were computed using `diagonal = TRUE`, this argument should be `TRUE` as well. Default: TRUE
#'
#' @details
#' When `b` is zero (both series are perfectly flat with no variation between
#' consecutive samples), the function returns `NA` because the normalisation
#' denominator is undefined.
#'
#' @return numeric value
#' @examples
#' #distance metric
#' d <- "euclidean"
#'
#' #use diagonals in least cost computations
#' diagonal <- TRUE
#'
#' #simulate two irregular time series
#' x <- zoo_simulate(
#'   name = "x",
#'   rows = 100,
#'   seasons = 2,
#'   seed = 1
#'   )
#'
#' y <- zoo_simulate(
#'   name = "y",
#'   rows = 80,
#'   seasons = 2,
#'   seed = 2
#'   )
#'
#' if(interactive()){
#'   zoo_plot(x = x)
#'   zoo_plot(x = y)
#' }
#'
#' #dynamic time warping
#'
#' #distance matrix
#' dist_matrix <- psi_distance_matrix(
#'   x = x,
#'   y = y,
#'   distance = d
#' )
#'
#' #cost matrix
#' cost_matrix <- psi_cost_matrix(
#'   dist_matrix = dist_matrix,
#'   diagonal = diagonal
#' )
#'
#' #least cost path
#' cost_path <- psi_cost_path(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix,
#'   diagonal = diagonal
#' )
#'
#' if(interactive()){
#'   utils_matrix_plot(
#'     m = cost_matrix,
#'     path = cost_path
#'     )
#' }
#'
#'
#' #computation of psi score
#'
#' #sum of distances in least cost path
#' a <- psi_cost_path_sum(
#'   path = cost_path
#'   )
#'
#' #auto sum of both time series
#' b <- psi_auto_sum(
#'   x = x,
#'   y = y,
#'   distance = d
#' )
#'
#' #dissimilarity score
#' psi_equation(
#'   a = a,
#'   b = b,
#'   diagonal = diagonal
#' )
#'
#' #full computation with distantia()
#' tsl <- list(
#'   x = x,
#'   y = y
#' )
#'
#' distantia(
#'   tsl = tsl,
#'   distance = d,
#'   diagonal = diagonal
#' )$psi
#'
#' if(interactive()){
#'   distantia_dtw_plot(
#'     tsl = tsl,
#'     distance = d,
#'     diagonal = diagonal
#'   )
#' }
#' @autoglobal
#' @export
#' @family psi_demo
psi_equation <- function(
    a = NULL,
    b = NULL,
    diagonal = TRUE
    ){

  if(!is.numeric(a)){
    stop("distantia::psi_equation(): argument 'a' must be numeric.", call. = FALSE)
  }

  if(!is.numeric(b)){
    stop("distantia::psi_equation(): argument 'b' must be numeric.", call. = FALSE)
  }

  if(is.logical(diagonal) == FALSE){
    stop("distantia::psi_equation(): argument 'diagonal' must be logical.", call. = FALSE)
  }

  psi_equation_cpp(
    a,
    b,
    diagonal
  )

}

#' Convert List of Data Frames to List of Zoo Objects
#'
#'
#' @param x (required, list of data frames). Default: NULL.
#' @param time_column (required, column name) Name of the column representing time, if any. Default: NULL.
#' @return A named list of zoo objects.
#' @examples
#' x <- utils_prepare_zoo_list(
#'   x = list(
#'     spain = fagus_dynamics[fagus_dynamics$name == "Spain", ],
#'     sweden = fagus_dynamics[fagus_dynamics$name == "Sweden", ]
#'   ),
#'   time_column = "time"
#' )
#' @autoglobal
#' @export
#' @family internal
utils_prepare_zoo_list <- function(
    x = NULL,
    time_column = NULL
){

  #skip if x has zoo objects
  if(zoo::is.zoo(x)){
    return(x)
  }

  if(is.list(x)){
    if(all(unlist(lapply(x, class)) == "zoo")){
      return(x)
    }
  }


  if(
    utils_check_list_class(
      x = x,
      expected_class = "data.frame"
    ) == FALSE
  ){

    stop("distantia::utils_prepare_zoo_list(): argument 'x' must be a list of data frames.", call. = FALSE)

  }

  #at least 2 elements
  if(length(x) == 1){
    stop("distantia::utils_prepare_zoo_list(): argument 'x' must be a list with at least two elements.", call. = FALSE)
  }

  #elements must be named
  if(any(is.null(names(x)))){
    stop("distantia::utils_prepare_zoo_list(): all elements in the list 'x' must be named.", call. = FALSE)
  }

  # + convert to zoo
  ###################
  x <- lapply(
    X = x,
    FUN = function(i){

      #get index
      i.index <- attributes(i)$index

      if(is.null(i.index) && time_column %in% colnames(i)){

        i.index <- i[[time_column]]
        i[[time_column]] <- NULL

      }

      #to numeric if integer
      if(is.integer(i.index)){
        i.index <- as.numeric(i.index)
      }

      #get numeric columns only
      i <- i[, sapply(i, is.numeric), drop = FALSE]

      #convert to zoo
      i.zoo <- zoo::zoo(
        x = i,
        order.by = i.index
      )

      i.zoo

    }
  )

  tsl_names_set(
    tsl = x
  )

}

test_that("`utils_prepare_zoo_list()` works", {
  x <- utils_prepare_zoo_list(
    x = list(
      spain = fagus_dynamics[fagus_dynamics$name == "Spain", ],
      sweden = fagus_dynamics[fagus_dynamics$name == "Sweden", ]
    ),
    time_column = "time"
  )

  expect_equal(class(x), "list")
  expect_equal(names(x), c("spain", "sweden"))
})

#' Resample Time Series Lists to a New Time
#'
#' @description
#'
#'
#' \strong{Objective}
#'
#' Time series resampling interpolates new values for time steps not available in the original time series. This operation is useful to:
#' \itemize{
#'   \item Transform irregular time series into regular.
#'   \item Align time series with different temporal resolutions.
#'   \item Increase (upsampling) or decrease (downsampling) the temporal resolution of a time series.
#' }
#'
#' Time series resampling \strong{should not be used} to extrapolate new values outside of the original time range of the time series, or to increase the resolution of a time series by a factor of two or more. These operations are known to produce non-sensical results.
#'
#' \strong{Warning}: This function resamples time series lists \strong{with overlapping times}. Please check such overlap by assessing the columns "begin" and "end" of the data frame resulting from `df <- tsl_time(tsl = tsl)`. Resampling will be limited by the shortest time series in your time series list. To resample non-overlapping time series, please subset the individual components of `tsl` one by one either using [tsl_subset()] or the syntax `tsl = my_tsl[[i]]`.
#'
#' \strong{Methods}
#'
#' This function offers three methods for time series interpolation:
#'
#' \itemize{
#'   \item "linear" (default): interpolation via piecewise linear regression as implemented in [zoo::na.approx()].
#'   \item "spline": cubic smoothing spline regression as implemented in [stats::smooth.spline()].
#'   \item "loess": local polynomial regression fitting as implemented in [stats::loess()].
#' }
#'
#' These methods are used to fit models `y ~ x` where `y` represents the values of a univariate time series and `x` represents a numeric version of its time.
#'
#' The functions [utils_optimize_spline()] and [utils_optimize_loess()] are used under the hood to optimize the complexity of the methods "spline" and "loess" by finding the configuration that minimizes the root mean squared error (RMSE) between  observed and predicted `y`. However, when the argument `max_complexity = TRUE`, the complexity optimization is ignored, and a maximum complexity model is used instead.
#'
#' \strong{New time}
#'
#' The argument `new_time` offers several alternatives to help define the new time of the resulting time series:
#'
#' \itemize{
#'   \item `NULL`: the target time series (`x`) is resampled to a regular time within its original time range and number of observations.
#'   \item `zoo object`: a zoo object to be used as template for resampling. Useful when the objective is equalizing the frequency of two separate zoo objects.
#'   \item `time series list`: a time series list to be used as template. The range of overlapping dates and the average resolution are used to generate the new resampling time. This method cannot be used to align two time series lists, unless the template is resampled beforehand.
#'   \item `time vector`: a time vector of a class compatible with the time in `x`.
#'   \item `keyword`: character string defining a resampling keyword, obtained via `zoo_time(x, keywords = "resample")$keywords`.
#'   \item `numeric`: a single number representing the desired interval between consecutive samples in the units of `x` (relevant units can be obtained via `zoo_time(x)$units`).
#' }
#'
#'
#' \strong{Step by Step}
#'
#' The steps to resample a time series list are:
#'
#' \enumerate{
#'   \item The time interpolation range is computed from the intersection of all times in `tsl`. This step ensures that no extrapolation occurs during resampling, but it also makes resampling of non-overlapping time series impossible.
#'   \item If `new_time` is provided, any values of `new_time` outside of the minimum and maximum interpolation times are removed to avoid extrapolation. If `new_time` is not provided, a regular time within the interpolation time range with the length of the shortest time series in `tsl` is generated.
#'   \item For each univariate time series, a model `y ~ x`, where `y` is the time series and `x` is its own time coerced to numeric is fitted.
#'   \itemize{
#'    \item If `max_complexity == FALSE`, the model with the complexity that minimizes the root mean squared error between the observed and predicted `y` is returned.
#'    \item If `max_complexity == TRUE` and `method = "spline"` or `method = "loess"`, the first valid model closest to a maximum complexity is returned.
#'
#'   }
#'   \item The fitted model is predicted over `new_time` to generate the resampled time series.
#' }
#'
#'
#' \strong{Other Details}
#'
#' Please use this operation with care, as there are limits to the amount of resampling that can be done without distorting the data. The safest option is to keep the distance between new time points within the same magnitude of the distance between the old time points.
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param new_time (required, zoo object, time series list, character string, time vector, numeric) New time to resample to. If a time vector is provided, it must be of a class compatible with the time of `tsl`.  If a zoo object or time series list is provided, its time is used as a template to resample `tsl`. Valid resampling keywords (see [tsl_time()]) are allowed. Numeric values are interpreted as interval widths in the time units of the time series. If NULL, irregular time series are predicted into a regular version of their own time. Default: NULL
#' @param method (optional, character string) Name of the method to resample the time series. One of "linear", "spline" or "loess". Default: "linear".
#' @param max_complexity (required, logical). Only relevant for methods "spline" and "loess". If TRUE, model optimization is ignored, and the a model of maximum complexity (an overfitted model) is used for resampling. Default: FALSE
#'
#' @return time series list
#' @export
#' @autoglobal
#' @seealso [zoo_resample()]
#' @examples
#'
#' #generate irregular time series
#' tsl <- tsl_simulate(
#'   n = 2,
#'   rows = 100,
#'   irregular = TRUE
#' )
#'
#' if(interactive()){
#'   tsl_plot(tsl)
#' }
#'
#'
#' #range of times between samples
#' tsl_time_summary(tsl)[
#'   c(
#'     "units",
#'     "resolution_min",
#'     "resolution_max"
#'     )
#'   ]
#'
#' #resample to regular using linear interpolation
#' tsl_regular <- tsl_resample(
#'   tsl = tsl
#' )
#'
#' if(interactive()){
#'   tsl_plot(tsl_regular)
#' }
#'
#' #check new resolution
#' tsl_time_summary(tsl_regular)[
#'   c(
#'     "units",
#'     "resolution_min",
#'     "resolution_max"
#'   )
#' ]
#'
#' #resample using keywords
#'
#' #valid resampling keywords
#' tsl_time_summary(
#'   tsl = tsl,
#'   keywords = "resample"
#' )$keywords
#'
#' #by month
#' tsl_months <- tsl_resample(
#'   tsl = tsl,
#'   new_time = "months"
#' )
#'
#' if(interactive()){
#'   tsl_plot(tsl_months)
#' }
#'
#' #by week
#' tsl_weeks <- tsl_resample(
#'   tsl = tsl,
#'   new_time = "weeks"
#' )
#'
#' if(interactive()){
#'   tsl_plot(tsl_weeks)
#' }
#'
#' #resample using time interval
#'
#' #get relevant units
#' tsl_time(tsl)$units
#'
#' #resampling to 15 days intervals
#' tsl_15_days <- tsl_resample(
#'   tsl = tsl,
#'   new_time = 15 #days
#' )
#'
#' tsl_time_summary(tsl_15_days)[
#'   c(
#'     "units",
#'     "resolution_min",
#'     "resolution_max"
#'   )
#' ]
#'
#' if(interactive()){
#'   tsl_plot(tsl_15_days)
#' }
#'
#' #Aligning two time series lists
#'
#' #two time series lists with different time ranges
#' tsl1 <- tsl_simulate(
#'   n = 2,
#'   rows = 80,
#'   time_range = c("2010-01-01", "2020-01-01"),
#'   irregular = TRUE
#' )
#'
#' tsl2 <- tsl_simulate(
#'   n = 2,
#'   rows = 120,
#'   time_range = c("2005-01-01", "2024-01-01"),
#'   irregular = TRUE
#' )
#'
#' #check time features
#' tsl_time_summary(tsl1)[
#'   c(
#'     "begin",
#'     "end",
#'     "resolution_min",
#'     "resolution_max"
#'   )
#' ]
#'
#' tsl_time_summary(tsl2)[
#'   c(
#'     "begin",
#'     "end",
#'     "resolution_min",
#'     "resolution_max"
#'   )
#' ]
#'
#' #tsl1 to regular
#' tsl1_regular <- tsl_resample(
#'   tsl = tsl1
#' )
#'
#' #tsl2 resampled to time of tsl1_regular
#' tsl2_regular <- tsl_resample(
#'   tsl = tsl2,
#'   new_time = tsl1_regular
#' )
#'
#' #check alignment
#' tsl_time_summary(tsl1_regular)[
#'   c(
#'     "begin",
#'     "end",
#'     "resolution_min",
#'     "resolution_max"
#'   )
#' ]
#'
#' tsl_time_summary(tsl2_regular)[
#'   c(
#'     "begin",
#'     "end",
#'     "resolution_min",
#'     "resolution_max"
#'   )
#' ]
#'
#' @family tsl_processing
tsl_resample <- function(
    tsl = NULL,
    new_time = NULL,
    method = "linear",
    max_complexity = FALSE
){

  old_time <- tsl_time(
    tsl = tsl,
    keywords = "resample"
  )

  #process new_time
  new_time <- utils_new_time(
    tsl = tsl,
    new_time = new_time,
    keywords = "resample"
  )

  #subset to original range
  new_time <- new_time[
    new_time >= min(old_time$begin) &
      new_time <= max(old_time$end)
  ]

  #compare old and new new_time intervals
  time_interval <- as.numeric(mean(diff(new_time)))
  old_time_interval <- as.numeric(mean(old_time$resolution))

  if(
    time_interval < (old_time_interval/10) ||
    time_interval > (old_time_interval*10)
  ){
    warning("The time intervals of 'new_time' and 'x' differ in one order of magnitude or more. The output time series might be highly distorted.")
  }

  #progress bar
  p <- progressr::progressor(along = tsl)

  #resample
  tsl <- future.apply::future_lapply(
    X = tsl,
    FUN = function(x){

      p()

      x <- zoo_resample(
        x = x,
        new_time = new_time,
        method = method,
        max_complexity = max_complexity
      )

    }
  )

  tsl <- tsl_names_set(
    tsl = tsl
  )

  tsl


}

#' Momentum Data Frame to Wide Format
#'
#' @description
#' Transforms a data frame returned by [momentum()] into wide format with the following columns:
#' \itemize{
#'   \item `most_similar`: name of the variable with the highest contribution to similarity (most negative value in the `importance` column) for each pair of time series.
#'   \item `most_dissimilar`: name of the variable with the highest contribution to dissimilarity (most positive value in the `importance` column) for each pair of time series.
#'   \item `importance__variable_name`: contribution to similarity (negative values) or dissimilarity (positive values) of the given variable.
#'   \item `psi_only_with__variable_name`: dissimilarity of the two time series when only using the given variable.
#'   \item `psi_without__variable_name`: dissimilarity of the two time series when removing the given variable.
#' }
#'
#'
#' @inheritParams momentum_aggregate
#' @param sep (required, character string) Separator between the importance metric name and the time series variable name. Default: "__".
#'
#' @return data frame
#' @export
#' @examples
#'
#' tsl <- tsl_initialize(
#'   x = distantia::albatross,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' #importance data frame
#' df <- momentum(
#'   tsl = tsl
#' )
#'
#' df
#'
#' #to wide format
#' df_wide <- momentum_to_wide(
#'   df = df
#' )
#'
#' df_wide
#'
#' @autoglobal
#' @family momentum_support
momentum_to_wide <- function(
    df = NULL,
    sep = "__"
){

  df_type <- attributes(df)$type

  if(!(
    df_type %in% c(
      "momentum_df"
    )
  )
  ){
    stop("distantia::momentum_to_wide(): argument 'df' must be the output of distantia::momentum().", call. = FALSE)
  }

  #aggregate to keep the important columns only
  df <- momentum_aggregate(
    df = df
  )

  #create columns most_similar and most_dissimilar
  select_variable <- function(importance, variable, f = which.max) {
    variable[f(importance)]
  }

  most_dissimilar <- stats::aggregate(
    importance ~ x + y,
    data = df,
    FUN = function(importance) {
      select_variable(
        importance,
        variable = df$variable[df$x == df$x[1] & df$y == df$y[1]],
        f = which.max
      )
    }
  )

  names(most_dissimilar)[3] <- "most_dissimilarity"

  most_similar <- stats::aggregate(
    importance ~ x + y,
    data = df,
    FUN = function(importance) {
      select_variable(
        importance,
        variable = df$variable[df$x == df$x[1] & df$y == df$y[1]],
        f = which.min
      )
    }
  )

  names(most_similar)[3] <- "most_similarity"

  #merge new columns
  new_columns <- merge(
    most_similar,
    most_dissimilar
  )

  # merge with df
  df <- merge(
    df,
    new_columns,
    by = c("x", "y"),
    all.x = TRUE
  )

  columns_psi <- c(
    "x",
    "y",
    "psi",
    "most_similarity",
    "most_dissimilarity"
  )

  columns_variable <- setdiff(
    x = colnames(df),
    y = c(columns_psi, "effect", "variable")
  )

  #fixed part of the data frame
  df_wide_psi <- unique(
    df[, columns_psi]
  )

  #variable to wide
  df_wide_variable <- stats::reshape(
    data = df[, c(
      "x",
      "y",
      "variable",
      columns_variable
      )],
    timevar = "variable",
    idvar = c("x", "y"),
    direction = "wide",
    sep = sep
  )

  #colnames for ordering
  id_columns <- colnames(df_wide_psi)

  importance_columns <- setdiff(
    x = colnames(df_wide_variable),
    y = id_columns
  ) |>
    sort()


  #merge
  df <- merge(
    x = df_wide_psi,
    y = df_wide_variable
  )

  #reorder columns
  df <- df[, c(id_columns, importance_columns)]

  rownames(df) <- NULL

  df

}

#' Auto Sum
#'
#' @description
#' Demonstration function to compute the sum of distances between consecutive samples in two time series
#'
#' @param x (required, zoo object or numeric matrix) univariate or multivariate time series with no NAs. Default: NULL.
#' @param y (required, zoo object or numeric matrix) a time series with the same number of columns as `x` and no NAs. Default: NULL.
#' @inheritParams distantia
#' @return numeric vector
#' @examples
#' #distance metric
#' d <- "euclidean"
#'
#' #simulate two irregular time series
#' x <- zoo_simulate(
#'   name = "x",
#'   rows = 100,
#'   seasons = 2,
#'   seed = 1
#' )
#'
#' y <- zoo_simulate(
#'   name = "y",
#'   rows = 80,
#'   seasons = 2,
#'   seed = 2
#' )
#'
#' if(interactive()){
#'   zoo_plot(x = x)
#'   zoo_plot(x = y)
#' }
#'
#' #auto sum of distances
#' psi_auto_sum(
#'   x = x,
#'   y = y,
#'   distance = d
#' )
#'
#' #same as:
#' x_sum <- psi_auto_distance(
#'   x = x,
#'   distance = d
#' )
#'
#' y_sum <- psi_auto_distance(
#'   x = y,
#'   distance = d
#' )
#'
#' x_sum + y_sum
#'
#' @export
#' @autoglobal
#' @family psi_demo
psi_auto_sum <- function(
    x = NULL,
    y = NULL,
    distance = "euclidean"
){

  #managing distance method
  distance <- utils_check_distance_args(
    distance = distance
  )[1]

  auto_sum_full_cpp(
    y = x,
    x = y,
    distance = distance
  )

}

#' Dynamic Time Warping Dissimilarity Analysis of Time Series Lists
#'
#' @description
#'
#' A minimalistic but slightly faster version of [distantia()] that computes dynamic time warping dissimilarity scores using diagonal least cost paths.
#'
#' @inheritParams distantia
#'
#' @return data frame with columns:
#' \itemize{
#'   \item `x`: time series name.
#'   \item `y`: time series name.
#'   \item `distance`: name of the distance metric.
#'   \item `psi`: psi dissimilarity of the sequences `x` and `y`.
#' }
#' @export
#' @autoglobal
#' @examples
#'
#'#load fagus_dynamics as tsl
#'#global centering and scaling
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl,
#'     guide_columns = 3
#'     )
#' }
#'
#' #dynamic time warping dissimilarity analysis
#' df_dtw <- distantia_dtw(
#'   tsl = tsl,
#'   distance = "euclidean"
#' )
#'
#' df_dtw[, c("x", "y", "psi")]
#'
#' #visualize dynamic time warping
#' if(interactive()){
#'
#'   distantia_dtw_plot(
#'     tsl = tsl[c("Spain", "Sweden")],
#'     distance = "euclidean"
#'   )
#'
#' }
#'
#'
#' @family distantia
distantia_dtw <- function(
    tsl = NULL,
    distance = "euclidean",
    bandwidth = 1
){


  #check input arguments
  args <- utils_check_args_distantia(
    tsl = tsl,
    distance = distance,
    bandwidth = bandwidth
  )

  tsl       <- args$tsl
  distance  <- args$distance[1]
  bandwidth <- args$bandwidth

  df <- utils_tsl_pairs(
    tsl = tsl
  )

  #iterate over pairs of time series
  df$psi <- foreach::foreach(
    i = seq_len(nrow(df)),
    .combine = "c",
    .errorhandling = "pass",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

   psi_dtw_cpp(
     x         = tsl[[df[i, "x"]]],
     y         = tsl[[df[i, "y"]]],
     distance  = distance,
     diagonal  = TRUE,   # DTW uses diagonal moves; psi_equation_cpp adds +1 correction accordingly
     bandwidth = bandwidth
    )

  }

  df <- df[order(df$psi), ]

  #add type
  attr(
    x = df,
    which = "type"
  ) <- "distantia_df"

  df

}

#' Check Input Arguments of `momentum()`
#'
#' @inheritParams momentum
#' @return list.
#' @export
#' @autoglobal
#' @family internal
utils_check_args_momentum <- function(
    tsl = NULL,
    distance = NULL,
    diagonal = NULL,
    bandwidth = NULL,
    lock_step = NULL,
    robust = NULL
){

  # tsl ----
  utils_check_args_tsl(
    tsl = tsl,
    min_length = 2
  )

  ## count and handle NA ----
  na_count <- tsl_count_NA(
    tsl = tsl
  ) |>
    unlist()

  if(sum(na_count) > 0){
    message(
      "distantia::utils_check_args_momentum(): ",
      sum(na_count),
      " NA cases were imputed with distantia::tsl_handle_NA() in these time series: '",
      paste(names(na_count[na_count > 0]), collapse = "', '"),
      "'."
    )
  }

  tsl <- tsl_handle_NA(
    tsl = tsl
  )

  ## subset cols -----
  tsl <- tsl_subset(
    tsl = tsl,
    numeric_cols = TRUE,
    shared_cols = TRUE
  )

  ## diagnose tsl ----
  tsl_diagnose(
    tsl = tsl,
    full = FALSE
  )

  #distance ----
  if(!is.null(distance)){

    distance <- utils_check_distance_args(
      distance = distance
    )

  }


  logical_message <- "must be logical. Accepted values are: TRUE, FALSE, c(TRUE, FALSE)."


  #diagonal ----
  if(!is.null(diagonal)){

    if(any(is.logical(diagonal) == FALSE)){
      stop("distantia::utils_check_args_momentum(): argument 'diagonal'", logical_message, ".", call. = FALSE)
    }

  }

  #bandwidth
  if(!is.null(bandwidth)){

    bandwidth[bandwidth > 1] <- 1
    bandwidth[bandwidth < 0] <- 0
    bandwidth <- unique(bandwidth)

  }

  #lock_step ----
  if(!is.null(lock_step)){

    if(any(is.logical(lock_step) == FALSE)){
      stop("distantia::utils_check_args_momentum(): Argument 'lock_step'", logical_message)
    }

  }


  #robust ----
  if(!is.null(robust)){

    if(any(is.logical(robust) == FALSE)){
      stop("distantia::utils_check_args_momentum(): argument 'robust'", logical_message)
    }
    if(length(robust) > 1){
      robust <- robust[1]
      message("distantia::utils_check_args_momentum(): argument 'robust' may either be TRUE or FALSE. Using the first option available in the input vector: '", robust, "'.")
    }

  }

  list(
    tsl = tsl,
    distance = distance,
    diagonal = diagonal,
    bandwidth = bandwidth,
    lock_step = lock_step,
    robust = robust
  )

}

#' Check Input Arguments of `distantia()`
#'
#' @inheritParams distantia
#' @return list.
#' @export
#' @autoglobal
#' @family internal
utils_check_args_distantia <- function(
    tsl = NULL,
    distance = NULL,
    diagonal = NULL,
    bandwidth = NULL,
    lock_step = NULL,
    repetitions = NULL,
    permutation = NULL,
    block_size = NULL,
    seed = NULL
){

  # tsl ----
  utils_check_args_tsl(
    tsl = tsl,
    min_length = 2
  )

  ## count and handle NA ----
  na_count <- tsl_count_NA(
    tsl = tsl
  ) |>
    unlist()

  if(sum(na_count) > 0){
    message(
      "distantia::utils_check_args_distantia(): ",
      sum(na_count),
      " NA cases were imputed with distantia::tsl_handle_NA() in these time series: '",
      paste(names(na_count[na_count > 0]), collapse = "', '"),
      "'."
    )
  }

  tsl <- tsl_handle_NA(
    tsl = tsl
  )

  ## subset cols -----
  tsl <- tsl_subset(
    tsl = tsl,
    numeric_cols = TRUE,
    shared_cols = TRUE
  )

  ## diagnose tsl ----
  tsl_diagnose(
    tsl = tsl,
    full = FALSE
  )

  #distance ----
  if(!is.null(distance)){

    distance <- utils_check_distance_args(
      distance = distance
    )

  }


  logical_message <- "must be logical. Accepted values are: TRUE, FALSE, c(TRUE, FALSE)."


  #diagonal ----
  if(!is.null(diagonal)){

    if(any(is.logical(diagonal) == FALSE)){
      stop("distantia::utils_check_args_distantia(): argument 'diagonal'", logical_message, ".", call. = FALSE)
    }

  }

  #bandwidth
  if(!is.null(bandwidth)){

    bandwidth[bandwidth > 1] <- 1
    bandwidth[bandwidth < 0] <- 0
    bandwidth <- unique(bandwidth)

  }

  #lock_step ----
  if(!is.null(lock_step)){

    if(any(is.logical(lock_step) == FALSE)){
      stop("distantia::utils_check_args_distantia(): Argument 'lock_step'", logical_message)
    }

  }

  #repetitions ----
  if(is.null(repetitions)) repetitions <- 0L
  if(!is.null(repetitions)){

    repetitions <- as.integer(repetitions)
    if(any(is.integer(repetitions) == FALSE)){
      stop("distantia::utils_check_args_distantia(): argument 'repetitions' must be a integer or a numeric vector.", call. = FALSE)
    }


    ##permutation ----
    if(repetitions > 0){

      if(is.null(permutation)){
        permutation <- "restricted_by_row"
      }

      permutation <- match.arg(
        arg = permutation,
        choices = c(
          "restricted_by_row",
          "restricted",
          "free_by_row",
          "free"
        ),
        several.ok = TRUE
      )

      #block_size ----
      block_size <- utils_block_size(
        tsl = tsl,
        block_size = block_size
      )

      #seed ----
      seed <- as.integer(seed)
      if(any(is.integer(seed) == FALSE)){
        stop("distantia::utils_check_args_distantia(): argument 'seed' must be a integer or a numeric vector.", call. = FALSE)
      }

    } else {

      permutation <- NULL
      block_size <- NULL
      seed <- NULL

    }

  }

  list(
    tsl = tsl,
    distance = distance,
    diagonal = diagonal,
    bandwidth = bandwidth,
    lock_step = lock_step,
    repetitions = repetitions,
    permutation = permutation,
    block_size = block_size,
    seed = seed
  )

}

#' Checks Least Cost Path
#'
#' @param path (required, data frame) least cost path generated with [psi_cost_path()]. This data frame must have the attribute `type == "cost_path"`, and must have been computed from the same time series used to compute the matrix `m`. Default: NULL.
#' @param arg_name (optional, character string) name of the argument being checked. Default: NULL.
#'
#' @return data frame.
#' @export
#' @autoglobal
#' @family internal
utils_check_args_path <- function(
    path = NULL,
    arg_name = "path"
){

  if(is.null(path)){
    stop(
      "distantia::utils_check_args_path() argument ",
      arg_name,
      " must not be NULL.",
      call. = FALSE
    )
  }

  if(
    inherits(x = path, what = "data.frame") == FALSE |
    any(
      is.null(
        attributes(path)[c(
          "y_name",
          "x_name",
          "type",
          "distance"
        )]
      )
    )
  ){
    stop(
      "distantia::utils_check_args_path(): argument ",
      arg_name,
      " must be a data frame resulting from distantia::psi_cost_path().", call. = FALSE
    )
  }

  path

}

#' Checks Input Matrix
#'
#' @param m (required, matrix) distance or cost matrix resulting from [psi_distance_matrix()] or [psi_cost_matrix()]. Default: NULL.
#' @param arg_name (optional, character string) name of the argument being checked. Default: NULL.
#'
#' @return matrix.
#' @export
#' @autoglobal
#' @family internal
utils_check_args_matrix <- function(
    m = NULL,
    arg_name = "m"
){

  if(is.null(m)){
    stop(
      "distantia::utils_check_args_matrix(): argument ",
      arg_name,
      " must not be NULL.",
      call. = FALSE
    )
  }

  if(is.list(m)){

    if(length(m) > 1){
      warning(
        "distantia::utils_check_args_matrix(): Argument ",
        arg_name,
        " is a list with several matrices, but only the first one will be used. Please use the notation m = m[[index]] to select a different one.", call. = FALSE)
    }

    m <- m[[1]]

  }

  if(
    inherits(x = m, what = "matrix") == FALSE){
    stop(
      "distantia::utils_check_args_matrix(): Argument ",
      arg_name,
      " must be of the class 'matrix'.", call. = FALSE)
  }

  m

}



#' Structural Check for Time Series Lists
#'
#' @description
#' Internal function to check that a time series list is a list of zoo objects and has a minimum number of objects. For a more comprehensive test, use [tsl_diagnose()].
#'
#'
#' @param tsl (required, list) list of zoo objects. Default: NULL.
#' @param min_length (required, positive integer) minimum number of zoo objects in `tsl`. Default: 2.
#'
#' @return Error messages, if any.
#' @export
#' @autoglobal
#' @family internal
utils_check_args_tsl <- function(
    tsl = NULL,
    min_length = 2
){

  if(is.null(tsl)){
    stop("distantia::utils_check_args_tsl(): argument 'tsl' must not be NULL.", call. = FALSE)
  }

  #check list class
  if(!is.list(tsl)){
    stop("distantia::utils_check_args_tsl(): argument 'tsl' must be a list.", call. = FALSE)
  }

  #check class of components
  tsl.class <- lapply(
    X = tsl,
    FUN = class
  ) |>
    unlist() |>
    unique()

  if(tsl.class != "zoo"){
    stop("distantia::utils_check_args_tsl(): argument 'tsl' must be a list of zoo objects.", call. = FALSE)
  }

  #check min length
  min_length <- abs(as.integer(min_length))
  if(length(tsl) < min_length){
    stop("distantia::utils_check_args_tsl(): argument 'tsl' must be a time series list with at least ", min_length, " zoo objects.", call. = FALSE)
  }

}

#' Checks Argument x
#'
#' @param x (required, zoo object) zoo time series. Default: NULL.
#' @param arg_name (optional, character string) name of the argument being checked. Default: NULL.
#'
#' @return zoo object.
#' @export
#' @autoglobal
#' @family internal
utils_check_args_zoo <- function(
    x = NULL,
    arg_name = "x"
){

  if(zoo::is.zoo(x) == FALSE){
    stop("distantia::utils_check_args_zoo(): argument ", arg_name, " must be a time series of class 'zoo'.", call. = FALSE)
  }

  x


}

#' Check Distance Argument
#'
#' @param distance (optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset `distances`. Default: "euclidean".
#'
#' @return character vector.
#' @export
#' @autoglobal
#' @examples
#'
#' utils_check_distance_args(
#'   distance = c(
#'     "euclidean",
#'     "euc"
#'    )
#'   )
#' @family internal
utils_check_distance_args <- function(
    distance = NULL
){

  #checking distance values
  distance <- match.arg(
    arg = distance,
    choices = c(
      distances$name,
      distances$abbreviation
    ),
    several.ok = TRUE
  )

  #select rows of distances df
  distances_df <- distances[
    distances$name %in% distance |
      distances$abbreviation %in% distance,
  ]

  if(nrow(distances_df) == 0){
    stop("distantia::utils_check_distance_args(): Please use distance names or abbreviations available in the data frame 'distances'.", call. = FALSE)
  }

  distances_df$name

}

#' Rolling Window Smoothing of Zoo Time Series
#'
#' @description
#' Just a fancy wrapper for [zoo::rollapply()].
#'
#' @param x (required, zoo object) Time series to smooth Default: NULL
#' @param window (optional, integer) Smoothing window width, in number of cases. Default: 3
#' @param f (optional, quoted or unquoted function name) Name of a function to aggregate numeric vectors. Typical examples are `mean`, `max`,`min`, `median`, and `quantile`. Default: `mean`.
#' @param ... (optional, additional arguments) additional arguments to `f`.
#'
#' @return zoo object
#' @autoglobal
#' @export
#' @family zoo_functions
#' @examples
#' x <- zoo_simulate()
#'
#' x_smooth <- zoo_smooth_window(
#'   x = x,
#'   window = 5,
#'   f = mean
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(x_smooth)
#' }
zoo_smooth_window <- function(
    x = NULL,
    window = 3,
    f = mean,
    ...
){

  if(zoo::is.zoo(x) == FALSE){
    stop("distantia::zoo_aggregate(): argument 'x' must be a zoo object.", call. = FALSE)
  }


  if(is.character(f)){

    f <- tryCatch(
      match.fun(f),
      error = function(e) NULL
    )

  }

  if(is.function(f) == FALSE){
    stop("distantia::zoo_aggregate(): Argument 'f' must be a function name. Examples of valid options are: 'mean', 'median', 'max', 'min', and 'sd'.", call. = FALSE)
  }

  #set window to even
  window <- as.integer(window)
  if((window %% 2) == 0){
    window <- window + 1
  }

  y <- zoo::rollapply(
    data = x,
    width = window,
    # fill = rep(x = "extend", times = window),
    fill = "extend",
    FUN = f,
    by.column = TRUE,
    partial = TRUE,
    ... = ...
  )

  zoo_name_set(
    x = y,
    name = attributes(x)$name
  )

}

#' Get Time Series Names from Time Series Lists
#'
#' @description
#'
#' A time series list has two sets of names: the names of the list items (as returned by `names(tsl)`), and the names of the contained zoo objects, as stored in their attribute "name". These names should ideally be the same, for the sake of data consistency. This function extracts either set of names.
#'
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param zoo (optional, logical) If TRUE, the attributes "name" of the zoo objects are returned. Default: TRUE
#'
#' @return list
#' @export
#' @autoglobal
#' @examples
#' #initialize a time series list
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#'
#' #get names of zoo objects
#' tsl_names_get(
#'   tsl = tsl,
#'   zoo = TRUE
#' )
#'
#' #get list names only
#' tsl_names_get(
#'   tsl = tsl,
#'   zoo = FALSE
#'   )
#'
#' #same as
#' names(tsl)
#' @family tsl_management
tsl_names_get <- function(
    tsl = NULL,
    zoo = TRUE
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  out <- names(tsl)

  if(zoo == TRUE){

    out <- sapply(
      X = tsl,
      FUN = function(x){
        y <- attributes(x)$name
        if(is.null(y)){
          return("")
        }
        y
      }
    )

    if(is.list(out)){
      out <- unlist(out)
    }

  }

  out

}


#' Set Time Series Names in a Time Series List
#'
#' @description
#' Sets the names of a time series list and the internal names of the zoo objects inside, stored in their attribute "name".
#'
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param names (optional, character vector) names to set. Must be of the same length of `x`. If NULL, and the list `x` has names, then the names of the zoo objects inside of the list are taken from the names of the list elements.
#'
#' @return time series list
#' @export
#'
#' @examples
#' #simulate time series list
#' tsl <- tsl_simulate(n = 3)
#'
#' #assess validity
#' tsl_diagnose(
#'   tsl = tsl
#' )
#'
#' #list and zoo names (default)
#' tsl_names_get(
#'   tsl = tsl
#' )
#'
#' #list names
#' tsl_names_get(
#'   tsl = tsl,
#'   zoo = FALSE
#' )
#'
#' #renaming list items and zoo objects
#' #------------------------------------
#' tsl <- tsl_names_set(
#'   tsl = tsl,
#'   names = c("X", "Y", "Z")
#' )
#'
#' # check new names
#' tsl_names_get(
#'   tsl = tsl
#' )
#'
#' #fixing naming issues
#' #------------------------------------
#'
#' #creating a invalid time series list
#' names(tsl)[2] <- "B"
#'
#' # check names
#' tsl_names_get(
#'   tsl = tsl
#' )
#'
#' #validate tsl
#' #returns NOT VALID
#' #recommends a solution
#' tsl_diagnose(
#'   tsl = tsl
#' )
#'
#' #fix issue with tsl_names_set()
#' #uses names of zoo objects for the list items
#' tsl <- tsl_names_set(
#'   tsl = tsl
#' )
#'
#' #validate again
#' tsl_diagnose(
#'   tsl = tsl
#' )
#'
#' #list names
#' tsl_names_get(
#'   tsl = tsl
#' )
#' @family tsl_management
tsl_names_set <- function(
    tsl = NULL,
    names = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #function to deduplicate names
  deduplicate <- function(x = NULL){

    if(
      !is.null(x) &&
      sum(duplicated(x)) > 0
    ){

      x <- utils_clean_names(
        x = x
      )

    }

    x

  }

  #user provided names if available
  if(!is.null(names)){

    names <- deduplicate(
      x = as.character(names)
    )

  } else {
    #generate names from input or LETTERS

    #get names from tsl list
    names <- tsl_names_get(
      tsl = tsl,
      zoo = FALSE
    ) |>
      deduplicate()

    #get names from zoo objects
    if(is.null(names)){

      #use zoo names as fallback
      names <- tsl_names_get(
        tsl = tsl,
        zoo = TRUE
      ) |>
        deduplicate()

    }

    #get names from LETTERS
    if(is.null(names)){

      n <- length(tsl)

      #more names than LETTERS
      if(n > length(LETTERS)){

        names <- c(
          LETTERS,
          as.vector(
            outer(
              X = LETTERS,
              Y = LETTERS,
              FUN = paste0
            )
          )
        )[seq_len(n)]

        #fewer names than LETTERS
      } else {

        names <- LETTERS[seq_len(n)]

      }

    }

  }

  #setting names

  #name list elements
  names(tsl) <- names

  #name zoo objects
  tsl <- Map(
    f = function(y, name) {

      zoo_name_set(
        x = y,
        name = name
      )

    },
    tsl,
    names
  )

  tsl

}

#' Clean Time Series Names in a Time Series List
#'
#' @description
#' Combines [utils_clean_names()] and [tsl_names_set()] to help clean, abbreviate, capitalize, and add a suffix or a prefix to time series list names.
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param lowercase (optional, logical) If TRUE, all names are coerced to lowercase. Default: FALSE
#' @param separator (optional, character string) Separator when replacing spaces and dots. Also used to separate `suffix` and `prefix` from the main word. Default: "_".
#' @param capitalize_first (optional, logical) Indicates whether to capitalize the first letter of each name. Default: FALSE.
#' @param capitalize_all (optional, logical) Indicates whether to capitalize all letters of each name. Default: FALSE.
#' @param length (optional, integer) Minimum length of abbreviated names. Names are abbreviated via [abbreviate()]. Default: NULL.
#' @param suffix (optional, character string) Suffix for the clean names. Default: NULL.
#' @param prefix (optional, character string) Prefix for the clean names. Default: NULL.
#'
#' @return time series list
#'
#' @autoglobal
#' @export
#' @examples
#' #initialize time series list
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #original names
#' tsl_names_get(
#'   tsl = tsl
#' )
#'
#' #abbreviate names
#' #---------------------------
#' tsl_clean <- tsl_names_clean(
#'   tsl = tsl,
#'   capitalize_first = TRUE,
#'   length = 4 #abbreviate to 4 characters
#' )
#'
#' #new names
#' tsl_names_get(
#'   tsl = tsl_clean
#' )
#'
#' #suffix and prefix
#' #---------------------------
#' tsl_clean <- tsl_names_clean(
#'   tsl = tsl,
#'   capitalize_all = TRUE,
#'   separator = "_",
#'   suffix = "fagus",
#'   prefix = "country"
#' )
#'
#' #new names
#' tsl_names_get(
#'   tsl = tsl_clean
#' )
#' @family tsl_management
tsl_names_clean <- function(
    tsl = NULL,
    lowercase = FALSE,
    separator = "_",
    capitalize_first = FALSE,
    capitalize_all = FALSE,
    length = NULL,
    suffix = NULL,
    prefix = NULL
){

  zoo_names <- tsl_names_get(
    tsl = tsl,
    zoo = TRUE
  )

  tsl_names <- utils_clean_names(
    x = zoo_names,
    lowercase = lowercase,
    separator = separator,
    capitalize_first = capitalize_first,
    capitalize_all = capitalize_all,
    length = length,
    suffix = suffix,
    prefix = prefix
  )

  names(tsl_names) <- NULL

  tsl <- tsl_names_set(
    tsl = tsl,
    names = tsl_names
  )

  tsl


}


#' Tests Naming Issues in Time Series Lists
#'
#' @param tsl (required, list) Time series list. Default: NULL
#'
#' @return logical
#' @export
#' @autoglobal
#' @examples
#' #creating three zoo time series
#'
#' #one with NA values
#' x <- zoo_simulate(
#'   name = "x",
#'   cols = 1,
#'   na_fraction = 0.1
#'   )
#'
#' #with different number of columns
#' #with repeated name
#' y <- zoo_simulate(
#'   name = "x",
#'   cols = 2
#'   )
#'
#' #with different time class
#' z <- zoo_simulate(
#'   name = "z",
#'   cols = 1,
#'   time_range = c(1, 100)
#'   )
#'
#' #adding a few structural issues
#'
#' #changing the column name of x
#' colnames(x) <- c("b")
#'
#' #converting z to vector
#' z <- zoo::zoo(
#'   x = runif(nrow(z)),
#'   order.by = zoo::index(z)
#' )
#'
#' #storing zoo objects in a list
#' #with mismatched names
#' tsl <- list(
#'   a = x,
#'   b = y,
#'   c = z
#' )
#'
#' #running full diagnose
#' tsl_names_test(
#'   tsl = tsl
#'   )
#' @family tsl_management
tsl_names_test <- function(
    tsl = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #get object names
  names_tsl <- names(tsl)

  names_zoo <- sapply(
    X = tsl,
    FUN = function(x){
      y <- attributes(x)$name
      if(is.null(y)){
        return("")
      }
      y
    }
  )

  #problematic conditions
  names_issues <- c(
    is.null(names_tsl),
    duplicated(names_tsl),
    duplicated(names_zoo),
    names_zoo == "",
    names_zoo != names_tsl
  )

  #test outcome
  #TRUE if no issues
  #FALSE if issues
  !any(names_issues)

}

test_that("`tsl_initialize()` works", {

  #from data frame
  tsl <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  )

  expect_equal(
    unique(sapply(X = tsl, FUN = class)), "zoo"
    )

  expect_true(
    all(tsl_names_get(tsl = tsl, zoo = TRUE) %in% c("Germany", "Spain", "Sweden")
        )
    )

  #from zoo objects
  tsl <- tsl_initialize(
    x = list(
      x = zoo_simulate(seed = 1),
      y = zoo_simulate(seed = 2)
      )
    )

  expect_true(length(tsl) == 2)
  expect_true(all(tsl_names_get(tsl) %in% c("x", "y")))

  #from wide data frame
  df <- stats::reshape(
    data = fagus_dynamics[, c(
      "name", "time",
      "evi"
    )], timevar = "name", idvar = "time", direction = "wide",
    sep = "_"
  )

  tsl <- tsl_initialize(x = df, time_column = "time")

  expect_equal(
    unique(unlist(tsl_colnames_get(tsl))),
    "x"
    )

  #from vector list
  vector_list <- list(
    a = cumsum(stats::rnorm(n = 50)),
    b = cumsum(stats::rnorm(n = 70)),
    c = cumsum(stats::rnorm(n = 20))
  )

  tsl <- tsl_initialize(x = vector_list)

  expect_equal(
    unique(unlist(tsl_colnames_get(tsl))),
    "x"
  )

  #from matrix list
  matrix_list <- list(
    a = matrix(runif(30), nrow = 10, ncol = 3),
    b = matrix(runif(80), nrow = 20, ncol = 4)
  )

  tsl <- tsl_initialize(x = matrix_list)

  expect_equal(
    tsl_colnames_get(tsl = tsl)$a, c("x1", "x2", "x3")
    )

  expect_equal(
    tsl_colnames_get(tsl = tsl)$b, c("x1", "x2", "x3", "x4")
  )

})

test_that("`tsl_initialize()` errors informatively on single-element list", {

  expect_error(
    tsl_initialize(
      x = list(a = zoo_simulate(seed = 1))
    ),
    regexp = "at least 2 time series",
    fixed = FALSE
  )

})

test_that("`tsl_names` works", {

  tsl <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  )

  tsl_clean <- tsl_names_clean(
    tsl = tsl,
    capitalize_first = TRUE,
    length = 4
  )

  expect_true(
    all(tsl_names_get(tsl = tsl_clean) == c("Grmn", "Span", "Swdn")
        )
    )

  tsl <- tsl_names_set(tsl = tsl, names = c("X", "Y", "Z"))

  expect_true(all(tsl_names_get(tsl = tsl) %in% c("X", "Y", "Z")))

})

#' Repair Issues in Time Series Lists
#'
#' @description
#' A Time Series List (`tsl` for short) is a list of zoo time series. This type of object, not defined as a class, is used throughout the `distantia` package to contain time series data ready for processing and analysis.
#'
#' The structure and values of a `tsl` must fulfill several general conditions:
#'
#' Structure:
#' \itemize{
#'   \item The list names match the attributes "name" of the zoo time series.
#'   \item All zoo time series must have at least one shared column name.
#'   \item Data in univariate zoo time series (as extracted by `zoo::coredata(x)`) must be of the class "matrix". Univariate zoo time series are often represented as vectors, but this breaks several subsetting and transformation operations implemented in this package.
#' }
#'
#' Values (optional, when `full = TRUE`):
#' \itemize{
#'   \item All time series have at least one shared numeric column.
#'   \item There are no NA, Inf, or NaN values in the time series.
#' }
#'
#' This function analyzes a `tsl` and tries to fix all possible issues to make it comply with the conditions listed above without any user input. Use with care, as it might modify your data.
#'
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param full (optional, logical) If TRUE, a full repair (structure and values) is triggered. Otherwise, only the data structure is repaired. Default: TRUE
#'
#' @return time series list
#' @export
#' @autoglobal
#' @examples
#' #creating three zoo time series
#'
#' #one with NA values
#' x <- zoo_simulate(
#'   name = "x",
#'   cols = 1,
#'   na_fraction = 0.1
#'   )
#'
#' #with different number of columns
#' #with repeated name
#' y <- zoo_simulate(
#'   name = "x",
#'   cols = 2
#'   )
#'
#' #with different time class
#' z <- zoo_simulate(
#'   name = "z",
#'   cols = 1,
#'   time_range = c(1, 100)
#'   )
#'
#' #adding a few structural issues
#'
#' #changing the column name of x
#' colnames(x) <- c("b")
#'
#' #converting z to vector
#' z <- zoo::zoo(
#'   x = runif(nrow(z)),
#'   order.by = zoo::index(z)
#' )
#'
#' #storing zoo objects in a list
#' #with mismatched names
#' tsl <- list(
#'   a = x,
#'   b = y,
#'   c = z
#' )
#'
#' #running full diagnose
#' tsl_diagnose(
#'   tsl = tsl,
#'   full = TRUE
#'   )
#'
#' tsl <- tsl_repair(tsl)
#' @family tsl_management
tsl_repair <- function(
    tsl = NULL,
    full = TRUE
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  if(!is.logical(full)){
    stop("distantia::tsl_repair(): argument 'full' must be TRUE to run a full repair, and FALSE otherwise.", call. = FALSE)
  }

  if(is.null(tsl)){
    stop("distantia::tsl_repair(): argument 'tsl' must not be NULL.", call. = FALSE)
  }

  if(!is.list(tsl)){
    stop("distantia::tsl_repair(): argument 'tsl' must be a list.", call. = FALSE)
  }

  #all possible issues
  all_issues <- list(
    tsl_objects_zoo = "  - coerced objects in 'tsl' to the class 'zoo'.",
    zoo_no_matrix = "  - converted univariate zoo vectors to matrix.",
    tsl_names_issues =  "  - fixed naming issues.",
    zoo_non_numeric_columns = "  - removed non-numeric columns from time series.",
    zoo_no_colnames = "  - REPAIR FAILED: cannot repair missing column names in zoo time series.",
    zoo_no_shared_columns = "  - REPAIR FAILED: no valid shared column names found across all time series.",
    zoo_duplicated_colnames = "  - deduplicated column names of zoo time series.",
    zoo_shared_columns = "  - removed exclusive columns not shared across time series.",
    zoo_NA_cases = "  - interpolated NA cases in zoo objects with distantia::tsl_handle_NA()."
  )

  # initialize feedback
  issues_structure <- vector()
  issues_values <- vector()

  #structural fixes

  #zoo objects

  # elements in tsl are not zoo objects
  if(
    !all(
      sapply(
        X = tsl,
        FUN = zoo::is.zoo
      )
    )
  ){

    issues_structure <- c(
      issues_structure,
      all_issues[["tsl_objects_zoo"]]
    )

    tsl <- lapply(
      X = tsl,
      FUN = zoo::zoo
    )

  } else {

    #univariate zoo to matrix

    #all zoo objects are matrices
    zoo.is.not.matrix <- sapply(
      X = tsl,
      FUN = function(x) !is.matrix(zoo::coredata(x))
    ) |>
      any()

    if(zoo.is.not.matrix){

      issues_structure <- c(
        issues_structure,
        all_issues[["zoo_no_matrix"]]
      )

      #convert vectors to matrix
      tsl <- lapply(
        X = tsl,
        FUN = zoo_vector_to_matrix
      )

    }

  }

  #naming issues
  if(
    tsl_names_test(tsl = tsl) == FALSE
    ){

    issues_structure <- c(
      issues_structure,
      all_issues[["tsl_names_issues"]]
    )

    tsl <- tsl_names_set(
      tsl = tsl
    ) |>
      suppressMessages()

  }

  # zoo colnames
  zoo_colnames_all <- tsl_colnames_get(
    tsl = tsl,
    names = "all"
  )

  if(any(unlist(zoo_colnames_all) == "unnamed")){

    #cannot repair this automatically
    issues_structure <- c(
      issues_structure,
      all_issues[["zoo_no_colnames"]]
    )

  }

  #test for duplicated colnames
  zoo_colnames_duplicated <- lapply(
    X = zoo_colnames_all,
    FUN = duplicated
  ) |>
    unlist() |>
    any()

  if(zoo_colnames_duplicated){

    issues_structure <- c(
      issues_structure,
      all_issues[["zoo_duplicated_colnames"]]
    )

    tsl <- tsl_colnames_clean(
      tsl = tsl
    )

  }

  zoo_colnames_shared <- tsl_colnames_get(
    tsl = tsl,
    names = "shared"
  )

  #there are no shared columns
  if(
    all(is.na(zoo_colnames_shared))
    ){

    issues_structure <- c(
      issues_structure,
      all_issues[["zoo_no_shared_columns"]]
    )

  }

  zoo_colnames_exclusive <- tsl_colnames_get(
    tsl = tsl,
    names = "exclusive"
  ) |>
    unlist() |>
    unique()

  if(any(!is.na(zoo_colnames_exclusive))){

    issues_structure <- c(
      issues_structure,
      all_issues[["zoo_shared_columns"]]
    )

    tsl <- tsl_subset(
      tsl = tsl,
      numeric_cols = FALSE,
      shared_cols = TRUE
    ) |>
      suppressWarnings()

  }

  #full repair
  if(full == TRUE){

    #all columns in zoo objects are numeric
    zoo.columns.numeric <- sapply(
      X = tsl,
      FUN = function(x){

        apply(
          X = zoo::coredata(x),
          FUN = is.numeric,
          MARGIN = 2
        ) |>
          all()

      }

    ) |>
      all()

    if(zoo.columns.numeric == FALSE){

      issues_values <- c(
        issues_values,
        all_issues[["zoo_non_numeric_columns"]]
      )

      tsl <- tsl_subset(
        tsl = tsl,
        numeric_cols = TRUE,
        shared_cols = FALSE
      )

    }

    # NA values
    na.count <- tsl_count_NA(
      tsl = tsl
    ) |>
      unlist() |>
      sum()

    if(na.count > 0){

      issues_values <- c(
        issues_values,
        all_issues[["zoo_NA_cases"]]
      )

      tsl <- tsl_handle_NA(
        tsl = tsl,
        na_action = "impute"
      )

    }

  }

  #end of full == TRUE



  if(length(issues_structure) > 0){

    message(
      "distantia::tsl_repair(): repairs in TSL structure:\n",
      "--------------------------------------------------\n\n",
      paste(
        issues_structure,
        collapse = "\n\n"
      )
    )

  }

  if(length(issues_values) > 0){

    message(
      "\ndistantia::tsl_repair(): repairs in TSL values:\n",
      "-------------------------------------------------\n\n",
      paste(
        issues_values,
        collapse = "\n\n"
      )
    )

  }

  message("\n\n")

  tsl_diagnose(
    tsl = tsl,
    full = TRUE
  )

  tsl

}

#' Simulate a Zoo Time Series
#'
#' @description
#' Generates simulated zoo time series.
#'
#' @param name (optional, character string) Name of the zoo object, to be stored in the attribute "name". Default: "A"
#' @param cols (optional, integer) Number of time series. Default: 5
#' @param rows (optional, integer) Length of the time series. Minimum is 10, but maximum is not limited. Very large numbers might crash the R session. Default: 100
#' @param time_range (optional character or numeric vector) Interval of the time series. Either a character vector with dates in format YYYY-MM-DD or or a numeric vector. If there is a mismatch between `time_range` and `rows` (for example, the number of days in `time_range` is smaller than `rows`), the upper value in `time_range` is adapted to `rows`. Default: c("2010-01-01", "2020-01-01")
#' @param data_range (optional, numeric vector of length 2) Extremes of the simulated time series values. The simulated time series are independently adjusted to random values within the provided range. Default: c(0, 1)
#' @param seasons (optional, integer) Number of seasons in the resulting time series. The maximum number of seasons is computed as `floor(rows/3)`. Default: 0
#' @param na_fraction (optional, numeric) Value between 0 and 0.5 indicating the approximate fraction of NA data in the simulated time series. Default: 0.
#' @param independent (optional, logical) If TRUE, each new column in a simulated time series is averaged with the previous column. Irrelevant when `cols <= 2`, and hard to perceive in the output when `seasons > 0`. Default: FALSE
#' @param irregular (optional, logical) If TRUE, the time series is created with 20 percent more rows, and a random 20 percent of rows are removed at random. Default: TRUE
#' @param seed (optional, integer) Random seed used to simulate the zoo object. Default: NULL
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#'
#' #generates a different time series on each execution when 'seed = NULL'
#' x <- zoo_simulate()
#'
#' #returns a zoo object
#' class(x)
#'
#' #time series names are uppercase letters
#' #this attribute is not defined in the zoo class and might be lost during data transformations
#' attributes(x)$name
#'
#' #column names are lowercase letters
#' names(x)
#'
#' #plotting methods
#' if(interactive()){
#'
#'   #plot time series with default zoo method
#'   plot(x)
#'
#'   #plot time series with distantia
#'   zoo_plot(
#'     x = x,
#'     xlab = "Date",
#'     ylab = "Value",
#'     title = "My time series"
#'   )
#'
#' }
#' @family simulate_time_series
zoo_simulate <- function(
    name = "A",
    cols = 5,
    rows = 100,
    time_range = c("2010-01-01", "2020-01-01"),
    data_range = c(0, 1),
    seasons = 0,
    na_fraction = 0,
    independent = FALSE,
    irregular = TRUE,
    seed = NULL
){

  # columns ----
  cols <- as.integer(cols[1])

  if(cols > length(letters)){

    cols.names <- c(
      letters,
      as.vector(
        outer(
          X = letters,
          Y = letters,
          FUN = paste0
        )
      )
    )[seq_len(cols)]

  } else {

    cols.names <- letters[seq_len(cols)]

  }

  # rows ----
  rows <- as.integer(rows[1])

  rows <- ifelse(
    test = rows < 10,
    yes = 10,
    no = rows
  )

  # seed ----
  if(is.null(seed)){
    seed <- sample.int(
      n = .Machine$integer.max,
      size = 1
    )
  }
  set.seed(as.integer(seed[1]))

  # data_range ----
  if(length(data_range) == 1 || length(unique(data_range)) == 1){
    stop("distantia::zoo_simulate(): argument 'data_range' must have two different values.", call. = FALSE)
  }
  if(!is.numeric(data_range)){
    data_range <- c(0, 1)
  }

  data_range_min <- min(data_range)
  data_range_max <- max(data_range)


  # time range ----
  if(length(time_range) != 2){
    stop("distantia::zoo_simulate(): argument 'time_range' must be a vector of length 2.", call. = FALSE)
  }

  time_range <- utils_as_time(
    x = time_range
  ) |>
    range()

  # irregular ----
  if(!is.logical(irregular)){
    irregular <- FALSE
  }

  #generating regular or irregular time vector
  time_vector <- seq(
    from = min(time_range),
    to = max(time_range),
    length.out = rows * ifelse(
      test = irregular == TRUE,
      yes = 2,
      no = 1
    )
  ) |>
    sample(size = rows) |>
    sort()

  # na_fraction ----
  na_fraction <- as.numeric(na_fraction[1])
  if(na_fraction < 0){
    na_fraction <- 0
  }
  if(na_fraction > 0.5){
    na_fraction <- 0.5
  }

  # independent ----
  if(!is.logical(independent)){
    independent <- FALSE
  }

  # create empty matrix
  m <- m_copy <- matrix(
    data = NA,
    nrow = rows,
    ncol = cols,
    dimnames = list(
      as.character(time_vector),
      cols.names
    )
  )

  #distribution of sd for rnorm
  sd_vector <- seq(
    from = ifelse(
      test = independent == TRUE,
      yes = 0.01,
      no = 0.5
    ),
    to = 1,
    by = 0.01
  )

  #generate template
  seasons <- min(
    floor(rows / 3),
    as.integer(abs(seasons))
    )

  template <- if(seasons > 0){
    #seasonal template
    sin(x = 2 * pi * seq_len(rows) / (rows / seasons))
  } else {
    #neutral template
    rep(x = 0, times = rows)
  }

  #iterate over columns of the empty matrix
  for(i in seq_len(ncol(m))){

    #cumulative sum of random values from a normal distribution
    values.i <- stats::rnorm(
      n = rows,
      sd = sample(
        x = sd_vector,
        size = 1
      )
    ) |>
      cumsum()

    #add to template
    values.i <- values.i + (template * 2)

    #store them in unaltered copy for next iteration
    m_copy[, i] <- values.i

    #create dependence with previous values
    if(independent == FALSE && i > 1){

      #create dependency
      values.i <- (values.i + m_copy[, i - 1]) / 2

    }

    # add seasonal component
    if(seasons > 0) {

      values.i <- values.i +
        sin(
          x = 2 * pi * seq_len(rows) / (rows / seasons)
        )

    }

    #rescale to data_range
    #but adding some random variability
    #minimum within the 1/third of the lower range values
    #maximum within the 1/third of the upper range values
    values.i <- utils_rescale_vector(
      x = values.i,
      new_min = stats::runif(
        n = 1,
        min = data_range_min,
        max = data_range_max/3
      ),
      new_max = stats::runif(
        n = 1,
        min = data_range_max - (data_range_max/3),
        max = data_range_max
      ),
    )

    #save values in matrix column
    m[, i] <- values.i

  }

  #apply na_fraction
  m[
    sample.int(
      n = length(m),
      size = floor(length(m) * na_fraction)
    )
  ] <- NA

  #to zoo
  m <- zoo::zoo(
    x = m,
    order.by = time_vector
  )

  #set name
  m <- zoo_name_set(
    x = m,
    name = as.character(name)
  )

  m

}

test_that("`zoo_to_tsl()` works", {
  x <- zoo_simulate(seed = 1)

  tsl <- zoo_to_tsl(x = x)
  expect_equal(class(tsl), "list")
  expect_equal(class(tsl[[1]]), "zoo")
  expect_equal(names(tsl), "A")
  expect_equal(attributes(tsl[[1]])$name, "A")
})

test_that("`zoo_permute()` works", {

  x <- zoo_simulate(cols = 2, seed = 1)

  x_free <- zoo_permute(x = x, permutation = "free", repetitions = 2)

  expect_equal(
    length(x_free), 2
  )

  expect_equal(
    nrow(x), nrow(x_free[[1]])
  )

})

test_that("`utils_optimize_spline()` works", {

  xy <- zoo_simulate(cols = 1, rows = 30, seed = 1)

  m <- utils_optimize_spline(
    x = as.numeric(zoo::index(xy)),
    y = xy[,1]
  )

  expect_equal(class(m), "smooth.spline")

})

test_that("`zoo_resample()` works", {

  x <- zoo_simulate(cols = 2, rows = 50, time_range = c(
    "2010-01-01",
    "2020-01-01"
  ), irregular = TRUE,
  seed = 1
  )

  x_intervals <- diff(zoo::index(x))

  expect_true(length(unique(x_intervals)) > 1)

  new_time <- seq.Date(
    from = min(zoo::index(x)), to = max(zoo::index(x)),
    by = floor(min(x_intervals))
  )

  x_linear <- zoo_resample(x = x, new_time = new_time, method = "linear")

  expect_equal(
    new_time, zoo::index(x_linear)
  )

  x_spline <- zoo_resample(x = x, new_time = new_time, method = "spline")

  expect_equal(
    new_time, zoo::index(x_spline)
  )

  x_loess <- zoo_resample(x = x, new_time = new_time, method = "loess")

  expect_equal(
    new_time, zoo::index(x_loess)
  )


})

# The default formats follow the rules of the ISO 8601 international standard which expresses a day as "2001-02-03".

#' Ensure Correct Class for Time Arguments
#'
#' Guesses the class of a vector based on its elements. This function can handle numeric vectors, character vectors that can be coerced to either "Date" or "POSIXct" classes, and vectors already in "Date" or "POSIXct" classes.
#'
#' @param x (required, vector) Vectors of the classes 'numeric', 'Date', and 'POSIXct' are valid and returned without any transformation. Character vectors are analyzed to determine their more probable type, and are coerced to 'Date' or 'POSIXct' depending on their number of elements. Generally, any character vector representing an ISO 8601 standard, like "YYYY-MM-DD" or "YYYY-MM-DD HH:MM:SS" will be converted to a valid class. If a character vector cannot be coerced to date, it is returned as is. Default: NULL
#' @param to_class (optional, class) Options are: NULL, "numeric", "Date", and "POSIXct". If NULL, 'x' is returned as the most appropriate time class. Otherwise, 'x' is coerced to the given class. Default: NULL
#'
#' @return time vector
#'
#' @examples
#' # numeric
#' utils_as_time(
#'   x = c(-123120, 1200)
#'   )
#'
#' # character string to Date
#' utils_as_time(
#'   x = c("2022-03-17", "2024-02-05")
#'   )
#'
#' # incomplete character strings to Date
#' utils_as_time(
#'   x = c("2022", "2024")
#'   )
#'
#' utils_as_time(
#'   x = c("2022-02", "2024-03")
#'   )
#'
#' # character string to POSIXct
#' utils_as_time(
#'   x = c("2022-03-17 12:30:45", "2024-02-05 11:15:45")
#'   )
#'
#' # Date vector (returns the input)
#' utils_as_time(
#'   x = as.Date(c("2022-03-17", "2024-02-05"))
#'   )
#'
#' # POSIXct vector (returns the input)
#' utils_as_time(
#'   x = as.POSIXct(c("2022-03-17 12:30:45", "2024-02-05 11:15:45"))
#'   )
#'
#' @export
#' @autoglobal
#' @family internal_time_handling
utils_as_time <- function(
    x = NULL,
    to_class = NULL
){


  if(is.null(to_class)){
    to_class <- ""
  }

  #x is valid time class
  if(utils_is_time(x) == TRUE){

    if(to_class == "Date"){
      x <- as.Date(x)
    }

    if(to_class == "POSIXct"){
      x <- as.POSIXct(x)
    }

    if(to_class == "numeric"){
      x <- as.numeric(x)
    }

    return(x)
  }

  #coerce factors to character
  if(is.factor(x)){
    x <- as.character(x)
  }

  #valid time class
  time_class <- sapply(
    X = x,
    FUN = base::class
  ) |>
    unique()

  # x is character
  if(time_class == "character"){

    #slash to hyphen
    x <- gsub(
      pattern = "/",
      replacement = "-",
      x = x
    )

    #count hyphens and colons
    x_hyphens <- lengths(
      regmatches(
        x = x,
        m = gregexpr(
          pattern = "-",
          text = x
          )
        )
      ) |>
      unique() |>
      max()


    x_colons <- lengths(
      regmatches(
        x = x,
        m = gregexpr(
          pattern = ":",
          text = x
        )
      )
    ) |>
      unique() |>
      max()

    #Date and POSIXct
    if(x_hyphens == 2){

      #POSIXct
      if(x_colons > 1){

        x <- as.POSIXct(
          x = x,
          tryFormats = c(
            "%Y-%m-%d %H:%M:%S",
            "%Y-%m-%d %H:%M:%OS",
            "%Y-%m-%d %H:%M"
          )
        )

      } else {

        #converting to Date
        x <- as.Date(
          x = x,
          format = "%Y-%m-%d"
        )

      }


    } else {

      warning(
        "Argument 'x' of class 'character' must have the format '%Y-%m-%d' for class 'Date', or '%Y-%m-%d %H:%M:%S' for class 'POSIXct'",
        call. = FALSE
      )

    }

  }

  if(to_class == "Date"){
    x <- as.Date(x)
  }

  if(to_class == "POSIXct"){
    x <- as.POSIXct(x)
  }

  if(to_class == "numeric"){
    x <- as.numeric(x)
  }

  x

}

#' Color Guide for Matrix Plot
#'
#' @description
#' Plot a color legend for a distance or cost matrix for multi-panel plots or external image editors.
#'
#' @param m (required, numeric matrix) distance or cost matrix generated by [psi_distance_matrix()] or [psi_cost_matrix()], but any numeric matrix will work. Default: NULL
#' @param matrix_color (optional, character vector) vector of colors. Default: NULL
#' @param breaks (optional, numeric vector) vector of breaks for the color guide. Default: NULL
#' @param title (optional, character string) guide title. Default: NULL
#' @param text_cex (optional, numeric) multiplier for the text size. Default: 1
#' @return Plot
#' @examples
#' #prepare time series list
#' tsl <- tsl_simulate(
#'   n = 2,
#'   independent = TRUE
#' )
#'
#' #distance matrix between time series
#' dm <- psi_distance_matrix(
#'   x = tsl[[1]],
#'   y = tsl[[2]]
#' )
#'
#' if(interactive()){
#'   utils_matrix_guide(m = dm)
#' }
#' @export
#' @autoglobal
#' @family internal_plotting
utils_matrix_guide <- function(
    m = NULL,
    matrix_color = NULL,
    breaks = NULL,
    title = NULL,
    text_cex = 1
){

  #check m
  m <- utils_check_args_matrix(m = m)

  title_cex <- 0.8 * text_cex
  axis_labels_cex <- 0.7 * text_cex
  title_distance <- 0.75

  if(inherits(x = m, what = "matrix") == FALSE){
    stop("distantia::utils_matrix_guide(): argument 'm' must be a distance or cost matrix resulting from distantia::psi_distance_matrix() or distantia::psi_cost_matrix().", call. = FALSE)
  }

  if(is.null(matrix_color)){
    matrix_color = color_continuous(
      n = 100
    )
  }

  if(is.null(breaks)){
    breaks <- utils_color_breaks(
      m = m,
      n = length(matrix_color)
    )
  }

  #guide values
  guide_values <- (
    breaks[1:(length(breaks)-1)] +
      breaks[2:length(breaks)]
  )/2

  #guide title
  if(is.null(title)){

    distance <- attributes(m)$distance

    if(is.character(distance)){
      distance <- tools::toTitleCase(distance)
    }

    type <- attributes(m)$type

    if("distance" %in% type){

      title <- ifelse(
        test = is.null(distance),
        yes = "Distance",
        no = paste(distance , "\ndistance")
      )

    }

    if("cost" %in% type){

      title <- ifelse(
        test = is.null(distance),
        yes = "Cost",
        no = paste(distance , "\ncost")
      )

    }

  }

  graphics::image(
    x = 1:2,
    y = breaks,
    z = matrix(
      guide_values,
      nrow = 1,
      ncol = length(guide_values)
    ),
    xaxt = "n",
    yaxt = "n",
    xlab = "",
    ylab = "",
    col = matrix_color,
    breaks = breaks
  )

  graphics::title(
    main = title,
    cex.main = title_cex,
    line = title_distance,
    adj = 0
  )

  graphics::axis(
    side = 4,
    las = 2,
    cex.axis = axis_labels_cex
  )

}

#' Handles Time Column in a List of Data Frames
#'
#' @param x (required, named list of data frames). Default: NULL.
#' @param time_column (optional if `lock_step = FALSE`, and required otherwise, column name) Name of numeric column representing time. Default: NULL.
#' @param lock_step (optional, logical) If `TRUE`, all input sequences are subset to their common times according to the values in the `time_column`. Default: `FALSE`.
#'
#' @return List of data frames.
#' @export
#' @autoglobal
#' @family internal
utils_prepare_time <- function(
    x = NULL,
    time_column = NULL,
    lock_step = FALSE
){



  #skip if x has zoo objects
  if(zoo::is.zoo(x)){
    return(x)
  }

  if(is.list(x)){
    if(all(unlist(lapply(x, class)) == "zoo")){
      return(x)
    }
  }

  #if no time column, add "row_id"
  if(is.null(time_column)){

    if(lock_step == TRUE){
      stop("distantia::utils_prepare_time(): argument 'time_column' cannot be NULL when 'lock_step' is TRUE'.", call. = FALSE)
    }

    time_column <- "row_id"

    x <- lapply(
      X = x,
      FUN = function(x){
        x[[time_column]] <- seq_len(nrow(x))
        return(x)
      }
    )

  }

  #check that the time column is numeric
  x.time.numeric <- lapply(
    X = x,
    FUN = function(x) is.numeric(x[[time_column]])
  ) |>
    unlist() |>
    unique()

  #check if it can be coerced to numeric
  if(all(x.time.numeric == FALSE)){

    x.time.numeric <- lapply(
      X = x,
      FUN = function(x) {
        x[[time_column]] <- as.numeric(x[[time_column]])
        is.numeric(x[[time_column]])
      }
    ) |>
      unlist()

  }

  #names of elements with no time
  x.time.numeric <- names(x.time.numeric[!x.time.numeric])

  #ERROR if time missing from any element
  if(length(x.time.numeric) > 0){

    stop("distantia::utils_prepare_time(): The time column '", time_column, "' is not numeric in these elements of 'x': ", paste(x.time.numeric, collapse = ", "), call. = FALSE)

    #check that time column is in all elements
    x.no.time <- lapply(
      X = x,
      FUN = function(x){
        time_column %in% colnames(x)
      }
    ) |>
      unlist()

    #names of elements with no time
    x.no.time <- names(x.no.time[!x.no.time])

    #ERROR if time missing from any element
    if(length(x.no.time) > 0){

      stop("distantia::utils_prepare_time(): these time series in 'x' do not have the time column '", time_column, "': ", paste(x.no.time, collapse = ", "), call. = FALSE)

    }

    #arrange all elements by time
    x <- lapply(
      X = x,
      FUN = function(x){
        x[order(x[[time_column]]), ]
      }
    )

    #apply lock step
    if(lock_step == TRUE){

      times <- lapply(
        X = x,
        FUN = function(x) x[[time_column]]
      ) |>
        unlist() |>
        table() |>
        as.data.frame(stringsAsFactors = FALSE)

      time_common <- times[times$Freq == length(x), "Var1"]

      times_common <- as.numeric(names(times)[times == length(x)])

      x <- lapply(
        X = x,
        FUN = function(x) x[x[[time_column]] %in% times_common, ]
      )

    }

  }

  #fix duplicated times
  for(i in seq_len(length((x)))){

    x.i <- x[[i]]
    x.i.name <- names(x)[i]
    x.i.time <- x.i[[time_column]]
    x.i.time.dupes <- duplicated(x.i.time)

    if(any(x.i.time.dupes == TRUE)){

      #interpolate duplicated indices
      x.i.time.new <- x.i.time
      x.i.time.new[x.i.time.dupes] <- NA
      x.i.time.new <- zoo::na.approx(
        object = x.i.time.new,
        na.rm = FALSE
        )

      message(
        "distantia::utils_prepare_time():  duplicated time indices in '",
        x.i.name,
        "':\n",
        paste0(
          "- value ",
          x.i.time[x.i.time.dupes],
          " replaced with ",
          x.i.time.new[x.i.time.dupes],
          collapse = "\n"
        ),
        "."
      )

      x.i[[time_column]] <- x.i.time.new

      x[[i]] <- x.i

    }

  }

  #set time attribute
  #remove the time column
  x <- lapply(
    X = x,
    FUN = function(x){
      attr(x = x, which = "index") <- x[[time_column]]
      x[[time_column]] <- NULL
      return(x)
    }
  )

  x

}

#' Default Continuous Color Palette
#'
#' @description
#' Uses the function [grDevices::hcl.colors()] to generate a continuous color palette.
#'
#'
#' @param n (required, integer) number of colors to generate. Default = NULL
#' @param palette (required, character string) Argument `palette` of [grDevices::hcl.colors()]. Default: "Zissou 1"
#' @param rev (optional, logical) If TRUE, the color palette is reversed. Default: FALSE
#' @return color vector
#' @export
#' @autoglobal
#' @examples
#'
#' color_continuous(n = 20)
#'
#' @family internal_plotting
color_continuous <- function(
    n = 5,
    palette = "Zissou 1",
    rev = FALSE
){

  grDevices::hcl.colors(
    n = n,
    palette = palette,
    rev = rev
  )

}

#' Random or Restricted Permutation of Zoo Time Series
#'
#' @description
#' Fast permutation of zoo time series for null model testing using a fast and efficient C++ implementations of different restricted and free permutation methods.
#'
#' The available permutation methods are:
#'
#' \itemize{
#'   \item "free" (see [permute_free_cpp()]): Unrestricted and independent re-shuffling of individual cases across rows and columns. Individual values are relocated to a new row and column within the dimensions of the original matrix.
#' \item "free_by_row" (see [permute_free_by_row_cpp()]): Unrestricted re-shuffling of complete rows. Each individual row is given a new random row number, and the data  matrix is re-ordered accordingly.
#' \item "restricted" (see [permute_restricted_cpp()]): Data re-shuffling across rows and columns is restricted to blocks of contiguous rows. The algorithm divides the data matrix into a set of blocks of contiguous rows, and individual cases are then assigned to a new row and column within their original block.
#' \item "restricted_by_row" (see [permute_restricted_by_row_cpp()]): Re-shuffling of complete rows is restricted to blocks of contiguous rows. The algorithm divides the data matrix into a set of blocks of contiguous rows, each individual row is given a new random row number within its original block, and the block is reordered accordingly to generate the permuted output.
#' }
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @param x (required, zoo object) zoo time series. Default: NULL
#' @param repetitions (optional, integer) number of permutations to compute. Large numbers may compromise your R session. Default: 1
#' @param permutation (optional, character string) permutation method. Valid values are listed below from higher to lower induced randomness:
#' \itemize{
#'   \item "free": unrestricted re-shuffling of individual cases across rows and columns. Ignores block_size.
#'   \item "free_by_row": unrestricted re-shuffling of complete rows. Ignores block size.
#'   \item "restricted": restricted shuffling across rows and columns within blocks of rows.
#'   \item "restricted_by_row": restricted re-shuffling of rows within blocks.
#' }
#' @param block_size (optional, integer) Block size in number of rows for restricted permutations. Only relevant when permutation methods are "restricted" or "restricted_by_row". A block of size `n` indicates that the original data is pre-divided into blocks of such size, and a given row can only be permuted within their original block. If NULL, defaults to the rounded one tenth of the number of rows in `x`. Default: NULL.
#' @param seed (optional, integer) initial random seed to use during permutations. Default: 1
#'
#' @return Time Series List
#' @export
#' @autoglobal
#' @examples
#'
#' #simulate zoo time series
#' x <- zoo_simulate(cols = 2)
#'
#' if(interactive()){
#'   zoo_plot(x)
#' }
#'
#' #free
#' x_free <- zoo_permute(
#'   x = x,
#'   permutation = "free",
#'   repetitions = 2
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = x_free,
#'     guide = FALSE
#'     )
#' }
#'
#' #free by row
#' x_free_by_row <- zoo_permute(
#'   x = x,
#'   permutation = "free_by_row",
#'   repetitions = 2
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = x_free_by_row,
#'     guide = FALSE
#'   )
#' }
#'
#' #restricted
#' x_restricted <- zoo_permute(
#'   x = x,
#'   permutation = "restricted",
#'   repetitions = 2
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = x_restricted,
#'     guide = FALSE
#'   )
#' }
#'
#' #restricted by row
#' x_restricted_by_row <- zoo_permute(
#'   x = x,
#'   permutation = "restricted_by_row",
#'   repetitions = 2
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = x_restricted_by_row,
#'     guide = FALSE
#'   )
#' }
#'
#' @family zoo_functions
zoo_permute <- function(
    x = NULL,
    repetitions = 1L,
    permutation = "restricted_by_row",
    block_size = NULL,
    seed = 1L
){

  #check x
  x <- utils_check_args_zoo(x = x)

  permutation <- match.arg(
    arg = permutation,
    choices = c(
      "restricted_by_row",
      "restricted",
      "free_by_row",
      "free"
    ),
    several.ok = FALSE
  )

  if (permutation == "restricted_by_row") {
    f <- permute_restricted_by_row_cpp
  } else if (permutation == "free_by_row") {
    f <- permute_free_by_row_cpp
  } else if (permutation == "restricted") {
    f <- permute_restricted_cpp
  } else if (permutation == "free") {
    f <- permute_free_cpp
  }

  block_size <- utils_block_size(
    tsl = list(x),
    block_size = block_size
  )

  repetitions <- seq_len(repetitions)

  p <- progressr::progressor(along = repetitions)

  permutations <- foreach::foreach(
    i = repetitions,
    .errorhandling = "pass",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

    p()

    x_name <- attributes(x)$name

    x_permuted <- f(
      x = as.matrix(x),
      block_size = block_size,
      seed = seed + i
    )

    x_permuted <- zoo::zoo(
      x = x_permuted,
      order.by = zoo::index(x)
    )

    x_permuted <- zoo_name_set(
      x = x_permuted,
      name = paste0(
        x_name,
        "_",
        as.character(i)
      )
    )

    return(x_permuted)

  }

  if(is.list(permutations) == FALSE){

    permutations <- list(permutations)

  }

  names(permutations) <- as.character(repetitions)

  # permutations <- tsl_names_set(
  #   tsl = permutations
  # )

  permutations

}

#' Smoothing of Time Series Lists
#'
#' @description
#' Rolling-window and exponential smoothing of Time Series Lists.
#'
#' Rolling-window smoothing computes a statistic over a fixed-width window of consecutive cases and replaces each central value with the computed statistic. It is commonly used to mitigate noise in high-frequency time series.
#'
#' Exponential smoothing computes each value as the weighted average of the current value and past smoothed values. This method is useful for reducing noise in time series data while preserving the overall trend.
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @inheritParams zoo_smooth_window
#' @param alpha (required, numeric) Exponential smoothing factor in the range (0, 1]. Determines the weight of the current value relative to past values. If not NULL, the arguments `window` and `f` are ignored, and exponential smoothing is performed instead. Default: NULL
#'
#' @return time series list
#' @export
#' @autoglobal
#' @family tsl_processing
#' @examples
#'
#'
#' tsl <- tsl_simulate(n = 2)
#'
#' #rolling window smoothing
#' tsl_smooth <- tsl_smooth(
#'   tsl = tsl,
#'   window = 5,
#'   f = mean
#' )
#'
#' if(interactive()){
#'   tsl_plot(tsl)
#'   tsl_plot(tsl_smooth)
#' }
#'
#' #exponential smoothing
#' tsl_smooth <- tsl_smooth(
#'   tsl = tsl,
#'   alpha = 0.2
#' )
#'
#' if(interactive()){
#'   tsl_plot(tsl)
#'   tsl_plot(tsl_smooth)
#' }
#'
tsl_smooth <- function(
    tsl = NULL,
    window = 3,
    f = mean,
    alpha = NULL,
    ...
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  if(!is.null(alpha)){
    if (alpha <= 0 || alpha > 1) stop("distantia::tsl_smooth(): 'alpha' must be in the range (0, 1].")
  }

  #progress bar
  p <- progressr::progressor(along = tsl)

  #parallelized loop
  tsl <- future.apply::future_lapply(
    X = tsl,
    FUN = function(x, ...){

      p()

      if(is.null(alpha)){

        x <- zoo_smooth_window(
          x = x,
          window = window,
          f = f,
          ... = ...
        )

      } else {

        x <- zoo_smooth_exponential(
          x = x,
          alpha = alpha
        )

      }

    },
    ... = ...
  )

  tsl <- tsl_names_set(
    tsl = tsl
  )

  tsl

}

#' Contribution of Individual Variables to Time Series Dissimilarity
#'
#' @description
#'
#' This function measures the contribution of individual variables to dissimilarity between pairs of time series to help answer the question *what makes two time series more or less similar?*
#'
#' Three key values are required to assess individual variable contributions:
#' \itemize{
#'   \item **psi**: dissimilarity when all variables are considered.
#'   \item **psi_only_with**: dissimilarity when using only the target variable.
#'   \item **psi_without**: dissimilarity when removing the target variable.
#' }
#'
#' The values `psi_only_with` and `psi_without` can be computed in two different ways, defined by the argument `robust`.
#' \itemize{
#'   \item `robust = FALSE`: This method replicates the importance algorithm released with the first version of the package, and it is only recommended when the goal is to compare new results with previous studies. It normalizes `psi_only_with` and `psi_without` using the least cost path obtained from the individual variable. As different variables may have different least cost paths for the same time series, normalization values may change from variable to variable, making individual importance scores harder to compare.
#'   \item `robust = TRUE` (default, recommended): This is a novel version of the importance algorithm that yields more stable and comparable solutions. It uses the least cost path of the complete time series to normalize `psi_only_with` and `psi_without`, making importance scores of separate variables fully comparable.
#' }
#'
#' The individual importance score of each variable (column "importance" in the output data frame) is based on different expressions depending on the `robust` argument, even when `lock_step = TRUE`:
#' \itemize{
#'   \item `robust = FALSE`: Importance is computed as `((psi - psi_without) * 100)/psi` and interpreted as "change in similarity when a variable is removed".
#'   \item `robust = TRUE`: Importance is computed as `((psi_only_with - psi_without) * 100)/psi` and interpreted as "relative dissimilarity induced by the variable expressed as a percentage".
#' }
#'
#' In either case, positive values indicate that the variable contributes to dissimilarity, while negative values indicate a net contribution to similarity.
#'
#' This function allows computing dissimilarity between pairs of time series using different combinations of arguments at once. For example, when the argument `distance` is set to `c("euclidean", "manhattan")`, the output data frame will show two dissimilarity scores for each pair of time series, one based on euclidean distances, and another based on manhattan distances. The same happens for most other parameters.
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @inheritParams distantia
#' @param robust (required, logical). If TRUE (default), importance scores are computed using the least cost path of the complete time series as reference. Setting it to FALSE allows to replicate importance scores of the previous versions of this package. This option is irrelevant when `lock_step = TRUE`. Default: TRUE
#' @return A data frame with the following columns:
#' \itemize{
#'   \item `x`: name of the time series `x`.
#'   \item `y`: name of the time series `y`.
#'   \item `psi`: psi score of `x` and `y`.
#'   \item `variable`: name of the individual variable.
#'   \item `importance`: importance score of the variable.
#'   \item `effect`: interpretation of the "importance" column, with the values "increases similarity" and "decreases similarity".
#'   \item `psi_only_with`: psi score of the variable.
#'   \item `psi_without`: psi score without the variable.
#'   \item `psi_difference`: difference between `psi_only_with` and `psi_without`.
#'   \item `distance`: name of the distance metric.
#'   \item `diagonal`: value of the argument `diagonal`.
#'   \item `lock_step`: value of the argument `lock_step`.
#'   \item `robust`: value of the argument `robust`.
#' }
#' @export
#' @autoglobal
#' @examples
#'
#' #progress bar
#' # progressr::handlers(global = TRUE)
#'
#' tsl <- tsl_initialize(
#'   x = distantia::albatross,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' df <- momentum(
#'   tsl = tsl,
#'   lock_step = TRUE #to speed-up example
#'   )
#'
#' #focus on important columns
#' df[, c(
#'   "x",
#'   "y",
#'   "variable",
#'   "importance",
#'   "effect"
#'   )]
#'
#' @family momentum
#'
momentum <- function(
    tsl = NULL,
    distance = "euclidean",
    diagonal = TRUE,
    bandwidth = 1,
    lock_step = FALSE,
    robust = TRUE
){

  #check input arguments
  args <- utils_check_args_momentum(
    tsl = tsl,
    distance = distance,
    diagonal = diagonal,
    bandwidth = bandwidth,
    lock_step = lock_step,
    robust = robust
  )

  tsl <- args$tsl
  distance <- args$distance
  diagonal <- args$diagonal
  bandwidth <- args$bandwidth
  lock_step <- args$lock_step
  robust <- args$robust

  #lock-step check
  if(any(lock_step == TRUE)){

    #count rows in time series
    row_counts <- tsl |>
      tsl_nrow() |>
      unlist() |>
      unique()

    if(length(row_counts) > 1){

      message(
        "distantia::momentum(): argument 'lock_step' is TRUE, but time series in 'tsl' do not have the same number of rows. Setting 'lock_step' to FALSE."
      )

      lock_step <- FALSE

    }

  }



  #stop if tsl is univariate
  tsl_ncol <- tsl_ncol(tsl = tsl) |>
    unlist() |>
    unique()

  if(1 %in% tsl_ncol){
    stop(
      "distantia::momentum(): variable contribution analysis requires multivariate time series, but 'tsl' contains univariate time series.",
      call. = FALSE
    )
  }

  #tsl pairs
  df <- utils_tsl_pairs(
    tsl = tsl,
    args_list = list(
      distance = distance,
      diagonal = diagonal,
      bandwidth = bandwidth,
      lock_step = lock_step,
      robust = robust
    )
  )

  iterations <- seq_len(nrow(df))

  p <- progressr::progressor(along = iterations)



  #iterate over pairs of time series
  df <- foreach::foreach(
    i = iterations,
    .combine = "rbind",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

    p()

    df.i <- df[i, ]

    x <- tsl[[df.i$x]]
    y <- tsl[[df.i$y]]

    if(df.i$lock_step == TRUE){

      df.i$robust <- TRUE

      importance.i <- importance_ls_cpp(
        x = x,
        y = y,
        distance = df.i$distance
      )

      #replacing importance with psi_drop
      if(df.i$robust == FALSE){
        importance.i$importance <- importance.i$psi_drop
      }

      importance.i$psi_drop <- NULL

    } else {

      if(df.i$robust == TRUE){

        importance.i <- importance_dtw_cpp(
          x = x,
          y = y,
          distance = df.i$distance,
          diagonal = df.i$diagonal,
          bandwidth = df.i$bandwidth
        )

      } else {

        importance.i <- importance_dtw_legacy_cpp(
          x = x,
          y = y,
          distance = df.i$distance,
          diagonal = df.i$diagonal
        )

      }

    } #end of importance i

    #set NaN to zero for constant pairs of sequences
    importance.i[is.na(importance.i)] <- 0

    importance.i <- merge(
      x = df.i,
      y = importance.i
    )

    return(importance.i)

  } #end of loop

  #interpretation
  df$effect <- ifelse(
    test = df$importance > 0,
    yes = "decreases similarity",
    no = "increases similarity"
  )

  df <- df[
    , c(
      "x",
      "y",
      "psi",
      "variable",
      "importance",
      "effect",
      "psi_difference",
      "psi_without",
      "psi_only_with",
      "distance",
      "diagonal",
      "bandwidth",
      "lock_step",
      "robust"
    )
  ]

  #remove dtw arguments if only lock-step was used
  if(
    "lock_step" %in% colnames(df) &&
    sum(df[["lock_step"]]) == nrow(df)
    ){
    df$diagonal <- NULL
    df$bandwidth <- NULL
  }

  attr(
    x = df,
    which = "type"
  ) <- "momentum_df"

  df

}

#' Dictionary of Time Keywords
#'
#' @description
#' Called by [utils_time_keywords_translate()] to generate a data frame that helps translate misnamed or abbreviated time keywords, like "day", "daily", or "d", into correct ones such as "days".
#'
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#' df <- utils_time_keywords_dictionary()
#' @family internal_time_handling
utils_time_keywords_dictionary <- function(){

  # millennia ----
  millennia <- c(
    "millennia",
    "milennia",
    "millenia",
    "millennium",
    "millenium",
    "milennium",
    "milenium",
    "milenio",
    "thousands",
    "thousand",
    "1000 years",
    "1000 y",
    "1000years",
    "1000y"
  )

  millennia <- c(millennia, toupper(millennia))

  # centuries ----
  centuries <- c(
    "centuries",
    "century",
    "centenio",
    "100 years",
    "100years",
    "100 y",
    "100y"
  )

  centuries <- c(centuries, toupper(centuries))

  # decades ----
  decades <- c(
    "decade",
    "decades",
    "decad",
    "10 years",
    "10years",
    "10 y",
    "10y"
  )

  decades <- c(decades, toupper(decades))

  # years ----
  years <- c(
    "years",
    "year",
    "ys",
    "y"
  )

  years <- c(years, toupper(years))

  # quarters ----
  quarters <- c(
    "quarters",
    "quarter",
    "quarterly",
    "q",
    "3 months",
    "3months"
  )

  quarters <- c(quarters, toupper(quarters))

  # months ----
  months <- c(
    "months",
    "month",
    "monthly",
    "m",
    "30 days",
    "30days",
    "30 d",
    "30d"
  )

  months <- c(months, toupper(months))

  # weeks ----
  weeks <- c(
    "weeks",
    "week",
    "weekly",
    "w",
    "7 days",
    "7days",
    "7 d",
    "7d"
  )

  weeks <- c(weeks, toupper(weeks))

  # days ----
  days <- c(
    "days",
    "day",
    "daily",
    "d",
    "24 hours",
    "24hours",
    "24 h",
    "24h"
  )

  days <- c(days, toupper(days))

  # hours ----
  hours <- c(
    "hours",
    "hour",
    "hourly",
    "h",
    "60 minutes",
    "60minutes",
    "60 min",
    "60min",
    "60 m",
    "60m"
  )

  hours <- c(hours, toupper(hours))

  # minutes ----
  minutes <- c(
    "minutes",
    "minute",
    "min",
    "m",
    "60 seconds",
    "60seconds",
    "60 secs",
    "60secs",
    "60 s",
    "60s"
  )

  minutes <- c(minutes, toupper(minutes))

  # seconds ----
  seconds <- c(
    "seconds",
    "second",
    "sec",
    "s"
  )

  seconds <- c(seconds, toupper(seconds))

  dictionary <- list(
    millennia = millennia,
    centuries = centuries,
    decades = decades,
    years = years,
    quarters = quarters,
    months = months,
    weeks = weeks,
    days = days,
    hours = hours,
    minutes = minutes,
    seconds = seconds
  ) |>
    utils::stack()

  dictionary$ind <- as.character(dictionary$ind)

  names(dictionary) <- c("pattern", "keyword")

  dictionary

}

#' Convert Individual Zoo Objects to Time Series List
#'
#' @description
#' Internal function to wrap a zoo object into a time series list.
#'
#' @param x (required, zoo object) Time series. Default: NULL
#'
#' @return time series list of length one.
#' @export
#' @autoglobal
#' @examples
#' #create zoo object
#' x <- zoo_simulate()
#' class(x)
#'
#' #to time series list
#' tsl <- zoo_to_tsl(
#'   x = x
#' )
#'
#' class(tsl)
#' class(tsl[[1]])
#' names(tsl)
#' attributes(tsl[[1]])$name
#' @family zoo_functions
zoo_to_tsl <- function(
    x = NULL
){

  if(!zoo::is.zoo(x)){
    stop("distantia::zoo_to_tsl(): argument 'x' must be a zoo time series.", call. = FALSE)
  }

  x_name <- attributes(x)$name
  if(is.null(x_name)){
    x_name <- ""
  }

  x <- zoo_name_set(
    x = x,
    name = x_name
  )

  out_tsl <- list(x)

  names(out_tsl) <- x_name

  out_tsl

}

#' Lock-Step Distance
#'
#' @description
#' Demonstration function to compute the lock-step distance between two univariate or multivariate time series.
#'
#' This function does not accept NA data in the matrices `x` and `y`.
#'
#' @param x (required, zoo object or numeric matrix) a time series with no NAs. Default: NULL
#' @param y (zoo object or numeric matrix) a time series with the same columns as `x` and no NAs. Default: NULL
#' @inheritParams distantia
#' @return numeric
#' @examples
#' #distance metric
#' d <- "euclidean"
#'
#' #simulate two time series
#' #of the same length
#' x <- zoo_simulate(
#'   name = "x",
#'   rows = 100,
#'   seasons = 2,
#'   seed = 1
#' )
#'
#' y <- zoo_simulate(
#'   name = "y",
#'   rows = 100,
#'   seasons = 2,
#'   seed = 2
#' )
#'
#' if(interactive()){
#'   zoo_plot(x = x)
#'   zoo_plot(x = y)
#' }
#'
#' #sum of distances
#' #between pairs of samples
#' psi_distance_lock_step(
#'   x = x,
#'   y = y,
#'   distance = d
#' )
#' @export
#' @autoglobal
#' @family psi_demo
psi_distance_lock_step <- function(
    x = NULL,
    y = NULL,
    distance = "euclidean"
){

  distance <- utils_check_distance_args(
    distance = distance
  )[1]

  if(nrow(as.matrix(x)) != nrow(as.matrix(y))){
    stop(
      "distantia::psi_distance_lock_step(): 'x' and 'y' must have the same number of rows for lock-step comparison.",
      call. = FALSE
    )
  }

  d <- distance_ls_cpp(
    x,
    y,
    distance
  )

  d

}

#' Aggregate `momentum()` Data Frames Across Parameter Combinations
#'
#' @description
#'
#' The function [momentum()] allows variable importance assessments based on several combinations of arguments at once. For example, when the argument `distance` is set to `c("euclidean", "manhattan")`, the output data frame will show two importance scores for each pair of compared time series and variable, one based on euclidean distances, and another based on manhattan distances.
#'
#' This function computes importance statistics across combinations of parameters.
#'
#' If there are no different combinations of arguments in the input data frame, no aggregation happens, but all parameter columns are removed.
#'
#' @param df (required, data frame) Output of [momentum()], [momentum_ls()], or [momentum_dtw()]. Default: NULL
#' @param f (optional, function) Function to summarize importance scores (for example, `mean`) when there are several combinations of parameters in `df`. Ignored when there is a single combination of arguments in the input. Default: `mean`
#' @param ... (optional, arguments of `f`) Further arguments to pass to the function `f`.
#'
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#' #three time series
#' #climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl,
#'     guide_columns = 3
#'     )
#' }
#'
#' #momentum with multiple parameter combinations
#' #-------------------------------------
#' df <- momentum(
#'   tsl = tsl,
#'   distance = c("euclidean", "manhattan"),
#'   lock_step = TRUE
#' )
#'
#' df[, c(
#'   "x",
#'   "y",
#'   "distance",
#'   "importance"
#' )]
#'
#' #aggregation using means
#' df <- momentum_aggregate(
#'   df = df,
#'   f = mean
#' )
#'
#' df
#' @family momentum_support
momentum_aggregate <- function(
    df = NULL,
    f = mean,
    ...
){

  df_type <- attributes(df)$type

  if(df_type != "momentum_df"){
    stop("distantia::momentum_aggregate(): argument 'df' must be the output of 'momentum()'.", call. = FALSE)
  }

  if(!is.function(f)){
    stop("distantia::momentum_aggregate(): argument 'f' must be a function such as mean, median, mean, max, and the likes.", call. = FALSE)
  }

  # split ----
  df_list <- utils_distantia_df_split(
    df = df
  )

  if(length(df_list) == 1){
    df$distance <- NULL
    df$diagonal <- NULL
    df$bandwidth <- NULL
    df$lock_step <- NULL
    df$robust <- NULL
    return(df)
  }

  if("robust" %in% colnames(df)){

    if(length(unique(df$robust)) == 2){

      warning("distantia::momentum_aggregate(): Column 'robust' has the values TRUE and FALSE. The aggregation of importance scores computed with both options is not supported. Cases where 'df$robust == FALSE' will be ignored.", call. = FALSE)

      df <- df[df$robust == TRUE, ]

      df$robust <- NULL

    }

  }

  df_aggregated <- stats::aggregate(
    x = df,
    by = psi ~ x + y + variable,
    FUN = f,
    ... = ...
  )

  df_aggregated$importance <- stats::aggregate(
    x = df,
    by = importance ~ x + y + variable,
    FUN = f,
    ... = ...
  )$importance

  if("psi_without" %in% colnames(df)){

    df_aggregated$psi_without <- stats::aggregate(
      x = df,
      by = psi_without ~ x + y + variable,
      FUN = f,
      ... = ...
    )$psi_without

  }

  if("psi_only_with" %in% colnames(df)){

    df_aggregated$psi_only_with <- stats::aggregate(
      x = df,
      by = psi_only_with ~ x + y + variable,
      FUN = f,
      ... = ...
    )$psi_only_with

  }

  df <- df_aggregated

  #add type
  attr(
    x = df,
    which = "type"
  ) <- df_type

  df

}

#' Plot Zoo Time Series
#'
#' @param x (required, zoo object) zoo time series. Default: NULL
#' @param line_color (optional, character vector) vector of colors for the distance or cost matrix. If NULL, uses an appropriate palette generated with [grDevices::palette.colors()]. Default: NULL
#' @param line_width (optional, numeric vector) Width of the time series lines. Default: 1
#' @param xlim (optional, numeric vector) Numeric vector with the limits of the x axis. Default: NULL
#' @param ylim (optional, numeric vector) Numeric vector with the limits of the y axis. Default: NULL
#' @param title (optional, character string) Main title of the plot. If NULL, it's set to the name of the time series. Default: NULL
#' @param xlab (optional, character string) Title of the x axis. Disabled if `subpanel` or `vertical` are TRUE. If NULL, the word "Time" is used. Default: NULL
#' @param ylab (optional, character string) Title of the y axis. Disabled if `subpanel` or `vertical` are TRUE. If NULL, it is left empty. Default: NULL
#' @param text_cex (optional, numeric) Multiplicator of the text size. Default: 1
#' @param guide (optional, logical) If TRUE, plots a legend. Default: TRUE
#' @param guide_position (optional, vector of xy coordinates or character string). This is a condensed version of the `x` and `y` arguments of the [graphics::legend()] function. Coordinates (in the range 0 1) or keyword to position the legend. Accepted keywords are: "bottomright", "bottom", "bottomleft", "left", "topleft", "top", "topright", "right" and "center". Default: "topright".
#' @param guide_cex (optional, numeric) Size of the guide's text and separation between the guide's rows. Default: 0.7.
#' @param vertical (optional, logical) For internal use within the package in multipanel plots. Switches the plot axes. Disabled if `subpanel = FALSE`. Default: FALSE
#' @param subpanel (optional, logical) For internal use within the package in multipanel plots. Strips down the plot for a sub-panel. Default: FALSE
#'
#' @return A plot.
#' @export
#'
#' @examples
#'
#' #simulate zoo time series
#' x <- zoo_simulate()
#'
#' if(interactive()){
#'
#'   zoo_plot(
#'     x = x,
#'     xlab = "Date",
#'     ylab = "Value",
#'     title = "My time series"
#'   )
#'
#' }
#' @keywords plotting
#' @family zoo_functions
zoo_plot <- function(
    x = NULL,
    line_color = NULL,
    line_width = 1,
    xlim = NULL,
    ylim = NULL,
    title = NULL,
    xlab = NULL,
    ylab = NULL,
    text_cex = 1,
    guide = TRUE,
    guide_position = "topright",
    guide_cex = 0.8,
    vertical = FALSE,
    subpanel = FALSE
){

  #check x
  x <- utils_check_args_zoo(x = x)

  #if standalone zoo plot
  if(subpanel == FALSE && guide == TRUE){

    #get user's par
    old.par <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(old.par))

    #plotting areas
    plt_all <- graphics::par()$plt

    plt_lines <- c(
      plt_all[1],
      0.70,
      plt_all[3],
      plt_all[4]
      )

    plt_guide <- c(
      0.70,
      plt_all[2],
      plt_all[3],
      plt_all[4]
      )

  }

  #axis title
  axis_title_distance <- 1.4
  axis_title_cex <- 0.9 * text_cex

  #axis labels
  axis_labels_cex <- 0.6 * text_cex

  #title
  main_title_distance <- 1.2
  main_title_cex <- 1.2 * text_cex

  if(subpanel == FALSE){

    axis_title_distance <- 2.2
    axis_labels_cex <- 0.8 * text_cex
    vertical <- FALSE

  }

  if(vertical == TRUE){

    axis_title_distance <- 2.4

  }

  df <- as.data.frame(x)

  #time column
  df.time <- zoo::index(x)

  #name
  name <- attributes(x)$name
  if(is.null(name)){
    name <- ""
  }

  #names of columns to plot
  x_colnames <- colnames(df)

  #default palette
  if(is.null(names(line_color))){
    line_color <- utils_line_color(
      x = x,
      line_color = line_color
    )
  }

  line_color <- line_color[names(line_color) %in% x_colnames]

  #line_width
  if(length(line_width) == 1){
    line_width <- rep(
      x = line_width,
      times = length(x_colnames)
    )
  }

  if(length(line_width) != length(x_colnames)){
    if(line_width[1] < line_width[length(line_width)]){
      line_width <- seq(
        from = min(line_width),
        to = max(line_width),
        length.out = length(x_colnames)
      )
    } else {
      line_width <- seq(
        from = max(line_width),
        to = min(line_width),
        length.out = length(x_colnames)
      )
    }
  }

  #data limits
  if(is.null(ylim)){
    ylim <- range(df, na.rm = TRUE)
  }
  if(is.null(xlim)){
    xlim <- range(zoo::index(x))
  }

  if(subpanel == TRUE){
    ylim[which.max(ylim)] <- ylim[which.max(ylim)] + (diff(ylim)/5)
  }

  #first plot
  if(vertical == FALSE){

    plot.x <- df.time
    plot.y <- df[, names(line_color)[1]]
    x.axis.side <- 1
    y.axis.side <- 2
    name.x <- name
    name.y <- NULL

    if(!is.null(xlim)){
      xlim <- utils_coerce_time_class(
        x = xlim,
        to = class(zoo::index(x))
      )
    }


  } else {

    plot.x <- df[, names(line_color)[1]]
    plot.y <- df.time
    x.axis.side <- 3
    y.axis.side <- 2
    xlim <- rev(ylim)
    ylim <- NULL
    name.x <- NULL
    name.y <- name

  }

  if(guide == TRUE){
    graphics::par(
      plt = graphics::par()$plt
    )
  }

  #first plot
  graphics::plot(
    x = plot.x,
    y = plot.y,
    xlab = "",
    type = "l",
    ylab = "",
    xaxt = "n",
    yaxt = "n",
    frame.plot = FALSE,
    ylim = ylim,
    xlim = xlim,
    lwd = line_width[1],
    col = line_color[1]
  )

  if("Date" %in% class(plot.x)){
    graphics::axis.Date(
      side = x.axis.side,
      las = 1,
      cex.axis = axis_labels_cex
    )
  }

  if("POSIXct" %in% class(plot.x)){
    graphics::axis.POSIXct(
      side = x.axis.side,
      las = 1,
      cex.axis = axis_labels_cex
    )
  }

  if(any(c("numeric", "integer") %in% class(plot.x))){
    graphics::axis(
      side = x.axis.side,
      las = 1,
      cex.axis = axis_labels_cex
    )
  }

  if("Date" %in% class(plot.y)){
    graphics::axis.Date(
      side = y.axis.side,
      las = 1,
      cex.axis = axis_labels_cex
    )
  }

  if("POSIXct" %in% class(plot.y)){
    graphics::axis.POSIXct(
      side = y.axis.side,
      las = 1,
      cex.axis = axis_labels_cex
    )
  }

  if(any(c("numeric", "integer") %in% class(plot.y))){
    graphics::axis(
      side = y.axis.side,
      las = 1,
      cex.axis = axis_labels_cex
    )
  }


  #add all the other lines
  if(length(x_colnames) > 1){

    for(i in seq(from = 2, to = length(x_colnames), by = 1)){

      if(vertical == FALSE){
        line.x <- df.time
        line.y <- df[, names(line_color)[i]]
      } else {
        line.x <- df[, names(line_color)[i]]
        line.y <- df.time
      }

      graphics::lines(
        x = line.x,
        y = line.y,
        lwd = line_width[i],
        col = line_color[i]
      )

    }

  }


  #decoration
  if(subpanel == FALSE){

    graphics::title(
      main = ifelse(
        test = is.null(title),
        yes = name,
        no = title
      ),
      cex.main = main_title_cex,
      line = main_title_distance
    )

    graphics::title(
      xlab = ifelse(
        test = is.null(xlab),
        yes = "",
        no = xlab
      ),
      line = axis_title_distance,
      cex.lab = axis_title_cex
    )

    graphics::title(
      ylab = ifelse(
        test = is.null(ylab),
        yes = "",
        no = ylab
      ),
      line = axis_title_distance,
      cex.lab = axis_title_cex
    )

  } else {

    graphics::title(
      xlab = name.x,
      ylab = name.y,
      cex.lab = axis_title_cex,
      line = axis_title_distance
    )


  }

  if(guide == TRUE){

    graphics::par(
      plt = graphics::par()$plt,
      new = TRUE,
      mgp = c(0, 0.5, 0),
      bty = "n"
    )

    utils_line_guide(
      x = x,
      position = guide_position,
      line_color = line_color,
      line_width = line_width,
      text_cex = guide_cex,
      subpanel = TRUE
    )

  }

}

#' Simulate a Time Series List
#'
#' @description
#' Generates simulated time series lists for testing and learning.
#'
#' This function supports progress bars generated by the `progressr` package, and accepts a parallelization setup via [future::plan()] (see examples).
#'
#' @param n (optional, integer) Number of time series to simulate. Default: 2.
#' @param cols (optional, integer) Number of columns of each time series. Default: 5
#' @param rows (optional, integer) Length of each time series. Minimum is 10, but maximum is not limited. Very large numbers might crash the R session. Default: 100
#' @param time_range (optional character or numeric vector) Time interval of the time series. Either a character vector with dates in format YYYY-MM-DD or a numeric vector. If there is a mismatch between `time_range` and `rows` (for example, the number of days in `time_range` is smaller than `rows`), the upper value in `time_range` is adapted to `rows`. Default: c("2010-01-01", "2020-01-01")
#' @param data_range (optional, numeric vector of length 2) Extremes of the time series values. Default: c(0, 1)
#' @param seasons (optional, integer) Number of seasons in the resulting time series. The maximum number of seasons is computed as `floor(rows/3)`. Default: 0
#' @param na_fraction (optional, numeric) Value between 0 and 0.5 indicating the approximate fraction of NA data in the simulated time series. Default: 0.
#' @param independent (optional, logical) If TRUE, each new column in a simulated time series is averaged with the previous column to generate dependency across columns, and each new simulated time series is weighted-averaged with a time series template to generate dependency across time series. Irrelevant when `cols < 2` or `n < 2`, and hard to perceive in the output when `seasons > 0`. Default: FALSE
#' @param irregular (optional, logical) If TRUE, the time intervals between consecutive samples and the number of rows are irregular. Default: TRUE
#' @param seed (optional, integer) Random seed used to simulate the zoo object. If NULL (default), a seed is selected at random. Default: NULL
#'
#' @return time series list
#' @export
#' @autoglobal
#' @examples
#'
#' # generates a different time series list on each iteration when seed = NULL
#' tsl <- tsl_simulate(
#'   n = 2,
#'   seasons = 4
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl
#'   )
#' }
#'
#' # generate 3 independent time series
#' tsl_independent <- tsl_simulate(
#'   n = 3,
#'   cols = 3,
#'   independent = TRUE
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl_independent
#'   )
#' }
#'
#' # generate 3 independent time series
#' tsl_dependent <- tsl_simulate(
#'   n = 3,
#'   cols = 3,
#'   independent = FALSE
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl_dependent
#'   )
#' }
#'
#' # with seasons
#' tsl_seasons <- tsl_simulate(
#'   n = 3,
#'   cols = 3,
#'   seasons = 4,
#'   independent = FALSE
#' )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl_seasons
#'   )
#' }
#' 
#' @family simulate_time_series
tsl_simulate <- function(
    n = 2,
    cols = 5,
    rows = 100,
    time_range = c("2010-01-01", "2020-01-01"),
    data_range = c(0, 1),
    seasons = 0,
    na_fraction = 0,
    independent = FALSE,
    irregular = TRUE,
    seed = NULL
){

  # n ----
  n <- as.integer(abs(n))[1]
  if(n <= 0){
    n <- 1
  }

  # na_fraction ----
  na_fraction <- as.numeric(na_fraction[1])
  if(na_fraction < 0){
    na_fraction <- 0
  }
  if(na_fraction > 0.5){
    na_fraction <- 0.5
  }

  # seed ----
  if(is.null(seed)){
    seed <- sample.int(
      n = .Machine$integer.max,
      size = 1
      )
  }
  set.seed(seed)

  # if irregular, increase row count
  if(!is.logical(irregular)){
    irregular <- FALSE
  }

  # time range ----
  if(length(time_range) != 2){
    stop("distantia::tsl_simulate(): argument 'time_range' must be a vector of length 2.")
  }

  time_range <- utils_as_time(
    x = time_range
  ) |>
    range()

  # generate template
  if(independent == FALSE){

    template <- zoo_simulate(
      cols = cols,
      rows = rows,
      time_range = time_range,
      data_range = data_range,
      seasons = seasons,
      na_fraction = 0,
      independent = independent,
      irregular = FALSE,
      seed = seed
    )

    template_weight <- seq(
      from = 0.9,
      to = 0,
      length.out = n + 1
    )

    #loop args
    arg_irregular <- FALSE
    arg_na_fraction <- 0

  } else {

    arg_irregular <- irregular
    arg_na_fraction <- na_fraction

  }

  #progress bar
  p <- progressr::progressor(along = seq_len(n))



  #generate tsl
  tsl <- foreach::foreach(
    i = seq_len(n),
    .errorhandling = "pass",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

    #call to progress bar
    p()

    #regular zoo with no NA
    zoo.i <- zoo_simulate(
      cols = cols,
      rows = rows,
      time_range = time_range,
      data_range = data_range,
      seasons = seasons,
      na_fraction = arg_na_fraction,
      independent = independent,
      irregular = arg_irregular,
      seed = NULL
    )

    #handle independence
    if(independent == FALSE){

      #add weighted template
      zoo.i <-
        zoo.i * (1 - template_weight[i]) +
        template * template_weight[i]

    }

    #regular or irregular rows
    if(irregular == TRUE){

      if(rows > 10){

        #compute rows to keep
        rows.i <- sample(
          x = rows,
          size = rows - sample.int(n = rows/4, size = 1)
        ) |>
          sort()

        #subsetting
        zoo.i <- zoo.i[rows.i]

      }


    } else {

      rows.i <- seq_len(rows)

    }

    #regular or irregular time
    zoo::index(x = zoo.i) <- seq(
      from = min(time_range),
      to = max(time_range),
      length.out = length(rows.i) * ifelse(
        test = irregular == TRUE,
        yes = 4,
        no = 1
      )
    ) |>
      sample(size = length(rows.i)) |>
      sort()

    #apply na_fraction
    zoo.i[
      sample.int(
        n = length(zoo.i),
        size = floor(length(zoo.i) * na_fraction)
      )
    ] <- NA

    zoo.i

  }

  # names ----
  if(n > length(LETTERS)){

    names_tsl <- c(
      LETTERS,
      as.vector(
        outer(
          X = LETTERS,
          Y = LETTERS,
          FUN = paste0
        )
      )
    )[seq_len(n)]

  } else {

    names_tsl <- LETTERS[seq_len(n)]

  }

  names(tsl) <- names_tsl

  tsl <- tsl_names_set(
    tsl = tsl,
    names = names_tsl
  )

  tsl

}

test_that("`tsl_count_NA()` works", {

  tsl <- tsl_simulate(,
                      seed = 1)

  expect_equal(
    sum(unlist(tsl_count_NA(tsl = tsl))),
    0
    )

  tsl <- tsl_simulate(,
                      seed = 1)
  tsl[[1]][1, 1] <- Inf
  tsl[[1]][2, 1] <- -Inf
  tsl[[1]][3, 1] <- NaN
  tsl[[1]][4, 1] <- NaN

  expect_true(
    sum(unlist(tsl_count_NA(tsl = tsl))) == 4
  )

  tsl <- tsl_simulate(na_fraction = 0.3,
                      seed = 1)

  expect_true(
    sum(unlist(tsl_count_NA(tsl = tsl))) > 0
  )

})

#' Convert Data Frame to a List of Data Frames
#'
#' @param x (required, data frame) Input data frame. Default: NULL.
#' @param name_column (optional, column name) Column name used to split `x` into a list of data frames. If omitted, the split is done by column instead. Default: NULL.
#' @param time_column (optional, column name) Name of the column representing time, if any. Default: NULL.
#' @return List of data frames.
#' @export
#' @autoglobal
#' @family internal
utils_prepare_df <- function(
    x = NULL,
    name_column = NULL,
    time_column = NULL
){

  if(inherits(x = x, what = "data.frame") == FALSE){
    return(x)
  }

  if(nrow(x) < 3){
    stop("distantia::utils_prepare_df(): number of rows in 'x' is not enough for a dissimilarity analysis.", call. = FALSE)
  }

  #drop geometry column, if any
  x <- utils_drop_geometry(
    df = x
  )

  if(is.null(time_column)){
    time_column <- NA
  } else {
    if(!(time_column %in% names(x))){
      stop("distantia::utils_prepare_df(): Column '", time_column, "' is not a column name of the data frame 'x'.", call. = FALSE)
    }
  }

  if(is.null(name_column)){
    name_column <- NA
  } else {
    if(!(name_column %in% names(x))){
      stop("distantia::utils_prepare_df(): column '", name_column, "' is not a column name of the data frame 'x'.", call. = FALSE)
    }
  }

  if(time_column %in% colnames(x)){
    x.time <- x[[time_column]]
    x[[time_column]] <- NULL
  }

  if(name_column %in% colnames(x)){
    x.id <- x[[name_column]]
    x[[name_column]] <- NULL
  }

  x <- x[, sapply(x, is.numeric), drop = FALSE]

  if(exists(x = "x.id")){

    if(exists(x = "x.time")){
      x[[time_column]] <- x.time
    }

    x[[name_column]] <- x.id

    x <- split(
      x = x[, colnames(x) != name_column],
      f = x[[name_column]]
    )

  } else {

    x <- lapply(
      X = x,
      FUN = function(x) {
        x <- data.frame(
          x = x
        )
        if(exists(x = "x.time")){
          x[[time_column]] <- x.time
        }
        return(x)
      }
    )

  }

  x

}

#' Time Delay Between Time Series
#'
#' @description
#' This function computes an approximation to the time-delay between pairs of time series as the difference between observations connected by the dynamic time warping path.
#'
#' Given a pair of time series `x` and `y`, and the time of their samples in the dynamic time warping `time(x)` and `time(y)`, when the argument `directional` is `TRUE`, the time delay is computed as follows:
#'\itemize{
#' \item Time delay from `x` to `y`: `time(y) - time(x)`.
#' \item Time delay from `y` to `x`: `time(x) - time(y)`
#'}
#' In such case, two rows per pair of time series are returned. Otherwise, the time delay is computed as `abs(time(y) - time(x))`, and only one row per pair of time series is returned.
#'
#' If the time series have more than 30 observations, 5% of cases are omitted at each extreme of the warping path to avoid overestimating time delays due to early misalignments.
#'
#' The function returns a data frame with the names of the time series in columns *x* and *y*, and summary statistics of the time delay. The mode and median are generally the most accurate time-delay metrics.
#'
#' This function requires scaled and detrended time series. It may yield non-sensical results in case of degenerate warping paths. Plotting dubious results with [distantia_dtw_plot()] is a good approach to identify these cases.
#'
#' @inheritParams distantia
#' @param directional (optional, logical) If TRUE, a directional time delay is computed as `x to y` and `y to x`, resulting in two rows per pair of time series. Otherwise, the absolute magnitude of the delay between `x` and `y` is returned as a single row per pair. Default: TRUE
#'
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#' #load two long-term temperature time series
#' #local scaling to focus on shape rather than values
#' #polynomial detrending to make them stationary
#' tsl <- tsl_init(
#'   x = cities_temperature[
#'     cities_temperature$name %in% c("London", "Kinshasa"),
#'     ],
#'   name = "name",
#'   time = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_local
#'   ) |>
#'   tsl_transform(
#'     f = f_detrend_poly,
#'     degree = 35 #data years
#'   )
#'
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl,
#'     guide = FALSE
#'   )
#' }
#'
#' #compute shifts
#' df_shift <- distantia_time_delay(
#'   tsl = tsl,
#'   directional = TRUE
#' )
#'
#' df_shift
#' @family distantia_support
distantia_time_delay <- function(
    tsl = NULL,
    distance = "euclidean",
    bandwidth = 1,
    directional = FALSE
){

  #check input arguments
  args <- utils_check_args_distantia(
    tsl = tsl,
    distance = distance[1]
  )

  tsl <- args$tsl
  distance <- args$distance

  #tsl pairs
  df <- utils_tsl_pairs(
    tsl = tsl,
    args_list = list(
      distance = distance
    )
  )

  #function to compute modal of differences
  delay_modal <- function(x) {
    ux <- unique(na.omit(x))
    y <- ux[which.max(tabulate(match(x, ux)))][1]
    class(y) <- class(x)
    attr(x = y, which = "units") <- attributes(x)$units
    y
  }

  iterations <- seq_len(nrow(df))

  p <- progressr::progressor(along = iterations)

  #iterate over pairs of time series
  df_delay <- foreach::foreach(
    i = iterations,
    .combine = "rbind",
    .errorhandling = "pass",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

    p()

    df.i <- df[i, ]

    #compute least cost path
    cost_path.i <- cost_path_cpp(
      x = tsl[[df.i[["x"]]]],
      y = tsl[[df.i[["y"]]]],
      distance = distance,
      diagonal = TRUE,
      weighted = TRUE,
      ignore_blocks = FALSE,
      bandwidth = bandwidth
    )

    #remove duplicates
    cost_path.i <- cost_path.i[
      !(duplicated(cost_path.i[["x"]]) |
          duplicated(cost_path.i[["y"]])),
      c("x", "y")
    ]

    #compute 5% padding
    padding <- ceiling(nrow(cost_path.i)/100) * 5

    #compute n to assess if we apply padding
    n <- nrow(cost_path.i) - (padding * 2)

    #apply padding when 30 or more cases
    if(n >= 30){

      cost_path.i <- cost_path.i[padding:(nrow(cost_path.i) - padding),]

    }

    #check time series units
    delay.units.x <- zoo_time(x = tsl[[df.i[["x"]]]])$units
    delay.units.y <- zoo_time(x = tsl[[df.i[["y"]]]])$units

    #compare by sample if units do not match
    if(delay.units.x != delay.units.y){

      warning(
        "distantia::distantia_time_delay(): time series '",
        df[["x"]], "' and '", df[["y"]], "' have different time units (",
        delay.units.x,
        " vs ",
        delay.units.y,
        ". Computing time delay with units 'samples'."
      )

      #change default units
      delay.units <- "samples"

      #sample columns
      colname_x <- "x"
      colname_y <- "y"

    } else {

      #create time columns
      cost_path.i[["x_time"]] <- zoo::index(
        x = tsl[[df.i[["x"]]]]
      )[cost_path.i[["x"]]]

      cost_path.i[["y_time"]] <- zoo::index(
        x = tsl[[df.i[["y"]]]]
      )[cost_path.i[["y"]]]

      #default delay units
      delay.units <- delay.units.x

      #default time column names
      colname_x <- "x_time"
      colname_y <- "y_time"

    }

    #time difference from x to y
    if(directional == FALSE){

      delay.x_to_y <- abs(as.numeric(cost_path.i[[colname_x]] - cost_path.i[[colname_y]]))

    } else {

      delay.x_to_y <- as.numeric(cost_path.i[[colname_y]] - cost_path.i[[colname_x]])

      delay.y_to_x <- as.numeric(cost_path.i[[colname_x]] - cost_path.i[[colname_y]])

    }

    #add units
    df.i[["units"]] <- delay.units

    df.i[["min"]] <- min(
      x = delay.x_to_y,
      na.rm = TRUE
    )

    df.i[["q1"]] <- stats::quantile(
      x = delay.x_to_y,
      na.rm = TRUE,
      probs = 0.25
    )

    #compute stats x to y
    df.i[["median"]] <- stats::median(
      x = delay.x_to_y,
      na.rm = TRUE
    )

    df.i[["modal"]] <- delay_modal(x = delay.x_to_y)

    df.i[["mean"]] <- mean(
      x = delay.x_to_y,
      na.rm = TRUE
    )

    df.i[["q3"]] <- stats::quantile(
      x = delay.x_to_y,
      na.rm = TRUE,
      probs = 0.75
    )

    df.i[["max"]] <- max(
      x = delay.x_to_y,
      na.rm = TRUE
    )

    #compute stats y to x
    if(directional == TRUE){

      df.j <- df.i
      df.j$x <- df.i$y
      df.j$y <- df.i$x


      df.j[["min"]] <- min(
        x = delay.y_to_x,
        na.rm = TRUE
      )

      df.j[["q1"]] <- stats::quantile(
        x = delay.y_to_x,
        na.rm = TRUE,
        probs = 0.25
      )

      df.j[["median"]] <- stats::median(
        x = delay.y_to_x,
        na.rm = TRUE
      )

      df.j[["modal"]] <- delay_modal(x = delay.y_to_x)


      df.j[["mean"]] <- mean(
        x = delay.y_to_x,
        na.rm = TRUE
      )

      df.j[["q3"]] <- stats::quantile(
        x = delay.y_to_x,
        na.rm = TRUE,
        probs = 0.75
      )

      df.j[["max"]] <- max(
        x = delay.y_to_x,
        na.rm = TRUE
      )

      df.i <- rbind(
        df.i,
        df.j
      )

    }

    df.i

  }

  rownames(df_delay) <- NULL

  #add type
  attr(
    x = df_delay,
    which = "type"
  ) <- "time_delay_df"

  df_delay

}

#' Rescale Numeric Vector to a New Data Range
#' @param x (required, numeric vector) Numeric vector. Default: `NULL`
#' @param new_min (optional, numeric) New minimum value. Default: `0`
#' @param new_max (optional, numeric) New maximum value. Default: `1`
#' @param old_min (optional, numeric) Old minimum value. Default: `NULL`
#' @param old_max (optional, numeric) Old maximum value. Default: `NULL`
#' @return numeric vector
#' @examples
#'
#'  out <- utils_rescale_vector(
#'    x = stats::rnorm(100),
#'    new_min = 0,
#'    new_max = 100,
#'    )
#'
#'  out
#'
#' @export
#' @autoglobal
#' @family tsl_processing_internal
utils_rescale_vector <- function(
    x = NULL,
    new_min = 0,
    new_max = 1,
    old_min = NULL,
    old_max = NULL
){

  if(!is.vector(x) || !is.numeric(x)){
    stop("distantia::utils_rescale_vector(): x must be a numeric vector.", call. = FALSE)
  }

  if(is.null(old_min)){
    old_min <- min(x, na.rm = TRUE)
  }

  if(is.null(old_max)){
    old_max <- max(x, na.rm = TRUE)
  }

  if(old_min == old_max){
    return(rep(new_min, length(x)))
  }

  ((x - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min


}

test_that("`utils_rescale_vector()` works", {
  out <- utils_rescale_vector(
    x = stats::rnorm(100), new_min = 0,
    new_max = 100,
  )
  expect_equal(range(out), c(0, 100))
})

test_that("`utils_rescale_vector()` handles min == max (constant vector)", {
  out <- utils_rescale_vector(x = rep(5, 10), new_min = 0, new_max = 1)
  expect_true(all(out == 0))
  expect_false(any(is.nan(out)))
  expect_false(any(is.infinite(out)))
})

#' K-Means Clustering of Dissimilarity Analysis Data Frames
#'
#' @description
#' This function combines the dissimilarity scores computed by [distantia()], the K-means clustering method implemented in [stats::kmeans()], and the clustering optimization method implemented in [utils_cluster_hclust_optimizer()] to help group together time series with similar features.
#'
#' When `clusters = NULL`, the function [utils_cluster_hclust_optimizer()] is run underneath to perform a parallelized grid search to find the  number of clusters maximizing the overall silhouette width of the clustering solution (see [utils_cluster_silhouette()]).
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @inheritParams distantia_aggregate
#' @param clusters (required, integer) Number of groups to generate. If NULL (default), [utils_cluster_kmeans_optimizer()] is used to find the number of clusters that maximizes the mean silhouette width of the clustering solution (see [utils_cluster_silhouette()]). Default: NULL
#' @param seed (optional, integer) Random seed to be used during the K-means computation. Default: 1
#' @return list:
#' \itemize{
#'   \item `cluster_object`: kmeans object for further analyses and custom plotting.
#'   \item `clusters`: integer, number of clusters.
#'   \item `silhouette_width`: mean silhouette width of the clustering solution.
#'   \item `df`: data frame with time series names, their cluster label, and their individual silhouette width scores.
#'   \item `d`: psi distance matrix used for clustering.
#'   \item `optimization`: only if `clusters = NULL`, data frame with optimization results from [utils_cluster_hclust_optimizer()].
#' }
#' @export
#' @autoglobal
#' @examples
#'
#' #weekly covid prevalence in California
#' tsl <- tsl_initialize(
#'   x = covid_prevalence,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #subset 10 elements to accelerate example execution
#' tsl <- tsl_subset(
#'   tsl = tsl,
#'   names = 1:10
#' )
#'
#' if(interactive()){
#'   #plotting first three time series
#'   tsl_plot(
#'     tsl = tsl[1:3],
#'     guide_columns = 3
#'   )
#' }
#'
#' #dissimilarity analysis
#' distantia_df <- distantia(
#'   tsl = tsl,
#'   lock_step = TRUE
#' )
#'
#' #hierarchical clustering
#' #automated number of clusters
#' distantia_kmeans <- distantia_cluster_kmeans(
#'   df = distantia_df,
#'   clusters = NULL
#' )
#'
#' #names of the output object
#' names(distantia_kmeans)
#'
#' #kmeans object
#' distantia_kmeans$cluster_object
#'
#' #distance matrix used for clustering
#' distantia_kmeans$d
#'
#' #number of clusters
#' distantia_kmeans$clusters
#'
#' #clustering data frame
#' #group label in column "cluster"
#' distantia_kmeans$df
#'
#' #mean silhouette width of the clustering solution
#' distantia_kmeans$silhouette_width
#'
#' #kmeans plot
#' # factoextra::fviz_cluster(
#' #   object = distantia_kmeans$cluster_object,
#' #   data = distantia_kmeans$d,
#' #   repel = TRUE
#' # )
#' @family distantia_support
distantia_cluster_kmeans <- function(
    df = NULL,
    clusters = NULL,
    seed = 1
){

  df <- distantia_aggregate(
    df = df,
    f = mean
  )

  d <- distantia_matrix(
    df = df
  )[[1]]

  #optimize kmeans
  if(is.null(clusters)){

    optimization_df <- utils_cluster_kmeans_optimizer(
      d = d,
      seed = seed
    )

    clusters <- optimization_df[
      which.max(optimization_df$silhouette_mean),
      "clusters"
    ]

  }

  if(clusters < 2){
    clusters <- 2
  }

  if(clusters >= nrow(d)){
    clusters <- nrow(d) - 1
  }

  set.seed(seed)

  k <- stats::kmeans(
    x = d,
    centers = clusters,
    algorithm = "Hartigan-Wong",
    nstart = nrow(df)
  )

  k_silhouette <- utils_cluster_silhouette(
    labels = k$cluster,
    d = d,
    mean = FALSE
  )

  out <- list(
    cluster_object = k,
    clusters = clusters,
    silhouette_width = mean(k_silhouette$silhouette_width),
    df = k_silhouette,
    d = d

  )

  if(exists("optimization_df")){
    out$optimization <- optimization_df
  }

  out

}

#' Optimize Spline Models for Time Series Resampling
#'
#' @description
#'
#' Internal function used in [zoo_resample()]. It finds the optimal `df` parameter of a smoothing spline model `y ~ x` fitted with [stats::smooth.spline()] that minimizes the root mean squared error (RMSE) between observations and predictions, and returns a model fitted with such `df`.
#'
#' @param x (required, numeric vector) predictor, a time vector coerced to numeric. Default: NULL
#' @param y (required, numeric vector) response, a column of a zoo object. Default: NULL
#' @param max_complexity (required, logical). If TRUE, RMSE optimization is ignored, and the model of maximum complexity is returned. Default: FALSE
#'
#' @return Object of class "smooth.spline".
#' @export
#' @autoglobal
#' @examples
#'
#' #zoo time series
#' xy <- zoo_simulate(
#'   cols = 1,
#'   rows = 30
#' )
#'
#' #optimize spline model
#' m <- utils_optimize_spline(
#'   x = as.numeric(zoo::index(xy)), #predictor
#'   y = xy[, 1] #response
#' )
#'
#' print(m)
#'
#' #plot observation
#' plot(
#'   x = zoo::index(xy),
#'   y = xy[, 1],
#'   col = "forestgreen",
#'   type = "l",
#'   lwd = 2
#'   )
#'
#' #plot prediction
#' points(
#'   x = zoo::index(xy),
#'   y = stats::predict(
#'     object = m,
#'     x = as.numeric(zoo::index(xy))
#'   )$y,
#'   col = "red"
#' )
#' 
#' @family tsl_processing_internal
utils_optimize_spline <- function(
    x = NULL,
    y = NULL,
    max_complexity = FALSE
){

  if(!is.numeric(x)){
    stop("distantia::utils_optimize_spline(): argument 'x' must be a numeric vector.", call. = FALSE)
  }

  if(!is.numeric(y)){
    stop("distantia::utils_optimize_spline(): argument 'y' must be a numeric vector.", call. = FALSE)
  }

  if(length(x) != length(y)){
    stop("distantia::utils_optimize_spline(): arguments 'x' and 'y' must be of the same length.", call. = FALSE)
  }

  names(x) <- NULL
  names(y) <- NULL

  model_df <- data.frame(
    x = x,
    y = y
  )

  complexity_space <- seq(
    from = floor(nrow(model_df)/2),
    to = nrow(model_df),
    by = 1
  )

  #max complexity switch
  if(max_complexity == TRUE){

    complexity_value <- max(complexity_space)

  } else {



    rmse <- foreach::foreach(
      complexity_value = complexity_space,
      .combine = "c",
      .errorhandling = "pass",
      .options.future = list(seed = FALSE)
    ) %dofuture% {

      m <- tryCatch({
        stats::smooth.spline(
          x = model_df$x,
          y = model_df$y,
          df = complexity_value,
          all.knots = TRUE
        )
      }, error = function(e) {
        NA
      }, warning = function(w) {
        NA
      })

      if(
        inherits(
          x = m,
          what = "smooth.spline"
        ) == FALSE
      ){
        return(NA)
      }

      #return rmse
      return(
        sqrt(
          mean(
            (model_df$y - stats::predict(
              object = m,
              x = model_df$x)$y
            )^2
          )
        )
      )

    }

    #set errors to NA
    rmse[!is.numeric(rmse)] <- max(rmse, na.rm = TRUE)

    #select df minimizing rmse
    complexity_value <- complexity_space[which.min(rmse)]

  }

  #new model
  m <- tryCatch({
    stats::smooth.spline(
      x = model_df$x,
      y = model_df$y,
      df = complexity_value,
      all.knots = TRUE
    )
  }, error = function(e) {
    NA
  }, warning = function(w) {
    NA
  })

  m

}

#' Subset Time Series Lists by Time Series Names, Time, and/or Column Names
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param names (optional, character or numeric vector) Character vector of names or numeric vector with list indices. If NULL, all time series are kept. Default: NULL
#' @param colnames (optional, character vector) Column names of the zoo objects in `tsl`. If NULL, all columns are returned. Default: NULL
#' @param time (optional, numeric vector) time vector of length two used to subset rows by time. If NULL, all rows in `tsl` are preserved. Default: NULL
#' @param numeric_cols (optional, logical) If TRUE, only the numeric columns of the zoo objects are returned. Default: TRUE
#' @param shared_cols (optional, logical) If TRUE, only columns shared across all zoo objects are returned. Default: TRUE
#'
#' @return time series list
#' @export
#' @autoglobal
#' @examples
#' #initialize time series list
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #checking available dimensions
#'
#' #names
#' tsl_names_get(
#'   tsl = tsl
#' )
#'
#' #colnames
#' tsl_colnames_get(
#'   tsl = tsl
#' )
#'
#' #time
#' tsl_time(
#'   tsl = tsl
#' )[, c("name", "begin", "end")]
#'
#' #subset
#' tsl_new <- tsl_subset(
#'   tsl = tsl,
#'   names = c("Sweden", "Germany"),
#'   colnames = c("rainfall", "temperature"),
#'   time = c("2010-01-01", "2015-01-01")
#' )
#'
#' #check new dimensions
#'
#' #names
#' tsl_names_get(
#'   tsl = tsl_new
#' )
#'
#' #colnames
#' tsl_colnames_get(
#'   tsl = tsl_new
#' )
#'
#' #time
#' tsl_time(
#'   tsl = tsl_new
#' )[, c("name", "begin", "end")]
#' @family tsl_management
tsl_subset <- function(
    tsl = NULL,
    names = NULL,
    colnames = NULL,
    time = NULL,
    numeric_cols = TRUE,
    shared_cols = TRUE
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #coerce zoo vectors to matrices
  tsl <- future.apply::future_lapply(
    X = tsl,
    FUN = zoo_vector_to_matrix
  )

  # subset names ----
  if(!is.null(names)){

    ## names is a character ----
    if(is.character(names)){

      names <- match.arg(
        arg = names,
        choices = names(tsl),
        several.ok = TRUE
      ) |>
        unique()

    }

    # names is numeric ----
    if(is.numeric(names)){

      names <- names[
        names > 0 &
          names <= length(tsl)
      ] |>
        as.integer() |>
        unique()

    }

    tsl <- tsl[names]

  }

  #get names to track missing time series
  zoo_names <- tsl_names_get(
    tsl = tsl,
    zoo = TRUE
  )

  # subset colnames ----
  if(!is.null(colnames)){

    #get all variable names
    all_colnames <- tsl_colnames_get(
      tsl = tsl,
      names = "all"
    ) |>
      unlist() |>
      unique()

    colnames <- match.arg(
      arg = colnames,
      choices = all_colnames,
      several.ok = TRUE
    ) |>
      unique()

    tsl <- future.apply::future_lapply(
      X = tsl,
      FUN = function(x){

        x_name <- attributes(x)$name

        x <- x[, colnames, drop = FALSE]

        zoo_name_set(
          x = x,
          name = x_name
        )

      }
    )

  }

  # subset numeric cols
  if(numeric_cols == TRUE){

    #returns NA if no columns are numeric
    tsl <- future.apply::future_lapply(
      X = tsl,
      FUN = function(x){

        x.num.cols <- apply(
          X = x,
          MARGIN = 2,
          FUN = is.numeric
        )

        if(all(x.num.cols) == FALSE){
          return(NA)
        }

        y <- x[
          ,
          x.num.cols,
          drop = FALSE
        ]

        zoo_name_set(
          x = y,
          name = attributes(x)$name
        )

      }
    )

    #names of elements with no numeric columns
    tsl.na <- names(tsl)[is.na(tsl)]

    if(length(tsl.na) > 0){

      tsl <- tsl[!names(tsl) %in% tsl.na]

      warning(
        "distantia::tsl_subset(): the following zoo object/s in 'tsl' have no numeric columns and have been removed: ",
        paste(tsl.na, collapse = ", "),
        ".",
        call. = FALSE
        )

    }

  }

  #subset shared cols
  if(
    shared_cols == TRUE &&
    length(tsl) > 1
    ){

    exclusive_cols <- tsl_colnames_get(
      tsl = tsl,
      names = "exclusive"
    ) |>
      unlist() |>
      unique() |>
      stats::na.omit()

    shared_cols <- tsl_colnames_get(
      tsl = tsl,
      names = "shared"
    ) |>
      unlist() |>
      unique() |>
      stats::na.omit()

    if(length(shared_cols) == 0){

      warning("distantia::tsl_subset(): zoo objects within 'tsl' have no shared columns. Ignoring subsetting of shared columns.", call. = FALSE)

    } else {

      tsl <- future.apply::future_lapply(
        X = tsl,
        FUN = function(x){

          y <- x[
            ,
            colnames(x) %in% shared_cols,
            drop = FALSE
          ]

          zoo_name_set(
            x = y,
            name = attributes(x)$name
          )

        }
      )

    }

  }

  # subset time ----
  if(!is.null(time)){

    if(length(time) < 2){
      stop("distantia::tsl_subset(): argument 'time' must be of length 2.", call. = FALSE)
    }

    if(length(time) > 2){
      time <- range(time)
    }

    tsl_time_df <- tsl_time(
      tsl = tsl
    )

    tsl_time_range <- range(
      c(
        tsl_time_df$begin,
        tsl_time_df$end
      )
    )

    time <- utils_coerce_time_class(
      x = time,
      to = tsl_time_df$class[1]
    ) |>
      range()

    if(
      max(time) < min(tsl_time_range) ||
      min(time) > max(tsl_time_range)
    ){

      warning(
        "distantia::tsl_subset(): Argument time must be a vector of class '",
        tsl_time_df$class[1],
        " ' with values between ",
        min(tsl_time_range),
        " and ",
        max(tsl_time_range),
        ". Ignoring subset by time.",
        call. = FALSE
      )

    } else {

      #subset by time
      tsl <- future.apply::future_lapply(
        X = tsl,
        FUN = function(x){

          y <- stats::window(
            x = x,
            start = min(time),
            end = max(time)
          )

          if(nrow(y) == 0){
            return(NULL)
          }

          zoo_name_set(
            x = y,
            name = attributes(x)$name
          )

        }

      )

      #remove NULL elements
      tsl <- Filter(Negate(is.null), tsl)

      tsl_removed <- setdiff(
        x = zoo_names,
        y = tsl_names_get(
          tsl = tsl,
          zoo = TRUE
        )
      )

      if(length(tsl_removed) > 0){
        warning(
          "distantia::tsl_subset(): There following time series do not overlap with 'time' and have been removed: ",
          paste0(tsl_removed, collapse = ", "),
          ".",
          call. = FALSE
        )

      }

    }

  }

  tsl

}

#' Split Dissimilarity Analysis Data Frames by Combinations of Arguments
#'
#' @description
#' Internal function to split a distantia data frame by groups of the arguments 'distance', 'diagonal', and 'lock_step'.
#'
#' @param df (required, data frame) Output of [distantia()] or [distantia_aggregate()]. Default: NULL.
#'
#' @return A list.
#' @export
#' @autoglobal
#' @examples
#' #Three time series
#' #Climate and NDVI in Fagus sylvatica stands in Spain, Germany, and Sweden
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #Dissimilarity analysis with four combinations of parameters
#' df <- distantia(
#'   tsl = tsl,
#'   distance = c(
#'     "euclidean",
#'     "manhattan"
#'     ),
#'   lock_step = c(
#'     TRUE,
#'     FALSE
#'   )
#' )
#'
#' #split by combinations of parameters
#' df_split <- utils_distantia_df_split(
#'   df = df
#' )
#'
#' #print output
#' df_split
#'
#' #class and length of the output
#' class(df_split)
#' length(df_split)
#' @family internal
utils_distantia_df_split <- function(
    df = NULL
){

  df_type <- attributes(df)$type

  #required columns
  required_columns <- c(
    "distance",
    "diagonal",
    "lock_step"
  )

  if(any(!(required_columns %in% names(df)))){
    return(list(df))
  }

  # detect groups ----
  df_groups <- unique(
    df[, required_columns]
  )

  df_groups$group <- seq_len(nrow(df_groups))

  rownames(df_groups) <- NULL

  #add group to df ----
  df <- merge(
    x = df,
    y = df_groups
  )

  # split ----
  df_list <- split(
    x = df,
    f = df$group
  )

  #set attributes
  for(i in seq_len(length(df_list))){

    attr(
      x = df_list[[i]],
      which = "type"
    ) <- df_type

  }

  df_list

}

#' Transform Raw Time Series Data to Time Series List
#'
#' @description
#'
#' Most functions in this package take a **time series list** (or **tsl** for short) as main input. A `tsl` is a list of zoo time series objects (see [zoo::zoo()]). There is not a formal class for `tsl` objects, but there are requirements these objects must follow to ensure the stability of the package functionalities (see [tsl_diagnose()]). These requirements are:
#' \itemize{
#'   \item There are no NA, Inf, -Inf, or NaN cases in the zoo objects (see [tsl_count_NA()] and [tsl_handle_NA()]).
#'   \item All zoo objects must have at least one common column name to allow time series comparison (see [tsl_colnames_get()]).
#'   \item All zoo objects have a character attribute "name" identifying the object. This attribute is not part of the zoo class, but the package ensures that this attribute is not lost during data manipulations.
#'   \item Each element of the time series list is named after the zoo object it contains (see [tsl_names_get()], [tsl_names_set()] and [tsl_names_clean()]).
#'   \item The time series list contains two zoo objects or more.
#' }
#'
#' The function [tsl_initialize()] (and its alias [tsl_init()]) is designed to convert the following data structures to a time series list:
#'
#' \itemize{
#'   \item Long data frame: with an ID column to separate time series, and a time column that can be of the classes "Date", "POSIXct", "integer", or "numeric". The resulting zoo objects and list elements are named after the values in the ID column.
#'   \item Wide data frame: each column is a time series representing the same variable observed at the same time in different places. Each column is converted to a separate zoo object and renamed.
#'   \item List of vectors: an object like `list(a = runif(10), b = runif(10))` is converted to a time series list with as many zoo objects as vectors are defined in the original list.
#'   \item List of matrices: a list containing matrices, such as `list(a = matrix(runif(30), 10, 3), b = matrix(runif(36), 12, 3))`.
#'   \item List of zoo objects: a list with zoo objects, such as `list(a = zoo_simulate(), b = zoo_simulate())`.
#' }
#'
#' @param x (required, list or data frame) Matrix or data frame in long format, list of vectors, list of matrices, or list of zoo objects. Default: NULL.
#' @param name_column (optional, column name) Column naming individual time series. Numeric names are converted to character with the prefix "X". Default: NULL
#' @param time_column (optional if `lock_step = FALSE`, and required otherwise, character string) Name of the column representing time, if any. Default: NULL.
#' @param lock_step (optional, logical) If TRUE, all input sequences are subsetted to their common times according to the values in `time_column`.
#' @return list of zoo objects (a time series list)
#' @examples
#' #long data frame
#' #---------------------
#' data("fagus_dynamics")
#'
#' #name_column is name
#' #time column is time
#' str(fagus_dynamics)
#'
#' #to tsl
#' #each group in name_column is a different time series
#' tsl <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #check validity (no messages or errors if valid)
#' tsl_diagnose(tsl)
#'
#' #class of contained objects
#' lapply(X = tsl, FUN = class)
#'
#' #get list and zoo names (between double quotes)
#' tsl_names_get(
#'   tsl = tsl,
#'   zoo = TRUE
#'   )
#'
#' #plot tsl
#' if(interactive()){
#'   tsl_plot(tsl)
#' }
#'
#' #list of zoo objects
#' #--------------------
#' x <- zoo_simulate()
#' y <- zoo_simulate()
#'
#' tsl <- tsl_initialize(
#'   x = list(
#'     x = x,
#'     y = y
#'   )
#' )
#'
#' #plot
#' if(interactive()){
#'   tsl_plot(tsl)
#' }
#'
#'
#' #wide data frame
#' #--------------------
#' #wide data frame
#' #each column is same variable in different places
#' df <- stats::reshape(
#'   data = fagus_dynamics[, c(
#'     "name",
#'     "time",
#'     "evi"
#'   )],
#'   timevar = "name",
#'   idvar = "time",
#'   direction = "wide",
#'   sep = "_"
#' )
#'
#' str(df)
#'
#' #to tsl
#' #key assumptions:
#' #all columns but "time" represent
#' #the same variable in different places
#' #all time series are of the same length
#' tsl <- tsl_initialize(
#'   x = df,
#'   time_column = "time"
#'   )
#'
#' #colnames are forced to be the same...
#' tsl_colnames_get(tsl)
#'
#' #...but can be changed
#' tsl <- tsl_colnames_set(
#'   tsl = tsl,
#'   names = "evi"
#' )
#' tsl_colnames_get(tsl)
#'
#' #plot
#' if(interactive()){
#'   tsl_plot(tsl)
#' }
#'
#'
#' #list of vectors
#' #---------------------
#' #create list of vectors
#' vector_list <- list(
#'   a = cumsum(stats::rnorm(n = 50)),
#'   b = cumsum(stats::rnorm(n = 70)),
#'   c = cumsum(stats::rnorm(n = 20))
#' )
#'
#' #to tsl
#' #key assumptions:
#' #all vectors represent the same variable
#' #in different places
#' #time series can be of different lengths
#' #no time column, integer indices are used as time
#' tsl <- tsl_initialize(
#'   x = vector_list
#' )
#'
#' #plot
#' if(interactive()){
#'   tsl_plot(tsl)
#' }
#'
#' #list of matrices
#' #-------------------------
#' #create list of matrices
#' matrix_list <- list(
#'   a = matrix(runif(30), nrow = 10, ncol = 3),
#'   b = matrix(runif(80), nrow = 20, ncol = 4)
#' )
#'
#' #to tsl
#' #key assumptions:
#' #each matrix represents a multivariate time series
#' #in a different place
#' #all multivariate time series have the same columns
#' #no time column, integer indices are used as time
#' tsl <- tsl_initialize(
#'   x = matrix_list
#' )
#'
#' #check column names
#' tsl_colnames_get(tsl = tsl)
#'
#' #remove exclusive column
#' tsl <- tsl_subset(
#'   tsl = tsl,
#'   shared_cols = TRUE
#'   )
#' tsl_colnames_get(tsl = tsl)
#'
#' #plot
#' if(interactive()){
#'   tsl_plot(tsl)
#' }
#'
#' #list of zoo objects
#' #-------------------------
#' zoo_list <- list(
#'   a = zoo_simulate(),
#'   b = zoo_simulate()
#' )
#'
#' #looks like a time series list! But...
#' tsl_diagnose(tsl = zoo_list)
#'
#' #let's set the names
#' zoo_list <- tsl_names_set(tsl = zoo_list)
#'
#' #check again: it's now a valid time series list
#' tsl_diagnose(tsl = zoo_list)
#'
#' #to do all this in one go:
#' tsl <- tsl_initialize(
#'   x = list(
#'     a = zoo_simulate(),
#'     b = zoo_simulate()
#'   )
#' )
#'
#' #plot
#' if(interactive()){
#'   tsl_plot(tsl)
#' }
#'
#' @autoglobal
#' @export
#' @family tsl_initialize
tsl_initialize <- function(
    x = NULL,
    name_column = NULL,
    time_column = NULL,
    lock_step = FALSE
){

  if(is.null(x)){
    stop("distantia::tsl_initialize(): argument 'x' must not be NULL")
  }

  if(is.list(x) && !is.data.frame(x) && length(x) < 2){
    stop(
      "distantia::tsl_initialize(): argument 'x' must have at least 2 time series.",
      call. = FALSE
    )
  }

  x <- utils_prepare_matrix(
    x = x
  )

  x <- utils_prepare_df(
    x = x,
    name_column = name_column,
    time_column = time_column
  )

  x <- utils_prepare_vector_list(
    x = x
  )

  x <- utils_prepare_matrix_list(
    x = x
  )

  x <- utils_prepare_time(
    x = x,
    time_column = time_column,
    lock_step = lock_step
  )

  tsl <- utils_prepare_zoo_list(
    x = x,
    time_column = time_column
  )

  tsl <- tsl_names_set(
    tsl = tsl
    )

  tsl_diagnose(
    tsl = tsl,
    full = TRUE
  )

  tsl

}

#' @rdname tsl_initialize
#' @export
tsl_init <- tsl_initialize

#' Data Frame to Distance Matrix
#'
#' @param df (required, data frame) Data frame with numeric columns to transform into a distance matrix. Default: NULL
#' @inheritParams tsl_initialize
#' @inheritParams distantia
#' @return square matrix
#' @export
#' @autoglobal
#' @examples
#' #compute distance matrix
#' m <- distance_matrix(
#'   df = cities_coordinates,
#'   name_column = "name",
#'   distance = "euclidean"
#' )
#'
#' #get data used to compute the matrix
#' attributes(m)$df
#'
#' #check matrix
#' m
#'
#' @family distances
distance_matrix <- function(
    df = NULL,
    name_column = NULL,
    distance = "euclidean"
){

  # df <- fagus_coordinates
  # name_column <- "name"

  if(is.null(df)){
    stop(
      "distantia::d_matrix(): argument 'df' must not be NULL.",
      call. = FALSE
    )
  }

  #check distance arg
  distance <- utils_check_distance_args(
    distance = distance
  )

  #drop geometry
  df <- distantia::utils_drop_geometry(
    df = df
  )

  #subset names
  if(
    !is.null(name_column) &&
    name_column %in% colnames(df)
  ){

    name <- as.character(df[[name_column]])
    df[[name_column]] <- NULL

  } else {
    name <- NULL
  }


  #subset numeric columns
  df <- df[,
           sapply(
             X = df,
             FUN = is.numeric
             ),
           drop = FALSE
           ]

  if(ncol(df) == 0){
    stop(
      "distantia::d_matrix(): argument 'df' must have at least one numeric column.",
      call. = FALSE
    )
  }

  #coerce to matrix
  df_matrix <- as.matrix(df)

  #compute distance matrix
  m <- distance_matrix_cpp(
    x = df_matrix,
    y = df_matrix,
    distance = distance
  )

  #add col and row names
  if(!is.null(name)){
    dimnames(m) <- list(
      as.character(name),
      as.character(name)
    )
  }

  #attributes
  attr(x = m, which = "type") <- "distance"
  attr(x = m, which = "distance") <- distance
  attr(x = m, which = "df") <- df


  m

}

test_that("`distance_matrix()` works", {
  m <- distance_matrix(
    df = cities_coordinates,
    name_column = "name",
    distance = "euclidean"
  )
  expect_equal(attributes(m)$distance, "euclidean")
  expect_equal(ncol(m), nrow(cities_coordinates))
  expect_equal(nrow(m), nrow(cities_coordinates))
})

#' Summary Statistics of Dissimilarity Data Frame
#' @description
#' Computes summary statistics from the output of [distantia()], returning a data frame with one row per time series and summary statistics of its dissimilarity scores with all other time series.
#'
#' @inheritParams distantia_aggregate
#'
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#' tsl <- tsl_simulate(
#'   n = 5,
#'   irregular = FALSE
#'   )
#'
#' df <- distantia(
#'   tsl = tsl,
#'   lock_step = TRUE
#'   )
#'
#' df_stats <- distantia_stats(df = df)
#'
#' df_stats
#' @family distantia_support
#' 
distantia_stats <- function(
    df = NULL
){

  df_type <- attributes(df)$type

  if(df_type != "distantia_df"){
    stop("distantia::distantia_stats(): argument 'df' must be the output of distantia().", call. = FALSE)
  }

  df <- distantia_aggregate(
    df = df
  )

  #disaggregate pairs
  df <- rbind(
    data.frame(
      name = df[["x"]],
      psi = df[["psi"]]
    ),
    data.frame(
      name = df[["y"]],
      psi = df[["psi"]]
    )
  )

  #warn and drop NA psi before computing stats
  df_has_na <- is.na(df[["psi"]])
  if(any(df_has_na)){
    warning(
      "distantia::distantia_stats(): ",
      sum(df_has_na),
      " NA psi value(s) excluded from summary statistics.",
      call. = FALSE
    )
    df <- df[!df_has_na, ]
  }

  #stats functions
  q1 <- function(x = NULL){
    stats::quantile(
      x = x,
      probs = 0.25,
      na.rm = TRUE
    )
  }

  q3 <- function(x = NULL){
    stats::quantile(
      x = x,
      probs = 0.75,
      na.rm = TRUE
    )
  }

  f <- c(
    mean,
    min,
    q1,
    stats::median,
    q3,
    max,
    sd
  )

  names(f) <- c(
    "mean",
    "min",
    "q1",
    "median",
    "q3",
    "max",
    "sd"
  )

  #progress bar and iterator
  p <- progressr::progressor(along = f)

  #computing stats
  stats_list <- foreach::foreach(
    i = seq_len(length(f)),
    .errorhandling = "pass",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

    p()

    df.i <- stats::aggregate(
      psi ~ name,
      data = df,
      FUN = f[[i]]
    )

    colnames(df.i) <- c(
      "name", names(f)[i]
    )

    df.i

  }

  df_stats <- Reduce(
    f = function(x, y){
      merge(x, y, by = "name")
    },
    x = stats_list
  )

  df_stats$range <- df_stats$max - df_stats$min

  df_stats

}

#' Handles Line Colors for Sequence Plots
#'
#' @description
#' Internal function that can be used to better understand how line colors are handled within other plotting functions.
#'
#'
#' @param x (required, sequence) zoo object or time series list. Default: NULL
#' @param line_color (optional, character vector) vector of colors for the time series columns. Selected palette depends on the number of columns to plot. Default: NULL
#'
#' @return color vector
#' @export
#' @family internal_plotting
utils_line_color <- function(
    x = NULL,
    line_color = NULL
){

  # get colnames for any length of x
  if(!is.matrix(x)){

    x_colnames <- lapply(
      X = x,
      FUN = colnames
    ) |>
      unlist() |>
      unique() |>
      sort()

  } else {

    #names of columns to plot
    x_colnames <- colnames(x)

  }

  #generate default colors
  if(is.null(line_color)){

    line_color <- color_discrete(
      n = length(x_colnames)
    )

  }


  #if more colors than required are provided
  if(length(line_color) > length(x_colnames)){

    #if no names, subset by length
    if(is.null(names(line_color))){

      line_color <- line_color[
        seq(
          from = 1,
          to = length(line_color),
          length.out = length(x_colnames))
      ]

      #if names, subset by name
    } else {

      line_color <- line_color[names(line_color) %in% x_colnames]

    }

  }

  #if fewer colors than required are provided
  if(length(line_color) < length(x_colnames)){

    stop(
      "distantia::utils_line_color(): There are fewer colors (",
      length(line_color),
      ") than unique columns (",
      length(x_colnames),
      "). Argument 'line_color' must be a vector with ",
      length(x_colnames),
      " colors.",
      call. = FALSE
    )

  }

  names(line_color) <- x_colnames

  line_color

}

#' Distantia Boxplot
#'
#' @description
#' Boxplot of a data frame returned by [distantia()] showing the summary statistics of psi scores for each time series against all others.
#'
#' @inheritParams distantia_aggregate
#' @param fill_color (optional, character vector) boxplot fill color. Default: NULL
#' @param f (optional, function) function used to aggregate the input data frame and arrange the boxes. One of `mean` or `median`. Default: `median`.
#' @param text_cex (optional, numeric) Multiplier of the text size. Default: 1
#'
#' @return boxplot
#' @export
#' @autoglobal
#' @examples
#' tsl <- tsl_initialize(
#'   x = distantia::albatross,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_global
#'   )
#'
#' df <- distantia(
#'   tsl = tsl,
#'   lock_step = TRUE
#'   )
#'
#' distantia_boxplot(
#'   df = df,
#'   text_cex = 1.5
#'   )
#' @family distantia_support
distantia_boxplot <- function(
    df = NULL,
    fill_color = NULL,
    f = median,
    text_cex = 1
){

  old_par <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old_par))

  if(attributes(df)$type != "distantia_df"){
    stop(
      "distantia::distantia_boxplot(): argument 'df' must be the output of distantia::distantia()",
      call. = FALSE
    )
  }

  #aggregate df if needed
  df <- distantia_aggregate(
    df = df,
    f = f
  )

  #boxplot
  utils_boxplot_common(
    variable = c(df$x, df$y),
    value = c(df$psi, df$psi),
    fill_color = fill_color,
    f = f,
    main = "Dissimilarity Summary",
    xlab = "Psi",
    ylab = "",
    text_cex = text_cex
  )

}

#' Guide for Time Series Plots
#'
#' @param x (required, sequence) a zoo object or a time series list. Default: NULL
#' @param position (optional, vector of xy coordinates or character string). This is a condensed version of the `x` and `y` arguments of the [graphics::legend()] function. Coordinates (in the range 0 1) or keyword to position the legend. Accepted keywords are: "bottomright", "bottom", "bottomleft", "left", "topleft", "top", "topright", "right" and "center". Default: "topright".
#' @param line_color (optional, character vector) vector of colors for the time series columns. If NULL, uses the palette "Zissou 1" provided by the function [grDevices::hcl.colors()]. Default: NULL
#' @param line_width (optional, numeric vector) Widths of the time series lines. Default: 1
#' @param length (optional, numeric) maps to the argument `seg.len` of [graphics::legend()]. Length of the lines drawn in the legend. Default: 1
#' @param text_cex (optional, numeric) Multiplier of the text size. Default: 0.7
#' @param guide_columns (optional, integer) number of columns for the legend items. Default: 1.
#' @param subpanel (optional, logical) internal argument used when generating the multipanel plot produced by [distantia_dtw_plot()].
#'
#' @return plot
#' @examples
#' x <- zoo_simulate()
#'
#' if(interactive()){
#'
#'   zoo_plot(x, guide = FALSE)
#'
#'   utils_line_guide(
#'     x = x,
#'     position = "right"
#'   )
#'
#' }
#'
#' @autoglobal
#' @export
#' @family internal_plotting
utils_line_guide <- function(
    x = NULL,
    position = "topright",
    line_color = NULL,
    line_width = 1,
    length = 1,
    text_cex = 0.7,
    guide_columns = 1,
    subpanel = FALSE
){

  title_cex <- 0.8 * text_cex
  axis_labels_cex <- 0.7 * text_cex
  title_distance <- 0.75

  #default palette
  line_color <- utils_line_color(
    x = x,
    line_color = line_color
  )

  #width
  if(length(line_width) == 1){
    line_width <- rep(
      x = line_width,
      times = length(line_color)
    )
  }

  if(length(line_width) != length(line_color)){
    if(line_width[1] < line_width[length(line_width)]){
      line_width <- seq(
        from = min(line_width),
        to = max(line_width),
        length.out = length(line_color)
      )
    } else {
      line_width <- seq(
        from = max(line_width),
        to = min(line_width),
        length.out = length(line_color)
      )
    }
  }


  if(is.numeric(position)){
    position_x <- position[1]
    position_y <- position[2]
  } else {
    position_x <- position
    position_y <- NULL
  }

  #plot legend
  if(subpanel == TRUE){

    graphics::plot(
      NULL,
      xaxt = 'n',
      yaxt = 'n',
      bty = 'n',
      frame.plot = FALSE,
      ylab = '',
      xlab = '',
      xlim = 0:1,
      ylim = 0:1
    )

  }

  xcoords <- c(0, cumsum(graphics::strwidth(names(line_color), cex = 0.5))[-length(names(line_color))])
  secondvector <- (1:length(names(line_color)))-1
  textwidths <- xcoords/secondvector
  textwidths[1] <- 0

  #when not a subpanel, let the function figure it out by itself
  if(subpanel == TRUE){
    textwidths <- NA
  }

  graphics::legend(
    x = position_x,
    y = position_y,
    legend = names(line_color),
    col = line_color,
    border = "black",
    lwd = line_width,
    bty = 'n',
    cex = text_cex,
    ncol = guide_columns,
    horiz = FALSE,
    title = NULL,
    y.intersp = text_cex,
    x.intersp = text_cex,
    xjust = 0.5,
    text.width = textwidths,
    seg.len = length
  )



}

#' Distance Methods
#'
#' A data frame with the names, abbreviations, and expressions of the distance metrics implemented in the package.
#'
#' @docType data
#' @keywords datasets
#' @family distances
#' @name distances
#' @usage data(distances)
#' @format A data frame with 5 columns and 10 rows.
"distances"

#' Time Series of Covid Prevalence in California Counties
#'
#' Dataset with Covid19 maximum weekly prevalence in California counties between 2020 and 2024, from healthdata.gov.
#'
#' County polygons and additional data for this dataset are in [covid_counties].
#'
#' The full dataset at daily resolution can be downloaded from [https://github.com/BlasBenito/distantia/blob/main/data_full/covid_prevalence.rda](https://github.com/BlasBenito/distantia/blob/main/data_full/covid_prevalence.rda) (use the "Download raw file" button).
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name covid_prevalence
#' @usage data(covid_prevalence)
#' @examples
#' #Convert to time series list
#' tsl <- tsl_initialize(
#'   x = covid_prevalence,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #time series plot
#' if(interactive()){
#'
#'  #Subset to avoid margin errors
#'  tsl_plot(
#'   tsl = tsl_subset(
#'     tsl = tsl,
#'     names = 1:4
#'     ),
#'   guide = FALSE
#'   )
#'
#' }
#' @format A data frame with 3 columns and 51,048 rows.
"covid_prevalence"

#' County Coordinates of the Covid Prevalence Dataset
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name covid_counties
#' @usage data(covid_counties)
#' @format An sf data frame with county polygons and census data.
"covid_counties"

#' Time Series Data from Three Fagus sylvatica Stands
#'
#' A data frame with 648 rows representing enhanced vegetation index, rainfall and temperature in three stands of Fagus sylvatica in Spain, Germany, and Sweden.
#'
#' Site coordinates for this dataset are in [fagus_coordinates].
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name fagus_dynamics
#' @usage data(fagus_dynamics)
#' @examples
#' data("fagus_dynamics")
#'
#' #Convert to time series list
#' fagus <- tsl_initialize(
#'   x = fagus_dynamics,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #time series plot
#' if(interactive()){
#'
#'  tsl_plot(
#'   tsl = fagus
#'   )
#'
#' }
#'
#' @format A data frame with 5 columns and 648 rows.
"fagus_dynamics"

#' Site Coordinates of Fagus sylvatica Stands
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name fagus_coordinates
#' @usage data(fagus_coordinates)
#' @format An sf data frame with 3 rows and 4 columns.
"fagus_coordinates"


#' Pollen Counts of Nine Interglacial Sites in Central Europe
#'
#' @description
#' Pollen counts of nine interglacial sites in central Europe.
#'
#' Site coordinates for this dataset are in [eemian_coordinates].
#'
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name eemian_pollen
#' @usage data(eemian_pollen)
#' @examples
#' data("eemian_pollen")
#'
#' #Convert to time series list
#' tsl <- tsl_initialize(
#'   x = eemian_pollen,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #time series plot
#' if(interactive()){
#'
#'  tsl_plot(
#'   tsl = tsl_subset(
#'     tsl = tsl,
#'     names = 1:3
#'     ),
#'   columns = 2,
#'   guide_columns = 2
#'   )
#'
#' }
#'
#' @format A data frame with 24 columns and 376 rows.
"eemian_pollen"

#' Site Coordinates of Nine Interglacial Sites in Central Europe
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name eemian_coordinates
#' @usage data(eemian_coordinates)
#' @format An sf data frame with 4 columns and 9 rows.
"eemian_coordinates"

#' Long Term Monthly Temperature in 20 Major Cities
#'
#' @description
#' Average temperatures between 1975 and 2010 of 20 major cities of the world. [Source](https://www.kaggle.com/datasets/berkeleyearth/climate-change-earth-surface-temperature-data?resource=download&select=GlobalLandTemperaturesByMajorCity.csv).
#'
#' Site coordinates for this dataset are in [cities_coordinates].
#'
#' The full dataset with 100 cities can be downloaded from [https://github.com/BlasBenito/distantia/blob/main/data_full/cities_temperature.rda](https://github.com/BlasBenito/distantia/blob/main/data_full/cities_temperature.rda) (use the "Download raw file" button).
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name cities_temperature
#' @usage data(cities_temperature)
#' @examples
#' data("cities_temperature")
#'
#' #Convert to time series list
#' cities <- tsl_initialize(
#'   x = cities_temperature,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #time series plot
#' if(interactive()){
#'
#'  #Only four cities are shown
#'  tsl_plot(
#'   tsl = tsl_subset(
#'     tsl = tsl,
#'     names = 1:4
#'     ),
#'   guide = FALSE
#'   )
#'
#' }
#' @format A data frame with 3 columns and 52,100 rows.
"cities_temperature"


#' Coordinates of 100 Major Cities
#'
#' @description
#' City coordinates and several environmental variables for the dataset `cities_temperature`.
#'
#' The full dataset with 100 cities can be downloaded from [https://github.com/BlasBenito/distantia/blob/main/data_full/cities_coordinates.rda](https://github.com/BlasBenito/distantia/blob/main/data_full/cities_coordinates.rda) (use the "Download raw file" button).
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name cities_coordinates
#' @usage data(cities_coordinates)
#' @format An sf data frame with 5 columns and 100 rows.
"cities_coordinates"


#' Flight Path Time Series of Albatrosses in the Pacific
#'
#' @description
#' Daily mean flight path data of four individuals of Waved Albatross (Phoebastria irrorata) captured via GPS during the summer of 2008. An sf data frame with columns name, time, latitude, longitude, ground speed, heading, and (uncalibrated) temperature.
#'
#' The full dataset at hourly resolution can be downloaded from [https://github.com/BlasBenito/distantia/blob/main/data_full/albatross.rda](https://github.com/BlasBenito/distantia/blob/main/data_full/albatross.rda) (use the "Download raw file" button).
#' @references \doi{10.5441/001/1.3hp3s250}
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name albatross
#' @usage data(albatross)
#' @examples
#' #Load as TSL
#' #Scale all variables
#' #Aggregate to daily resolution
#' #Align all time series to same temporal span
#' tsl <- tsl_initialize(
#'   x = albatross,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_transform(
#'     f = f_scale_local
#'   ) |>
#'   tsl_aggregate(
#'     new_time = "days"
#'   )
#'
#' if(interactive()){
#'   tsl_plot(
#'     tsl = tsl,
#'     guide_columns = 5
#'     )
#' }
#' @format A data frame.
"albatross"



#' Rainfall and Temperature in The Americas
#'
#' @description
#' Monthly temperature and rainfall between 2009 and 2019 in 72 hexagonal cells covering The Americas.
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name honeycomb_climate
#' @usage data(honeycomb_climate)
"honeycomb_climate"


#' Hexagonal Grid
#'
#' @description
#' Sf data frame with hexagonal grid of the dataset [honeycomb_climate].
#'
#' @docType data
#' @keywords datasets
#' @family example_data
#' @name honeycomb_polygons
#' @usage data(honeycomb_polygons)
"honeycomb_polygons"

#' Get Number of Columns in Time Series Lists
#'
#' @param tsl (required, list) Time series list. Default: NULL
#'
#' @return list
#' @export
#' @examples
#' #Initialize time series list
#' tsl <- tsl_simulate(
#'   n = 2,
#'   cols = 6
#' )
#'
#' #Number of columns per zoo object
#' tsl_ncol(tsl = tsl)
#' @autoglobal
#' @family tsl_management
tsl_ncol <- function(
    tsl = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  lapply(
    X = tsl,
    FUN = function(x){
      ncol(x)
    }
  )

}

#' Auto Breaks for Matrix Plotting Functions
#'
#' @param m (required, numeric matrix) distance or cost matrix generated by [psi_distance_matrix()] or [psi_cost_matrix()], but any numeric matrix will work. Default: NULL
#' @param n (required, integer) number of colors to compute the breaks for. Default: 100
#'
#' @return numeric vector
#' @autoglobal
#' @export
#' @family internal_plotting
utils_color_breaks <- function(
    m = NULL,
    n = 100
    ){

  m <- utils_check_args_matrix(m = m)

  m.range <- range(m, na.rm = TRUE)

  if(n == 1L) return(c(m.range[1] - 0.5, m.range[2] + 0.5))

  a <- seq(
    from = m.range[1],
    to = m.range[2],
    length.out = n
  )

  b <- (a[2]- a[1])/2

  breaks <- c(a[1] - b, a + b)

  breaks

}

#' Number of Decimal Places
#'
#' Computes the number of decimal places in a numeric value.
#'
#' @param x (required, numeric) Numeric value. Default: NULL
#'
#' @return integer
#' @export
#' @autoglobal
#' @examples
#' utils_digits(x = 0.234)
#' @family internal
utils_digits <- function(
    x = NULL
    ){

  out <- 0

  if(abs(x - round(x)) > .Machine$double.eps^0.5){

    #remove terminal zeros
    x <- sub(
      pattern = '0+$',
      replacement = '',
      x = as.character(x)
    )

    #number of decimals
    out <- nchar(
      x = strsplit(
        x = x,
        split = ".",
        fixed = TRUE
      )[[1]][[2]]
    )

  }

  out

}

test_that("`utils_digits()` works", {
  expect_equal(utils_digits(x = 0), 0)
  expect_equal(utils_digits(x = 0.2), 1)
  expect_equal(utils_digits(x = 0.23), 2)
  expect_equal(utils_digits(x = 0.234), 3)
})

#' Data Frame with Supported Time Units
#'
#' @description
#' Returns a data frame with the names of the supported time units, the classes that can handle each time unit, and the threshold used to identify which time units can be used when aggregating a time series.
#'
#' @param all_columns (optional, logical) If TRUE, all columns are returned. Default: FALSE
#' @param class (optional, class name). Used to filter rows and columns. Accepted values are "numeric", "Date", and "POSIXct". Default: NULL
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#'
#' df <- utils_time_units()
#' head(df)
#' @family internal_time_handling
utils_time_units <- function(
    all_columns = FALSE,
    class = NULL
    ){

  df <- data.frame(
    factor = c(
      1000,
      100,
      10,
      NA,
      NA,
      NA,
      NA,
      NA,
      NA,
      NA,
      NA,
      as.numeric("1e8"),
      as.numeric("1e7"),
      as.numeric("1e6"),
      as.numeric("1e5"),
      as.numeric("1e4"),
      as.numeric("1e3"),
      as.numeric("1e2"),
      as.numeric("1e1"),
      as.numeric("1"),
      as.numeric("1e-1"),
      as.numeric("1e-2"),
      as.numeric("1e-3"),
      as.numeric("1e-4"),
      as.numeric("1e-5"),
      as.numeric("1e-6"),
      as.numeric("1e-7"),
      as.numeric("1e-8")
    ),
    keyword = c(
      FALSE,
      FALSE,
      FALSE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE
    ),
    base_units = c(
      "days",
      "days",
      "days",
      "days",
      "days",
      "days",
      "days",
      "days",
      "hours",
      "mins",
      "secs",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "numeric"
    ),
    units = c(
      "millennia",
      "centuries",
      "decades",
      "years",
      "quarters",
      "months",
      "weeks",
      "days",
      "hours",
      "minutes",
      "seconds",
      "1e8",
      "1e7",
      "1e6",
      "1e5",
      "1e4",
      "1e3",
      "1e2",
      "1e1",
      "1",
      "1e-1",
      "1e-2",
      "1e-3",
      "1e-4",
      "1e-5",
      "1e-6",
      "1e-7",
      "1e-8"
    ),
    Date = c(
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE
    ),
    POSIXct = c(
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE
    ),
    numeric = c(
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE
    ),
    integer = c(
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE,
      TRUE
    ),
    #used in zoo_time() only
    threshold = c(
      365000,
      36500,
      3650,
      365,
      90,
      30,
      7,
      1,
      24,
      60,
      60,
      1e8,
      1e7,
      1e6,
      1e5,
      1e4,
      1e3,
      1e2,
      1e1,
      1,
      1e-1,
      1e-2,
      1e-3,
      1e-4,
      1e-5,
      1e-6,
      1e-7,
      1e-8
    )
  )

  if(!is.null(class)){

    if("POSIXct" %in% class){
      class <- "POSIXct"
    }

    if(class %in% c(
      "Date",
      "POSIXct",
      "numeric",
      "integer"
    )){

      df <- df[
        df[[class]] == TRUE,
        c("factor", "keyword", "base_units", "units", class, "threshold")]

    }

  }

  if(all_columns == FALSE){
    df$factor <- NULL
    df$threshold <- NULL
    df$keyword <- NULL
  }

  df
}

test_that("`utils_time_units()` works", {
  df <- utils_time_units()
  expect_equal(class(df), "data.frame")
  expect_equal(colnames(df), c("base_units", "units", "Date", "POSIXct", "numeric", "integer"))
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' (C++) Sum Distances Between Consecutive Samples in a Time Series
#' @description Computes the cumulative sum of distances between consecutive
#' samples in a univariate or multivariate time series.
#' NA values should be removed before using this function.
#' @param x (required, numeric matrix) univariate or multivariate time series.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean"
#' @return numeric
#' @examples
#' #simulate a time series
#' x <- zoo_simulate()
#'
#' #compute auto distance
#' auto_distance_cpp(
#'   x = x,
#'   distance = "euclidean"
#'   )
#' @export
#' @family Rcpp_auto_sum
auto_distance_cpp <- function(x, distance = "euclidean") {
    .Call(`_distantia_auto_distance_cpp`, x, distance)
}

#' (C++) Subset Matrix by Rows
#' @description Subsets a time series matrix to the coordinates of a trimmed
#' least-cost path when blocks are ignored during a dissimilarity analysis.
#' @param m (required, numeric matrix) a univariate or multivariate time series.
#' @param rows (required, integer vector) vector of rows to subset from a
#' least-cost path data frame.
#' @return numeric matrix
#' @examples
#' #simulate a time series
#' m <- zoo_simulate(seed = 1)
#'
#' #sample some rows
#' rows <- sample(
#'   x = nrow(m),
#'   size = 10
#'   ) |>
#'   sort()
#'
#' #subset by rows
#' m_subset <- subset_matrix_by_rows_cpp(
#'   m = m,
#'   rows = rows
#'   )
#'
#' #compare with original
#' m[rows, ]
#'
#' @export
#' @family Rcpp_auto_sum
subset_matrix_by_rows_cpp <- function(m, rows) {
    .Call(`_distantia_subset_matrix_by_rows_cpp`, m, rows)
}

#' (C++) Sum Distances Between All Consecutive Samples in Two Time Series
#' @description Computes the cumulative auto sum of autodistances of two time series.
#' The output value is used as normalization factor when computing dissimilarity scores.
#' @param x (required, numeric matrix) univariate or multivariate time series.
#' @param y (required, numeric matrix) univariate or multivariate time series
#' with the same number of columns as 'x'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean"
#' @return numeric
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #auto sum
#' auto_sum_full_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#' @export
#' @family Rcpp_auto_sum
auto_sum_full_cpp <- function(x, y, distance = "euclidean") {
    .Call(`_distantia_auto_sum_full_cpp`, x, y, distance)
}

#' (C++) Sum Distances Between All Consecutive Samples in the Least Cost Path Between Two Time Series
#' @description Computes the cumulative auto sum of auto-distances of two time series
#' for the coordinates of a trimmed least cost path. The output value is used
#' as normalization factor when computing dissimilarity scores.
#' @param x (required, numeric matrix) univariate or multivariate time series.
#' @param y (required, numeric matrix) univariate or multivariate time series
#' with the same number of columns as 'x'.
#' @param path (required, data frame) least-cost path produced by [cost_path_orthogonal_cpp()].
#' Default: NULL
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @return numeric
#' @export
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' #least cost matrix
#' cost_matrix <- cost_matrix_orthogonal_cpp(
#'   dist_matrix = dist_matrix
#' )
#'
#' #least cost path
#' cost_path <- cost_path_orthogonal_cpp(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' nrow(cost_path)
#'
#' #remove blocks from least-cost path
#' cost_path_trimmed <- cost_path_trim_cpp(
#'   path = cost_path
#' )
#'
#' nrow(cost_path_trimmed)
#'
#' #auto sum
#' auto_sum_path_cpp(
#'   x = x,
#'   y = y,
#'   path = cost_path_trimmed,
#'   distance = "euclidean"
#' )
#' @export
#' @family Rcpp_auto_sum
auto_sum_path_cpp <- function(x, y, path, distance = "euclidean") {
    .Call(`_distantia_auto_sum_path_cpp`, x, y, path, distance)
}

#' (C++) Sum Distances Between Consecutive Samples in Two Time Series
#' @description Sum of auto-distances of two time series.
#' This function switches between [auto_sum_full_cpp()] and [auto_sum_path_cpp()]
#' depending on the value of the argument `ignore_blocks`.
#' @param x (required, numeric matrix) of same number of columns as 'y'.
#' @param y (required, numeric matrix) of same number of columns as 'x'.
#' @param path (required, data frame) output of [cost_path_orthogonal_cpp()].
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean"
#' @param ignore_blocks (optional, logical). If TRUE, blocks of consecutive path
#' coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.
#' @return numeric
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' #least cost matrix
#' cost_matrix <- cost_matrix_orthogonal_cpp(
#'   dist_matrix = dist_matrix
#' )
#'
#' #least cost path
#' cost_path <- cost_path_orthogonal_cpp(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' nrow(cost_path)
#'
#' #remove blocks from least-cost path
#' cost_path_trimmed <- cost_path_trim_cpp(
#'   path = cost_path
#' )
#'
#' nrow(cost_path_trimmed)
#'
#' #auto sum
#' auto_sum_cpp(
#'   x = x,
#'   y = y,
#'   path = cost_path_trimmed,
#'   distance = "euclidean",
#'   ignore_blocks = FALSE
#' )
#' @export
#' @family Rcpp_auto_sum
auto_sum_cpp <- function(x, y, path, distance = "euclidean", ignore_blocks = FALSE) {
    .Call(`_distantia_auto_sum_cpp`, x, y, path, distance, ignore_blocks)
}

#' (C++) Compute Orthogonal and Diagonal Least Cost Matrix from a Distance Matrix
#' @description Computes the least cost matrix from a distance matrix.
#' Considers diagonals during computation of least-costs.
#' @param dist_matrix (required, distance matrix). Square distance matrix, output of [distance_matrix_cpp()].
#' @return Least cost matrix.
#' @export
#' @family Rcpp_matrix
cost_matrix_diagonal_cpp <- function(dist_matrix) {
    .Call(`_distantia_cost_matrix_diagonal_cpp`, dist_matrix)
}

#' (C++) Compute Orthogonal and Weighted Diagonal Least Cost Matrix from a Distance Matrix
#' @description Computes the least cost matrix from a distance matrix.
#' Weights diagonals by a factor of 1.414214 (square root of 2) with respect to orthogonal paths.
#' @param dist_matrix (required, distance matrix). Distance matrix.
#' @return Least cost matrix.
#' @export
#' @family Rcpp_matrix
cost_matrix_diagonal_weighted_cpp <- function(dist_matrix) {
    .Call(`_distantia_cost_matrix_diagonal_weighted_cpp`, dist_matrix)
}

#' (C++) Compute Orthogonal Least Cost Matrix from a Distance Matrix
#' @description Computes the least cost matrix from a distance matrix.
#' @param dist_matrix (required, distance matrix). Output of [distance_matrix_cpp()].
#' @return Least cost matrix.
#' @export
#' @family Rcpp_matrix
cost_matrix_orthogonal_cpp <- function(dist_matrix) {
    .Call(`_distantia_cost_matrix_orthogonal_cpp`, dist_matrix)
}

#' (C++) Least Cost Path for Sequence Slotting
#' @description Computes a least-cost matrix from a distance matrix.
#' This version differs from [cost_path_orthogonal_cpp()] in the way it solves ties.
#' In the case of a tie, [cost_path_orthogonal_cpp()] uses the first neighbor satisfying
#' the minimum distance condition, while this function selects the neighbor
#' that changes the axis of movement within the least-cost matrix. This function
#' is not used anywhere within the package, but was left here for future reference.
#' @param dist_matrix (required, numeric matrix). Distance matrix between two
#' time series.
#' @param cost_matrix (required, numeric matrix). Least-cost matrix generated from
#' `dist_matrix`.
#' @return data frame
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' #least cost matrix
#' cost_matrix <- cost_matrix_orthogonal_cpp(
#'   dist_matrix = dist_matrix
#' )
#'
#' #least cost path
#' cost_path <- cost_path_slotting_cpp(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' cost_path
#' @export
#' @family Rcpp_cost_path
cost_path_slotting_cpp <- function(dist_matrix, cost_matrix) {
    .Call(`_distantia_cost_path_slotting_cpp`, dist_matrix, cost_matrix)
}

#' (C++) Orthogonal Least Cost Path
#' @description
#' Computes an orthogonal least-cost path within a cost matrix. Each steps within
#' the least-cost path either moves in the x or the y direction, but never diagonally.
#' @param dist_matrix (required, numeric matrix). Distance matrix between two
#' time series.
#' @param cost_matrix (required, numeric matrix). Cost matrix generated from
#' `dist_matrix`.
#' @param bandwidth (required, numeric) Size of the Sakoe-Chiba band at
#' both sides of the diagonal used to constrain the least cost path. Expressed
#' as a fraction of the number of matrix rows and columns. Unrestricted by default.
#' Default: 1
#' @return data frame
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' #least cost matrix
#' cost_matrix <- cost_matrix_orthogonal_cpp(
#'   dist_matrix = dist_matrix
#' )
#'
#' #least cost path
#' cost_path <- cost_path_orthogonal_cpp(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' cost_path
#' @export
#' @family Rcpp_cost_path
cost_path_orthogonal_bandwidth_cpp <- function(dist_matrix, cost_matrix, bandwidth = 1) {
    .Call(`_distantia_cost_path_orthogonal_bandwidth_cpp`, dist_matrix, cost_matrix, bandwidth)
}

#' (C++) Orthogonal Least Cost Path
#' @description
#' Computes an orthogonal least-cost path within a cost matrix. Each steps within
#' the least-cost path either moves in the x or the y direction, but never diagonally.
#' @param dist_matrix (required, numeric matrix). Distance matrix between two
#' time series.
#' @param cost_matrix (required, numeric matrix). Cost matrix generated from
#' `dist_matrix`.
#' @return data frame
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' #least cost matrix
#' cost_matrix <- cost_matrix_orthogonal_cpp(
#'   dist_matrix = dist_matrix
#' )
#'
#' #least cost path
#' cost_path <- cost_path_orthogonal_cpp(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' cost_path
#' @export
#' @family Rcpp_cost_path
cost_path_orthogonal_cpp <- function(dist_matrix, cost_matrix) {
    .Call(`_distantia_cost_path_orthogonal_cpp`, dist_matrix, cost_matrix)
}

#' (C++) Orthogonal and Diagonal Least Cost Path Restricted by Sakoe-Chiba band
#' @description Computes the least cost matrix from a distance matrix.
#' Considers diagonals during computation of least-costs. In case of ties,
#' diagonals are favored.
#' @param dist_matrix (required, numeric matrix). Distance matrix between two
#' time series.
#' @param cost_matrix (required, numeric matrix). Cost matrix generated from
#' `dist_matrix`.
#' @param bandwidth (required, numeric) Size of the Sakoe-Chiba band at
#' both sides of the diagonal used to constrain the least cost path. Expressed
#' as a fraction of the number of matrix rows and columns. Unrestricted by default.
#' Default: 1
#' @return data frame
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' #least cost matrix
#' cost_matrix <- cost_matrix_orthogonal_cpp(
#'   dist_matrix = dist_matrix
#' )
#'
#' #least cost path
#' cost_path <- cost_path_diagonal_cpp(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' cost_path
#' @export
#' @family Rcpp_cost_path
cost_path_diagonal_bandwidth_cpp <- function(dist_matrix, cost_matrix, bandwidth = 1) {
    .Call(`_distantia_cost_path_diagonal_bandwidth_cpp`, dist_matrix, cost_matrix, bandwidth)
}

#' (C++) Orthogonal and Diagonal Least Cost Path
#' @description Computes the least cost matrix from a distance matrix.
#' Considers diagonals during computation of least-costs. In case of ties,
#' diagonals are favored.
#' @param dist_matrix (required, numeric matrix). Distance matrix between two
#' time series.
#' @param cost_matrix (required, numeric matrix). Cost matrix generated from
#' `dist_matrix`.
#' @return data frame
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' #least cost matrix
#' cost_matrix <- cost_matrix_orthogonal_cpp(
#'   dist_matrix = dist_matrix
#' )
#'
#' #least cost path
#' cost_path <- cost_path_diagonal_cpp(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' cost_path
#' @export
#' @family Rcpp_cost_path
cost_path_diagonal_cpp <- function(dist_matrix, cost_matrix) {
    .Call(`_distantia_cost_path_diagonal_cpp`, dist_matrix, cost_matrix)
}

#' (C++) Remove Blocks from a Least Cost Path
#' @param path (required, data frame) least-cost path produced by [cost_path_orthogonal_cpp()].
#' @return data frame
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' #least cost matrix
#' cost_matrix <- cost_matrix_orthogonal_cpp(
#'   dist_matrix = dist_matrix
#' )
#'
#' #least cost path
#' cost_path <- cost_path_slotting_cpp(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' nrow(cost_path)
#'
#' #remove blocks from least-cost path
#' cost_path_trimmed <- cost_path_trim_cpp(
#'   path = cost_path
#' )
#'
#' nrow(cost_path_trimmed)
#' @export
#' @family Rcpp_cost_path
cost_path_trim_cpp <- function(path) {
    .Call(`_distantia_cost_path_trim_cpp`, path)
}

#' (C++) Sum Distances in a Least Cost Path
#' @param path (required, data frame) least-cost path produced by [cost_path_orthogonal_cpp()].
#' @return numeric
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' #least cost matrix
#' cost_matrix <- cost_matrix_orthogonal_cpp(
#'   dist_matrix = dist_matrix
#' )
#'
#' #least cost path
#' cost_path <- cost_path_slotting_cpp(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' cost_path_sum_cpp(
#'   path = cost_path
#'   )
#' @export
#' @family Rcpp_cost_path
cost_path_sum_cpp <- function(path) {
    .Call(`_distantia_cost_path_sum_cpp`, path)
}

#' Least Cost Path
#' @description Least cost path between two time series \code{x} and \code{y}.
#' NA values must be removed from \code{x} and \code{y} before using this function.
#' If the selected distance function is "chi" or "cosine", pairs of zeros should
#' be either removed or replaced with pseudo-zeros (i.e. 0.00001).
#' @param x (required, numeric matrix) multivariate time series.
#' @param y (required, numeric matrix) multivariate time series
#' with the same number of columns as 'x'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @param diagonal (optional, logical). If TRUE, diagonals are included in the
#' computation of the cost matrix. Default: TRUE.
#' @param weighted (optional, logical). Only relevant when diagonal is TRUE. When TRUE,
#' diagonal cost is weighted by y factor of 1.414214 (square root of 2). Default: TRUE.
#' @param ignore_blocks (optional, logical). If TRUE, blocks of consecutive path
#' coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.
#' @param bandwidth (required, numeric) Size of the Sakoe-Chiba band at
#' both sides of the diagonal used to constrain the least cost path. Expressed
#' as a fraction of the number of matrix rows and columns. Unrestricted by default.
#' Default: 1
#' @return data frame
#' @export
#' @family Rcpp_cost_path
cost_path_cpp <- function(x, y, distance = "euclidean", diagonal = TRUE, weighted = TRUE, ignore_blocks = FALSE, bandwidth = 1) {
    .Call(`_distantia_cost_path_cpp`, x, y, distance, diagonal, weighted, ignore_blocks, bandwidth)
}

#' (C++) Distance Matrix of Two Time Series
#' @description Computes the distance matrix between the rows of two matrices
#' \code{y} and \code{x} representing regular or irregular time series with the same number of
#' columns. NA values should be removed before using this function. If the selected distance function is "chi" or "cosine", pairs of zeros should
#' be either removed or replaced with pseudo-zeros (i.e. 0.00001).
#' @param x (required, numeric matrix) univariate or multivariate time series.
#' @param y (required, numeric matrix) univariate or multivariate time series
#' with the same number of columns as 'x'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @return numeric matrix
#' @examples
#' #simulate two time series
#' x <- zoo_simulate(seed = 1)
#' y <- zoo_simulate(seed = 2)
#'
#' #distance matrix
#' dist_matrix <- distance_matrix_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#' @export
#' @family Rcpp_matrix
#' @name distance_matrix_cpp
distance_matrix_cpp <- function(x, y, distance = "euclidean") {
    .Call(`_distantia_distance_matrix_cpp`, x, y, distance)
}

#' (C++) Sum of Pairwise Distances Between Cases in Two Aligned Time Series
#' @description Computes the lock-step sum of distances between two regular
#' and aligned time series. NA values should be removed before using this function.
#' If the selected distance function is "chi" or "cosine", pairs of zeros should
#' be either removed or replaced with pseudo-zeros (i.e. 0.00001).
#' @param x (required, numeric matrix) univariate or multivariate time series.
#' @param y (required, numeric matrix) univariate or multivariate time series
#' with the same number of columns and rows as 'x'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @return numeric
#' @examples
#' #simulate two regular time series
#' x <- zoo_simulate(
#'   seed = 1,
#'   irregular = FALSE
#'   )
#' y <- zoo_simulate(
#'   seed = 2,
#'   irregular = FALSE
#'   )
#'
#' #distance matrix
#' dist_matrix <- distance_ls_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#' @export
#' @family Rcpp_matrix
distance_ls_cpp <- function(x, y, distance = "euclidean") {
    .Call(`_distantia_distance_ls_cpp`, x, y, distance)
}

#' (C++) Chebyshev Distance Between Two Vectors
#' @description Computed as: \code{max(abs(x - y))}. Cannot handle NA values.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_chebyshev_cpp(x = runif(100), y = runif(100))
#' @export
#' @family Rcpp_distance_methods
distance_chebyshev_cpp <- function(x, y) {
    .Call(`_distantia_distance_chebyshev_cpp`, x, y)
}

#' (C++) Jaccard Distance Between Two Binary Vectors
#' @description Computes the Jaccard distance between two binary vectors.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_jaccard_cpp(x = c(0, 1, 0, 1), y = c(1, 1, 0, 0))
#' @export
#' @family Rcpp_distance_methods
distance_jaccard_cpp <- function(x, y) {
    .Call(`_distantia_distance_jaccard_cpp`, x, y)
}

#' (C++) Manhattan Distance Between Two Vectors
#' @description Computed as: \code{sum(abs(x - y))}. Cannot handle NA values.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_manhattan_cpp(x = runif(100), y = runif(100))
#' @export
#' @family Rcpp_distance_methods
distance_manhattan_cpp <- function(x, y) {
    .Call(`_distantia_distance_manhattan_cpp`, x, y)
}

#' (C++) Euclidean Distance Between Two Vectors
#' @description Computed as: \code{sqrt(sum((x - y)^2)}. Cannot handle NA values.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_euclidean_cpp(x = runif(100), y = runif(100))
#' @export
#' @family Rcpp_distance_methods
distance_euclidean_cpp <- function(x, y) {
    .Call(`_distantia_distance_euclidean_cpp`, x, y)
}

#' (C++) Hellinger Distance Between Two Vectors
#' @description Computed as: \code{sqrt(1/2 * sum((sqrt(x) - sqrt(y))^2))}.
#' Cannot handle NA values.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_hellinger_cpp(x = runif(100), y = runif(100))
#' @export
#' @family Rcpp_distance_methods
distance_hellinger_cpp <- function(x, y) {
    .Call(`_distantia_distance_hellinger_cpp`, x, y)
}

#' (C++) Normalized Chi Distance Between Two Vectors
#' @description Computed as:
#' \code{xy <- x + y}
#' \code{y. <- y / sum(y)}
#' \code{x. <- x / sum(x)}
#' \code{sqrt(sum(((x. - y.)^2) / (xy / sum(xy))))}.
#' Cannot handle NA values. When \code{x} and \code{y} have zeros in the same
#' position, \code{NaNs} are produced. Please replace these zeros with
#' pseudo-zeros (i.e. 0.0001) if you wish to use this distance metric.
#' @examples distance_chi_cpp(x = runif(100), y = runif(100))
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @export
#' @family Rcpp_distance_methods
distance_chi_cpp <- function(x, y) {
    .Call(`_distantia_distance_chi_cpp`, x, y)
}

#' (C++) Canberra Distance Between Two Binary Vectors
#' @description Computes the Canberra distance between two binary vectors.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_canberra_cpp(c(0, 1, 0, 1), c(1, 1, 0, 0))
#' @export
#' @family Rcpp_distance_methods
distance_canberra_cpp <- function(x, y) {
    .Call(`_distantia_distance_canberra_cpp`, x, y)
}

#' (C++) Russell-Rao Distance Between Two Binary Vectors
#' @description Computes the Russell-Rao distance between two binary vectors.
#' @param x (required, numeric). Binary vector of 1s and 0s.
#' @param y (required, numeric) Binary vector of 1s and 0s of same length as `x`.
#' @return numeric
#' @examples distance_russelrao_cpp(c(0, 1, 0, 1), c(1, 1, 0, 0))
#' @export
#' @family Rcpp_distance_methods
distance_russelrao_cpp <- function(x, y) {
    .Call(`_distantia_distance_russelrao_cpp`, x, y)
}

#' (C++) Cosine Dissimilarity Between Two Vectors
#' @description Computes the cosine dissimilarity between two numeric vectors.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_cosine_cpp(c(0.2, 0.4, 0.5), c(0.1, 0.8, 0.2))
#' @export
#' @family Rcpp_distance_methods
distance_cosine_cpp <- function(x, y) {
    .Call(`_distantia_distance_cosine_cpp`, x, y)
}

#' (C++) Hamming Distance Between Two Binary Vectors
#' @description Computes the Hamming distance between two binary vectors.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_hamming_cpp(c(0, 1, 0, 1), c(1, 1, 0, 0))
#' @export
#' @family Rcpp_distance_methods
distance_hamming_cpp <- function(x, y) {
    .Call(`_distantia_distance_hamming_cpp`, x, y)
}

#' (C++) Bray-Curtis Distance Between Two Vectors
#' @description Computes the Bray-Curtis distance, suitable for species abundance data.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_bray_curtis_cpp(x = runif(100), y = runif(100))
#' @export
#' @family Rcpp_distance_methods
distance_bray_curtis_cpp <- function(x, y) {
    .Call(`_distantia_distance_bray_curtis_cpp`, x, y)
}

#' (C++) Sørensen Distance Between Two Binary Vectors
#' @description Computes the Sørensen distance, suitable for presence/absence data.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @return numeric
#' @examples distance_sorensen_cpp(x = c(0, 1, 1, 0), y = c(1, 1, 0, 0))
#' @export
#' @family Rcpp_distance_methods
distance_sorensen_cpp <- function(x, y) {
    .Call(`_distantia_distance_sorensen_cpp`, x, y)
}

update_path_dist_cpp <- function(x, y, path, distance = "euclidean") {
    .Call(`_distantia_update_path_dist_cpp`, x, y, path, distance)
}

select_column_cpp <- function(x, column_index) {
    .Call(`_distantia_select_column_cpp`, x, column_index)
}

delete_column_cpp <- function(x, column_index) {
    .Call(`_distantia_delete_column_cpp`, x, column_index)
}

#' (C++) Contribution of Individual Variables to the Dissimilarity Between Two Aligned Time Series
#' @description Computes the contribution of individual variables to the
#' similarity/dissimilarity between two aligned multivariate time series.
#' This function generates a data frame with the following columns:
#' \itemize{
#'   \item variable: name of the individual variable for which the importance
#'   is being computed, from the column names of the arguments `x` and `y`.
#'   \item psi: global dissimilarity score `psi` of the two time series.
#'   \item psi_only_with: dissimilarity between `x` and `y` computed from the given variable alone.
#'   \item psi_without: dissimilarity between `x` and `y` computed from all other variables.
#'   \item psi_difference: difference between `psi_only_with` and `psi_without`.
#'   \item importance: contribution of the variable to the similarity/dissimilarity
#'   between `x` and `y`, computed as `(psi_difference * 100) / psi_all`.
#'   Positive scores represent contribution to dissimilarity,
#'   while negative scores represent contribution to similarity.
#' }
#' @param x (required, numeric matrix) multivariate time series.
#' @param y (required, numeric matrix) multivariate time series
#' with the same number of columns and rows as 'x'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @return data frame
#' @examples
#' #simulate two regular time series
#' x <- zoo_simulate(
#'   seed = 1,
#'   irregular = FALSE
#'   )
#'
#' y <- zoo_simulate(
#'   seed = 2,
#'   irregular = FALSE
#'   )
#'
#' #same number of rows
#' nrow(x) == nrow(y)
#'
#' #compute importance
#' df <- importance_ls_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' df
#' @family Rcpp_importance
#' @export
importance_ls_cpp <- function(x, y, distance = "euclidean") {
    .Call(`_distantia_importance_ls_cpp`, x, y, distance)
}

#' (C++) Contribution of Individual Variables to the Dissimilarity Between Two Time Series (Legacy Version)
#' @description Computes the contribution of individual variables to the
#' similarity/dissimilarity between two irregular multivariate time series.
#' In opposition to the robust version, least-cost paths for each combination
#' of variables are computed independently, which makes the results of individual
#' variables harder to compare. This function should only be used when the objective is
#' replicating importance scores generated with previous versions of the package `distantia`.
#' This function generates a data frame with the following columns:
#' \itemize{
#'   \item variable: name of the individual variable for which the importance
#'   is being computed, from the column names of the arguments `x` and `y`.
#'   \item psi: global dissimilarity score `psi` of the two time series.
#'   \item psi_only_with: dissimilarity between `x` and `y` computed from the given variable alone.
#'   \item psi_without: dissimilarity between `x` and `y` computed from all other variables.
#'   \item psi_difference: difference between `psi_only_with` and `psi_without`.
#'   \item importance: contribution of the variable to the similarity/dissimilarity
#'   between `x` and `y`, computed as `((psi_all - psi_without) * 100) / psi_all`.
#'   Positive scores represent contribution to dissimilarity,
#'   while negative scores represent contribution to similarity.
#' }
#' @param x (required, numeric matrix) multivariate time series.
#' @param y (required, numeric matrix) multivariate time series
#' with the same number of columns as 'x'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @param diagonal (optional, logical). If TRUE, diagonals are included in the
#' computation of the cost matrix. Default: TRUE.
#' @param weighted (optional, logical). Only relevant when diagonal is TRUE. When TRUE,
#' diagonal cost is weighted by y factor of 1.414214 (square root of 2). Default: TRUE.
#' @param ignore_blocks (optional, logical). If TRUE, blocks of consecutive path
#' coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.
#' @param bandwidth (required, numeric) Size of the Sakoe-Chiba band at
#' both sides of the diagonal used to constrain the least cost path. Expressed
#' as a fraction of the number of matrix rows and columns. Unrestricted by default.
#' Default: 1
#' @return data frame
#' @examples
#' #simulate two regular time series
#' x <- zoo_simulate(
#'   seed = 1,
#'   rows = 100
#'   )
#'
#' y <- zoo_simulate(
#'   seed = 2,
#'   rows = 150
#'   )
#'
#' #different number of rows
#' #this is not a requirement though!
#' nrow(x) == nrow(y)
#'
#' #compute importance
#' df <- importance_dtw_legacy_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' df
#' @family Rcpp_importance
#' @export
importance_dtw_legacy_cpp <- function(y, x, distance = "euclidean", diagonal = FALSE, weighted = TRUE, ignore_blocks = FALSE, bandwidth = 1) {
    .Call(`_distantia_importance_dtw_legacy_cpp`, y, x, distance, diagonal, weighted, ignore_blocks, bandwidth)
}

#' (C++) Contribution of Individual Variables to the Dissimilarity Between Two Time Series (Robust Version)
#' @description Computes the contribution of individual variables to the
#' similarity/dissimilarity between two irregular multivariate time series.
#' In opposition to the legacy version, importance computation is
#' performed taking the least-cost path of the whole sequence as reference. This
#' operation makes the importance scores of individual variables fully comparable.
#' This function generates a data frame with the following columns:
#' \itemize{
#'   \item variable: name of the individual variable for which the importance
#'   is being computed, from the column names of the arguments `x` and `y`.
#'   \item psi: global dissimilarity score `psi` of the two time series.
#'   \item psi_only_with: dissimilarity between `x` and `y` computed from the given variable alone.
#'   \item psi_without: dissimilarity between `x` and `y` computed from all other variables.
#'   \item psi_difference: difference between `psi_only_with` and `psi_without`.
#'   \item importance: contribution of the variable to the similarity/dissimilarity
#'   between `x` and `y`, computed as `(psi_difference * 100) / psi_all`.
#'   Positive scores represent contribution to dissimilarity,
#'   while negative scores represent contribution to similarity.
#' }
#' @param x (required, numeric matrix) multivariate time series.
#' @param y (required, numeric matrix) multivariate time series
#' with the same number of columns as 'x'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @param diagonal (optional, logical). If TRUE, diagonals are included in the
#' computation of the cost matrix. Default: TRUE.
#' @param weighted (optional, logical). Only relevant when diagonal is TRUE. When TRUE,
#' diagonal cost is weighted by y factor of 1.414214 (square root of 2). Default: TRUE.
#' @param ignore_blocks (optional, logical). If TRUE, blocks of consecutive path
#' coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.
#' @param bandwidth (required, numeric) Size of the Sakoe-Chiba band at
#' both sides of the diagonal used to constrain the least cost path. Expressed
#' as a fraction of the number of matrix rows and columns. Unrestricted by default.
#' Default: 1
#' @return data frame
#' @examples
#' #simulate two regular time series
#' x <- zoo_simulate(
#'   seed = 1,
#'   rows = 100
#'   )
#'
#' y <- zoo_simulate(
#'   seed = 2,
#'   rows = 150
#'   )
#'
#' #different number of rows
#' #this is not a requirement though!
#' nrow(x) == nrow(y)
#'
#' #compute importance
#' df <- importance_dtw_cpp(
#'   x = x,
#'   y = y,
#'   distance = "euclidean"
#' )
#'
#' df
#' @family Rcpp_importance
#' @export
importance_dtw_cpp <- function(x, y, distance = "euclidean", diagonal = TRUE, weighted = TRUE, ignore_blocks = FALSE, bandwidth = 1) {
    .Call(`_distantia_importance_dtw_cpp`, x, y, distance, diagonal, weighted, ignore_blocks, bandwidth)
}

#' (C++) Restricted Permutation of Complete Rows Within Blocks
#' @description Divides a sequence in blocks of a given size and permutes rows
#' within these blocks.
#' Larger block sizes increasingly disrupt the data structure over time.
#' @param x (required, numeric matrix). Numeric matrix to permute.
#' @param block_size (optional, integer) block size in number of rows.
#' Minimum value is 2, and maximum value is nrow(x).
#' @param seed (optional, integer) random seed to use.
#' @return numeric matrix
#' @family Rcpp_permutation
#' @export
permute_restricted_by_row_cpp <- function(x, block_size, seed = 1L) {
    .Call(`_distantia_permute_restricted_by_row_cpp`, x, block_size, seed)
}

#' (C++) Unrestricted Permutation of Complete Rows
#' @description Unrestricted shuffling of rows within the whole sequence.
#' @param x (required, numeric matrix). Numeric matrix to permute.
#' @param block_size (optional, integer) this function ignores this argument and sets it to x.nrow().
#' @param seed (optional, integer) random seed to use.
#' @return numeric matrix
#' @family Rcpp_permutation
#' @export
permute_free_by_row_cpp <- function(x, block_size, seed = 1L) {
    .Call(`_distantia_permute_free_by_row_cpp`, x, block_size, seed)
}

#' (C++) Restricted Permutation of Cases Within Blocks
#' @description Divides a sequence or time series in blocks and permutes cases
#' within these blocks. This function does not preserve rows, and should not be
#' used if the sequence has dependent columns.
#' Larger block sizes increasingly disrupt the data structure over time.
#' @param x (required, numeric matrix). Numeric matrix to permute.
#' @param block_size (optional, integer) block size in number of rows.
#' Minimum value is 2, and maximum value is nrow(x).
#' @param seed (optional, integer) random seed to use.
#' @return numeric matrix
#' @family Rcpp_permutation
#' @export
permute_restricted_cpp <- function(x, block_size, seed = 1L) {
    .Call(`_distantia_permute_restricted_cpp`, x, block_size, seed)
}

#' (C++) Unrestricted Permutation of Cases
#' @description Unrestricted shuffling of cases within the whole sequence.
#' @param x (required, numeric matrix). Numeric matrix to permute.
#' @param block_size (optional, integer) this function ignores this argument and sets it to x.nrow().
#' @param seed (optional, integer) random seed to use.
#' @return numeric matrix
#' @family Rcpp_permutation
#' @export
permute_free_cpp <- function(x, block_size, seed = 1L) {
    .Call(`_distantia_permute_free_cpp`, x, block_size, seed)
}

#' (C++) Equation of the Psi Dissimilarity Score
#' @description Equation to compute the `psi` dissimilarity score
#' (Birks and Gordon 1985). Psi is computed as \eqn{\psi = (2a / b) - 1},
#' where \eqn{a} is the sum of distances between the relevant samples of two
#' time series, and \eqn{b} is the cumulative sum of distances between
#' consecutive samples in the two time series.
#' If `a` is computed with dynamic time warping, and diagonals are used in the
#' computation of the least cost path, then one is added to the result of the equation above.
#' @param a (required, numeric) output of [cost_path_sum_cpp()] on a least cost path.
#' @param b (required, numeric) auto sum of both sequences,
#' result of [auto_sum_cpp()].
#' @param diagonal (optional, logical). Must be TRUE when diagonals are used in
#' dynamic time warping and for lock-step distances. Default: FALSE.
#' @return numeric
#' @family Rcpp_dissimilarity_analysis
#' @export
psi_equation_cpp <- function(a, b, diagonal = TRUE) {
    .Call(`_distantia_psi_equation_cpp`, a, b, diagonal)
}

#' (C++) Psi Dissimilarity Score of Two Aligned Time Series
#' @description Computes the psi dissimilarity score between two time series
#' observed at the same times. Time series \code{y} and \code{x} with the same
#' number of columns and rows. NA values should be removed before using this function.
#' If the selected distance function is "chi" or "cosine", pairs of zeros should
#' be either removed or replaced with pseudo-zeros (i.e. 0.00001).
#' @param x (required, numeric matrix) of same number of columns as 'y'.
#' @param y (required, numeric matrix) of same number of columns as 'x'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @return numeric
#' @family Rcpp_dissimilarity_analysis
#' @export
psi_ls_cpp <- function(x, y, distance = "euclidean") {
    .Call(`_distantia_psi_ls_cpp`, x, y, distance)
}

#' (C++) Null Distribution of the Dissimilarity Scores of Two Aligned Time Series
#' @description Applies permutation methods to compute null distributions for
#' the psi scores of two time series observed at the same times.
#' NA values should be removed before using this function.
#' If the selected distance function is "chi" or "cosine", pairs of zeros should
#' be either removed or replaced with pseudo-zeros (i.e. 0.00001).
#' @param x (required, numeric matrix) of same number of columns as 'y'.
#' @param y (required, numeric matrix) of same number of columns as 'x'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @param repetitions (optional, integer) number of null psi values to generate. Default: 100
#' @param permutation (optional, character) permutation method. Valid values are listed below from higher to lower randomness:
#' \itemize{
#'   \item "free": unrestricted shuffling of rows and columns. Ignores block_size.
#'   \item "free_by_row": unrestricted shuffling of complete rows. Ignores block size.
#'   \item "restricted": restricted shuffling of rows and columns within blocks.
#'   \item "restricted_by_row": restricted shuffling of rows within blocks.
#' }
#' @param block_size (optional, integer) block size in rows for
#' restricted permutation. A block size of 3 indicates that a row can only be permuted
#' within a block of 3 adjacent rows. Minimum value is 2. Default: 3.
#' @param seed (optional, integer) initial random seed to use for replicability. Default: 1
#' @return numeric vector
#' @family Rcpp_dissimilarity_analysis
#' @export
psi_null_ls_cpp <- function(x, y, distance = "euclidean", repetitions = 100L, permutation = "restricted_by_row", block_size = 3L, seed = 1L) {
    .Call(`_distantia_psi_null_ls_cpp`, x, y, distance, repetitions, permutation, block_size, seed)
}

#' (C++) Psi Dissimilarity Score of Two Time-Series
#' @description Computes the psi score of two time series \code{y} and \code{x}
#' with the same number of columns.
#' NA values should be removed before using this function.
#' If the selected distance function is "chi" or "cosine", pairs of zeros should
#' be either removed or replaced with pseudo-zeros (i.e. 0.00001).
#' @param y (required, numeric matrix) time series.
#' @param x (required, numeric matrix) of same number of columns as 'y'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @param diagonal (optional, logical). If TRUE, diagonals are included in the
#' computation of the cost matrix. Default: TRUE.
#' @param weighted (optional, logical). Only relevant when diagonal is TRUE. When TRUE,
#' diagonal cost is weighted by y factor of 1.414214 (square root of 2). Default: TRUE.
#' @param ignore_blocks (optional, logical). If TRUE, blocks of consecutive path
#' coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.
#' @param bandwidth (required, numeric) Size of the Sakoe-Chiba band at
#' both sides of the diagonal used to constrain the least cost path. Expressed
#' as a fraction of the number of matrix rows and columns. Unrestricted by default.
#' @return numeric
#' @family Rcpp_dissimilarity_analysis
#' @export
psi_dtw_cpp <- function(x, y, distance = "euclidean", diagonal = TRUE, weighted = TRUE, ignore_blocks = FALSE, bandwidth = 1) {
    .Call(`_distantia_psi_dtw_cpp`, x, y, distance, diagonal, weighted, ignore_blocks, bandwidth)
}

#' (C++) Null Distribution of Dissimilarity Scores of Two Time Series
#' @description Applies permutation methods to compute null distributions for
#' the psi scores of two time series.
#' NA values should be removed before using this function.
#' If the selected distance function is "chi" or "cosine", pairs of zeros should
#' be either removed or replaced with pseudo-zeros (i.e. 0.00001).
#' @param y (required, numeric matrix).
#' @param x (required, numeric matrix) of same number of columns as 'y'.
#' @param distance (optional, character string) distance name from the "names"
#' column of the dataset `distances` (see `distances$name`). Default: "euclidean".
#' @param diagonal (optional, logical). If TRUE, diagonals are included in the
#' computation of the cost matrix. Default: FALSE.
#' @param weighted (optional, logical). If TRUE, diagonal is set to TRUE, and
#' diagonal cost is weighted by a factor of 1.414214 (square root of 2). Default: FALSE.
#' @param ignore_blocks (optional, logical). If TRUE, blocks of consecutive path
#' coordinates are trimmed to avoid inflating the psi distance. This argument
#' has nothing to do with block_size!. Default: FALSE.
#' @param bandwidth (required, numeric) Size of the Sakoe-Chiba band at
#' both sides of the diagonal used to constrain the least cost path. Expressed
#' as a fraction of the number of matrix rows and columns. Unrestricted by default.
#' Default: 1
#' @param repetitions (optional, integer) number of null psi values to generate. Default: 100
#' @param permutation (optional, character) permutation method. Valid values are listed below from higher to lower randomness:
#' \itemize{
#'   \item "free": unrestricted shuffling of rows and columns. Ignores block_size.
#'   \item "free_by_row": unrestricted shuffling of complete rows. Ignores block size.
#'   \item "restricted": restricted shuffling of rows and columns within blocks.
#'   \item "restricted_by_row": restricted shuffling of rows within blocks.
#' }
#' @param block_size (optional, integer) block size in rows for
#' restricted permutation. A block size of 3 indicates that a row can only be permuted
#' within a block of 3 adjacent rows. Minimum value is 2. Default: 3.
#' @param seed (optional, integer) initial random seed to use for replicability. Default: 1
#' @return numeric vector
#' @family Rcpp_dissimilarity_analysis
#' @export
psi_null_dtw_cpp <- function(x, y, distance = "euclidean", diagonal = TRUE, weighted = TRUE, ignore_blocks = FALSE, bandwidth = 1, repetitions = 100L, permutation = "restricted_by_row", block_size = 3L, seed = 1L) {
    .Call(`_distantia_psi_null_dtw_cpp`, x, y, distance, diagonal, weighted, ignore_blocks, bandwidth, repetitions, permutation, block_size, seed)
}

test_that("Functions `distantia()`, `distantia_ls()`, and `distantia_dtw()` work.", {

  #short multivariate tsl for testing
  tsl <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  ) |>
    tsl_subset(
      time = c("2010-01-01", "2011-01-01")
    ) |>
    tsl_transform(
      f = f_scale_global
      )

  #params
  distance = c("euclidean", "manhattan")
  permutation = c(
    "restricted_by_row",
    "restricted",
    "free_by_row",
    "free"
    )
  diagonal <- c(TRUE, FALSE)
  bandwidth <- c(1, 0.5)
  lock_step <- c(TRUE, FALSE)
  block_size <- c(3, 6)
  repetitions <- 10
  seed <- c(1, 2)

  #distantia output only
  df <- distantia(
    tsl = tsl,
    distance = distance,
    diagonal = diagonal,
    bandwidth = bandwidth,
    lock_step = lock_step,
    permutation = permutation,
    block_size = block_size,
    repetitions = repetitions,
    seed = seed
  )

  expect_true(all(c("x", "y", "distance", "diagonal", "bandwidth", "lock_step", "repetitions", "permutation", "block_size", "seed", "psi", "p_value", "null_mean", "null_sd") %in% colnames(df)))

  #check for invalid combinations with lock_step
  expect_true(
    nrow(
      df[df$lock_step == TRUE & !is.na(df$diagonal),]
    ) == 0
  )

  expect_true(
    nrow(
      df[df$lock_step == TRUE & !is.na(df$bandwidth),]
    ) == 0
  )

  #check for invalid combinations of permutation and block_size
  expect_true(
    nrow(
      df[df$permutation %in% c("free", "free_by_row") & !is.na(df$block_size),]
    ) == 0
  )

  expect_true(all(names(tsl) %in% unique(c(df$x, df$y))))

  expect_true(
    all(df[df$x == "Germany" & df$y == "Sweden", "psi"] < df[df$x == "Spain" & df$y == "Sweden", "psi"])
  )

  #reproduce p-values
  psi_null <- psi_null_dtw_cpp(
    x = tsl[["Spain"]],
    y = tsl[["Sweden"]],
    distance = "euclidean",
    repetitions = 10,
    permutation = "restricted_by_row",
    block_size = 3,
    seed = 1
  )

  expect_equal(
    mean(psi_null),
    df[
      df$x == "Spain" &
        df$y == "Sweden" &
        df$distance == "euclidean" &
        df$diagonal == TRUE &
        df$bandwidth == 1 &
        df$lock_step == FALSE &
        df$permutation == "restricted_by_row" &
        df$block_size == 3 &
        df$seed == 1,
      "null_mean"
      ]
    )

 #compare with distantia with no repetitions
 df_distantia <- distantia(
   tsl = tsl,
   repetitions = 0
 )

 df_distantia <- df_distantia[, c("x", "y", "psi")]

 rownames(df_distantia) <- NULL

 df_test <- df[df$lock_step == FALSE & df$diagonal == TRUE & df$bandwidth == 1 & df$permutation == "free" & df$seed == 1 & df$distance == "euclidean", c("x", "y", "psi")]

 rownames(df_test) <- NULL

 expect_true(all.equal(df_distantia, df_test))

 #compare with distantia ls
  df_ls <- distantia_ls(
    tsl = tsl
  )

  expect_true(all(c("x", "y", "distance", "lock_step", "psi") %in% colnames(df_ls)))
  expect_true(all(df_ls$lock_step == TRUE))
  expect_equal(df_ls$distance[1], "euclidean")

  df_ls <- df_ls[, c("x", "y", "psi")]

  rownames(df_ls) <- NULL

  df_test <- df[df$lock_step == TRUE & df$permutation == "free" & df$seed == 1 & df$distance == "euclidean", c("x", "y", "psi")]

  rownames(df_test) <- NULL

  expect_true(all.equal(df_ls, df_test))

  #compare with distantia_dtw
  df_dtw <- distantia_dtw(
    tsl = tsl
  )

  df_dtw <- df_dtw[, c("x", "y", "psi")]

  rownames(df_dtw) <- NULL

  df_test <- df[df$lock_step == FALSE & df$diagonal == TRUE & df$bandwidth == 1 & df$permutation == "free" & df$seed == 1 & df$distance == "euclidean", c("x", "y", "psi")]

  rownames(df_test) <- NULL

  expect_true(all.equal(df_dtw, df_test))

  #test distantia aggregate
  df_agg <- distantia_aggregate(
    df = df
  )

  expect_true(
    nrow(df_agg) == 3
  )

  expect_true(all(c("x", "y", "psi") %in% colnames(df_agg)))

  #test distantia stats
  df_stats <- distantia_stats(
    df = df
  )

  expect_true(
    nrow(df_stats) == 3
  )

  expect_true(all(c("name", "mean", "min", "q1", "median", "q3", "max", "sd", "range") %in% colnames(df_stats)))

  #testing self-comparison
  tsl <- tsl_init(
    x = list(
      x = tsl[[1]],
      y = tsl[[1]]
    )
  )

  out <- distantia_ls(tsl = tsl)$psi

  testthat::expect_equal(out, 0)

  out <- distantia_dtw(tsl = tsl)$psi

  testthat::expect_equal(out, 0)

})

test_that("psi_ls_cpp() uses diagonal=TRUE sign convention (psi = 2a/b)", {

  x_mat <- matrix(c(0, 1, 2), ncol = 1)
  y_mat <- matrix(c(3, 4, 5), ncol = 1)
  # a = 9: sum of row-wise euclidean distances (3+3+3)
  # b = 4: auto-sum of consecutive diffs (x: 1+1=2, y: 1+1=2)
  # diagonal=TRUE:  (2*9/4) - 1 + 1 = 4.5
  # diagonal=FALSE: (2*9/4) - 1     = 3.5
  expect_equal(psi_ls_cpp(x_mat, y_mat, "euclidean"), 4.5)
  expect_equal(psi_equation_cpp(9, 4, FALSE), 3.5)

})

test_that("distantia() input edge cases are handled safely", {

  tsl <- tsl_initialize(
    x = fagus_dynamics,
    name_column = "name",
    time_column = "time"
  ) |> tsl_subset(time = c("2010-01-01", "2011-01-01"))

  # Item 1: invalid distance string errors at R level (not C++)
  expect_error(
    distantia(tsl = tsl, distance = "not_a_distance"),
    regexp = "arg"
  )

  # Item 3: repetitions = NULL behaves like repetitions = 0
  out_null <- distantia(tsl = tsl, repetitions = NULL)
  out_zero <- distantia(tsl = tsl, repetitions = 0)
  expect_equal(out_null, out_zero)

  # Item 5a: pure lock-step output lacks diagonal/bandwidth columns
  out_ls <- distantia(tsl = tsl, lock_step = TRUE)
  expect_false("diagonal" %in% colnames(out_ls))
  expect_false("bandwidth" %in% colnames(out_ls))

  # Item 5b: psi column is always numeric
  expect_type(out_zero$psi, "double")
  expect_type(out_ls$psi, "double")

})

# Generated by roxyglobals: do not edit by hand

utils::globalVariables(c(
  "distances", # <distance>
  "i", # <distantia>
  "median", # <distantia_boxplot>
  "i", # <distantia_dtw>
  "i", # <distantia_ls>
  "sd", # <distantia_model_frame>
  "sd", # <distantia_stats>
  "i", # <distantia_stats>
  "i", # <distantia_time_delay>
  "i", # <momentum>
  "median", # <momentum_boxplot>
  "i", # <momentum_dtw>
  "i", # <momentum_ls>
  "sd", # <momentum_model_frame>
  "sd", # <momentum_stats>
  "i", # <momentum_stats>
  "i", # <tsl_simulate>
  "i", # <tsl_stats>
  "na.omit", # <tsl_stats>
  "median", # <utils_boxplot_common>
  "distances", # <utils_check_distance_args>
  "i", # <utils_cluster_hclust_optimizer>
  "i", # <utils_cluster_kmeans_optimizer>
  "i", # <zoo_permute>
  "i", # <zoo_resample>
  NULL
))

#' Convert Matrix to Data Frame
#'
#' @param x (required, matrix). Default: NULL.
#'
#' @return A data frame.
#' @export
#' @autoglobal
#' @family internal
utils_prepare_matrix <- function(
    x = NULL
){

  if(inherits(x = x, what = "matrix") == FALSE){
    return(x)
  }

  #coerce to numeric
  x.numeric <- apply(
    X = x,
    MARGIN = 2,
    FUN = as.numeric
  ) |>
    suppressWarnings() |>
    as.data.frame()

  #identify columns with all NA
  x.na.names <- apply(
    X = x.numeric,
    MARGIN = 2,
    FUN = function(x){
      if(sum(is.na(x)) == length(x)){
        return(TRUE)
      } else {
        FALSE
      }
    }
  )

  x.na.names <- names(x.na.names[x.na.names == TRUE])

  #convert x to data frame
  x <- as.data.frame(x)

  #add non numerics to x.numeric
  if(length(x.na.names) > 0){
    for(i in x.na.names){
      x.numeric[[i]] <- x[[i]]
    }
  }

  x.numeric


}

#' Get Column Names from Time Series Lists
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param names (optional, character string) Three different sets of column names can be requested:
#' \itemize{
#'   \item "all" (default): list with the column names in each zoo object in `tsl`. Unnamed columns are tagged with the string "unnamed".
#'   \item "shared": character vector with the shared column names in at least two zoo objects in `tsl`.
#'   \item "exclusive": list with names of exclusive columns (if any) in each zoo object in `tsl`.
#' }
#'
#' @return list
#' @export
#' @autoglobal
#' @examples
#' #generate example data
#' tsl <- tsl_simulate()
#'
#' #list all column names
#' tsl_colnames_get(
#'   tsl = tsl,
#'   names = "all"
#' )
#'
#' #change one column name
#' names(tsl[[1]])[1] <- "new_column"
#'
#' #all names again
#' tsl_colnames_get(
#'   tsl = tsl,
#'   names = "all"
#' )
#'
#' #shared column names
#' tsl_colnames_get(
#'   tsl = tsl,
#'   names = "shared"
#' )
#'
#' #exclusive column names
#' tsl_colnames_get(
#'   tsl = tsl,
#'   names = "exclusive"
#' )
#' @family tsl_management
tsl_colnames_get <- function(
    tsl = NULL,
    names = c(
      "all",
      "shared",
      "exclusive"
      )
){

  names <- match.arg(
    arg = names,
    choices = c(
      c(
        "all",
        "shared",
        "exclusive"
        )
    ),
    several.ok = FALSE
  )

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #all names
  all.names <- lapply(
    X = tsl,
    FUN = function(x){
      y <- colnames(x)
      if(is.null(y)){
        y <- "unnamed"
      }
      y
    }
  )

  if(names == "all"){
    return(all.names)
  }

  #shared names
  all.names.table <- all.names |>
    unlist() |>
    table()

  #returns shared names in at least 2 zoo objects
  shared.names <- names(
    all.names.table[all.names.table == length(tsl)]
    )

  #get exclusive names
  exclusive.names <- names(
    all.names.table[all.names.table == 1]
  )

  #subset all.names to shared names
  shared.names <- lapply(
    X = all.names,
    FUN = function(x){
      y <- x[x %in% shared.names]
      if(length(y) == 0){
        return(NA)
      }
      y
    }
  )

  if(names == "shared"){
    return(shared.names)
  }

  #convert shared names to vector
  shared.names.vector <- shared.names |>
    unlist() |>
    unique() |>
    stats::na.omit()

  #exclusive names
  exclusive.names <- lapply(
    X = all.names,
    FUN = function(x){
      y <- x[x %in% exclusive.names]
      if(length(y) == 0){
        return(NA)
      }
      y
    }
  )

  exclusive.names

}


#' Set Column Names in a Time Series List
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param names (required, list or character vector):
#' \itemize{
#'   \item list: with same names as 'tsl', containing a vector of new column names for each time series in 'tsl'.
#'   \item character vector: vector of new column names assigned by position.
#' }
#'
#' @return time series list
#' @export
#'
#' @examples
#' tsl <- tsl_simulate(
#'   cols = 3
#'   )
#'
#' tsl_colnames_get(
#'   tsl = tsl
#'   )
#'
#' #using a vector
#' #extra names are ignored
#' tsl <- tsl_colnames_set(
#'   tsl = tsl,
#'   names = c("x", "y", "z", "zz")
#' )
#'
#' tsl_colnames_get(
#'   tsl = tsl
#' )
#'
#' #using a list
#' #extra names are ignored too
#' tsl <- tsl_colnames_set(
#'   tsl = tsl,
#'   names = list(
#'     A = c("A", "B", "C"),
#'     B = c("X", "Y", "Z", "ZZ")
#'   )
#' )
#'
#' tsl_colnames_get(
#'   tsl = tsl
#' )
#' @family tsl_management
tsl_colnames_set <- function(
    tsl = NULL,
    names = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #coerce zoo vectors to matrices
  tsl <- lapply(
    X = tsl,
    FUN = zoo_vector_to_matrix
  )

  #computing minimum names length
  min.names.length <- lapply(
    X = tsl,
    FUN = ncol
  ) |>
    unlist() |>
    max()

  #names is a character vector
  #convert to list
  if(
    is.vector(names) &&
    is.character(names)
  ){

    if(length(names) < min.names.length){
      stop("distantia::tsl_colnames_set(): argument 'names' must be a character vector of length ", min.names.length, ".", call. = FALSE)
    }

    #convert to list
    names <- lapply(
      X = tsl,
      FUN = function(x){

        names[seq_len(ncol(x))]

      }
    )

  }

  #names is a list
  if(is.list(names)){

    if(length(names) != length(tsl)){
      stop("distantia::tsl_colnames_set(): arguments 'names' and 'tsl' must be lists of the same length.", call. = FALSE)
    }

    if(any(base::names(names) != base::names(tsl))){
      stop("distantia::tsl_colnames_set(): arguments 'names' and 'tsl' must be lists with the same names.", call. = FALSE)
    }

    #test lengths
    test.length <- Map(
      f = function(x, name) {
        length(name) >= ncol(x)
      },
      tsl,
      names
    ) |>
      unlist() |>
      any()

    if(!any(test.length)){
      stop("distantia::tsl_colnames_set(): length of each element in 'names' must match the number of columns of each element in 'tsl'.", call. = FALSE)
    }


  }

  #rename zoo columns
  tsl <- Map(
    f = function(y, name) {
      colnames(y) <- name[seq_len(length(colnames(y)))]
      y
    },
    tsl,
    names
  )

  tsl

}

#' Clean Column Names in Time Series Lists
#'
#' @description
#' Uses the function [utils_clean_names()] to simplify and normalize messy column names in a time series list.
#'
#' The cleanup operations are applied in the following order:
#' \itemize{
#'   \item Remove leading and trailing whitespaces.
#'   \item Generate syntactically valid names with [base::make.names()].
#'   \item Replaces dots and spaces with the `separator`.
#'   \item Coerces names to lowercase.
#'   \item If `capitalize_first = TRUE`, the first letter is capitalized.
#'   \item If `capitalize_all = TRUE`, all letters are capitalized.
#'   \item If argument `length` is provided, [base::abbreviate()] is used to abbreviate the new column names.
#'   \item If `suffix` is provided, it is added at the end of the column name using the separator.
#'   \item If `prefix` is provided, it is added at the beginning of the column name using the separator.
#' }
#'
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param lowercase (optional, logical) If TRUE, all names are coerced to lowercase. Default: FALSE
#' @param separator (optional, character string) Separator when replacing spaces and dots. Also used to separate `suffix` and `prefix` from the main word. Default: "_".
#' @param capitalize_first (optional, logical) Indicates whether to capitalize the first letter of each name. Default: FALSE.
#' @param capitalize_all (optional, logical) Indicates whether to capitalize all letters of each name. Default: FALSE.
#' @param length (optional, integer) Minimum length of abbreviated names. Names are abbreviated via [abbreviate()]. Default: NULL.
#' @param suffix (optional, character string) String to append to the column names. Default: NULL.
#' @param prefix (optional, character string) String to prepend to the column names. Default: NULL.
#'
#' @return time series list
#'
#' @examples
#' #generate example data
#' tsl <- tsl_simulate(cols = 3)
#'
#' #list all column names
#' tsl_colnames_get(
#'   tsl = tsl
#' )
#'
#' #rename columns
#' tsl <- tsl_colnames_set(
#'   tsl = tsl,
#'   names = c(
#'   "New name 1",
#'   "new Name 2",
#'   "NEW NAME 3"
#'   )
#' )
#'
#' #check new names
#' tsl_colnames_get(
#'   tsl = tsl,
#'   names = "all"
#' )
#'
#' #clean names
#' tsl <- tsl_colnames_clean(
#'   tsl = tsl
#' )
#'
#' tsl_colnames_get(
#'   tsl = tsl
#' )
#'
#' #abbreviated
#' tsl <- tsl_colnames_clean(
#'   tsl = tsl,
#'   capitalize_first = TRUE,
#'   length = 6,
#'   suffix = "clean"
#' )
#'
#' tsl_colnames_get(
#'   tsl = tsl
#' )
#' @autoglobal
#' @export
#' @family tsl_management
tsl_colnames_clean <- function(
    tsl = NULL,
    lowercase = FALSE,
    separator = "_",
    capitalize_first = FALSE,
    capitalize_all = FALSE,
    length = NULL,
    suffix = NULL,
    prefix = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #clean names
  tsl <- lapply(
    X = tsl,
    FUN = function(x){

      x <- zoo_vector_to_matrix(
        x = x,
        name = NULL
      )

      x_colnames <- colnames(x)

      if(is.null(x_colnames)){
        x_colnames <- rep(
          x = "unnamed",
          times = ncol(x)
          )
      }

      x_colnames <- utils_clean_names(
        x = x_colnames,
        lowercase = lowercase,
        separator = separator,
        capitalize_first = capitalize_first,
        capitalize_all = capitalize_all,
        length = length,
        suffix = suffix,
        prefix = prefix
      )

      base::names(x_colnames) <- NULL

      colnames(x) <- x_colnames

      x

    }
  )

  tsl

}


#' Append Suffix to Column Names of Time Series List
#'
#' @inheritParams tsl_colnames_clean
#' @return time series list
#' @export
#' @autoglobal
#' @examples
#' tsl <- tsl_simulate()
#'
#' tsl_colnames_get(tsl = tsl)
#'
#' tsl <- tsl_colnames_suffix(
#'   tsl = tsl,
#'   suffix = "_my_suffix"
#' )
#'
#' tsl_colnames_get(tsl = tsl)
#' @family tsl_management
tsl_colnames_suffix <- function(
    tsl = NULL,
    suffix = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  if(is.null(suffix)){
    return(tsl)
  }

  tsl <- future.apply::future_lapply(
    X = tsl,
    FUN = function(x){

      colnames(x) <- paste0(
        colnames(x),
        suffix
      )

      x

    }
  )

  tsl

}

#' Append Prefix to Column Names of Time Series List
#'
#' @inheritParams tsl_colnames_clean
#' @return time series list
#' @export
#' @autoglobal
#' @examples
#' tsl <- tsl_simulate()
#'
#' tsl_colnames_get(tsl = tsl)
#'
#' tsl <- tsl_colnames_prefix(
#'   tsl = tsl,
#'   prefix = "my_prefix_"
#' )
#'
#' tsl_colnames_get(tsl = tsl)
#' @family tsl_management
tsl_colnames_prefix <- function(
    tsl = NULL,
    prefix = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  if(is.null(prefix)){
    return(tsl)
  }

  tsl <- future.apply::future_lapply(
    X = tsl,
    FUN = function(x){

      colnames(x) <- paste0(
        prefix,
        colnames(x)
      )

      x

    }
  )

  tsl

}

test_that("`tsl_join()` works", {

  tsl_a <- tsl_simulate(
    n = 2,
    cols = 2,
    irregular = TRUE,
    seed = 1
    )

  tsl_b <- tsl_simulate(
    n = 3,
    cols = 2,
    irregular = TRUE,
    seed = 2
  ) |>
    tsl_colnames_set(names = c("c", "d"))

  #NA interpolation message
  expect_message(
    tsl <- tsl_join(tsl_a, tsl_b)
    )

  expect_true(
    all(
      tsl_colnames_get(tsl)$a == c("a", "b", "c", "d")
    )
  )
  expect_true(
    all(
      tsl_colnames_get(tsl)$b == c("a", "b", "c", "d")
    )
  )

})

test_that("`tsl_colnames_clean()` works", {

  tsl <- tsl_simulate(cols = 3,
                      seed = 1)

  # tsl_colnames_set ----
  tsl <- tsl_colnames_set(
    tsl = tsl,
    names = c("New name 1", "new Name 2", "NEW NAME 3")
    )

  expect_equal(
    unique(unlist(tsl_colnames_get(tsl = tsl, names = "all"))),
    c("New name 1", "new Name 2", "NEW NAME 3")
    )

  # tsl_colnames_clean ----
   tsl <- tsl_colnames_clean(tsl = tsl)

   expect_true(
     all(
     unique(unlist(tsl_colnames_get(tsl = tsl, names = "all"))) %in%
     utils_clean_names(x = c("New name 1", "new Name 2", "NEW NAME 3"))
     )
   )

  tsl <- tsl_colnames_clean(
    tsl = tsl,
    capitalize_first = TRUE,
    length = 6,
    suffix = "clean"
  )

  expect_true(
    all(
      unique(unlist(tsl_colnames_get(tsl = tsl, names = "all"))) %in%
        utils_clean_names(x = c("New name 1", "new Name 2", "NEW NAME 3"), length = 6, capitalize_first = TRUE, suffix = "clean")
    )
  )

  #tsl_colnames_get ----

  names(tsl[[1]])[1] <- "new_column"

  expect_true(
    "new_column" %in%
    unique(unlist(tsl_colnames_get(tsl = tsl, names = "all")))
    )

  expect_true(
    "new_column" %in%
      unique(unlist(tsl_colnames_get(tsl = tsl, names = "exclusive")))
  )

  expect_true(
    !"new_column" %in%
      unique(unlist(tsl_colnames_get(tsl = tsl, names = "shared")))
  )

  #tsl_colnames_prefix ----
  tsl <- tsl_colnames_prefix(tsl = tsl, prefix = "my_prefix_")

  expect_true(
    "my_prefix_new_column" %in%
      unique(unlist(tsl_colnames_get(tsl = tsl, names = "all")))
  )

  #tsl_colnames_set ----
  tsl <- tsl_colnames_set(
    tsl = tsl,
    names = c("x", "y", "z", "zz")
    )

  expect_equal(
    unique(unlist(tsl_colnames_get(tsl = tsl))), c("x", "y", "z")
    )


  tsl <- tsl_colnames_set(
    tsl = tsl,
    names = list(
      A = c("A", "B", "C"),
      B = c("X", "Y", "Z", "ZZ")
      )
    )
  expect_equal(tsl_colnames_get(tsl = tsl)$A, c("A", "B", "C"))
  expect_equal(tsl_colnames_get(tsl = tsl)$B, c("X", "Y", "Z"))

  #tsl_colnames_suffix
  tsl <- tsl_colnames_suffix(tsl = tsl, suffix = "_my_suffix")

  expect_equal(
    tsl_colnames_get(tsl = tsl)$A,
    paste0(c("A", "B", "C"), "_my_suffix")
    )
})

#' Data Frame with Pairs of Time Series in Time Series Lists
#'
#' @description
#' Internal function used in [distantia()] and [momentum()] to generate a data frame with combinations of time series and function arguments.
#'
#' @param tsl (required, list) Time series list. Default: NULL.
#' @param args_list (required, list) List of arguments to combine with the pairs of time series. Default: NULL.
#'
#' @return A data frame.
#' @export
#' @autoglobal
#' @family internal
utils_tsl_pairs <- function(
    tsl = NULL,
    args_list = NULL
){

  df_tsl <- data.frame(
    t(
      utils::combn(
        names(tsl),
        m = 2
      )
    )
  )

  names(df_tsl) <- c(
    "x",
    "y"
  )

  if(is.null(args_list)){
    return(df_tsl)
  }

  df_args <- expand.grid(
    args_list,
    stringsAsFactors = FALSE
  )

  df <- merge(
    x = df_tsl,
    y = df_args
  )

  #remove invalid combinations

  #remove invalid commbinations of permutation x block_size
  free <- c("free", "free_by_row")

  if(
    all(c("permutation", "block_size") %in% colnames(df)) &&
    any(free %in% df$permutation)
  ){

    df[df$permutation %in% free, "block_size"] <- NA

  }

  #remove invalid combinations of lock_step and bandwidth
  if(
    all(c("lock_step", "bandwidth") %in% colnames(df)) &&
    any(df$lock_step == TRUE)
  ){

    df[df$lock_step ==TRUE, "bandwidth"] <- NA

  }

  #remove invalid combinations of lock_step and diagonal
  if(
    all(c("lock_step", "diagonal") %in% colnames(df)) &&
    any(df$lock_step == TRUE)
  ){

    df[df$lock_step == TRUE, "diagonal"] <- NA

  }

  df <- unique(df)

  df


}

#' Set Name of a Zoo Time Series
#'
#' @description
#' Zoo time series do not have an attribute 'name'. However, within `distantia`, to keep data consistency in several plotting and analysis operations, an attribute 'name' is used for these objects. This function is a convenient wrapper of `attr(x = x, which = "name") <- "xxx"`.
#'
#'
#' @param x (required, zoo object) Zoo time series to analyze. Default: NULL.
#' @param name (required, character string) Name or new name of the zoo object. If NULL, `x` is returned as is. Default: NULL.
#'
#' @return zoo time series
#' @export
#' @autoglobal
#' @examples
#' #simulate zoo time series
#' x <- zoo_simulate()
#'
#' #get current name
#' zoo_name_get(x = x)
#'
#' #change name
#' x <- zoo_name_set(
#'   x = x,
#'   name = "My.New.name"
#' )
#'
#' zoo_name_get(x = x)
#'
#' #clean name
#' x <- zoo_name_clean(
#'   x = x,
#'   lowercase = TRUE
#' )
#'
#' zoo_name_get(x = x)
#' @family zoo_functions
zoo_name_set <- function(
    x = NULL,
    name = NULL
){

  if(!zoo::is.zoo(object = x)){
    stop("distantia::zoo_name_set(): argument 'x' must be a zoo time series.", call. = FALSE)
  }

  if(!is.character(name)){
    return(x)
  }

  attr(
    x = x,
    which = "name"
  ) <- name

  x

}


#' Get Name of a Zoo Time Series
#'
#' @description
#' Just a convenient wrapper of `attributes(x)$name`.
#'
#'
#' @param x (required, zoo object) Zoo time series to analyze. Default: NULL.
#'
#' @return character string
#' @export
#' @autoglobal
#' @examples
#' #simulate zoo time series
#' x <- zoo_simulate()
#'
#' #get current name
#' zoo_name_get(x = x)
#'
#' #change name
#' x <- zoo_name_set(
#'   x = x,
#'   name = "My.New.name"
#' )
#'
#' zoo_name_get(x = x)
#'
#' #clean name
#' x <- zoo_name_clean(
#'   x = x,
#'   lowercase = TRUE
#' )
#'
#' zoo_name_get(x = x)
#' @family zoo_functions
zoo_name_get <- function(
    x = NULL
){

  if(zoo::is.zoo(object = x) == FALSE){
    stop("distantia::zoo_name_get(): argument 'x' must be a zoo time series", call. = FALSE)
  }

  x_name <- attr(x, which = "name", exact = TRUE)

  if(is.null(x_name)){
    warning("distantia::zoo_name_get(): zoo object 'x' does not have the attribute 'name'.", call. = FALSE)
  }

  x_name

}

#' Clean Name of a Zoo Time Series
#'
#' @description
#' Combines [utils_clean_names()] and [zoo_name_set()] to help clean, abbreviate, capitalize, and add a suffix or a prefix to the name of a zoo object.
#'
#' @param x (required, zoo object) Zoo time series to analyze. Default: NULL.
#' @param lowercase (optional, logical) If TRUE, all names are coerced to lowercase. Default: FALSE
#' @param separator (optional, character string) Separator when replacing spaces and dots. Also used to separate `suffix` and `prefix` from the main word. Default: "_".
#' @param capitalize_first (optional, logical) Indicates whether to capitalize the first letter of each name. Default: FALSE.
#' @param capitalize_all (optional, logical) Indicates whether to capitalize all letters of each name. Default: FALSE.
#' @param length (optional, integer) Minimum length of abbreviated names. Names are abbreviated via [abbreviate()]. Default: NULL.
#' @param suffix (optional, character string) Suffix for the clean names. Default: NULL.
#' @param prefix (optional, character string)  Prefix for the clean names. Default: NULL.
#'
#' @return zoo time series
#'
#' @autoglobal
#' @export
#' @examples
#' #simulate zoo time series
#' x <- zoo_simulate()
#'
#' #get current name
#' zoo_name_get(x = x)
#'
#' #change name
#' x <- zoo_name_set(
#'   x = x,
#'   name = "My.New.name"
#' )
#'
#' zoo_name_get(x = x)
#'
#' #clean name
#' x <- zoo_name_clean(
#'   x = x,
#'   lowercase = TRUE
#' )
#'
#' zoo_name_get(x = x)
#' @family zoo_functions
zoo_name_clean <- function(
    x = NULL,
    lowercase = FALSE,
    separator = "_",
    capitalize_first = FALSE,
    capitalize_all = FALSE,
    length = NULL,
    suffix = NULL,
    prefix = NULL
){

  if(zoo::is.zoo(object = x) == FALSE){
    stop("distantia::zoo_name_clean(): argument 'x' must be a zoo time series", call. = FALSE)
  }

  x_name <- zoo_name_get(
    x = x
  ) |>
    suppressWarnings()

  if(is.null(x_name)){
    stop("distantia::zoo_name_clean(): Argument 'x' does not have the attribute 'name'. Please, set a name with 'distantia::zoo_name_set()' first.", call. = FALSE)
  }

  x_name_clean <- utils_clean_names(
    x = x_name,
    lowercase = lowercase,
    separator = separator,
    capitalize_first = capitalize_first,
    capitalize_all = capitalize_all,
    length = length,
    suffix = suffix,
    prefix = prefix
  )

  names(x_name_clean) <- NULL

  x <- zoo_name_set(
    x = x,
    name = x_name_clean
  )

  x


}

test_that("`zoo_name_clean()` works", {

  x <- zoo_simulate(seed = 1)
  expect_equal(zoo_name_get(x = x), "A")

  x <- zoo_name_set(x = x, name = "My.New.name")
  expect_equal(zoo_name_get(x = x), "My.New.name")

  x <- zoo_name_clean(x = x, lowercase = TRUE)
  expect_equal(zoo_name_get(x = x), "my_new_name")

})

test_that("`zoo_vector_to_matrix()` works", {
  x <- zoo::zoo(x = runif(100))
  expect_equal(is.matrix(zoo::coredata(x)), FALSE)
  y <- zoo_vector_to_matrix(x = x)
  expect_equal(is.matrix(zoo::coredata(y)), TRUE)
})

test_that("`zoo_vector_to_matrix()` preserves the 'name' attribute", {
  x <- zoo::zoo(x = runif(10))
  x <- zoo_name_set(x = x, name = "test_series")
  y <- zoo_vector_to_matrix(x = x, name = "col1")
  expect_equal(zoo_name_get(x = y), "test_series")
  expect_equal(colnames(zoo::coredata(y)), "col1")
})

#' Sum of Distances in Least Cost Path
#'
#' @description
#' Demonstration function to sum the distances of a least cost path.
#'
#'
#' @param path (required, data frame) least cost path produced by [psi_cost_path()]. Default: NULL
#' @return numeric value
#' @examples
#' #distance metric
#' d <- "euclidean"
#'
#' #simulate two irregular time series
#' x <- zoo_simulate(
#'   name = "x",
#'   rows = 100,
#'   seasons = 2,
#'   seed = 1
#' )
#'
#' y <- zoo_simulate(
#'   name = "y",
#'   rows = 80,
#'   seasons = 2,
#'   seed = 2
#' )
#'
#' if(interactive()){
#'   zoo_plot(x = x)
#'   zoo_plot(x = y)
#' }
#'
#' #distance matrix
#' dist_matrix <- psi_distance_matrix(
#'   x = x,
#'   y = y,
#'   distance = d
#' )
#'
#' #orthogonal least cost matrix
#' cost_matrix <- psi_cost_matrix(
#'   dist_matrix = dist_matrix
#' )
#'
#' #orthogonal least cost path
#' cost_path <- psi_cost_path(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix
#' )
#'
#' #sum of distances in cost path
#' psi_cost_path_sum(
#'   path = cost_path
#' )
#' @autoglobal
#' @export
#' @family psi_demo
psi_cost_path_sum <- function(path = NULL){

  path <- utils_check_args_path(
    path = path
  )

  cost_path_sum_cpp(
    path = path
  )

}

#' Compute Silhouette Width of a Clustering Solution
#'
#' @description
#'
#' The silhouette width is a measure of how similar an object is to its own cluster (cohesion) compared to other clusters (separation).
#'
#' There are some general guidelines to interpret the individual silhouette widths of the clustered objects (as returned by this function when `mean = FALSE`):
#'
#' \itemize{
#'   \item Close to 1: object is well matched to its own cluster and poorly matched to neighboring clusters.
#'   \item Values near 0: the object is between two neighboring clusters.
#'   \item Values near -1: the object is likely to be assigned to the wrong cluster.
#' }
#'
#' When `mean = TRUE`, the overall mean of the silhouette widths of all objects is returned. These values should be interpreted as follows:
#'
#' \itemize{
#'  \item Higher than 0.7: robust clustering solution.
#'  \item Higher than 0.5: reasonable clustering solution.
#'  \item Higher than 0.25: weak clustering solution.
#'
#' }
#'
#' This metric may not perform well when the clusters have irregular shapes or sizes.
#'
#' This code was adapted from [https://svn.r-project.org/R-packages/trunk/cluster/R/silhouette.R](https://svn.r-project.org/R-packages/trunk/cluster/R/silhouette.R).
#'
#'
#' @param labels (required, integer vector) Labels resulting from a clustering algorithm applied to `d`. Must have the same length as columns and rows in `d`. Default: NULL
#' @param d (required, matrix) distance matrix typically resulting from [distantia_matrix()], but any other square matrix should work. Default: NULL
#' @param mean (optional, logical) If TRUE, the mean of the silhouette widths is returned. Default: FALSE
#'
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#' #weekly covid prevalence in three California counties
#' #load as tsl
#' #subset first 10 time series
#' #sum by month
#' tsl <- tsl_initialize(
#'   x = covid_prevalence,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_subset(
#'     names = 1:10
#'   ) |>
#'   tsl_aggregate(
#'     new_time = "months",
#'     method = max
#'   )
#'
#' #compute dissimilarity
#' distantia_df <- distantia(
#'   tsl = tsl,
#'   lock_step = TRUE
#' )
#'
#' #generate dissimilarity matrix
#' psi_matrix <- distantia_matrix(
#'   df = distantia_df
#' )
#'
#' #example with kmeans clustering
#' #------------------------------------
#'
#' #kmeans with 3 groups
#' psi_kmeans <- stats::kmeans(
#'   x = as.dist(psi_matrix[[1]]),
#'   centers = 3
#' )
#'
#' #case-wise silhouette width
#' utils_cluster_silhouette(
#'   labels = psi_kmeans$cluster,
#'   d = psi_matrix
#' )
#'
#' #overall silhouette width
#' utils_cluster_silhouette(
#'   labels = psi_kmeans$cluster,
#'   d = psi_matrix,
#'   mean = TRUE
#' )
#'
#'
#' #example with hierarchical clustering
#' #------------------------------------
#'
#' #hierarchical clustering
#' psi_hclust <- stats::hclust(
#'   d = as.dist(psi_matrix[[1]])
#' )
#'
#' #generate labels for three groups
#' psi_hclust_labels <- stats::cutree(
#'   tree = psi_hclust,
#'   k = 3,
#' )
#'
#' #case-wise silhouette width
#' utils_cluster_silhouette(
#'   labels = psi_hclust_labels,
#'   d = psi_matrix
#' )
#'
#' #overall silhouette width
#' utils_cluster_silhouette(
#'   labels = psi_hclust_labels,
#'   d = psi_matrix,
#'   mean = TRUE
#' )
#' @family distantia_support
utils_cluster_silhouette <- function(
    labels = NULL,
    d = NULL,
    mean = FALSE
){

  if(is.list(d)){
    d <- d[[1]]
  }

  if(!is.matrix(d)){
    stop("distantia::utils_cluster_silhouette(): argument 'd' must be a matrix.", call. = FALSE)
  }

  if(nrow(d) != ncol(d)){
    stop("distantia::utils_cluster_silhouette(): argument 'd' must be a square distance matrix.", call. = FALSE)
  }

  if(length(labels) != ncol(d)){
    stop("distantia::utils_cluster_silhouette(): argument 'labels' must have the same length as ncol(d) or nrow(d).", call. = FALSE)
  }

  if(length(unique(labels)) == 1L){
    if(isTRUE(mean)) return(NA_real_)
    label_names <- names(labels)
    if(is.null(label_names)) label_names <- as.character(seq_along(labels))
    return(data.frame(
      name = label_names,
      cluster = unname(labels),
      silhouette_width = NA_real_,
      stringsAsFactors = FALSE
    ))
  }

  labels_length <- length(labels)

  labels_groups <- sort(unique(labels))

  output <- matrix(
    data = NA,
    nrow = labels_length,
    ncol = 3,
    dimnames = list(
      names(labels),
      c("cluster","neighbor","silhouette_width")
    )
  )

  # j-th cluster:
  for(i in seq_len(length(labels_groups))){

    cluster_i_members <- labels == labels_groups[i]

    cluster_i_size <- sum(cluster_i_members)

    output[cluster_i_members, "cluster"] <- labels_groups[i]

    ## minimal distances to points in all other clusters:
    cluster_i_distances <- rbind(
      apply(
        X = d[
          !cluster_i_members,
          cluster_i_members,
          drop = FALSE
          ],
        MARGIN = 2,
        FUN = function(r) tapply(
          X = r,
          INDEX = labels[!cluster_i_members],
          FUN = mean
          )
      )
    )

    cluster_i_min_distance <- apply(
      X = cluster_i_distances,
      MARGIN = 2,
      FUN = which.min
      )

    output[cluster_i_members,"neighbor"] <- labels_groups[-i][cluster_i_min_distance]

    if(cluster_i_size > 1){

      #cluster cohesion
        cluster_i_cohesion <- colSums(
          d[cluster_i_members, cluster_i_members]
          )/(cluster_i_size - 1)

        #cluster separation
        cluster_i_separation <- cluster_i_distances[
          cbind(
            cluster_i_min_distance,
            seq(along = cluster_i_min_distance)
            )
          ]

        cluster_i_silhouette <- ifelse(
          test = cluster_i_cohesion != cluster_i_separation,
          yes = (cluster_i_separation - cluster_i_cohesion) / pmax(cluster_i_separation, cluster_i_cohesion),
          no = 0
          )

    } else {
      cluster_i_silhouette <- 0
    }

    output[
      cluster_i_members,
      "silhouette_width"
      ] <- cluster_i_silhouette

  }

  if(mean == FALSE){

    output <- output[, c("cluster", "silhouette_width")] |>
      as.data.frame()

    output$name <- rownames(output)

    rownames(output) <- NULL

    output <- output[, c("name", "cluster", "silhouette_width")]

  } else {
    output <- mean(output[, "silhouette_width"])
  }

  output

}

#' Default Block Size for Restricted Permutation in Dissimilarity Analyses
#'
#' @param tsl (required, list) Time series list. Default: NULL.
#' @param block_size (optional, integer vector) Row block sizes for restricted permutation tests. Used only when the permutation method is "restricted" or "restricted_by_row". A block of size `n` indicates that a row can only be permuted within a block of `n` adjacent rows. If NULL, defaults to 20 percent of the rows of the shortest time series in `tsl`. Minimum value is 2; maximum value is 50 percent of the rows of the shortest time series in `tsl`. Default: NULL.
#'
#' @return An integer.
#' @export
#' @autoglobal
#' @family distantia_support
utils_block_size <- function(
    tsl = NULL,
    block_size = NULL
){

  min_tsl_row <- tsl |>
    tsl_nrow() |>
    unlist() |>
    min()

  #default value
  default_block_size <- floor(min_tsl_row / 5)

  if(is.null(block_size)){

    return(default_block_size)

  }

  block_size <- as.integer(block_size)

  if(any(!is.integer(block_size))){

    return(default_block_size)

  }

  #remove extremes
  block_size <- block_size[
    block_size >= 2 &
      block_size <= floor(min_tsl_row / 2)
    ]

  block_size

}

#' Dissimilarity Model Frame
#'
#' @description
#' This function generates a model frame for statistical or machine learning analysis from these objects:
#'
#' \itemize{
#'   \item: Dissimilarity data frame generated by [distantia()], [distantia_ls()], [distantia_dtw()], or [distantia_time_delay()]. The output model frame will have as many rows as this data frame.
#'   \item: Data frame with static descriptors of the time series. These descriptors are converted to distances between pairs of time series via [distance_matrix()].
#'   \item: List defining composite predictors. This feature allows grouping together predictors that have a common meaning. For example, `composite_predictors = list(temperature = c("temperature_mean", "temperature_min", "temperature_max")` generates a new predictor named "temperature", which results from computing the multivariate distances between the vectors of temperature variables of each pair of time series. Predictors in one of such groups will be scaled before distance computation if their maximum standard deviations differ by a factor of 10 or more.
#' }
#'
#'
#' The resulting data frame contains the following columns:
#' \itemize{
#'   \item `x` and `y`: names of the pair of time series represented in the row.
#'   \item response columns in `response_df`.
#'   \item predictors columns: representing the distance between the values of the given static predictor between `x` and `y`.
#'   \item (optional) `geographic_distance`: If `predictors_df` is an sf `sf` data frame, then geographic distances are computed via [sf::st_distance()].
#' }
#'
#' This function supports a parallelization setup via [future::plan()].
#'
#' @param response_df (required, data frame) output of [distantia()], [distantia_ls()], [distantia_dtw()], or [distantia_time_delay()]. Default: NULL
#' @param predictors_df (required, data frame or sf data frame) Data frame with numeric predictors for the model frame. Must have a column with the time series names in `response_df$x` and `response_df$y`. If `sf` data frame, the column "geographic_distance" with distances between pairs of time series is added to the model frame. Default: NULL
#' @param composite_predictors (optional, list) list defining composite predictors. For example, `composite_predictors = list(a = c("b", "c"))` uses the columns `"b"` and `"c"` from `predictors_df` to generate the predictor `a` as the multivariate distance between `"b"` and `"c"` for each pair of time series in `response_df`. Default: NULL
#' @param scale (optional, logical) if TRUE, all predictors are scaled and centered with [scale()]. Default: TRUE
#' @param distance (optional, string) Method to compute the distance between predictor values for all pairs of time series in `response_df`. Default: "euclidean".
#'
#' @return data frame: with attributes "predictors", "response", and "formula".
#' @export
#' @autoglobal
#' @examples
#'
#' #covid prevalence in California counties
#' tsl <- tsl_initialize(
#'   x = covid_prevalence,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   #subset to shorten example runtime
#'   tsl_subset(
#'     names = 1:5
#'   )
#'
#' #dissimilarity analysis
#' df <- distantia_ls(tsl = tsl)
#'
#' #combine several predictors
#' #into a new one
#' composite_predictors <- list(
#'   economy = c(
#'     "poverty_percentage",
#'     "median_income",
#'     "domestic_product"
#'     )
#' )
#'
#' #generate model frame
#' model_frame <- distantia_model_frame(
#'   response_df = df,
#'   predictors_df = covid_counties,
#'   composite_predictors = composite_predictors,
#'   scale = TRUE
#' )
#'
#' head(model_frame)
#'
#' #names of response and predictors
#' #and an additive formula
#' #are stored as attributes
#' attributes(model_frame)$predictors
#'
#' #if response_df is output of distantia():
#' attributes(model_frame)$response
#' attributes(model_frame)$formula
#'
#' #example of linear model
#' # model <- lm(
#' #   formula = attributes(model_frame)$formula,
#' #   data = model_frame
#' # )
#' #
#' # summary(model)
#'
#' @family distantia_support
distantia_model_frame <- function(
    response_df = NULL,
    predictors_df = NULL,
    composite_predictors = NULL,
    scale = TRUE,
    distance = "euclidean"
){

  distance <- utils_check_distance_args(
    distance = distance
  )

  #df ----
  if(is.null(response_df)){

    stop(
      "distantia::distantia/momentum_model_frame(): argument `response_df' must not be NULL.",
      call. = FALSE
    )

  }

  df_type_distantia <- c(
    "distantia_df",
    "time_delay_df"
  )

  df_type_momentum <- "momentum_df"

  df_types <- c(
    df_type_distantia,
    df_type_momentum
  )

  df_type <- attributes(response_df)$type

  if(
    is.null(df_type) ||
    !(df_type %in% df_types)
  ){
    stop("distantia::distantia/momentum_model_frame(): argument 'response_df' must be the output of distantia::distantia(), distantia::distantia_time_delay(), or distantia::momentum().", call. = FALSE)
  }


  if(df_type %in% df_type_distantia){

    f_name <- "distantia::distantia_model_frame(): "

    if(df_type == "distantia_df"){

      response_df <- distantia_aggregate(
        df = response_df
      )

    }
  }

  if(df_type == df_type_momentum){

    f_name <- "distantia::momentum_model_frame(): "

    response_df <- response_df |>
      momentum_aggregate() |>
      momentum_to_wide()

  }

  #add ID
  response_df$id <- seq_len(nrow(response_df))

  #predictors_df ----
  if(is.null(predictors_df)){

    stop(
      f_name, "argument `predictors_df' must not be NULL.",
      call. = FALSE
    )

  }

  if(!inherits(x = predictors_df, what = "data.frame")){

    stop(
      f_name, "argument `predictors_df' must be a data frame or an sf data frame.",
      call. = FALSE
    )

  }

  #identify names column in predictors response_df
  df_names <- unique(
    c(response_df$x, response_df$y)
  )

  predictors_name_column <- sapply(
    X = predictors_df,
    FUN = function(x){
      sum(df_names %in% x) == length(df_names)
    }
  )

  predictors_name_column <- names(predictors_name_column[which(predictors_name_column)])

  if(length(predictors_name_column) == 0){
    stop(
      f_name, "argument 'predictors_df' must have a column with the names in 'response_df$x' and 'response_df$y'.",
      call. = FALSE
    )
  }

  #numeric columns
  predictors_numeric_columns <- sapply(
    X = predictors_df,
    FUN = is.numeric
  )

  predictors_numeric_columns <- names(predictors_numeric_columns[which(predictors_numeric_columns)])

  #if no predictors and no sf, stop
  if(
    length(predictors_numeric_columns) == 0 &&
    is.null(attributes(predictors_df)$sf_column)
  ){

    stop(
      f_name, "argument 'predictors_df' must have at least one numeric column.",
      call. = FALSE
    )

  }

  #compute distances between predictors

  #default predictors list
  composite_predictors_default <- as.list(
    predictors_numeric_columns
  )

  names(composite_predictors_default) <- predictors_numeric_columns

  #predictors list
  if(is.null(composite_predictors)){

    composite_predictors <- composite_predictors_default

  } else {

    #add names if not named
    if(is.null(names(composite_predictors))){
      names(composite_predictors) <- sapply(
        X = composite_predictors,
        FUN = function(x){
          paste(x, collapse = "_")
        }
      )
    }

    composite_predictors_default <- composite_predictors_default[!(names(composite_predictors_default) %in% names(composite_predictors))]

    composite_predictors <- c(
      composite_predictors,
      composite_predictors_default
    )

  }

  #remove geometry
  predictors_df_no_geom <- utils_drop_geometry(
    df = predictors_df
  )

  #id df
  response_df.id <- response_df[, "id", drop = FALSE]

  #compute distance matrices
  model_frame_list <- future.apply::future_lapply(
    X = seq_len(length(composite_predictors)),
    FUN = function(x){

      #extract predictors columns
      df.i <- predictors_df_no_geom[
        ,
        unlist(composite_predictors[[x]]),
        drop = FALSE
      ]

      #decide if scaling is needed
      sd_ratios <- max(sapply(X = df.i, FUN = sd)) /
        min(sapply(X = df.i, FUN = sd))

      if(sd_ratios > 10){

        df.i <- df.i |>
          scale() |>
          as.data.frame()

      }

      #add name column
      df.i[[predictors_name_column]] <- predictors_df[[predictors_name_column]]

      #compute distance matrix
      m.i <- distantia::distance_matrix(
        df = df.i,
        name_column = predictors_name_column,
        distance = distance
      )

      #prepare df
      out.df.i <- response_df.id

      #extract matrix as vector
      variable.i <- mapply(
        FUN = function(a, b) m.i[a, b],
        response_df$x,
        response_df$y
      )

      names(variable.i) <- NULL

      out.df.i[[names(composite_predictors)[x]]] <- variable.i

      out.df.i

    },
    future.seed = NULL
  )

  names(model_frame_list) <- names(composite_predictors)

  predictors_numeric_columns <- c(
    predictors_numeric_columns,
    names(composite_predictors)
    )

  #add geographic distance if predictors_df is an sf data frame
  if(!is.null(attributes(predictors_df)$sf_column)){

    if(
      !any(
        requireNamespace("sf", quietly = TRUE),
        requireNamespace("lwgeom", quietly = TRUE)
      )
    ){
      stop(
        f_name, "this function requires installing the packages 'sf' and 'lwgeom'.",
        call. = FALSE
      )
    }

    m.i <- sf::st_distance(
      x = predictors_df
    )

    dimnames(m.i) <- list(
      as.character(predictors_df[[predictors_name_column]]),
      as.character(predictors_df[[predictors_name_column]])
    )

    variable.i <- mapply(
      FUN = function(a, b) m.i[a, b],
      response_df$x,
      response_df$y
    )

    names(variable.i) <- NULL

    new_distance_column <- "geographic_distance"

    if(new_distance_column %in% c(colnames(response_df), colnames(predictors_df))){
      new_distance_column <- paste0("_", new_distance_column, "_")
    }

    response_df.id[[new_distance_column]] <- variable.i

    predictors_numeric_columns <- c(predictors_numeric_columns, new_distance_column)

    #add to model frame list
    if(exists("model_frame_list")){
      model_frame_list[["geographic_distance"]] <- response_df.id
    } else {
      model_frame_list <- list(
        geographic_distance = response_df.id
      )
    }


  }

  #add distantia df
  model_frame_list <- c(
    list(response = response_df),
    model_frame_list
  )

  #join all data frames
  model_frame <- Reduce(
    f = function(x, y){
      merge(x, y, by = "id")
    },
    x = model_frame_list
  )

  model_frame$id <- NULL

  #scale
  if(scale == TRUE){

    predictors_numeric_columns <- unique(predictors_numeric_columns)

    #join response with scaled predictors
    model_frame <- cbind(

      #response data frame
      model_frame[,
                  setdiff(
                    x = colnames(model_frame),
                    y = predictors_numeric_columns
                    )
                  ],

      #predictors data frame
      model_frame[, predictors_numeric_columns, drop = FALSE] |>
        base::scale() |>
        as.data.frame()

    )

  }

  attr(
    x = model_frame,
    which = "predictors"
  ) <- predictors_numeric_columns

  if(df_type == "distantia_df"){

    attr(
      x = model_frame,
      which = "response"
    ) <- "psi"

    attr(
      x = model_frame,
      which = "formula"
    ) <- stats::as.formula(
      paste(
        "psi ~ ",
        paste(predictors_numeric_columns, collapse = " + ")
      )
    )

  }

  model_frame

}

#' @title distantia: A Toolset for Time Series Dissimilarity Analysis
#'
#' @description
#' Fast C++ implementation of Dynamic Time Warping for time series dissimilarity analysis, with applications in environmental monitoring, sensor data analysis, climate science, signal processing, pattern recognition, and financial data analysis. Built upon the ideas presented in Benito and Birks (2020) <doi:10.1111/ecog.04895>, the package provides tools for analyzing time series of varying lengths and structures, including irregular multivariate time series. Key features include individual variable contribution analysis, restricted permutation tests for statistical significance, and imputation of missing data via generalized additive models. Additionally, the package provides comprehensive tools for preparing and managing time series data.
#'
#' @docType package
#' @author Blas Benito <blasbenito@gmail.com>
#' @keywords internal
#' @importFrom Rcpp evalCpp
#' @useDynLib distantia
#' @name distantia-package
"_PACKAGE"

#' Handle NA Cases in Time Series Lists
#'
#' @description
#' Removes or imputes NA cases in time series lists. Imputation is done via interpolation against time via [zoo::na.approx()], and if there are still leading or trailing NA cases after NA interpolation, then [zoo::na.spline()] is applied as well to fill these gaps. Interpolated values are forced to fall within the observed data range.
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @param tsl (required, list) Time series list. Default: NULL
#' @param na_action (required, character) NA handling action. Available options are:
#' \itemize{
#'   \item "impute" (default): NA cases are interpolated from neighbors as a function of time (see [zoo::na.approx()] and [zoo::na.spline()]).
#'   \item "omit": rows with NA cases are removed.
#' }
#'
#' @return time series list
#' @export
#'
#' @examples
#'
#' #tsl with NA cases
#' tsl <- tsl_simulate(
#'   na_fraction = 0.25
#' )
#'
#' tsl_count_NA(tsl = tsl)
#'
#' if(interactive()){
#'   #issues warning
#'   tsl_plot(tsl = tsl)
#' }
#'
#' #omit NA (default)
#' #--------------------------------------
#'
#' #original row count
#' tsl_nrow(tsl = tsl)
#'
#' #remove rows with NA
#' tsl_no_na <- tsl_handle_NA(
#'   tsl = tsl,
#'   na_action = "omit"
#' )
#'
#' #count rows again
#' #large data loss in this case!
#' tsl_nrow(tsl = tsl_no_na)
#'
#' #count NA again
#' tsl_count_NA(tsl = tsl_no_na)
#'
#' if(interactive()){
#'   tsl_plot(tsl = tsl_no_na)
#' }
#'
#'
#' #impute NA with zoo::na.approx
#' #--------------------------------------
#'
#' #impute NA cases
#' tsl_no_na <- tsl_handle_NA(
#'   tsl = tsl,
#'   na_action = "impute"
#' )
#'
#' #count rows again
#' #large data loss in this case!
#' tsl_nrow(tsl = tsl_no_na)
#'
#' if(interactive()){
#'   tsl_plot(tsl = tsl_no_na)
#' }
#' @family tsl_management
tsl_handle_NA <- function(
    tsl = NULL,
    na_action = c(
      "impute",
      "omit"
      )
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #replaces Inf with Na
  tsl <- tsl_Inf_to_NA(
    tsl = tsl
  )

  #replaces NaN with NA
  tsl <- tsl_NaN_to_NA(
    tsl = tsl
  )

  na_action <- match.arg(
    arg = na_action[1],
    choices = c(
      "omit",
      "impute"
    ),
    several.ok = FALSE
  )

  #progress bar
  p <- progressr::progressor(along = tsl)

  if(na_action == "omit"){

    tsl <- future.apply::future_lapply(
      X = tsl,
      FUN = function(x){

        p()

        stats::na.omit(x)

      }
    )

  }


  if(na_action == "impute"){

    tsl <- future.apply::future_lapply(
      X = tsl,
      FUN = function(x){

        p()

        #check if x is integer
        x.integer <- is.integer(x)

        #get time
        x.index <- zoo::index(x)

        #find minimum and maximum to clamp interpolation bounds
        x.min <- lapply(
          X = x,
          FUN = min,
          na.rm = TRUE
        )

        x.max <- lapply(
          X = x,
          FUN = max,
          na.rm = TRUE
        )

        #interpolate with the given function
        x.interpolated <- zoo::na.approx(
          object = x,
          na.rm = FALSE
          )

        #remove leading or trailing NAs
        if(sum(is.na(x.interpolated)) > 0){

          x.interpolated <- zoo::na.spline(
            object = x.interpolated
          )

        }

        #setting minimum and maximum bounds
        for(i in seq_len(length(x.min))){
          x.interpolated[x.interpolated[, i] < x.min[[i]], i] <- x.min[[i]]
          x.interpolated[x.interpolated[, i] > x.max[[i]], i] <- x.max[[i]]
        }

        if(x.integer == TRUE){
          mode(x.interpolated) <- "integer"
        }

        x.interpolated <- zoo::zoo(
          x = x.interpolated,
          order.by = x.index
        )

        x.interpolated

      }
    )

  }

  na.count <- tsl_count_NA(
    tsl = tsl
  )

  tsl <- tsl_names_set(
    tsl = tsl
  )

  tsl

}


#' @rdname tsl_handle_NA
#' @export
#' @autoglobal
tsl_Inf_to_NA <- function(
    tsl = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  tsl <- future.apply::future_lapply(
    X = tsl,
    FUN = function(x){

      y <- as.matrix(x)

      y[is.infinite(y)] <- NA

      y <- zoo::zoo(
        x = y,
        order.by = zoo::index(x)
      )

      zoo_name_set(
        x = y,
        name = attributes(x)$name
      )

    }
  )

  tsl

}

#' @rdname tsl_handle_NA
#' @export
#' @autoglobal
tsl_NaN_to_NA <- function(
    tsl = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  tsl <- future.apply::future_lapply(
    X = tsl,
    FUN = function(x){
      x[is.nan(x)] <- NA
      x
    }
  )

  tsl

}

#' Clean Character Vector of Names
#'
#' Clean and format character vectors for use as column names or variable names.
#'
#' The cleanup operations are applied in the following order:
#' \itemize{
#'   \item Remove leading and trailing whitespaces.
#'   \item Generate syntactically valid names with [base::make.names()].
#'   \item Replace dots and spaces with the `separator`.
#'   \item Coerce names to lowercase.
#'   \item If argument `length` is provided, [base::abbreviate()] is used to abbreviate the new column names.
#'   \item If `suffix` is provided, it is added at the end of the column name using the separator.
#'   \item If `prefix` is provided, it is added at the beginning of the column name using the separator.
#'   \item If `capitalize_first = TRUE`, the first letter is capitalized.
#'   \item If `capitalize_all = TRUE`, all letters are capitalized.
#' }
#'
#' @param x (required, character vector) Names to be cleaned. Default: NULL.
#' @param lowercase (optional, logical) If TRUE, all names are coerced to lowercase. Default: FALSE.
#' @param separator (optional, character string) Separator when replacing spaces and dots and appending `suffix` and `prefix` to the main word. Default: "_".
#' @param capitalize_first (optional, logical) Capitalize the first letter of each name. Default: FALSE.
#' @param capitalize_all (optional, logical) Capitalize all letters of each name. Default: FALSE.
#' @param length (optional, integer) Minimum length of abbreviated names. Names are abbreviated via [abbreviate()]. Default: NULL.
#' @param suffix (optional, character string) String to append to the cleaned names. Default: NULL.
#' @param prefix (optional, character string) String to prepend to the cleaned names. Default: NULL.
#'
#' @return A character vector.
#' @autoglobal
#' @export
#' @examples
#' x <- c(
#'   "GerMany",
#'   "spain",
#'   "SWEDEN"
#' )
#'
#' #abbreviate names
#' #---------------------------
#' #abbreviate to 4 characters
#' utils_clean_names(
#'   x = x,
#'   capitalize_all = TRUE,
#'   length = 4
#' )
#'
#' #suffix and prefix
#' #---------------------------
#' utils_clean_names(
#'   x = x,
#'   capitalize_first = TRUE,
#'   separator = "_",
#'   prefix = "my_prefix",
#'   suffix = "my_suffix"
#' )
#' @family internal
utils_clean_names <- function(
    x = NULL,
    lowercase = FALSE,
    separator = "_",
    capitalize_first = FALSE,
    capitalize_all = FALSE,
    length = NULL,
    suffix = NULL,
    prefix = NULL
){

  if(!is.character(x)){
    stop("distantia::utils_clean_names(): input 'names' must be a character vector.", call. = FALSE)
  }

  if(anyNA(x)){
    stop("distantia::utils_clean_names(): Input 'names' contains missing values.", call. = FALSE)
  }

  x.old <- x

  #remove leading/trailing whitespaces
  x <- trimws(
    x = x,
    which = "both"
  )

  #generic R style clean-up
  x <- make.names(
    names = x,
    unique = TRUE
  )

  #dots and spaces to underscore
  x <- gsub(
    pattern = "\\s+|\\.{1,}",
    replacement = separator,
    x = x
  )

  #to lowercase
  if(lowercase == TRUE){
    x <- tolower(x)
  }

  #abbreviate
  if(is.numeric(length)){

    x <- abbreviate(
      names.arg = x,
      minlength = as.integer(length),
      method = "both.sides",
      named = FALSE
    )

  }

  #prefix and suffix
  if(!is.null(suffix)) {
    x <- paste(
      x,
      suffix,
      sep = separator
    )
  }

  # prefix
  if(!is.null(prefix)) {
    x <- paste(
      prefix,
      x,
      sep = separator
    )
  }

  #first letter to uppercase
  if(capitalize_first == TRUE){

    x <- paste0(
      toupper(
        substr(
          x = x,
          start = 1,
          stop = 1
        )
      ),
      substr(
        x = x,
        start = 2,
        stop = nchar(x)
      )
    )

  }

  #all letters to uppercase
  if(capitalize_all == TRUE){

    x <- toupper(x)

  }

  names(x) <- x.old

  x

}

test_that("`utils_clean_names()` works", {

  x <- c("GerMany", "spain", "SWEDEN")

  expect_true(
    all(
      utils_clean_names(x = x, capitalize_all = TRUE, length = 4) == c("GRMN", "SPAN", "SWED")
      )
    )

  expect_true(
    all(
      utils_clean_names(
        x = x, capitalize_first = TRUE, separator = "_",
        prefix = "my_prefix", suffix = "my_suffix"
      ) == c("My_prefix_GerMany_my_suffix", "My_prefix_spain_my_suffix", "My_prefix_SWEDEN_my_suffix")
    )
  )

})

#' Lists Available Transformation Functions
#'
#' @return character vector
#' @export
#' @autoglobal
#' @examples
#' f_list()
#' @family tsl_transformation
f_list <- function(){

  f_funs <-  ls(
    name = "package:distantia",
    pattern = "^f_"
  )

  f_funs[f_funs != "f_list"]

}

# Trend Adjustments ----

#' Data Transformation: Linear Trend of Zoo Time Series
#'
#' @description
#' Fits a linear model on each column of a zoo object using time as a predictor, and predicts the outcome.
#'
#'
#' @param x (required, zoo object) Zoo time series object to transform.
#' @param center (required, logical) If TRUE, the output is centered at zero. If FALSE, it is centered at the data mean. Default: TRUE
#' @param ... (optional, additional arguments) Ignored in this function.
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(cols = 2)
#'
#' y <- f_trend_linear(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_trend_linear <- function(
    x = NULL,
    center = TRUE,
    ...
) {

  x_data <- zoo::coredata(x, drop = FALSE)

  m <- stats::lm(
    formula = x_data ~ zoo::index(x)
  )

  #predict linear trend
  y <- stats::predict(m)

  if(center == TRUE){
    y <- y - mean(x_data)
  }

  y <- as.matrix(y)
  dimnames(y)[[2]] <- dimnames(x)[[2]]

  y <- zoo::zoo(
    x = y,
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )

  y

}

#' Data Transformation: Polynomial Linear Trend of Zoo Time Series
#'
#' @description
#' Fits a polynomial linear model on each column of a zoo object using time as a predictor, and predicts the outcome to return the polynomial trend of the time series. This method is a useful alternative to [f_trend_linear] when the overall trend of the time series does not follow a straight line.
#'
#'
#' @param x (required, zoo object) Zoo time series object to transform.
#' @param center (required, logical) If TRUE, the output is centered at zero. If FALSE, it is centered at the data mean. Default: TRUE
#' @param degree (optional, integer) Degree of the polynomial. Default: 2
#' @param ... (optional, additional arguments) Ignored in this function.
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(cols = 2)
#'
#' y <- f_trend_poly(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_trend_poly <- function(
    x = NULL,
    degree = 2,
    center = TRUE,
    ...
) {

  x_data <- zoo::coredata(x, drop = FALSE)

  m <- stats::lm(
    formula = x_data ~ stats::poly(
      x = zoo::index(x),
      degree = degree,
      raw = TRUE
    )
  )

  #predict linear trend
  y <- stats::predict(m)

  if(center == FALSE){
    y <- y + mean(x_data)
  }

  y <- as.matrix(y)
  dimnames(y)[[2]] <- dimnames(x)[[2]]

  # recreate zoo
  y <- zoo::zoo(
    x = y,
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )

  y

}


#' Data Transformation: Linear Detrending of Zoo Time Series
#'
#' @description
#' Fits a linear model on each column of a zoo object using time as a predictor, predicts the outcome, and subtracts it from the original data to return a detrended time series. This method may not be suitable if the input data is seasonal or has a complex trend, so be mindful of its limitations when applied.
#'
#'
#' @param x (required, zoo object) Zoo time series object to transform.
#' @param center (required, logical) If TRUE, the output is centered at zero. If FALSE, it is centered at the data mean. Default: TRUE
#' @param ... (optional, additional arguments) Ignored in this function.
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(cols = 2)
#'
#' y <- f_detrend_linear(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_detrend_linear <- function(
    x = NULL,
    center = TRUE,
    ...
) {

  x_data <- zoo::coredata(x, drop = FALSE)

  m <- stats::lm(
    formula = x_data ~ zoo::index(x)
  )

  #predict linear trend
  x_trend <- stats::predict(m)

  #subtract trend
  y <- x_data - x_trend

  if(center == FALSE){
    y <- y + mean(x_data)
  }

  y <- as.matrix(y)
  dimnames(y)[[2]] <- dimnames(x)[[2]]

  # recreate zoo
  y <- zoo::zoo(
    x = y,
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )

  y

}

#' Data Transformation: Polynomial Linear Detrending of Zoo Time Series
#'
#' @description
#' Fits a polynomial linear model on each column of a zoo object using time as a predictor, predicts the outcome, and subtracts it from the original data to return a detrended time series. This method is a useful alternative to [f_detrend_linear] when the overall trend of the time series does not follow a straight line.
#'
#'
#' @param x (required, zoo object) Zoo time series object to transform.
#' @param center (required, logical) If TRUE, the output is centered at zero. If FALSE, it is centered at the data mean. Default: TRUE
#' @param degree (optional, integer) Degree of the polynomial. Default: 2
#' @param ... (optional, additional arguments) Ignored in this function.
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(cols = 2)
#'
#' y <- f_detrend_poly(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_detrend_poly <- function(
    x = NULL,
    degree = 2,
    center = TRUE,
    ...
) {

  x_data <- zoo::coredata(x, drop = FALSE)

  m <- stats::lm(
    formula = x_data ~ stats::poly(
      x = zoo::index(x),
      degree = degree,
      raw = TRUE
      )
  )

  #predict linear trend
  x_trend <- stats::predict(m)

  #subtract trend
  y <- x_data - x_trend

  if(center == FALSE){
    y <- y + mean(x_data)
  }

  y <- as.matrix(y)
  dimnames(y)[[2]] <- dimnames(x)[[2]]

  # recreate zoo
  y <- zoo::zoo(
    x = y,
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )

  y

}


#' Data Transformation: Detrending and Differencing
#'
#' @description
#' Performs differencing to remove trends from a zoo time series, isolating short-term fluctuations by subtracting values at specified lags. The function preserves the original index and metadata, with an option to center the output around the mean of the original series. Suitable for preprocessing time series data to focus on random fluctuations unrelated to overall trends.
#'
#' @param x (required, zoo object) Zoo time series object to transform.
#' @param lag (optional, integer) Lag value for differencing. Default: 1
#' @param center (required, logical) If TRUE, the output is centered at zero. If FALSE, it is centered at the data mean. Default: TRUE
#' @param ... (optional, additional arguments) Ignored in this function.
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(cols = 2)
#'
#' y_lag1 <- f_detrend_difference(
#'   x = x,
#'   lag = 1
#' )
#'
#' y_lag5 <- f_detrend_difference(
#'   x = x,
#'   lag = 5
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y_lag1)
#'   zoo_plot(y_lag5)
#' }
#' @family tsl_transformation
f_detrend_difference <- function(
    x = NULL,
    lag = 1,
    center = TRUE,
    ...
) {

  x_data <- zoo::coredata(x, drop = FALSE)

  y <- diff(
    x = as.matrix(x_data),
    lag = as.integer(lag[1])
  )

  first.row <- matrix(
    data = 0,
    nrow = 1,
    ncol = ncol(x_data),
    dimnames = list(
      "1",
      colnames(x_data)
    )
  )

  y <- rbind(
    first.row,
    y
  )

  if(center == FALSE){
    y <- y + mean(x_data)
  }

  y <- as.matrix(y)
  dimnames(y)[[2]] <- dimnames(x)[[2]]

  y <- zoo::zoo(
    x = as.matrix(y),
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )

  y

}



# Compositional Transformations ----

#' Data Transformation: Rowwise Proportions
#'
#' @inheritParams f_hellinger
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(cols = 2)
#'
#' y <- f_proportion(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_proportion <- function(
    x = NULL,
    ...
){

  y <- sweep(
    x = x,
    MARGIN = 1,
    STATS = rowSums(x),
    FUN = "/"
  )

  y

}

#' Data Transformation: Rowwise Square Root of Proportions
#'
#' @inheritParams f_hellinger
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(cols = 2)
#'
#' y <- f_proportion_sqrt(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_proportion_sqrt <- function(
    x = NULL,
    ...
){

  f_proportion(x)^0.5

}

#' Data Transformation: Rowwise Percentages
#'
#' @inheritParams f_hellinger
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(cols = 2)
#'
#' y <- f_percent(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_percent <- function(
    x = NULL,
    ...
){

  f_proportion(x)*100

}

#' Data Transformation: Rowwise Hellinger Transformation
#'
#' @description
#' Transforms the input zoo object to proportions via [f_proportion] and then applies the Hellinger transformation.
#'
#'
#' @param x (required, zoo object) Zoo time series object to transform.
#' @param ... (optional, additional arguments) Ignored in this function.
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(
#'   cols = 5,
#'   data_range = c(0, 500)
#'   )
#'
#' y <- f_hellinger(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_hellinger <- function(
    x = NULL,
    ...
){

  y <- f_proportion(x = x)

  y <- sqrt(
    sweep(
      x = y,
      MARGIN = 1,
      STATS = rowSums(y),
      FUN = "/"
    )
  )

  y

}


#' Data Transformation: Rowwise Centered Log-Ratio
#'
#' @description
#' Centers log-transformed proportions by subtracting the geometric mean of the row.
#'
#' @inheritParams f_hellinger
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(
#'   cols = 5,
#'   data_range = c(0, 500)
#'   )
#'
#' y <- f_clr(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_clr <- function(
    x = NULL,
    ...
) {

  x.log <- f_log(x = x)

  y <- sweep(
    x = x.log,
    MARGIN = 1,
    STATS = rowMeans(x.log),
    FUN = "-"
  )

  y

}


#' Data Transformation: Log
#'
#' @description
#' Applies logarithmic transformation to data to reduce skewness.
#'
#' @inheritParams f_hellinger
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(
#'   cols = 5,
#'   data_range = c(0, 500)
#'   )
#'
#' y <- f_log(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_log <- function(
    x = NULL,
    ...
){

  y <- log(x)

  y <- as.matrix(y)

  y[is.infinite(y)] <- 0

  y <- zoo::zoo(
    x = y,
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )

  y

}

#' Data Transformation: Convert Zoo Object to Binary
#'
#' @description Converts a zoo object to binary (1 and 0) based on a given threshold.
#'
#' @inheritParams f_hellinger
#' @param threshold (required, numeric) Values greater than this number become 1, others become 0. Default: NULL (set to the mean of the time series)
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(
#'   data_range = c(0, 1)
#'   )
#'
#' y <- f_binary(
#'   x = x,
#'   threshold = 0.5
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_binary <- function(
    x = NULL,
    threshold = NULL
){

  if(is.null(threshold)){
    threshold <- mean(x)
  }

  y <- ifelse(
    test = x > threshold,
    yes = 1,
    no = 0
  )

  y <- zoo::zoo(
    x = y,
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )

  y

}


# Scale Adjustments ----

#' Data Transformation: Global Centering and Scaling
#'
#' @description
#' Scaling and/or centering by variable using the mean and standard deviation computed across all time series. Global scaling helps dynamic time warping take variable offsets between time series into account.
#'
#' @param x (required, zoo object) Zoo time series object to transform.
#' @param center (optional, logical or numeric vector) Triggers centering if TRUE. Default: TRUE
#' @param scale (optional, logical or numeric vector) Triggers scaling if TRUE. Default: TRUE
#' @param .global (optional, logical) Used to trigger global scaling within [tsl_transform()].
#' @param ... (optional, additional arguments) Ignored in this function.
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate()
#'
#' y <- f_scale_global(
#'   x = x
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_scale_global <- function(
    x = NULL,
    center = TRUE,
    scale = TRUE,
    .global,
    ...
){

  y <- f_scale_local(
    x = x,
    center = center,
    scale = scale
  )

}

#' Data Transformation: Local Centering and Scaling
#'
#' @description
#' Scaling and/or centering by variable and time series. Local scaling helps dynamic time warping focus entirely on shape comparisons.
#'
#' @inherit f_scale_global
#' @export
#' @autoglobal
#' @family tsl_transformation
f_scale_local <- function(
    x = NULL,
    center = TRUE,
    scale = TRUE,
    ...
){

  y <- scale(
    x = x,
    center = center,
    scale = scale
  )

  y <- as.matrix(y)

  y[is.nan(y)] <- 0

  y <- zoo::zoo(
    x = y,
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )

  y

}

#' @title Data Transformation: Local Rescaling to a New Range
#' @inherit f_rescale_global
#' @export
#' @autoglobal
#' @family tsl_transformation
f_rescale_local <- function(
    x = NULL,
    new_min = 0,
    new_max = 1,
    old_min = NULL,
    old_max = NULL,
    ...
){

  y <- zoo::coredata(x)

  if(!is.matrix(y)){
    y <- as.matrix(y)
  }

  if(length(new_min) < ncol(y)){
    new_min <- rep(x = new_min, times = ncol(y))
  }

  if(length(new_max) < ncol(y)){
    new_max <- rep(x = new_max, times = ncol(y))
  }

  for(i in seq_len(ncol(y))){

    y[, i] <- utils_rescale_vector(
      x =  y[, i],
      new_min = new_min[i],
      new_max = new_max[i],
      old_min = old_min[i],
      old_max = old_max[i]
    )

  }

  y <- zoo::zoo(
    x = y,
    order.by = zoo::index(x)
  ) |>
    zoo_name_set(
      name = attributes(x)$name
    )
  y

}


#' @title Data Transformation: Global Rescaling to a New Range
#' @param x (required, zoo object) Zoo time series object to transform.
#' @param new_min (optional, numeric) New minimum value. Default: `0`
#' @param new_max (optional, numeric) New maximum value. Default: `1`
#' @param old_min (optional, numeric) Old minimum value. Default: `NULL`
#' @param old_max (optional, numeric) Old maximum value. Default: `NULL`
#' @param .global (optional, logical) Used to trigger global scaling within [tsl_transform()].
#' @param ... (optional, additional arguments) Ignored in this function.
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' x <- zoo_simulate(cols = 2)
#'
#' y <- f_rescale_global(
#'   x = x,
#'   new_min = 0,
#'   new_max = 100
#' )
#'
#' if(interactive()){
#'   zoo_plot(x)
#'   zoo_plot(y)
#' }
#' @family tsl_transformation
f_rescale_global <- function(
    x = NULL,
    new_min = 0,
    new_max = 1,
    old_min = NULL,
    old_max = NULL,
    .global,
    ...
){

  y <- f_rescale_local(
    x = x,
    new_min = new_min,
    new_max = new_max,
    old_min = old_min,
    old_max = old_max
  )

  y

}

#' Translate User Time Keywords Into Valid Ones
#'
#' @description
#' Internal function to translate misnamed or abbreviated keywords into valid ones. Uses [utils_time_keywords_dictionary()] as reference dictionary.
#'
#'
#' @param keyword (optional, character string) A time keyword such as "day". Default: NULL
#'
#' @return A valid time keyword.
#' @export
#' @autoglobal
#' @examples
#' #millennia
#' utils_time_keywords_translate(
#'   keyword = "1000 years"
#' )
#'
#' utils_time_keywords_translate(
#'   keyword = "1000 y"
#' )
#'
#' utils_time_keywords_translate(
#'   keyword = "thousands"
#' )
#'
#' #years
#' utils_time_keywords_translate(
#'   keyword = "year"
#' )
#'
#' utils_time_keywords_translate(
#'   keyword = "y"
#' )
#'
#' #days
#' utils_time_keywords_translate(
#'   keyword = "d"
#' )
#'
#' utils_time_keywords_translate(
#'   keyword = "day"
#' )
#'
#' #seconds
#' utils_time_keywords_translate(
#'   keyword = "s"
#' )
#'
#' utils_time_keywords_translate(
#'   keyword = "sec"
#' )
#' @family internal_time_handling
utils_time_keywords_translate <- function(
    keyword = NULL
){

  if(is.null(keyword)){
    return(NULL)
  }

  if(!is.character(keyword) || length(keyword) > 1){
    return(keyword)
  }

  old_keyword <- keyword

  df <- utils_time_keywords_dictionary()

  if(keyword %in% df$keyword){
    return(keyword)
  }

  if(keyword %in% df$pattern){
    return(df[df$pattern == keyword, "keyword"])
  }

  keyword

}

test_that("`utils_time_keywords_translate()` works", {

  expect_equal(utils_time_keywords_translate(keyword = "1000 years"), "millennia")
  expect_equal(utils_time_keywords_translate(keyword = "1000 y"), "millennia")
  expect_equal(utils_time_keywords_translate(keyword = "thousands"), "millennia")
  expect_equal(utils_time_keywords_translate(keyword = "year"), "years")
  expect_equal(utils_time_keywords_translate(keyword = "y"), "years")
  expect_equal(utils_time_keywords_translate(keyword = "d"), "days")
  expect_equal(utils_time_keywords_translate(keyword = "day"), "days")
  expect_equal(utils_time_keywords_translate(keyword = "s"), "seconds")
  expect_equal(utils_time_keywords_translate(keyword = "sec"), "seconds")

})

#' Cost Matrix
#'
#' @description
#' Demonstration function to compute a cost matrix from a distance matrix.
#'
#' @param dist_matrix (required, numeric matrix). Distance matrix generated by [psi_distance_matrix()]. Default: NULL
#' @inheritParams distantia
#' @return numeric matrix
#' @examples
#' #distance metric
#' d <- "euclidean"
#'
#' #use diagonals in least cost computations
#' diagonal <- TRUE
#'
#' #simulate two irregular time series
#' x <- zoo_simulate(
#'   name = "x",
#'   rows = 100,
#'   seasons = 2,
#'   seed = 1
#' )
#'
#' y <- zoo_simulate(
#'   name = "y",
#'   rows = 80,
#'   seasons = 2,
#'   seed = 2
#' )
#'
#' if(interactive()){
#'   zoo_plot(x = x)
#'   zoo_plot(x = y)
#' }
#'
#' #distance matrix
#' dist_matrix <- psi_distance_matrix(
#'   x = x,
#'   y = y,
#'   distance = d
#' )
#'
#' #cost matrix
#' cost_matrix <- psi_cost_matrix(
#'   dist_matrix = dist_matrix,
#'   diagonal = diagonal
#' )
#'
#' if(interactive()){
#'   utils_matrix_plot(
#'     m = cost_matrix
#'     )
#' }
#' @autoglobal
#' @export
#' @family psi_demo
psi_cost_matrix <- function(
    dist_matrix = NULL,
    diagonal = TRUE
){

  dist_matrix <- utils_check_args_matrix(
    m = dist_matrix,
    arg_name = "dist_matrix"
  )

  if(is.logical(diagonal) == FALSE){
    stop("distantia::psi_cost_matrix(): argument 'diagonal' must be logical (TRUE or FALSE).", call. = FALSE)
  }

  #diagonal is TRUE
  if(diagonal == TRUE){

    m <- cost_matrix_diagonal_weighted_cpp(
      dist_matrix = dist_matrix
    )


  } else {

    m <- cost_matrix_orthogonal_cpp(
      dist_matrix = dist_matrix
    )

  }

  #adding names
  dimnames(m) <- dimnames(dist_matrix)

  #adding attributes

  attr(x = m, which = "y_time") <- attributes(dist_matrix)$y_time
  attr(x = m, which = "x_time") <- attributes(dist_matrix)$x_time
  attr(x = m, which = "y_name") <- attributes(dist_matrix)$y_name
  attr(x = m, which = "x_name") <- attributes(dist_matrix)$x_name
  attr(x = m, which = "type") <- "cost"
  attr(x = m, which = "distance") <- attributes(dist_matrix)$distance

  m

}

#' Resample Zoo Objects to a New Time
#'
#' @description
#'
#' \strong{Objective}
#'
#' Time series resampling involves interpolating new values for time steps not available in the original time series. This operation is useful to:
#' \itemize{
#'   \item Transform irregular time series into regular.
#'   \item Align time series with different temporal resolutions.
#'   \item Increase (upsampling) or decrease (downsampling) the temporal resolution of a time series.
#' }
#'
#' On the other hand, time series resampling \strong{should not be used} to extrapolate new values outside of the original time range of the time series, or to increase the resolution of a time series by a factor of two or more. These operations are known to produce non-sensical results.
#'
#' \strong{Methods}
#' This function offers three methods for time series interpolation:
#'
#' \itemize{
#'   \item "linear" (default): interpolation via piecewise linear regression as implemented in [zoo::na.approx()].
#'   \item "spline": cubic smoothing spline regression as implemented in [stats::smooth.spline()].
#'   \item "loess": local polynomial regression fitting as implemented in [stats::loess()].
#' }
#'
#' These methods are used to fit models `y ~ x` where `y` represents the values of a univariate time series and `x` represents a numeric version of its time.
#'
#' The functions [utils_optimize_spline()] and [utils_optimize_loess()] are used under the hood to optimize the complexity of the methods "spline" and "loess" by finding the configuration that minimizes the root mean squared error (RMSE) between  observed and predicted `y`. However, when the argument `max_complexity = TRUE`, the complexity optimization is ignored, and a maximum complexity model is used instead.
#'
#' \strong{New time}
#'
#' The argument `new_time` offers several alternatives to help define the new time of the resulting time series:
#'
#' \itemize{
#'   \item `NULL`: the target time series (`x`) is resampled to a regular time within its original time range and number of observations.
#'   \item `zoo object`: a zoo object to be used as template for resampling. Useful when the objective is equalizing the frequency of two separate zoo objects.
#'   \item `time vector`: a time vector of a class compatible with the time in `x`.
#'   \item `keyword`: character string defining a resampling keyword, obtained via `zoo_time(x, keywords = "resample")$keywords`.
#'   \item `numeric`: a single number representing the desired interval between consecutive samples in the units of `x` (relevant units can be obtained via `zoo_time(x)$units`).
#' }
#'
#' \strong{Step by Step}
#'
#' The steps to resample a time series list are:
#'
#' \enumerate{
#'   \item The time interpolation range taken from the index of the zoo object. This step ensures that no extrapolation occurs during resampling.
#'   \item If `new_time` is provided, any values of `new_time` outside of the minimum and maximum interpolation times are removed to avoid extrapolation. If `new_time` is not provided, a regular time within the interpolation time range of the zoo object is generated.
#'   \item For each univariate time time series, a model `y ~ x`, where `y` is the time series and `x` is its own time coerced to numeric is fitted.
#'   \itemize{
#'    \item If `max_complexity == FALSE` and `method = "spline"` or `method = "loess"`, the model with the complexity that minimizes the root mean squared error between the observed and predicted `y` is returned.
#'    \item If `max_complexity == TRUE` and `method = "spline"` or `method = "loess"`, the first valid model closest to a maximum complexity is returned.
#'   }
#'   \item The fitted model is predicted over `new_time` to generate the resampled time series.
#' }
#'
#' \strong{Other Details}
#'
#' Please use this operation with care, as there are limits to the amount of resampling that can be done without distorting the data. The safest option is to keep the distance between new time points within the same magnitude of the distance between the old time points.
#'
#' @param x (required, zoo object) Time series to resample. Default: NULL
#' @param new_time (optional, zoo object, keyword, or time vector) New time to resample `x` to. The available options are:
#' \itemize{
#'   \item NULL: a regular version of the time in `x` is generated and used for resampling.
#'   \item zoo object: the index of the given zoo object is used as template to resample `x`.
#'   \item time vector: a vector with new times to resample `x` to. If time in `x` is of class "numeric", this vector must be numeric as well. Otherwise, vectors of classes "Date" and "POSIXct" can be used indistinctly.
#'   \item keyword: a valid keyword returned by `zoo_time(x)$keywords`, used to generate a time vector with the relevant units.
#'   \item numeric of length 1: interpreted as new time interval, in the highest resolution units returned by `zoo_time(x)$units`.
#' }
#' @param method (optional, character string) Name of the method to resample the time series. One of "linear", "spline" or "loess". Default: "linear".
#' @param max_complexity (required, logical). Only relevant for methods "spline" and "loess". If TRUE, model optimization is ignored, and the a model of maximum complexity (an overfitted model) is used for resampling. Default: FALSE
#'
#'
#' @return zoo object
#' @export
#' @autoglobal
#' @examples
#' #simulate irregular time series
#' x <- zoo_simulate(
#'   cols = 2,
#'   rows = 50,
#'   time_range = c("2010-01-01", "2020-01-01"),
#'   irregular = TRUE
#'   )
#'
#' #plot time series
#' if(interactive()){
#'   zoo_plot(x)
#' }
#'
#' #intervals between samples
#' x_intervals <- diff(zoo::index(x))
#' x_intervals
#'
#' #create regular time from the minimum of the observed intervals
#' new_time <- seq.Date(
#'   from = min(zoo::index(x)),
#'   to = max(zoo::index(x)),
#'   by = floor(min(x_intervals))
#' )
#'
#' new_time
#' diff(new_time)
#'
#' #resample using piecewise linear regression
#' x_linear <- zoo_resample(
#'   x = x,
#'   new_time = new_time,
#'   method = "linear"
#' )
#'
#' #resample using max complexity splines
#' x_spline <- zoo_resample(
#'   x = x,
#'   new_time = new_time,
#'   method = "spline",
#'   max_complexity = TRUE
#' )
#'
#' #resample using max complexity loess
#' x_loess <- zoo_resample(
#'   x = x,
#'   new_time = new_time,
#'   method = "loess",
#'   max_complexity = TRUE
#' )
#'
#'
#' #intervals between new samples
#' diff(zoo::index(x_linear))
#' diff(zoo::index(x_spline))
#' diff(zoo::index(x_loess))
#'
#' #plotting results
#' if(interactive()){
#'
#'   par(mfrow = c(4, 1), mar = c(3,3,2,2))
#'
#'   zoo_plot(
#'     x,
#'     guide = FALSE,
#'     title = "Original"
#'     )
#'
#'   zoo_plot(
#'     x_linear,
#'     guide = FALSE,
#'     title = "Method: linear"
#'   )
#'
#'   zoo_plot(
#'     x_spline,
#'     guide = FALSE,
#'     title = "Method: spline"
#'     )
#'
#'   zoo_plot(
#'     x_loess,
#'     guide = FALSE,
#'     title = "Method: loess"
#'   )
#'
#' }
#'
#' @family zoo_functions
zoo_resample <- function(
    x = NULL,
    new_time = NULL,
    method = "linear",
    max_complexity = FALSE
){


  if(zoo::is.zoo(x) == FALSE){
    stop("distantia::zoo_resample(): argument 'x' must be a zoo object.", call. = FALSE)
  }

  method <- match.arg(
    arg = method,
    choices = c(
      "linear",
      "spline",
      "loess"
    ),
    several.ok = FALSE
  )

  # handle new time ----
  old_time <- zoo_time(
    x = x,
    keywords = "resample"
    )

  #new_time is NULL
  if(is.null(new_time)){

    #create regular new_time from x
    new_time <- seq(
      from = old_time$begin,
      to = old_time$end,
      length.out = old_time$rows
    )

  }

  #process new_time
  new_time <- utils_new_time(
    tsl = zoo_to_tsl(x = x),
    new_time = new_time,
    keywords = "resample"
  )

  #subset to original range
  new_time <- new_time[
    new_time >= old_time$begin &
      new_time <= old_time$end
  ]

  #compare old and new new_time intervals
  time_interval <- as.numeric(mean(diff(new_time)))
  old_time_interval <- as.numeric(mean(diff(zoo::index(x))))

  if(
    time_interval <= (old_time_interval/10) ||
    time_interval >= (old_time_interval*10)
  ){
    warning("distantia::zoo_resample(): The time intervals of 'new_time' and 'x' differ in one order of magnitude or more. The output time series might be highly distorted.", call. = FALSE)
  }

  #default method
  if(method == "linear"){

    y <- zoo::na.approx(
      object = x,
      xout = new_time
    )

    y <- zoo_name_set(
      x = y,
      name = attributes(x)$name
    )

    return(y)

  }

  #convert times to numeric
  new_time_numeric <- utils_coerce_time_class(
    x = new_time,
    to = "numeric"
  )

  old_time <- zoo::index(x)

  old_time_numeric <- utils_coerce_time_class(
    x = old_time,
    to = "numeric"
  )

  #interpolate


  y <- foreach::foreach(
    i = seq_len(ncol(x)),
    .combine = "cbind",
    .errorhandling = "pass",
    .options.future = list(seed = TRUE)
  ) %dofuture% {

    x.i <- as.numeric(x[, i])

    if(method == "spline"){
      f <- utils_optimize_spline
    }

    if(method == "loess"){
      f <- utils_optimize_loess
    }

    interpolation.model <- f(
      x = old_time_numeric,
      y = x.i,
      max_complexity = max_complexity
    )


    #if model fails, return NA
    if(
      inherits(
        x = interpolation.model,
        what = "loess"
      ) == FALSE &&
      inherits(
        x = interpolation.model,
        what = "smooth.spline"
      ) == FALSE
    ){

      return(rep(x = NA, times = length(new_time)))

    }

    #predict model
    if(method == "spline"){

      y.i <- stats::predict(
        object = interpolation.model,
        x = new_time_numeric
      )$y

    } else {

      y.i <- stats::predict(
        object = interpolation.model,
        newdata = data.frame(
          x = new_time_numeric
        )
      )

    }

    #set bounds
    y.i[y.i > max(x.i)] <- max(x.i)
    y.i[y.i < min(x.i)] <- min(x.i)

    return(y.i)

  }

  #rename columns
  colnames(y) <- colnames(x)

  #convert to zoo
  y <- zoo::zoo(
    x = y,
    order.by = new_time
  )

  #reset name
  zoo_name_set(
    x = y,
    name = attributes(x)$name
  )

}

#' Count NA Cases in Time Series Lists
#'
#' @description
#' Converts Inf, -Inf, and NaN to NA (via [tsl_Inf_to_NA()] and [tsl_NaN_to_NA()]), and counts the total number of NA cases in each time series.
#'
#'
#' @param tsl (required, list) Time series list. Default: NULL
#'
#' @return list
#' @export
#' @autoglobal
#' @examples
#' #tsl with no NA cases
#' tsl <- tsl_simulate()
#'
#' tsl_count_NA(tsl = tsl)
#'
#' #tsl with NA cases
#' tsl <- tsl_simulate(
#'   na_fraction = 0.3
#' )
#'
#' tsl_count_NA(tsl = tsl)
#'
#' #tsl with variety of empty cases
#' tsl <- tsl_simulate()
#' tsl[[1]][1, 1] <- Inf
#' tsl[[1]][2, 1] <- -Inf
#' tsl[[1]][3, 1] <- NaN
#' tsl[[1]][4, 1] <- NaN
#'
#' tsl_count_NA(tsl = tsl)
#'
#' @family tsl_management
tsl_count_NA <- function(
    tsl = NULL
){

  utils_check_args_tsl(
    tsl = tsl,
    min_length = 1
  )

  #replaces Inf with Na
  tsl <- tsl_Inf_to_NA(
    tsl = tsl
  )

  #replaces NaN with NA
  tsl <- tsl_NaN_to_NA(
    tsl = tsl
  )

  na_count_list <- future.apply::future_lapply(
    X = tsl,
    FUN = function(x){
      sum(is.na(x))
    }
  )


  na_count_list

}

#' Hierarchical Clustering of Dissimilarity Analysis Data Frames
#'
#' @description
#' This function combines the dissimilarity scores computed by [distantia()], the agglomerative clustering methods provided by [stats::hclust()], and the clustering optimization method implemented in [utils_cluster_hclust_optimizer()] to help group together time series with similar features.
#'
#' When `clusters = NULL`, the function [utils_cluster_hclust_optimizer()] is run underneath to perform a parallelized grid search to find the  number of clusters maximizing the overall silhouette width of the clustering solution (see [utils_cluster_silhouette()]). When `method = NULL` as well, the optimization also includes all methods available in [stats::hclust()] in the grid search.
#'
#' This function supports a parallelization setup via [future::plan()], and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#'
#' @inheritParams distantia_aggregate
#' @param clusters (required, integer) Number of groups to generate. If NULL (default), [utils_cluster_kmeans_optimizer()] is used to find the number of clusters that maximizes the mean silhouette width of the clustering solution (see [utils_cluster_silhouette()]). Default: NULL
#' @param method (optional, character string) Argument of [stats::hclust()] defining the agglomerative method. One of: "ward.D", "ward.D2", "single", "complete", "average" (= UPGMA), "mcquitty" (= WPGMA), "median" (= WPGMC) or "centroid" (= UPGMC). Unambiguous abbreviations are accepted as well. If NULL (default), [utils_cluster_hclust_optimizer()] finds the optimal method. Default: "complete".
#' @return list:
#' \itemize{
#'   \item `cluster_object`: hclust object for further analyses and custom plotting.
#'   \item `clusters`: integer, number of clusters.
#'   \item `silhouette_width`: mean silhouette width of the clustering solution.
#'   \item `df`: data frame with time series names, their cluster label, and their individual silhouette width scores.
#'   \item `d`: psi distance matrix used for clustering.
#'   \item `optimization`: only if `clusters = NULL`, data frame with optimization results from [utils_cluster_hclust_optimizer()].
#' }
#'
#' @export
#' @autoglobal
#' @examples
#'
#' #weekly covid prevalence in California
#' tsl <- tsl_initialize(
#'   x = covid_prevalence,
#'   name_column = "name",
#'   time_column = "time"
#' )
#'
#' #subset 10 elements to accelerate example execution
#' tsl <- tsl_subset(
#'   tsl = tsl,
#'   names = 1:10
#' )
#'
#' if(interactive()){
#'   #plotting first three time series
#'   tsl_plot(
#'     tsl = tsl[1:3],
#'     guide_columns = 3
#'   )
#' }
#'
#' #dissimilarity analysis
#' distantia_df <- distantia(
#'   tsl = tsl,
#'   lock_step = TRUE
#' )
#'
#' #hierarchical clustering
#' #automated number of clusters
#' #automated method selection
#' distantia_clust <- distantia_cluster_hclust(
#'   df = distantia_df,
#'   clusters = NULL,
#'   method = NULL
#' )
#'
#' #names of the output object
#' names(distantia_clust)
#'
#' #cluster object
#' distantia_clust$cluster_object
#'
#' #distance matrix used for clustering
#' distantia_clust$d
#'
#' #number of clusters
#' distantia_clust$clusters
#'
#' #clustering data frame
#' #group label in column "cluster"
#' #negatives in column "silhouette_width" higlight anomalous cluster assignation
#' distantia_clust$df
#'
#' #mean silhouette width of the clustering solution
#' distantia_clust$silhouette_width
#'
#' #plot
#' if(interactive()){
#'
#'   dev.off()
#'
#'   clust <- distantia_clust$cluster_object
#'   k <- distantia_clust$clusters
#'
#'   #tree plot
#'   plot(
#'     x = clust,
#'     hang = -1
#'   )
#'
#'   #highlight groups
#'   stats::rect.hclust(
#'     tree = clust,
#'     k = k,
#'     cluster = stats::cutree(
#'       tree = clust,
#'       k = k
#'     )
#'   )
#'
#' }
#'
#' @family distantia_support
distantia_cluster_hclust <- function(
    df = NULL,
    clusters = NULL,
    method = "complete"
){

  #aggregate df if needed
  df <- distantia_aggregate(
    df = df,
    f = mean
  )

  d <- distantia_matrix(
    df = df
  )[[1]]

  d_dist <- stats::as.dist(d)

  #optimize groups
  if(is.null(clusters)){

    optimization_df <- utils_cluster_hclust_optimizer(
      d = d,
      method = method
    )

    best_index <- which.max(optimization_df$silhouette_mean)

    clusters <- optimization_df[
      best_index,
      "clusters"
    ]

    method <- optimization_df[
      best_index,
      "method"
    ]

  }

  if(clusters < 2){
    clusters <- 2
  }

  if(clusters >= nrow(d)){
    clusters <- nrow(d) - 1
  }

  hc <- stats::hclust(
    d = d_dist,
    method = method
  )

  hc_groups <- stats::cutree(
    tree = hc,
    k = clusters,
  )

  hc_silhouette <- utils_cluster_silhouette(
    labels = hc_groups,
    d = d,
    mean = FALSE
  )

  out <- list(
    cluster_object = hc,
    clusters = clusters,
    silhouette_width = mean(hc_silhouette$silhouette_width),
    df = hc_silhouette,
    d = d_dist
  )

  if(exists("optimization_df")){
    out$optimization <- optimization_df
  }

  out

}

#' Distance Between Two Numeric Vectors
#'
#' @description
#' Computes the distance between two numeric vectors using a distance metric included in the data frame `distantia::distances`.
#' @param x (required, numeric vector).
#' @param y (required, numeric vector) of same length as `x`.
#' @param distance (optional, character string) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset `distances`. Default: "euclidean".
#' @return numeric value
#' @examples
#'
#' distance(
#'   x = runif(100),
#'   y = runif(100),
#'   distance = "euclidean"
#' )
#'
#' @autoglobal
#' @export
#' @family distances
distance <- function(
    x = NULL,
    y = NULL,
    distance = "euclidean"
    ){

  distance <- utils_check_distance_args(
    distance = distance
  )[1]

  if(is.null(x)){
    stop("distantia::distance(): argument 'x' must not be NULL.", call. = FALSE)
  }

  if(is.null(y)){
    stop("distantia::distance(): argument 'y' must not be NULL.", call. = FALSE)
  }

  if(length(x) != length(y)){
    stop("distantia::distance(): arguments 'x' and 'y' must be of the same length.", call. = FALSE)
  }

  #handling NA
  df <- data.frame(
    x = x,
    y = y
  ) |>
    stats::na.omit()

  #methods that don't accept two zeros in same position
  if(distance %in% c(
    "chi",
    "cos",
    "cosine"
  )
  ){

    pseudozero <- mean(x = c(df$x, df$y)) * 0.0001
    df[df == 0] <- pseudozero

  }

  #generate expression
  expression <- paste0(
    distances[
      distances$name %in% distance |
        distances$abbreviation %in% distance,
      "function_name"
    ],
    "(df$x, df$y)"
  )

  y <- eval(parse(text = expression))

  attr(x = y, which = "distance") <- distance

  y

}

test_that("`distance()` works", {
  expect_true(is.numeric(distance(x = runif(100), y = runif(100), distance = "euclidean")))

  expect_true(distance(x = 1, y = 1) == 0)
})

test_that("`distance()` works with bray_curtis full name and abbreviation", {
  x <- c(0.1, 0.4, 0.5)
  y <- c(0.2, 0.3, 0.5)
  expect_true(is.numeric(distance(x = x, y = y, distance = "bray_curtis")))
  expect_equal(
    distance(x = x, y = y, distance = "bray_curtis"),
    distance(x = x, y = y, distance = "bra")
  )
})

test_that("`distance()` works with sorensen full name and abbreviation", {
  x <- c(1, 0, 1, 0)
  y <- c(1, 1, 0, 0)
  expect_true(is.numeric(distance(x = x, y = y, distance = "sorensen")))
  expect_equal(
    distance(x = x, y = y, distance = "sorensen"),
    distance(x = x, y = y, distance = "sor")
  )
})

#' Least Cost Path
#'
#' @description
#' Demonstration function to compute the least cost path within a cost matrix
#'
#'
#' @param dist_matrix (required, numeric matrix) Distance matrix generated by [psi_distance_matrix()]. Default: NULL
#' @param cost_matrix (required, numeric matrix) Cost matrix generated from the distance matrix with [psi_cost_matrix()]. Default: NULL
#' @inheritParams distantia
#' @return data frame
#' @export
#' @examples
#' #distance metric
#' d <- "euclidean"
#'
#' #simulate two irregular time series
#' x <- zoo_simulate(
#'   name = "x",
#'   rows = 100,
#'   seasons = 2,
#'   seed = 1
#' )
#'
#' y <- zoo_simulate(
#'   name = "y",
#'   rows = 80,
#'   seasons = 2,
#'   seed = 2
#' )
#'
#' if(interactive()){
#'   zoo_plot(x = x)
#'   zoo_plot(x = y)
#' }
#'
#' #distance matrix
#' dist_matrix <- psi_distance_matrix(
#'   x = x,
#'   y = y,
#'   distance = d
#' )
#'
#' #diagonal least cost path
#' #------------------------
#'
#' cost_matrix <- psi_cost_matrix(
#'   dist_matrix = dist_matrix,
#'   diagonal = TRUE
#' )
#'
#' cost_path <- psi_cost_path(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix,
#'   diagonal = TRUE
#' )
#'
#' if(interactive()){
#'   utils_matrix_plot(
#'     m = cost_matrix,
#'     path = cost_path
#'     )
#' }
#'
#'
#' #orthogonal least cost path
#' #--------------------------
#' cost_matrix <- psi_cost_matrix(
#'   dist_matrix = dist_matrix,
#'   diagonal = FALSE
#' )
#'
#' cost_path <- psi_cost_path(
#'   dist_matrix = dist_matrix,
#'   cost_matrix = cost_matrix,
#'   diagonal = FALSE
#' )
#'
#' if(interactive()){
#'   utils_matrix_plot(
#'     m = cost_matrix,
#'     path = cost_path
#'   )
#' }
#' @autoglobal
#' @family psi_demo
psi_cost_path <- function(
    dist_matrix = NULL,
    cost_matrix = NULL,
    diagonal = TRUE,
    bandwidth = 1
){

  dist_matrix <- utils_check_args_matrix(
    m = dist_matrix,
    arg_name = "dist_matrix"
  )

  cost_matrix <- utils_check_args_matrix(
    m = cost_matrix,
    arg_name = "cost_matrix"
  )

  if(is.logical(diagonal) == FALSE){
    stop("distantia::psi_cost_path(): argument 'diagonal' must be logical (TRUE or FALSE).", call. = FALSE)
  }

  if(diagonal == FALSE){

    if(bandwidth >= 1){

      path <- cost_path_orthogonal_cpp(
        dist_matrix = dist_matrix,
        cost_matrix = cost_matrix
      )

    } else {

      path <- cost_path_orthogonal_bandwidth_cpp(
        dist_matrix = dist_matrix,
        cost_matrix = cost_matrix,
        bandwidth = bandwidth
      )

    }

  } else {

    if(bandwidth >= 1){

      path <- cost_path_diagonal_cpp(
        dist_matrix = dist_matrix,
        cost_matrix = cost_matrix
      )

    } else {

      path <- cost_path_diagonal_bandwidth_cpp(
        dist_matrix = dist_matrix,
        cost_matrix = cost_matrix,
        bandwidth = bandwidth
      )

    }

  }

  attr(x = path, which = "y_name") <- attributes(dist_matrix)$y_name
  attr(x = path, which = "x_name") <- attributes(dist_matrix)$x_name
  attr(x = path, which = "type") <- "cost_path"
  attr(x = path, which = "distance") <- attributes(dist_matrix)$distance


  path

}

#' Spatial Representation of `distantia()` Data Frames
#'
#' @description
#' Given an sf data frame with geometry types POLYGON, MULTIPOLYGON, or POINT representing time series locations, this function transforms the output of [distantia()], [distantia_ls()], [distantia_dtw()] or [distantia_time_delay()] to an sf data frame.
#'
#' If `network = TRUE`, the sf data frame is of type LINESTRING, with edges connecting time series locations. This output is helpful to build many-to-many dissimilarity maps (see examples).
#'
#' If `network = FALSE`, the sf data frame contains the geometry in the input `sf` argument. This output helps build one-to-many dissimilarity maps.
#'
#' @param df (required, data frame) Output of [distantia()] or [distantia_time_delay()]. Default: NULL
#' @param sf (required, sf data frame) Points or polygons representing the location of the time series in argument 'df'. It must have a column with all time series names in `df$x` and `df$y`. Default: NULL
#' @param network (optional, logical) If TRUE, the resulting sf data frame is of type LINESTRING and represents network edges. Default: TRUE
#'
#' @return sf data frame (LINESTRING geometry)
#' @export
#'
#' @examples
#' tsl <- distantia::tsl_initialize(
#'   x = distantia::covid_prevalence,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#' distantia::tsl_subset(
#'   names = c(
#'     "Los_Angeles",
#'     "San_Francisco",
#'     "Fresno",
#'     "San_Joaquin"
#'     )
#'  )
#'
#' df_psi <- distantia::distantia_ls(
#'   tsl = tsl
#' )
#'
#' #network many to many
#' sf_psi <- distantia::distantia_spatial(
#'   df = df_psi,
#'   sf = distantia::covid_counties,
#'   network = TRUE
#' )
#'
#' #network map
#' # mapview::mapview(
#' #   distantia::covid_counties,
#' #   col.regions = NA,
#' #   alpha.regions = 0,
#' #   color = "black",
#' #   label = "name",
#' #   legend = FALSE,
#' #   map.type = "OpenStreetMap"
#' # ) +
#' #   mapview::mapview(
#' #     sf_psi_subset,
#' #     layer.name = "Psi",
#' #     label = "edge_name",
#' #     zcol = "psi",
#' #     lwd = 3
#' #   ) |>
#' #   suppressWarnings()
#'
#' @autoglobal
#' @family distantia_support
distantia_spatial <- function(
    df = NULL,
    sf = NULL,
    network = TRUE
){

  if(
    requireNamespace(
      package = "sf",
      quietly = TRUE
    ) == FALSE){
    stop("distantia::distantia/momentum_spatial(): this function requires installing the package 'sf'.", call. = FALSE)
  }

  #check df
  if(is.null(df)){
    stop("distantia::distantia/momentum_spatial(): argument 'df' must be a data frame resulting from distantia::distantia() or distantia::momentum().", call. = FALSE)
  }

  df_type_distantia <- c(
    "distantia_df",
    "time_delay_df"
  )

  df_type_momentum <- "momentum_df"

  df_types <- c(
    df_type_distantia,
    df_type_momentum
  )

  df_type <- attributes(df)$type

  if(
    is.null(df_type) ||
    !(df_type %in% df_types)
    ){
    stop("distantia::distantia/momentum_spatial(): argument 'df' must be the output of distantia::distantia(), distantia::distantia_time_delay(), or distantia::momentum().", call. = FALSE)
  }

  if(df_type %in% df_type_distantia){

    f_name <- "distantia::distantia_spatial(): "

    if(df_type == "distantia_df"){

      df <- distantia_aggregate(
        df = df
      )

    }
  }

  if(df_type == df_type_momentum){

    f_name <- "distantia::momentum_spatial(): "

    df <- df |>
      momentum_aggregate() |>
      momentum_to_wide()

  }

  # sf ----
  if(is.null(sf)){
    stop(f_name, "argument 'sf' must be an sf data frame.", call. = FALSE)
  }

  ## find geometry ----
  if(inherits(x = sf, what = "sf") == FALSE){
    stop(f_name, "argument 'sf' must be an 'sf' data frame.", call. = FALSE)
  }

  ## find names column ----
  df_names <- unique(c(df$x, df$y))

  sf_names <- apply(
    X = sf,
    MARGIN = 2,
    FUN = function(x){
      if(all(df_names %in% x)){
        return(TRUE)
      } else {
        return(FALSE)
      }
    }
  )

  sf_names <- names(sf_names[sf_names == TRUE])

  if(length(sf_names) == 0){
    stop(
      f_name,
      "no column in 'sf' contains all time series names from 'df'. ",
      "Ensure 'sf' has a column with all time series names.",
      call. = FALSE
    )
  }

  sf_names <- sf_names[1]

  if(sf_names %in% colnames(df)){
    df[[sf_names]] <- NULL
  }

  ## remove common columns in sf
  sf_valid_columns <- setdiff(
    x = colnames(sf),
    y = colnames(df)
  )

  sf <- sf[, sf_valid_columns]

  #prepare y vs x mirror
  df_mirror <- df
  df_mirror$x <- df$y
  df_mirror$y <- df$x

  #prepare and return non-network output
  if(network == FALSE){

    #prepare rows of self x vs y
    df_self <- data.frame(
      x = unique(df$x),
      y = unique(df$x)
    )

    df_self_cols <- setdiff(
      x = colnames(df),
      y = c("x", "y")
    )

    for(col.i in df_self_cols){
      if(is.numeric(df[[col.i]])){
        df_self[[col.i]] <- 0
      } else {
        df_self[[col.i]] <- NA
      }
    }

    #join all rows
    df <- rbind(
      df,
      df_mirror,
      df_self
    )

    #remove duplicates
    df$xy <- paste(df$x, df$y)
    df <- df[!duplicated(df$xy), ]
    df$xy <- NULL

    #create join column
    df[[sf_names]] <- df$y

    #merge sf
    df_sf <- merge(
      x = sf,
      y = df,
      by = sf_names
    )

    #remove merge column
    df_sf[[sf_names]] <- NULL

    #order by names
    df_sf <- df_sf[order(df_sf$x, df_sf$y), ]

    rownames(df_sf) <- NULL

    return(df_sf)

  }

  #building network

  #check geometry type
  sf_geometry_type <- sf |>
    sf::st_geometry_type() |>
    unique() |>
    as.character()

  #if polygons, compute centroids
  if(any(c("POLYGON", "MULTIPOLYGON") %in% sf_geometry_type)){

    sf <- sf::st_set_geometry(
      sf,
      sf::st_centroid(
        x = sf::st_geometry(obj = sf),
        of_largest_polygon = TRUE
      )
    )

  }

  ## add id column ----
  sf$id <- seq_len(nrow(sf))

  ## map id column in df ----
  df <- merge(
    x = df,
    y = sf::st_drop_geometry(sf[, c("id", sf_names)]),
    by.x = "x",
    by.y = sf_names
  )

  colnames(df)[colnames(df) == "id"] <- "id_x"

  df <- merge(
    x = df,
    y = sf::st_drop_geometry(sf[, c("id", sf_names)]),
    by.x = "y",
    by.y = sf_names
  )

  colnames(df)[colnames(df) == "id"] <- "id_y"


  # generate network ----

  ## function to create lines
  points_to_line <- function(id_x, id_y, sf) {

    sf[c(id_x, id_y), ] |>
      sf::st_coordinates() |>
      sf::st_linestring() |>
      sf::st_sfc()

  }

  ## generate lines list ----
  sf_lines_list <- mapply(
    FUN = points_to_line,
    df$id_x,
    df$id_y,
    MoreArgs = list(sf = sf)
  )

  ## to sf ----
  df_sf <- df |>
    sf::st_sf(
      geometry = sf::st_sfc(sf_lines_list),
      crs = sf::st_crs(sf)
    )

  ## warn about zero-length linestrings (identical coordinates)
  zero_length <- !sf::st_is_valid(df_sf)
  if(any(zero_length)){
    bad_pairs <- paste(df$x[zero_length], "-", df$y[zero_length], collapse = ", ")
    warning(
      "distantia_spatial(): identical spatial coordinates produced degenerate (zero-length) linestrings for pair(s): ",
      bad_pairs, ". Check the 'sf' input.",
      call. = FALSE
    )
  }

  ## remove id columns
  df_sf$id_x <- NULL
  df_sf$id_y <- NULL

  ## create name column
  old_colnames <- colnames(df_sf)

  df_sf$edge_name <- paste(df_sf$x, "-", df_sf$y)

  #put name in first column
  df_sf <- df_sf[, c("edge_name", old_colnames)]

  #length
  df_sf$length <- sf::st_length(
    x = df_sf
  ) |>
    as.numeric()

  rownames(df_sf) <- NULL

  df_sf

}

#' Optimize the Silhouette Width of Hierarchical Clustering Solutions
#'
#' @description
#' Performs a parallelized grid search to find the number of clusters maximizing the overall silhouette width of the clustering solution (see [utils_cluster_silhouette()]). When `method = NULL`, the optimization also includes all methods available in [stats::hclust()] in the grid search. This function supports parallelization via [future::plan()] and a progress bar generated by the `progressr` package (see Examples).
#'
#' @param d (required, matrix) distance matrix typically resulting from [distantia_matrix()], but any other square matrix should work. Default: NULL
#' @param method (optional, character string) argument of [stats::hclust()] defining the agglomerative method. One of: "ward.D", "ward.D2", "single", "complete", "average" (= UPGMA), "mcquitty" (= WPGMA), "median" (= WPGMC), or "centroid" (= UPGMC). Unambiguous abbreviations are accepted as well.
#'
#' This function supports parallelization via [future::plan()] and progress bars provided by the package [progressr](https://CRAN.R-project.org/package=progressr).
#'
#' @return data frame
#' @export
#' @autoglobal
#' @examples
#'
#' #weekly covid prevalence
#' #in 10 California counties
#' #aggregated by month
#' tsl <- tsl_initialize(
#'   x = covid_prevalence,
#'   name_column = "name",
#'   time_column = "time"
#' ) |>
#'   tsl_subset(
#'     names = 1:10
#'   ) |>
#'   tsl_aggregate(
#'     new_time = "months",
#'     fun = max
#'   )
#'
#' if(interactive()){
#'   #plotting first three time series
#'   tsl_plot(
#'     tsl = tsl_subset(
#'       tsl = tsl,
#'       names = 1:3
#'     ),
#'     guide_columns = 3
#'   )
#' }
#'
#' #compute dissimilarity matrix
#' psi_matrix <- distantia(
#'   tsl = tsl,
#'   lock_step = TRUE
#' ) |>
#'   distantia_matrix()
#'
#' #optimize hierarchical clustering
#' hclust_optimization <- utils_cluster_hclust_optimizer(
#'   d = psi_matrix
#' )
#'
#' #best solution in first row
#' head(hclust_optimization)
#' 
#' @family distantia_support
utils_cluster_hclust_optimizer <- function(
    d = NULL,
    method = NULL
){

  if(is.list(d)){
    d <- d[[1]]
  }

  if(!is.matrix(d)){
    stop("distantia::utils_cluster_hclust_optimizer(): argument 'd' must be a matrix.", call. = FALSE)
  }

  if(nrow(d) != ncol(d)){
    stop("distantia::utils_cluster_hclust_optimizer(): argument 'd' must be a square distance matrix.", call. = FALSE)
  }

  d_dist <- stats::as.dist(d)

  clusters_vector <- seq(
    from = 2,
    to = nrow(d),
    by = 1
  )

  if(is.null(method)){
    methods_vector <- c(
      "ward.D",
      "ward.D2",
      "single",
      "complete",
      "average",
      "mcquitty",
      "median",
      "centroid"
    )
  } else {
    methods_vector <- method
  }

  df <- expand.grid(
    clusters = clusters_vector,
    method = methods_vector,
    stringsAsFactors = FALSE
  )
  df$silhouette_mean <- NA



  #clustering methods
  hc_methods <- foreach::foreach(
    i = methods_vector
  ) %dofuture% {

    tryCatch({
      stats::hclust(
        d = d_dist,
        method = i
      )
    }, error = function(e) {
      NA
    })

  }

  names(hc_methods) <- methods_vector
  hc_methods <- stats::na.omit(hc_methods)

  #clustering groups
  df <- df[
    df$method %in% names(hc_methods),
  ]

  #iterations
  iterations <- seq_len(nrow(df))

  p <- progressr::progressor(along = iterations)

  #parallelized optimization
  df$silhouette_mean <- foreach::foreach(
    i = iterations,
    .combine = "c"
  ) %dofuture% {

    # p()

    k_groups <- tryCatch({
      stats::cutree(
        tree = hc_methods[[df[i, "method"]]],
        k = df[i, "clusters"],
      )
    }, error = function(e) {
      NA
    })

    if(!is.integer(k_groups)){
      return(NA)
    }

    k_sil <- tryCatch({
      utils_cluster_silhouette(
        labels = k_groups,
        d = d,
        mean = TRUE
      )
    }, error = function(e) {
      NA
    })

    k_sil

  }

  df <- df[order(-df$silhouette_mean), ]

  rownames(df) <- NULL

  df

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_names.R
\name{tsl_names_clean}
\alias{tsl_names_clean}
\title{Clean Time Series Names in a Time Series List}
\usage{
tsl_names_clean(
  tsl = NULL,
  lowercase = FALSE,
  separator = "_",
  capitalize_first = FALSE,
  capitalize_all = FALSE,
  length = NULL,
  suffix = NULL,
  prefix = NULL
)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{lowercase}{(optional, logical) If TRUE, all names are coerced to lowercase. Default: FALSE}

\item{separator}{(optional, character string) Separator when replacing spaces and dots. Also used to separate \code{suffix} and \code{prefix} from the main word. Default: "_".}

\item{capitalize_first}{(optional, logical) Indicates whether to capitalize the first letter of each name. Default: FALSE.}

\item{capitalize_all}{(optional, logical) Indicates whether to capitalize all letters of each name. Default: FALSE.}

\item{length}{(optional, integer) Minimum length of abbreviated names. Names are abbreviated via \code{\link[=abbreviate]{abbreviate()}}. Default: NULL.}

\item{suffix}{(optional, character string) Suffix for the clean names. Default: NULL.}

\item{prefix}{(optional, character string) Prefix for the clean names. Default: NULL.}
}
\value{
time series list
}
\description{
Combines \code{\link[=utils_clean_names]{utils_clean_names()}} and \code{\link[=tsl_names_set]{tsl_names_set()}} to help clean, abbreviate, capitalize, and add a suffix or a prefix to time series list names.
}
\examples{
#initialize time series list
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
)

#original names
tsl_names_get(
  tsl = tsl
)

#abbreviate names
#---------------------------
tsl_clean <- tsl_names_clean(
  tsl = tsl,
  capitalize_first = TRUE,
  length = 4 #abbreviate to 4 characters
)

#new names
tsl_names_get(
  tsl = tsl_clean
)

#suffix and prefix
#---------------------------
tsl_clean <- tsl_names_clean(
  tsl = tsl,
  capitalize_all = TRUE,
  separator = "_",
  suffix = "fagus",
  prefix = "country"
)

#new names
tsl_names_get(
  tsl = tsl_clean
)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{distances}
\alias{distances}
\title{Distance Methods}
\format{
A data frame with 5 columns and 10 rows.
}
\usage{
data(distances)
}
\description{
A data frame with the names, abbreviations, and expressions of the distance metrics implemented in the package.
}
\seealso{
Other distances:
\code{\link[=distance]{distance()}},
\code{\link[=distance_matrix]{distance_matrix()}}
}
\concept{distances}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_prepare_matrix.R
\name{utils_prepare_matrix}
\alias{utils_prepare_matrix}
\title{Convert Matrix to Data Frame}
\usage{
utils_prepare_matrix(x = NULL)
}
\arguments{
\item{x}{(required, matrix). Default: NULL.}
}
\value{
A data frame.
}
\description{
Convert Matrix to Data Frame
}
\seealso{
Other internal:
\code{\link[=utils_boxplot_common]{utils_boxplot_common()}},
\code{\link[=utils_check_args_distantia]{utils_check_args_distantia()}},
\code{\link[=utils_check_args_matrix]{utils_check_args_matrix()}},
\code{\link[=utils_check_args_momentum]{utils_check_args_momentum()}},
\code{\link[=utils_check_args_path]{utils_check_args_path()}},
\code{\link[=utils_check_args_tsl]{utils_check_args_tsl()}},
\code{\link[=utils_check_args_zoo]{utils_check_args_zoo()}},
\code{\link[=utils_check_distance_args]{utils_check_distance_args()}},
\code{\link[=utils_check_list_class]{utils_check_list_class()}},
\code{\link[=utils_clean_names]{utils_clean_names()}},
\code{\link[=utils_digits]{utils_digits()}},
\code{\link[=utils_distantia_df_split]{utils_distantia_df_split()}},
\code{\link[=utils_prepare_df]{utils_prepare_df()}},
\code{\link[=utils_prepare_matrix_list]{utils_prepare_matrix_list()}},
\code{\link[=utils_prepare_time]{utils_prepare_time()}},
\code{\link[=utils_prepare_vector_list]{utils_prepare_vector_list()}},
\code{\link[=utils_prepare_zoo_list]{utils_prepare_zoo_list()}},
\code{\link[=utils_tsl_pairs]{utils_tsl_pairs()}}
}
\concept{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color_discrete.R
\name{color_discrete}
\alias{color_discrete}
\title{Default Discrete Color Palettes}
\usage{
color_discrete(n = NULL, rev = FALSE)
}
\arguments{
\item{n}{(required, integer) number of colors to generate. Default = NULL}

\item{rev}{(optional, logical) If TRUE, the color palette is reversed. Default: FALSE}
}
\value{
color vector
}
\description{
Uses the function \code{\link[grDevices:palette.colors]{grDevices::palette.colors()}} to generate discrete color palettes using the following rules:
\itemize{
\item \code{n <= 9}: "Okabe-Ito".
\item \code{n == 10}: "Tableau 10"
\item \code{n > 10 && n <= 12}: "Paired"
\item \code{n > 12 && n <= 26}: "Alphabet"
\item \code{n > 26 && n <= 36}: "Polychrome 36"
}
}
\examples{

color_discrete(n = 9)
}
\seealso{
Other internal_plotting:
\code{\link[=color_continuous]{color_continuous()}},
\code{\link[=utils_color_breaks]{utils_color_breaks()}},
\code{\link[=utils_line_color]{utils_line_color()}},
\code{\link[=utils_line_guide]{utils_line_guide()}},
\code{\link[=utils_matrix_guide]{utils_matrix_guide()}},
\code{\link[=utils_matrix_plot]{utils_matrix_plot()}}
}
\concept{internal_plotting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_to_df.R
\name{tsl_to_df}
\alias{tsl_to_df}
\title{Transform Time Series List to Data Frame}
\usage{
tsl_to_df(tsl = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}
}
\value{
data frame
}
\description{
Transform Time Series List to Data Frame
}
\examples{

tsl <- tsl_simulate(
  n = 3,
  rows = 10,
  time_range = c(
    "2010-01-01",
    "2020-01-01"
  ),
  irregular = FALSE
)

df <- tsl_to_df(
  tsl = tsl
)

names(df)
nrow(df)
head(df)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{cities_temperature}
\alias{cities_temperature}
\title{Long Term Monthly Temperature in 20 Major Cities}
\format{
A data frame with 3 columns and 52,100 rows.
}
\usage{
data(cities_temperature)
}
\description{
Average temperatures between 1975 and 2010 of 20 major cities of the world. \href{https://www.kaggle.com/datasets/berkeleyearth/climate-change-earth-surface-temperature-data?resource=download&select=GlobalLandTemperaturesByMajorCity.csv}{Source}.

Site coordinates for this dataset are in \link{cities_coordinates}.

The full dataset with 100 cities can be downloaded from \url{https://github.com/BlasBenito/distantia/blob/main/data_full/cities_temperature.rda} (use the "Download raw file" button).
}
\examples{
data("cities_temperature")

#Convert to time series list
cities <- tsl_initialize(
  x = cities_temperature,
  name_column = "name",
  time_column = "time"
)

#time series plot
if(interactive()){

 #Only four cities are shown
 tsl_plot(
  tsl = tsl_subset(
    tsl = tsl,
    names = 1:4
    ),
  guide = FALSE
  )

}
}
\seealso{
Other example_data:
\code{\link{albatross}},
\code{\link{cities_coordinates}},
\code{\link{covid_counties}},
\code{\link{covid_prevalence}},
\code{\link{eemian_coordinates}},
\code{\link{eemian_pollen}},
\code{\link{fagus_coordinates}},
\code{\link{fagus_dynamics}},
\code{\link{honeycomb_climate}},
\code{\link{honeycomb_polygons}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_count_NA.R
\name{tsl_count_NA}
\alias{tsl_count_NA}
\title{Count NA Cases in Time Series Lists}
\usage{
tsl_count_NA(tsl = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}
}
\value{
list
}
\description{
Converts Inf, -Inf, and NaN to NA (via \code{\link[=tsl_Inf_to_NA]{tsl_Inf_to_NA()}} and \code{\link[=tsl_NaN_to_NA]{tsl_NaN_to_NA()}}), and counts the total number of NA cases in each time series.
}
\examples{
#tsl with no NA cases
tsl <- tsl_simulate()

tsl_count_NA(tsl = tsl)

#tsl with NA cases
tsl <- tsl_simulate(
  na_fraction = 0.3
)

tsl_count_NA(tsl = tsl)

#tsl with variety of empty cases
tsl <- tsl_simulate()
tsl[[1]][1, 1] <- Inf
tsl[[1]][2, 1] <- -Inf
tsl[[1]][3, 1] <- NaN
tsl[[1]][4, 1] <- NaN

tsl_count_NA(tsl = tsl)

}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{fagus_dynamics}
\alias{fagus_dynamics}
\title{Time Series Data from Three Fagus sylvatica Stands}
\format{
A data frame with 5 columns and 648 rows.
}
\usage{
data(fagus_dynamics)
}
\description{
A data frame with 648 rows representing enhanced vegetation index, rainfall and temperature in three stands of Fagus sylvatica in Spain, Germany, and Sweden.
}
\details{
Site coordinates for this dataset are in \link{fagus_coordinates}.
}
\examples{
data("fagus_dynamics")

#Convert to time series list
fagus <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
)

#time series plot
if(interactive()){

 tsl_plot(
  tsl = fagus
  )

}

}
\seealso{
Other example_data:
\code{\link{albatross}},
\code{\link{cities_coordinates}},
\code{\link{cities_temperature}},
\code{\link{covid_counties}},
\code{\link{covid_prevalence}},
\code{\link{eemian_coordinates}},
\code{\link{eemian_pollen}},
\code{\link{fagus_coordinates}},
\code{\link{honeycomb_climate}},
\code{\link{honeycomb_polygons}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/color_continuous.R
\name{color_continuous}
\alias{color_continuous}
\title{Default Continuous Color Palette}
\usage{
color_continuous(n = 5, palette = "Zissou 1", rev = FALSE)
}
\arguments{
\item{n}{(required, integer) number of colors to generate. Default = NULL}

\item{palette}{(required, character string) Argument \code{palette} of \code{\link[grDevices:hcl.colors]{grDevices::hcl.colors()}}. Default: "Zissou 1"}

\item{rev}{(optional, logical) If TRUE, the color palette is reversed. Default: FALSE}
}
\value{
color vector
}
\description{
Uses the function \code{\link[grDevices:hcl.colors]{grDevices::hcl.colors()}} to generate a continuous color palette.
}
\examples{

color_continuous(n = 20)

}
\seealso{
Other internal_plotting:
\code{\link[=color_discrete]{color_discrete()}},
\code{\link[=utils_color_breaks]{utils_color_breaks()}},
\code{\link[=utils_line_color]{utils_line_color()}},
\code{\link[=utils_line_guide]{utils_line_guide()}},
\code{\link[=utils_matrix_guide]{utils_matrix_guide()}},
\code{\link[=utils_matrix_plot]{utils_matrix_plot()}}
}
\concept{internal_plotting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{cities_coordinates}
\alias{cities_coordinates}
\title{Coordinates of 100 Major Cities}
\format{
An sf data frame with 5 columns and 100 rows.
}
\usage{
data(cities_coordinates)
}
\description{
City coordinates and several environmental variables for the dataset \code{cities_temperature}.

The full dataset with 100 cities can be downloaded from \url{https://github.com/BlasBenito/distantia/blob/main/data_full/cities_coordinates.rda} (use the "Download raw file" button).
}
\seealso{
Other example_data:
\code{\link{albatross}},
\code{\link{cities_temperature}},
\code{\link{covid_counties}},
\code{\link{covid_prevalence}},
\code{\link{eemian_coordinates}},
\code{\link{eemian_pollen}},
\code{\link{fagus_coordinates}},
\code{\link{fagus_dynamics}},
\code{\link{honeycomb_climate}},
\code{\link{honeycomb_polygons}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_colnames.R
\name{tsl_colnames_get}
\alias{tsl_colnames_get}
\title{Get Column Names from Time Series Lists}
\usage{
tsl_colnames_get(tsl = NULL, names = c("all", "shared", "exclusive"))
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{names}{(optional, character string) Three different sets of column names can be requested:
\itemize{
\item "all" (default): list with the column names in each zoo object in \code{tsl}. Unnamed columns are tagged with the string "unnamed".
\item "shared": character vector with the shared column names in at least two zoo objects in \code{tsl}.
\item "exclusive": list with names of exclusive columns (if any) in each zoo object in \code{tsl}.
}}
}
\value{
list
}
\description{
Get Column Names from Time Series Lists
}
\examples{
#generate example data
tsl <- tsl_simulate()

#list all column names
tsl_colnames_get(
  tsl = tsl,
  names = "all"
)

#change one column name
names(tsl[[1]])[1] <- "new_column"

#all names again
tsl_colnames_get(
  tsl = tsl,
  names = "all"
)

#shared column names
tsl_colnames_get(
  tsl = tsl,
  names = "shared"
)

#exclusive column names
tsl_colnames_get(
  tsl = tsl,
  names = "exclusive"
)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_dtw.R
\name{distantia_dtw}
\alias{distantia_dtw}
\title{Dynamic Time Warping Dissimilarity Analysis of Time Series Lists}
\usage{
distantia_dtw(tsl = NULL, distance = "euclidean", bandwidth = 1)
}
\arguments{
\item{tsl}{(required, time series list) list of zoo time series. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}

\item{bandwidth}{(optional, numeric) Proportion of space at each side of the cost matrix diagonal (aka \emph{Sakoe-Chiba band}) defining a valid region for dynamic time warping, used to control the flexibility of the warping path. This method prevents degenerate alignments due to differences in magnitude between time series when the data is not properly scaled. If \code{1} (default), DTW is unconstrained. If \code{0}, DTW is fully constrained and the warping path follows the matrix diagonal. Recommended values may vary depending on the nature of the data. Ignored if \code{lock_step = TRUE}. Default: 1.}
}
\value{
data frame with columns:
\itemize{
\item \code{x}: time series name.
\item \code{y}: time series name.
\item \code{distance}: name of the distance metric.
\item \code{psi}: psi dissimilarity of the sequences \code{x} and \code{y}.
}
}
\description{
A minimalistic but slightly faster version of \code{\link[=distantia]{distantia()}} that computes dynamic time warping dissimilarity scores using diagonal least cost paths.
}
\examples{

#load fagus_dynamics as tsl
#global centering and scaling
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

if(interactive()){
  tsl_plot(
    tsl = tsl,
    guide_columns = 3
    )
}

#dynamic time warping dissimilarity analysis
df_dtw <- distantia_dtw(
  tsl = tsl,
  distance = "euclidean"
)

df_dtw[, c("x", "y", "psi")]

#visualize dynamic time warping
if(interactive()){

  distantia_dtw_plot(
    tsl = tsl[c("Spain", "Sweden")],
    distance = "euclidean"
  )

}


}
\seealso{
Other distantia:
\code{\link[=distantia]{distantia()}},
\code{\link[=distantia_dtw_plot]{distantia_dtw_plot()}},
\code{\link[=distantia_ls]{distantia_ls()}}
}
\concept{distantia}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{importance_dtw_legacy_cpp}
\alias{importance_dtw_legacy_cpp}
\title{(C++) Contribution of Individual Variables to the Dissimilarity Between Two Time Series (Legacy Version)}
\usage{
importance_dtw_legacy_cpp(
  y,
  x,
  distance = "euclidean",
  diagonal = FALSE,
  weighted = TRUE,
  ignore_blocks = FALSE,
  bandwidth = 1
)
}
\arguments{
\item{y}{(required, numeric matrix) multivariate time series
with the same number of columns as 'x'.}

\item{x}{(required, numeric matrix) multivariate time series.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}

\item{diagonal}{(optional, logical). If TRUE, diagonals are included in the
computation of the cost matrix. Default: TRUE.}

\item{weighted}{(optional, logical). Only relevant when diagonal is TRUE. When TRUE,
diagonal cost is weighted by y factor of 1.414214 (square root of 2). Default: TRUE.}

\item{ignore_blocks}{(optional, logical). If TRUE, blocks of consecutive path
coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.}

\item{bandwidth}{(required, numeric) Size of the Sakoe-Chiba band at
both sides of the diagonal used to constrain the least cost path. Expressed
as a fraction of the number of matrix rows and columns. Unrestricted by default.
Default: 1}
}
\value{
data frame
}
\description{
Computes the contribution of individual variables to the
similarity/dissimilarity between two irregular multivariate time series.
In opposition to the robust version, least-cost paths for each combination
of variables are computed independently, which makes the results of individual
variables harder to compare. This function should only be used when the objective is
replicating importance scores generated with previous versions of the package \code{distantia}.
This function generates a data frame with the following columns:
\itemize{
\item variable: name of the individual variable for which the importance
is being computed, from the column names of the arguments \code{x} and \code{y}.
\item psi: global dissimilarity score \code{psi} of the two time series.
\item psi_only_with: dissimilarity between \code{x} and \code{y} computed from the given variable alone.
\item psi_without: dissimilarity between \code{x} and \code{y} computed from all other variables.
\item psi_difference: difference between \code{psi_only_with} and \code{psi_without}.
\item importance: contribution of the variable to the similarity/dissimilarity
between \code{x} and \code{y}, computed as \code{((psi_all - psi_without) * 100) / psi_all}.
Positive scores represent contribution to dissimilarity,
while negative scores represent contribution to similarity.
}
}
\examples{
#simulate two regular time series
x <- zoo_simulate(
  seed = 1,
  rows = 100
  )

y <- zoo_simulate(
  seed = 2,
  rows = 150
  )

#different number of rows
#this is not a requirement though!
nrow(x) == nrow(y)

#compute importance
df <- importance_dtw_legacy_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)

df
}
\seealso{
Other Rcpp_importance:
\code{\link[=importance_dtw_cpp]{importance_dtw_cpp()}},
\code{\link[=importance_ls_cpp]{importance_ls_cpp()}}
}
\concept{Rcpp_importance}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{covid_counties}
\alias{covid_counties}
\title{County Coordinates of the Covid Prevalence Dataset}
\format{
An sf data frame with county polygons and census data.
}
\usage{
data(covid_counties)
}
\description{
County Coordinates of the Covid Prevalence Dataset
}
\seealso{
Other example_data:
\code{\link{albatross}},
\code{\link{cities_coordinates}},
\code{\link{cities_temperature}},
\code{\link{covid_prevalence}},
\code{\link{eemian_coordinates}},
\code{\link{eemian_pollen}},
\code{\link{fagus_coordinates}},
\code{\link{fagus_dynamics}},
\code{\link{honeycomb_climate}},
\code{\link{honeycomb_polygons}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_initialize.R
\name{tsl_initialize}
\alias{tsl_initialize}
\alias{tsl_init}
\title{Transform Raw Time Series Data to Time Series List}
\usage{
tsl_initialize(
  x = NULL,
  name_column = NULL,
  time_column = NULL,
  lock_step = FALSE
)

tsl_init(x = NULL, name_column = NULL, time_column = NULL, lock_step = FALSE)
}
\arguments{
\item{x}{(required, list or data frame) Matrix or data frame in long format, list of vectors, list of matrices, or list of zoo objects. Default: NULL.}

\item{name_column}{(optional, column name) Column naming individual time series. Numeric names are converted to character with the prefix "X". Default: NULL}

\item{time_column}{(optional if \code{lock_step = FALSE}, and required otherwise, character string) Name of the column representing time, if any. Default: NULL.}

\item{lock_step}{(optional, logical) If TRUE, all input sequences are subsetted to their common times according to the values in \code{time_column}.}
}
\value{
list of zoo objects (a time series list)
}
\description{
Most functions in this package take a \strong{time series list} (or \strong{tsl} for short) as main input. A \code{tsl} is a list of zoo time series objects (see \code{\link[zoo:zoo]{zoo::zoo()}}). There is not a formal class for \code{tsl} objects, but there are requirements these objects must follow to ensure the stability of the package functionalities (see \code{\link[=tsl_diagnose]{tsl_diagnose()}}). These requirements are:
\itemize{
\item There are no NA, Inf, -Inf, or NaN cases in the zoo objects (see \code{\link[=tsl_count_NA]{tsl_count_NA()}} and \code{\link[=tsl_handle_NA]{tsl_handle_NA()}}).
\item All zoo objects must have at least one common column name to allow time series comparison (see \code{\link[=tsl_colnames_get]{tsl_colnames_get()}}).
\item All zoo objects have a character attribute "name" identifying the object. This attribute is not part of the zoo class, but the package ensures that this attribute is not lost during data manipulations.
\item Each element of the time series list is named after the zoo object it contains (see \code{\link[=tsl_names_get]{tsl_names_get()}}, \code{\link[=tsl_names_set]{tsl_names_set()}} and \code{\link[=tsl_names_clean]{tsl_names_clean()}}).
\item The time series list contains two zoo objects or more.
}

The function \code{\link[=tsl_initialize]{tsl_initialize()}} (and its alias \code{\link[=tsl_init]{tsl_init()}}) is designed to convert the following data structures to a time series list:

\itemize{
\item Long data frame: with an ID column to separate time series, and a time column that can be of the classes "Date", "POSIXct", "integer", or "numeric". The resulting zoo objects and list elements are named after the values in the ID column.
\item Wide data frame: each column is a time series representing the same variable observed at the same time in different places. Each column is converted to a separate zoo object and renamed.
\item List of vectors: an object like \code{list(a = runif(10), b = runif(10))} is converted to a time series list with as many zoo objects as vectors are defined in the original list.
\item List of matrices: a list containing matrices, such as \code{list(a = matrix(runif(30), 10, 3), b = matrix(runif(36), 12, 3))}.
\item List of zoo objects: a list with zoo objects, such as \code{list(a = zoo_simulate(), b = zoo_simulate())}.
}
}
\examples{
#long data frame
#---------------------
data("fagus_dynamics")

#name_column is name
#time column is time
str(fagus_dynamics)

#to tsl
#each group in name_column is a different time series
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
)

#check validity (no messages or errors if valid)
tsl_diagnose(tsl)

#class of contained objects
lapply(X = tsl, FUN = class)

#get list and zoo names (between double quotes)
tsl_names_get(
  tsl = tsl,
  zoo = TRUE
  )

#plot tsl
if(interactive()){
  tsl_plot(tsl)
}

#list of zoo objects
#--------------------
x <- zoo_simulate()
y <- zoo_simulate()

tsl <- tsl_initialize(
  x = list(
    x = x,
    y = y
  )
)

#plot
if(interactive()){
  tsl_plot(tsl)
}


#wide data frame
#--------------------
#wide data frame
#each column is same variable in different places
df <- stats::reshape(
  data = fagus_dynamics[, c(
    "name",
    "time",
    "evi"
  )],
  timevar = "name",
  idvar = "time",
  direction = "wide",
  sep = "_"
)

str(df)

#to tsl
#key assumptions:
#all columns but "time" represent
#the same variable in different places
#all time series are of the same length
tsl <- tsl_initialize(
  x = df,
  time_column = "time"
  )

#colnames are forced to be the same...
tsl_colnames_get(tsl)

#...but can be changed
tsl <- tsl_colnames_set(
  tsl = tsl,
  names = "evi"
)
tsl_colnames_get(tsl)

#plot
if(interactive()){
  tsl_plot(tsl)
}


#list of vectors
#---------------------
#create list of vectors
vector_list <- list(
  a = cumsum(stats::rnorm(n = 50)),
  b = cumsum(stats::rnorm(n = 70)),
  c = cumsum(stats::rnorm(n = 20))
)

#to tsl
#key assumptions:
#all vectors represent the same variable
#in different places
#time series can be of different lengths
#no time column, integer indices are used as time
tsl <- tsl_initialize(
  x = vector_list
)

#plot
if(interactive()){
  tsl_plot(tsl)
}

#list of matrices
#-------------------------
#create list of matrices
matrix_list <- list(
  a = matrix(runif(30), nrow = 10, ncol = 3),
  b = matrix(runif(80), nrow = 20, ncol = 4)
)

#to tsl
#key assumptions:
#each matrix represents a multivariate time series
#in a different place
#all multivariate time series have the same columns
#no time column, integer indices are used as time
tsl <- tsl_initialize(
  x = matrix_list
)

#check column names
tsl_colnames_get(tsl = tsl)

#remove exclusive column
tsl <- tsl_subset(
  tsl = tsl,
  shared_cols = TRUE
  )
tsl_colnames_get(tsl = tsl)

#plot
if(interactive()){
  tsl_plot(tsl)
}

#list of zoo objects
#-------------------------
zoo_list <- list(
  a = zoo_simulate(),
  b = zoo_simulate()
)

#looks like a time series list! But...
tsl_diagnose(tsl = zoo_list)

#let's set the names
zoo_list <- tsl_names_set(tsl = zoo_list)

#check again: it's now a valid time series list
tsl_diagnose(tsl = zoo_list)

#to do all this in one go:
tsl <- tsl_initialize(
  x = list(
    a = zoo_simulate(),
    b = zoo_simulate()
  )
)

#plot
if(interactive()){
  tsl_plot(tsl)
}

}
\concept{tsl_initialize}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{auto_distance_cpp}
\alias{auto_distance_cpp}
\title{(C++) Sum Distances Between Consecutive Samples in a Time Series}
\usage{
auto_distance_cpp(x, distance = "euclidean")
}
\arguments{
\item{x}{(required, numeric matrix) univariate or multivariate time series.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean"}
}
\value{
numeric
}
\description{
Computes the cumulative sum of distances between consecutive
samples in a univariate or multivariate time series.
NA values should be removed before using this function.
}
\examples{
#simulate a time series
x <- zoo_simulate()

#compute auto distance
auto_distance_cpp(
  x = x,
  distance = "euclidean"
  )
}
\seealso{
Other Rcpp_auto_sum:
\code{\link[=auto_sum_cpp]{auto_sum_cpp()}},
\code{\link[=auto_sum_full_cpp]{auto_sum_full_cpp()}},
\code{\link[=auto_sum_path_cpp]{auto_sum_path_cpp()}},
\code{\link[=subset_matrix_by_rows_cpp]{subset_matrix_by_rows_cpp()}}
}
\concept{Rcpp_auto_sum}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psi_cost_path_sum.R
\name{psi_cost_path_sum}
\alias{psi_cost_path_sum}
\title{Sum of Distances in Least Cost Path}
\usage{
psi_cost_path_sum(path = NULL)
}
\arguments{
\item{path}{(required, data frame) least cost path produced by \code{\link[=psi_cost_path]{psi_cost_path()}}. Default: NULL}
}
\value{
numeric value
}
\description{
Demonstration function to sum the distances of a least cost path.
}
\examples{
#distance metric
d <- "euclidean"

#simulate two irregular time series
x <- zoo_simulate(
  name = "x",
  rows = 100,
  seasons = 2,
  seed = 1
)

y <- zoo_simulate(
  name = "y",
  rows = 80,
  seasons = 2,
  seed = 2
)

if(interactive()){
  zoo_plot(x = x)
  zoo_plot(x = y)
}

#distance matrix
dist_matrix <- psi_distance_matrix(
  x = x,
  y = y,
  distance = d
)

#orthogonal least cost matrix
cost_matrix <- psi_cost_matrix(
  dist_matrix = dist_matrix
)

#orthogonal least cost path
cost_path <- psi_cost_path(
  dist_matrix = dist_matrix,
  cost_matrix = cost_matrix
)

#sum of distances in cost path
psi_cost_path_sum(
  path = cost_path
)
}
\seealso{
Other psi_demo:
\code{\link[=psi_auto_distance]{psi_auto_distance()}},
\code{\link[=psi_auto_sum]{psi_auto_sum()}},
\code{\link[=psi_cost_matrix]{psi_cost_matrix()}},
\code{\link[=psi_cost_path]{psi_cost_path()}},
\code{\link[=psi_distance_lock_step]{psi_distance_lock_step()}},
\code{\link[=psi_distance_matrix]{psi_distance_matrix()}},
\code{\link[=psi_equation]{psi_equation()}}
}
\concept{psi_demo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_distantia_df_split.R
\name{utils_distantia_df_split}
\alias{utils_distantia_df_split}
\title{Split Dissimilarity Analysis Data Frames by Combinations of Arguments}
\usage{
utils_distantia_df_split(df = NULL)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=distantia]{distantia()}} or \code{\link[=distantia_aggregate]{distantia_aggregate()}}. Default: NULL.}
}
\value{
A list.
}
\description{
Internal function to split a distantia data frame by groups of the arguments 'distance', 'diagonal', and 'lock_step'.
}
\examples{
#Three time series
#Climate and NDVI in Fagus sylvatica stands in Spain, Germany, and Sweden
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
)

#Dissimilarity analysis with four combinations of parameters
df <- distantia(
  tsl = tsl,
  distance = c(
    "euclidean",
    "manhattan"
    ),
  lock_step = c(
    TRUE,
    FALSE
  )
)

#split by combinations of parameters
df_split <- utils_distantia_df_split(
  df = df
)

#print output
df_split

#class and length of the output
class(df_split)
length(df_split)
}
\seealso{
Other internal:
\code{\link[=utils_boxplot_common]{utils_boxplot_common()}},
\code{\link[=utils_check_args_distantia]{utils_check_args_distantia()}},
\code{\link[=utils_check_args_matrix]{utils_check_args_matrix()}},
\code{\link[=utils_check_args_momentum]{utils_check_args_momentum()}},
\code{\link[=utils_check_args_path]{utils_check_args_path()}},
\code{\link[=utils_check_args_tsl]{utils_check_args_tsl()}},
\code{\link[=utils_check_args_zoo]{utils_check_args_zoo()}},
\code{\link[=utils_check_distance_args]{utils_check_distance_args()}},
\code{\link[=utils_check_list_class]{utils_check_list_class()}},
\code{\link[=utils_clean_names]{utils_clean_names()}},
\code{\link[=utils_digits]{utils_digits()}},
\code{\link[=utils_prepare_df]{utils_prepare_df()}},
\code{\link[=utils_prepare_matrix]{utils_prepare_matrix()}},
\code{\link[=utils_prepare_matrix_list]{utils_prepare_matrix_list()}},
\code{\link[=utils_prepare_time]{utils_prepare_time()}},
\code{\link[=utils_prepare_vector_list]{utils_prepare_vector_list()}},
\code{\link[=utils_prepare_zoo_list]{utils_prepare_zoo_list()}},
\code{\link[=utils_tsl_pairs]{utils_tsl_pairs()}}
}
\concept{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_check_args.R
\name{utils_check_args_tsl}
\alias{utils_check_args_tsl}
\title{Structural Check for Time Series Lists}
\usage{
utils_check_args_tsl(tsl = NULL, min_length = 2)
}
\arguments{
\item{tsl}{(required, list) list of zoo objects. Default: NULL.}

\item{min_length}{(required, positive integer) minimum number of zoo objects in \code{tsl}. Default: 2.}
}
\value{
Error messages, if any.
}
\description{
Internal function to check that a time series list is a list of zoo objects and has a minimum number of objects. For a more comprehensive test, use \code{\link[=tsl_diagnose]{tsl_diagnose()}}.
}
\seealso{
Other internal:
\code{\link[=utils_boxplot_common]{utils_boxplot_common()}},
\code{\link[=utils_check_args_distantia]{utils_check_args_distantia()}},
\code{\link[=utils_check_args_matrix]{utils_check_args_matrix()}},
\code{\link[=utils_check_args_momentum]{utils_check_args_momentum()}},
\code{\link[=utils_check_args_path]{utils_check_args_path()}},
\code{\link[=utils_check_args_zoo]{utils_check_args_zoo()}},
\code{\link[=utils_check_distance_args]{utils_check_distance_args()}},
\code{\link[=utils_check_list_class]{utils_check_list_class()}},
\code{\link[=utils_clean_names]{utils_clean_names()}},
\code{\link[=utils_digits]{utils_digits()}},
\code{\link[=utils_distantia_df_split]{utils_distantia_df_split()}},
\code{\link[=utils_prepare_df]{utils_prepare_df()}},
\code{\link[=utils_prepare_matrix]{utils_prepare_matrix()}},
\code{\link[=utils_prepare_matrix_list]{utils_prepare_matrix_list()}},
\code{\link[=utils_prepare_time]{utils_prepare_time()}},
\code{\link[=utils_prepare_vector_list]{utils_prepare_vector_list()}},
\code{\link[=utils_prepare_zoo_list]{utils_prepare_zoo_list()}},
\code{\link[=utils_tsl_pairs]{utils_tsl_pairs()}}
}
\concept{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{covid_prevalence}
\alias{covid_prevalence}
\title{Time Series of Covid Prevalence in California Counties}
\format{
A data frame with 3 columns and 51,048 rows.
}
\usage{
data(covid_prevalence)
}
\description{
Dataset with Covid19 maximum weekly prevalence in California counties between 2020 and 2024, from healthdata.gov.
}
\details{
County polygons and additional data for this dataset are in \link{covid_counties}.

The full dataset at daily resolution can be downloaded from \url{https://github.com/BlasBenito/distantia/blob/main/data_full/covid_prevalence.rda} (use the "Download raw file" button).
}
\examples{
#Convert to time series list
tsl <- tsl_initialize(
  x = covid_prevalence,
  name_column = "name",
  time_column = "time"
)

#time series plot
if(interactive()){

 #Subset to avoid margin errors
 tsl_plot(
  tsl = tsl_subset(
    tsl = tsl,
    names = 1:4
    ),
  guide = FALSE
  )

}
}
\seealso{
Other example_data:
\code{\link{albatross}},
\code{\link{cities_coordinates}},
\code{\link{cities_temperature}},
\code{\link{covid_counties}},
\code{\link{eemian_coordinates}},
\code{\link{eemian_pollen}},
\code{\link{fagus_coordinates}},
\code{\link{fagus_dynamics}},
\code{\link{honeycomb_climate}},
\code{\link{honeycomb_polygons}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_colnames.R
\name{tsl_colnames_set}
\alias{tsl_colnames_set}
\title{Set Column Names in a Time Series List}
\usage{
tsl_colnames_set(tsl = NULL, names = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{names}{(required, list or character vector):
\itemize{
\item list: with same names as 'tsl', containing a vector of new column names for each time series in 'tsl'.
\item character vector: vector of new column names assigned by position.
}}
}
\value{
time series list
}
\description{
Set Column Names in a Time Series List
}
\examples{
tsl <- tsl_simulate(
  cols = 3
  )

tsl_colnames_get(
  tsl = tsl
  )

#using a vector
#extra names are ignored
tsl <- tsl_colnames_set(
  tsl = tsl,
  names = c("x", "y", "z", "zz")
)

tsl_colnames_get(
  tsl = tsl
)

#using a list
#extra names are ignored too
tsl <- tsl_colnames_set(
  tsl = tsl,
  names = list(
    A = c("A", "B", "C"),
    B = c("X", "Y", "Z", "ZZ")
  )
)

tsl_colnames_get(
  tsl = tsl
)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia-package.R
\docType{package}
\name{distantia-package}
\alias{distantia-package}
\title{distantia: A Toolset for Time Series Dissimilarity Analysis}
\description{
Fast C++ implementation of Dynamic Time Warping for time series dissimilarity analysis, with applications in environmental monitoring, sensor data analysis, climate science, signal processing, pattern recognition, and financial data analysis. Built upon the ideas presented in Benito and Birks (2020) \url{doi:10.1111/ecog.04895}, the package provides tools for analyzing time series of varying lengths and structures, including irregular multivariate time series. Key features include individual variable contribution analysis, restricted permutation tests for statistical significance, and imputation of missing data via generalized additive models. Additionally, the package provides comprehensive tools for preparing and managing time series data.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://blasbenito.github.io/distantia/}
  \item Report bugs at \url{https://github.com/BlasBenito/distantia/issues}
}

}
\author{
Blas Benito \href{mailto:blasbenito@gmail.com}{blasbenito@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/momentum_spatial.R
\name{momentum_spatial}
\alias{momentum_spatial}
\title{Spatial Representation of \code{momentum()} Data Frames}
\usage{
momentum_spatial(df = NULL, sf = NULL, network = TRUE)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=momentum]{momentum()}}, \code{\link[=momentum_ls]{momentum_ls()}}, or \code{\link[=momentum_dtw]{momentum_dtw()}}. Default: NULL}

\item{sf}{(required, sf data frame) Points or polygons representing the location of the time series in argument 'df'. It must have a column with all time series names in \code{df$x} and \code{df$y}. Default: NULL}

\item{network}{(optional, logical) If TRUE, the resulting sf data frame is of type LINESTRING and represents network edges. Default: TRUE}
}
\value{
sf data frame (LINESTRING geometry)
}
\description{
Given an sf data frame with geometry types POLYGON, MULTIPOLYGON, or POINT representing time series locations, this function transforms the output of \code{\link[=momentum]{momentum()}}, \code{\link[=momentum_ls]{momentum_ls()}}, or \code{\link[=momentum_dtw]{momentum_dtw()}} to an sf data frame.

If \code{network = TRUE}, the sf data frame is of type LINESTRING, with edges connecting time series locations. This output is helpful for building many-to-many importance maps (see examples).

If \code{network = FALSE}, the sf data frame contains the geometry from the input \code{sf} argument. This output helps build one-to-many importance maps.
}
\examples{
tsl <- distantia::tsl_initialize(
  x = distantia::eemian_pollen,
  name_column = "name",
  time_column = "time"
) |>
#reduce size to speed-up example runtime
distantia::tsl_subset(
  names = 1:3
  )

df_momentum <- distantia::momentum(
  tsl = tsl
)

#network many to many
sf_momentum <- distantia::momentum_spatial(
  df = df_momentum,
  sf = distantia::eemian_coordinates,
  network = TRUE
)

#network map
# mapview::mapview(
#   sf_momentum,
#   layer.name = "Importance - Abies",
#   label = "edge_name",
#   zcol = "importance__Abies",
#   lwd = 3
# ) |>
#   suppressWarnings()

}
\seealso{
Other momentum_support:
\code{\link[=momentum_aggregate]{momentum_aggregate()}},
\code{\link[=momentum_boxplot]{momentum_boxplot()}},
\code{\link[=momentum_model_frame]{momentum_model_frame()}},
\code{\link[=momentum_stats]{momentum_stats()}},
\code{\link[=momentum_to_wide]{momentum_to_wide()}}
}
\concept{momentum_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_check_args.R
\name{utils_check_args_distantia}
\alias{utils_check_args_distantia}
\title{Check Input Arguments of \code{distantia()}}
\usage{
utils_check_args_distantia(
  tsl = NULL,
  distance = NULL,
  diagonal = NULL,
  bandwidth = NULL,
  lock_step = NULL,
  repetitions = NULL,
  permutation = NULL,
  block_size = NULL,
  seed = NULL
)
}
\arguments{
\item{tsl}{(required, time series list) list of zoo time series. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}

\item{diagonal}{(optional, logical vector). If TRUE, diagonals are included in the dynamic time warping computation. Default: TRUE}

\item{bandwidth}{(optional, numeric) Proportion of space at each side of the cost matrix diagonal (aka \emph{Sakoe-Chiba band}) defining a valid region for dynamic time warping, used to control the flexibility of the warping path. This method prevents degenerate alignments due to differences in magnitude between time series when the data is not properly scaled. If \code{1} (default), DTW is unconstrained. If \code{0}, DTW is fully constrained and the warping path follows the matrix diagonal. Recommended values may vary depending on the nature of the data. Ignored if \code{lock_step = TRUE}. Default: 1.}

\item{lock_step}{(optional, logical vector) If TRUE, time series captured at the same times are compared sample wise (with no dynamic time warping). Requires time series in argument \code{tsl} to be fully aligned, or it will return an error. Default: FALSE.}

\item{repetitions}{(optional, integer vector) number of permutations to compute the p-value. If 0, p-values are not computed. Otherwise, the minimum is 2. The resolution of the p-values and the overall computation time depends on the number of permutations. Default: 0}

\item{permutation}{(optional, character vector) permutation method, only relevant when \code{repetitions} is higher than zero. Valid values are: "restricted_by_row", "restricted", "free_by_row", and "free". Default: "restricted_by_row".}

\item{block_size}{(optional, integer) Size of the row blocks for the restricted permutation test. Only relevant when permutation methods are "restricted" or "restricted_by_row" and \code{repetitions} is higher than zero. A block of size \code{n} indicates that a row can only be permuted within a block of \code{n} adjacent rows. If NULL, defaults to the rounded one tenth of the shortest time series in \code{tsl}. Default: NULL.}

\item{seed}{(optional, integer) initial random seed to use for replicability when computing p-values. Default: 1}
}
\value{
list.
}
\description{
Check Input Arguments of \code{distantia()}
}
\seealso{
Other internal:
\code{\link[=utils_boxplot_common]{utils_boxplot_common()}},
\code{\link[=utils_check_args_matrix]{utils_check_args_matrix()}},
\code{\link[=utils_check_args_momentum]{utils_check_args_momentum()}},
\code{\link[=utils_check_args_path]{utils_check_args_path()}},
\code{\link[=utils_check_args_tsl]{utils_check_args_tsl()}},
\code{\link[=utils_check_args_zoo]{utils_check_args_zoo()}},
\code{\link[=utils_check_distance_args]{utils_check_distance_args()}},
\code{\link[=utils_check_list_class]{utils_check_list_class()}},
\code{\link[=utils_clean_names]{utils_clean_names()}},
\code{\link[=utils_digits]{utils_digits()}},
\code{\link[=utils_distantia_df_split]{utils_distantia_df_split()}},
\code{\link[=utils_prepare_df]{utils_prepare_df()}},
\code{\link[=utils_prepare_matrix]{utils_prepare_matrix()}},
\code{\link[=utils_prepare_matrix_list]{utils_prepare_matrix_list()}},
\code{\link[=utils_prepare_time]{utils_prepare_time()}},
\code{\link[=utils_prepare_vector_list]{utils_prepare_vector_list()}},
\code{\link[=utils_prepare_zoo_list]{utils_prepare_zoo_list()}},
\code{\link[=utils_tsl_pairs]{utils_tsl_pairs()}}
}
\concept{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cost_path_cpp}
\alias{cost_path_cpp}
\title{Least Cost Path}
\usage{
cost_path_cpp(
  x,
  y,
  distance = "euclidean",
  diagonal = TRUE,
  weighted = TRUE,
  ignore_blocks = FALSE,
  bandwidth = 1
)
}
\arguments{
\item{x}{(required, numeric matrix) multivariate time series.}

\item{y}{(required, numeric matrix) multivariate time series
with the same number of columns as 'x'.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}

\item{diagonal}{(optional, logical). If TRUE, diagonals are included in the
computation of the cost matrix. Default: TRUE.}

\item{weighted}{(optional, logical). Only relevant when diagonal is TRUE. When TRUE,
diagonal cost is weighted by y factor of 1.414214 (square root of 2). Default: TRUE.}

\item{ignore_blocks}{(optional, logical). If TRUE, blocks of consecutive path
coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.}

\item{bandwidth}{(required, numeric) Size of the Sakoe-Chiba band at
both sides of the diagonal used to constrain the least cost path. Expressed
as a fraction of the number of matrix rows and columns. Unrestricted by default.
Default: 1}
}
\value{
data frame
}
\description{
Least cost path between two time series \code{x} and \code{y}.
NA values must be removed from \code{x} and \code{y} before using this function.
If the selected distance function is "chi" or "cosine", pairs of zeros should
be either removed or replaced with pseudo-zeros (i.e. 0.00001).
}
\seealso{
Other Rcpp_cost_path:
\code{\link[=cost_path_diagonal_bandwidth_cpp]{cost_path_diagonal_bandwidth_cpp()}},
\code{\link[=cost_path_diagonal_cpp]{cost_path_diagonal_cpp()}},
\code{\link[=cost_path_orthogonal_bandwidth_cpp]{cost_path_orthogonal_bandwidth_cpp()}},
\code{\link[=cost_path_orthogonal_cpp]{cost_path_orthogonal_cpp()}},
\code{\link[=cost_path_slotting_cpp]{cost_path_slotting_cpp()}},
\code{\link[=cost_path_sum_cpp]{cost_path_sum_cpp()}},
\code{\link[=cost_path_trim_cpp]{cost_path_trim_cpp()}}
}
\concept{Rcpp_cost_path}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_line_color.R
\name{utils_line_color}
\alias{utils_line_color}
\title{Handles Line Colors for Sequence Plots}
\usage{
utils_line_color(x = NULL, line_color = NULL)
}
\arguments{
\item{x}{(required, sequence) zoo object or time series list. Default: NULL}

\item{line_color}{(optional, character vector) vector of colors for the time series columns. Selected palette depends on the number of columns to plot. Default: NULL}
}
\value{
color vector
}
\description{
Internal function that can be used to better understand how line colors are handled within other plotting functions.
}
\seealso{
Other internal_plotting:
\code{\link[=color_continuous]{color_continuous()}},
\code{\link[=color_discrete]{color_discrete()}},
\code{\link[=utils_color_breaks]{utils_color_breaks()}},
\code{\link[=utils_line_guide]{utils_line_guide()}},
\code{\link[=utils_matrix_guide]{utils_matrix_guide()}},
\code{\link[=utils_matrix_plot]{utils_matrix_plot()}}
}
\concept{internal_plotting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_is_time.R
\name{utils_is_time}
\alias{utils_is_time}
\title{Test if Vector is a Valid Time Class}
\usage{
utils_is_time(x = NULL)
}
\arguments{
\item{x}{(required, vector) Vector to test. If the class of the vector elements is 'numeric', 'POSIXct', or 'Date', the function returns TRUE. Default: NULL.}
}
\value{
logical
}
\description{
Test if Vector is a Valid Time Class
}
\examples{

utils_is_time(
  x = c("2024-01-01", "2024-02-01")
)

utils_is_time(
  x = utils_as_time(
    x = c("2024-01-01", "2024-02-01")
    )
)
}
\seealso{
Other internal_time_handling:
\code{\link[=utils_as_time]{utils_as_time()}},
\code{\link[=utils_coerce_time_class]{utils_coerce_time_class()}},
\code{\link[=utils_new_time]{utils_new_time()}},
\code{\link[=utils_time_keywords]{utils_time_keywords()}},
\code{\link[=utils_time_keywords_dictionary]{utils_time_keywords_dictionary()}},
\code{\link[=utils_time_keywords_translate]{utils_time_keywords_translate()}},
\code{\link[=utils_time_units]{utils_time_units()}}
}
\concept{internal_time_handling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{psi_dtw_cpp}
\alias{psi_dtw_cpp}
\title{(C++) Psi Dissimilarity Score of Two Time-Series}
\usage{
psi_dtw_cpp(
  x,
  y,
  distance = "euclidean",
  diagonal = TRUE,
  weighted = TRUE,
  ignore_blocks = FALSE,
  bandwidth = 1
)
}
\arguments{
\item{x}{(required, numeric matrix) of same number of columns as 'y'.}

\item{y}{(required, numeric matrix) time series.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}

\item{diagonal}{(optional, logical). If TRUE, diagonals are included in the
computation of the cost matrix. Default: TRUE.}

\item{weighted}{(optional, logical). Only relevant when diagonal is TRUE. When TRUE,
diagonal cost is weighted by y factor of 1.414214 (square root of 2). Default: TRUE.}

\item{ignore_blocks}{(optional, logical). If TRUE, blocks of consecutive path
coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.}

\item{bandwidth}{(required, numeric) Size of the Sakoe-Chiba band at
both sides of the diagonal used to constrain the least cost path. Expressed
as a fraction of the number of matrix rows and columns. Unrestricted by default.}
}
\value{
numeric
}
\description{
Computes the psi score of two time series \code{y} and \code{x}
with the same number of columns.
NA values should be removed before using this function.
If the selected distance function is "chi" or "cosine", pairs of zeros should
be either removed or replaced with pseudo-zeros (i.e. 0.00001).
}
\seealso{
Other Rcpp_dissimilarity_analysis:
\code{\link[=psi_equation_cpp]{psi_equation_cpp()}},
\code{\link[=psi_ls_cpp]{psi_ls_cpp()}},
\code{\link[=psi_null_dtw_cpp]{psi_null_dtw_cpp()}},
\code{\link[=psi_null_ls_cpp]{psi_null_ls_cpp()}}
}
\concept{Rcpp_dissimilarity_analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_color_breaks.R
\name{utils_color_breaks}
\alias{utils_color_breaks}
\title{Auto Breaks for Matrix Plotting Functions}
\usage{
utils_color_breaks(m = NULL, n = 100)
}
\arguments{
\item{m}{(required, numeric matrix) distance or cost matrix generated by \code{\link[=psi_distance_matrix]{psi_distance_matrix()}} or \code{\link[=psi_cost_matrix]{psi_cost_matrix()}}, but any numeric matrix will work. Default: NULL}

\item{n}{(required, integer) number of colors to compute the breaks for. Default: 100}
}
\value{
numeric vector
}
\description{
Auto Breaks for Matrix Plotting Functions
}
\seealso{
Other internal_plotting:
\code{\link[=color_continuous]{color_continuous()}},
\code{\link[=color_discrete]{color_discrete()}},
\code{\link[=utils_line_color]{utils_line_color()}},
\code{\link[=utils_line_guide]{utils_line_guide()}},
\code{\link[=utils_matrix_guide]{utils_matrix_guide()}},
\code{\link[=utils_matrix_plot]{utils_matrix_plot()}}
}
\concept{internal_plotting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{importance_dtw_cpp}
\alias{importance_dtw_cpp}
\title{(C++) Contribution of Individual Variables to the Dissimilarity Between Two Time Series (Robust Version)}
\usage{
importance_dtw_cpp(
  x,
  y,
  distance = "euclidean",
  diagonal = TRUE,
  weighted = TRUE,
  ignore_blocks = FALSE,
  bandwidth = 1
)
}
\arguments{
\item{x}{(required, numeric matrix) multivariate time series.}

\item{y}{(required, numeric matrix) multivariate time series
with the same number of columns as 'x'.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}

\item{diagonal}{(optional, logical). If TRUE, diagonals are included in the
computation of the cost matrix. Default: TRUE.}

\item{weighted}{(optional, logical). Only relevant when diagonal is TRUE. When TRUE,
diagonal cost is weighted by y factor of 1.414214 (square root of 2). Default: TRUE.}

\item{ignore_blocks}{(optional, logical). If TRUE, blocks of consecutive path
coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.}

\item{bandwidth}{(required, numeric) Size of the Sakoe-Chiba band at
both sides of the diagonal used to constrain the least cost path. Expressed
as a fraction of the number of matrix rows and columns. Unrestricted by default.
Default: 1}
}
\value{
data frame
}
\description{
Computes the contribution of individual variables to the
similarity/dissimilarity between two irregular multivariate time series.
In opposition to the legacy version, importance computation is
performed taking the least-cost path of the whole sequence as reference. This
operation makes the importance scores of individual variables fully comparable.
This function generates a data frame with the following columns:
\itemize{
\item variable: name of the individual variable for which the importance
is being computed, from the column names of the arguments \code{x} and \code{y}.
\item psi: global dissimilarity score \code{psi} of the two time series.
\item psi_only_with: dissimilarity between \code{x} and \code{y} computed from the given variable alone.
\item psi_without: dissimilarity between \code{x} and \code{y} computed from all other variables.
\item psi_difference: difference between \code{psi_only_with} and \code{psi_without}.
\item importance: contribution of the variable to the similarity/dissimilarity
between \code{x} and \code{y}, computed as \code{(psi_difference * 100) / psi_all}.
Positive scores represent contribution to dissimilarity,
while negative scores represent contribution to similarity.
}
}
\examples{
#simulate two regular time series
x <- zoo_simulate(
  seed = 1,
  rows = 100
  )

y <- zoo_simulate(
  seed = 2,
  rows = 150
  )

#different number of rows
#this is not a requirement though!
nrow(x) == nrow(y)

#compute importance
df <- importance_dtw_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)

df
}
\seealso{
Other Rcpp_importance:
\code{\link[=importance_dtw_legacy_cpp]{importance_dtw_legacy_cpp()}},
\code{\link[=importance_ls_cpp]{importance_ls_cpp()}}
}
\concept{Rcpp_importance}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_cluster_silhouette.R
\name{utils_cluster_silhouette}
\alias{utils_cluster_silhouette}
\title{Compute Silhouette Width of a Clustering Solution}
\usage{
utils_cluster_silhouette(labels = NULL, d = NULL, mean = FALSE)
}
\arguments{
\item{labels}{(required, integer vector) Labels resulting from a clustering algorithm applied to \code{d}. Must have the same length as columns and rows in \code{d}. Default: NULL}

\item{d}{(required, matrix) distance matrix typically resulting from \code{\link[=distantia_matrix]{distantia_matrix()}}, but any other square matrix should work. Default: NULL}

\item{mean}{(optional, logical) If TRUE, the mean of the silhouette widths is returned. Default: FALSE}
}
\value{
data frame
}
\description{
The silhouette width is a measure of how similar an object is to its own cluster (cohesion) compared to other clusters (separation).

There are some general guidelines to interpret the individual silhouette widths of the clustered objects (as returned by this function when \code{mean = FALSE}):

\itemize{
\item Close to 1: object is well matched to its own cluster and poorly matched to neighboring clusters.
\item Values near 0: the object is between two neighboring clusters.
\item Values near -1: the object is likely to be assigned to the wrong cluster.
}

When \code{mean = TRUE}, the overall mean of the silhouette widths of all objects is returned. These values should be interpreted as follows:

\itemize{
\item Higher than 0.7: robust clustering solution.
\item Higher than 0.5: reasonable clustering solution.
\item Higher than 0.25: weak clustering solution.

}

This metric may not perform well when the clusters have irregular shapes or sizes.

This code was adapted from \url{https://svn.r-project.org/R-packages/trunk/cluster/R/silhouette.R}.
}
\examples{
#weekly covid prevalence in three California counties
#load as tsl
#subset first 10 time series
#sum by month
tsl <- tsl_initialize(
  x = covid_prevalence,
  name_column = "name",
  time_column = "time"
) |>
  tsl_subset(
    names = 1:10
  ) |>
  tsl_aggregate(
    new_time = "months",
    method = max
  )

#compute dissimilarity
distantia_df <- distantia(
  tsl = tsl,
  lock_step = TRUE
)

#generate dissimilarity matrix
psi_matrix <- distantia_matrix(
  df = distantia_df
)

#example with kmeans clustering
#------------------------------------

#kmeans with 3 groups
psi_kmeans <- stats::kmeans(
  x = as.dist(psi_matrix[[1]]),
  centers = 3
)

#case-wise silhouette width
utils_cluster_silhouette(
  labels = psi_kmeans$cluster,
  d = psi_matrix
)

#overall silhouette width
utils_cluster_silhouette(
  labels = psi_kmeans$cluster,
  d = psi_matrix,
  mean = TRUE
)


#example with hierarchical clustering
#------------------------------------

#hierarchical clustering
psi_hclust <- stats::hclust(
  d = as.dist(psi_matrix[[1]])
)

#generate labels for three groups
psi_hclust_labels <- stats::cutree(
  tree = psi_hclust,
  k = 3,
)

#case-wise silhouette width
utils_cluster_silhouette(
  labels = psi_hclust_labels,
  d = psi_matrix
)

#overall silhouette width
utils_cluster_silhouette(
  labels = psi_hclust_labels,
  d = psi_matrix,
  mean = TRUE
)
}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_smooth_window.R
\name{zoo_smooth_window}
\alias{zoo_smooth_window}
\title{Rolling Window Smoothing of Zoo Time Series}
\usage{
zoo_smooth_window(x = NULL, window = 3, f = mean, ...)
}
\arguments{
\item{x}{(required, zoo object) Time series to smooth Default: NULL}

\item{window}{(optional, integer) Smoothing window width, in number of cases. Default: 3}

\item{f}{(optional, quoted or unquoted function name) Name of a function to aggregate numeric vectors. Typical examples are \code{mean}, \code{max},\code{min}, \code{median}, and \code{quantile}. Default: \code{mean}.}

\item{...}{(optional, additional arguments) additional arguments to \code{f}.}
}
\value{
zoo object
}
\description{
Just a fancy wrapper for \code{\link[zoo:rollapply]{zoo::rollapply()}}.
}
\examples{
x <- zoo_simulate()

x_smooth <- zoo_smooth_window(
  x = x,
  window = 5,
  f = mean
)

if(interactive()){
  zoo_plot(x)
  zoo_plot(x_smooth)
}
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_smooth.R
\name{tsl_smooth}
\alias{tsl_smooth}
\title{Smoothing of Time Series Lists}
\usage{
tsl_smooth(tsl = NULL, window = 3, f = mean, alpha = NULL, ...)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{window}{(optional, integer) Smoothing window width, in number of cases. Default: 3}

\item{f}{(optional, quoted or unquoted function name) Name of a function to aggregate numeric vectors. Typical examples are \code{mean}, \code{max},\code{min}, \code{median}, and \code{quantile}. Default: \code{mean}.}

\item{alpha}{(required, numeric) Exponential smoothing factor in the range (0, 1]. Determines the weight of the current value relative to past values. If not NULL, the arguments \code{window} and \code{f} are ignored, and exponential smoothing is performed instead. Default: NULL}

\item{...}{(optional, additional arguments) additional arguments to \code{f}.}
}
\value{
time series list
}
\description{
Rolling-window and exponential smoothing of Time Series Lists.

Rolling-window smoothing computes a statistic over a fixed-width window of consecutive cases and replaces each central value with the computed statistic. It is commonly used to mitigate noise in high-frequency time series.

Exponential smoothing computes each value as the weighted average of the current value and past smoothed values. This method is useful for reducing noise in time series data while preserving the overall trend.

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{


tsl <- tsl_simulate(n = 2)

#rolling window smoothing
tsl_smooth <- tsl_smooth(
  tsl = tsl,
  window = 5,
  f = mean
)

if(interactive()){
  tsl_plot(tsl)
  tsl_plot(tsl_smooth)
}

#exponential smoothing
tsl_smooth <- tsl_smooth(
  tsl = tsl,
  alpha = 0.2
)

if(interactive()){
  tsl_plot(tsl)
  tsl_plot(tsl_smooth)
}

}
\seealso{
Other tsl_processing:
\code{\link[=tsl_aggregate]{tsl_aggregate()}},
\code{\link[=tsl_resample]{tsl_resample()}},
\code{\link[=tsl_stats]{tsl_stats()}},
\code{\link[=tsl_transform]{tsl_transform()}}
}
\concept{tsl_processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{psi_null_ls_cpp}
\alias{psi_null_ls_cpp}
\title{(C++) Null Distribution of the Dissimilarity Scores of Two Aligned Time Series}
\usage{
psi_null_ls_cpp(
  x,
  y,
  distance = "euclidean",
  repetitions = 100L,
  permutation = "restricted_by_row",
  block_size = 3L,
  seed = 1L
)
}
\arguments{
\item{x}{(required, numeric matrix) of same number of columns as 'y'.}

\item{y}{(required, numeric matrix) of same number of columns as 'x'.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}

\item{repetitions}{(optional, integer) number of null psi values to generate. Default: 100}

\item{permutation}{(optional, character) permutation method. Valid values are listed below from higher to lower randomness:
\itemize{
\item "free": unrestricted shuffling of rows and columns. Ignores block_size.
\item "free_by_row": unrestricted shuffling of complete rows. Ignores block size.
\item "restricted": restricted shuffling of rows and columns within blocks.
\item "restricted_by_row": restricted shuffling of rows within blocks.
}}

\item{block_size}{(optional, integer) block size in rows for
restricted permutation. A block size of 3 indicates that a row can only be permuted
within a block of 3 adjacent rows. Minimum value is 2. Default: 3.}

\item{seed}{(optional, integer) initial random seed to use for replicability. Default: 1}
}
\value{
numeric vector
}
\description{
Applies permutation methods to compute null distributions for
the psi scores of two time series observed at the same times.
NA values should be removed before using this function.
If the selected distance function is "chi" or "cosine", pairs of zeros should
be either removed or replaced with pseudo-zeros (i.e. 0.00001).
}
\seealso{
Other Rcpp_dissimilarity_analysis:
\code{\link[=psi_dtw_cpp]{psi_dtw_cpp()}},
\code{\link[=psi_equation_cpp]{psi_equation_cpp()}},
\code{\link[=psi_ls_cpp]{psi_ls_cpp()}},
\code{\link[=psi_null_dtw_cpp]{psi_null_dtw_cpp()}}
}
\concept{Rcpp_dissimilarity_analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/momentum.R
\name{momentum}
\alias{momentum}
\title{Contribution of Individual Variables to Time Series Dissimilarity}
\usage{
momentum(
  tsl = NULL,
  distance = "euclidean",
  diagonal = TRUE,
  bandwidth = 1,
  lock_step = FALSE,
  robust = TRUE
)
}
\arguments{
\item{tsl}{(required, time series list) list of zoo time series. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}

\item{diagonal}{(optional, logical vector). If TRUE, diagonals are included in the dynamic time warping computation. Default: TRUE}

\item{bandwidth}{(optional, numeric) Proportion of space at each side of the cost matrix diagonal (aka \emph{Sakoe-Chiba band}) defining a valid region for dynamic time warping, used to control the flexibility of the warping path. This method prevents degenerate alignments due to differences in magnitude between time series when the data is not properly scaled. If \code{1} (default), DTW is unconstrained. If \code{0}, DTW is fully constrained and the warping path follows the matrix diagonal. Recommended values may vary depending on the nature of the data. Ignored if \code{lock_step = TRUE}. Default: 1.}

\item{lock_step}{(optional, logical vector) If TRUE, time series captured at the same times are compared sample wise (with no dynamic time warping). Requires time series in argument \code{tsl} to be fully aligned, or it will return an error. Default: FALSE.}

\item{robust}{(required, logical). If TRUE (default), importance scores are computed using the least cost path of the complete time series as reference. Setting it to FALSE allows to replicate importance scores of the previous versions of this package. This option is irrelevant when \code{lock_step = TRUE}. Default: TRUE}
}
\value{
A data frame with the following columns:
\itemize{
\item \code{x}: name of the time series \code{x}.
\item \code{y}: name of the time series \code{y}.
\item \code{psi}: psi score of \code{x} and \code{y}.
\item \code{variable}: name of the individual variable.
\item \code{importance}: importance score of the variable.
\item \code{effect}: interpretation of the "importance" column, with the values "increases similarity" and "decreases similarity".
\item \code{psi_only_with}: psi score of the variable.
\item \code{psi_without}: psi score without the variable.
\item \code{psi_difference}: difference between \code{psi_only_with} and \code{psi_without}.
\item \code{distance}: name of the distance metric.
\item \code{diagonal}: value of the argument \code{diagonal}.
\item \code{lock_step}: value of the argument \code{lock_step}.
\item \code{robust}: value of the argument \code{robust}.
}
}
\description{
This function measures the contribution of individual variables to dissimilarity between pairs of time series to help answer the question \emph{what makes two time series more or less similar?}

Three key values are required to assess individual variable contributions:
\itemize{
\item \strong{psi}: dissimilarity when all variables are considered.
\item \strong{psi_only_with}: dissimilarity when using only the target variable.
\item \strong{psi_without}: dissimilarity when removing the target variable.
}

The values \code{psi_only_with} and \code{psi_without} can be computed in two different ways, defined by the argument \code{robust}.
\itemize{
\item \code{robust = FALSE}: This method replicates the importance algorithm released with the first version of the package, and it is only recommended when the goal is to compare new results with previous studies. It normalizes \code{psi_only_with} and \code{psi_without} using the least cost path obtained from the individual variable. As different variables may have different least cost paths for the same time series, normalization values may change from variable to variable, making individual importance scores harder to compare.
\item \code{robust = TRUE} (default, recommended): This is a novel version of the importance algorithm that yields more stable and comparable solutions. It uses the least cost path of the complete time series to normalize \code{psi_only_with} and \code{psi_without}, making importance scores of separate variables fully comparable.
}

The individual importance score of each variable (column "importance" in the output data frame) is based on different expressions depending on the \code{robust} argument, even when \code{lock_step = TRUE}:
\itemize{
\item \code{robust = FALSE}: Importance is computed as \code{((psi - psi_without) * 100)/psi} and interpreted as "change in similarity when a variable is removed".
\item \code{robust = TRUE}: Importance is computed as \code{((psi_only_with - psi_without) * 100)/psi} and interpreted as "relative dissimilarity induced by the variable expressed as a percentage".
}

In either case, positive values indicate that the variable contributes to dissimilarity, while negative values indicate a net contribution to similarity.

This function allows computing dissimilarity between pairs of time series using different combinations of arguments at once. For example, when the argument \code{distance} is set to \code{c("euclidean", "manhattan")}, the output data frame will show two dissimilarity scores for each pair of time series, one based on euclidean distances, and another based on manhattan distances. The same happens for most other parameters.

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{

#progress bar
# progressr::handlers(global = TRUE)

tsl <- tsl_initialize(
  x = distantia::albatross,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

df <- momentum(
  tsl = tsl,
  lock_step = TRUE #to speed-up example
  )

#focus on important columns
df[, c(
  "x",
  "y",
  "variable",
  "importance",
  "effect"
  )]

}
\seealso{
Other momentum:
\code{\link[=momentum_dtw]{momentum_dtw()}},
\code{\link[=momentum_ls]{momentum_ls()}}
}
\concept{momentum}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_names.R
\name{tsl_names_test}
\alias{tsl_names_test}
\title{Tests Naming Issues in Time Series Lists}
\usage{
tsl_names_test(tsl = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}
}
\value{
logical
}
\description{
Tests Naming Issues in Time Series Lists
}
\examples{
#creating three zoo time series

#one with NA values
x <- zoo_simulate(
  name = "x",
  cols = 1,
  na_fraction = 0.1
  )

#with different number of columns
#with repeated name
y <- zoo_simulate(
  name = "x",
  cols = 2
  )

#with different time class
z <- zoo_simulate(
  name = "z",
  cols = 1,
  time_range = c(1, 100)
  )

#adding a few structural issues

#changing the column name of x
colnames(x) <- c("b")

#converting z to vector
z <- zoo::zoo(
  x = runif(nrow(z)),
  order.by = zoo::index(z)
)

#storing zoo objects in a list
#with mismatched names
tsl <- list(
  a = x,
  b = y,
  c = z
)

#running full diagnose
tsl_names_test(
  tsl = tsl
  )
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_cluster_hclust.R
\name{distantia_cluster_hclust}
\alias{distantia_cluster_hclust}
\title{Hierarchical Clustering of Dissimilarity Analysis Data Frames}
\usage{
distantia_cluster_hclust(df = NULL, clusters = NULL, method = "complete")
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=distantia]{distantia()}}, \code{\link[=distantia_ls]{distantia_ls()}}, \code{\link[=distantia_dtw]{distantia_dtw()}}, or \code{\link[=distantia_time_delay]{distantia_time_delay()}}. Default: NULL}

\item{clusters}{(required, integer) Number of groups to generate. If NULL (default), \code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}} is used to find the number of clusters that maximizes the mean silhouette width of the clustering solution (see \code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}). Default: NULL}

\item{method}{(optional, character string) Argument of \code{\link[stats:hclust]{stats::hclust()}} defining the agglomerative method. One of: "ward.D", "ward.D2", "single", "complete", "average" (= UPGMA), "mcquitty" (= WPGMA), "median" (= WPGMC) or "centroid" (= UPGMC). Unambiguous abbreviations are accepted as well. If NULL (default), \code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}} finds the optimal method. Default: "complete".}
}
\value{
list:
\itemize{
\item \code{cluster_object}: hclust object for further analyses and custom plotting.
\item \code{clusters}: integer, number of clusters.
\item \code{silhouette_width}: mean silhouette width of the clustering solution.
\item \code{df}: data frame with time series names, their cluster label, and their individual silhouette width scores.
\item \code{d}: psi distance matrix used for clustering.
\item \code{optimization}: only if \code{clusters = NULL}, data frame with optimization results from \code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}}.
}
}
\description{
This function combines the dissimilarity scores computed by \code{\link[=distantia]{distantia()}}, the agglomerative clustering methods provided by \code{\link[stats:hclust]{stats::hclust()}}, and the clustering optimization method implemented in \code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}} to help group together time series with similar features.

When \code{clusters = NULL}, the function \code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}} is run underneath to perform a parallelized grid search to find the  number of clusters maximizing the overall silhouette width of the clustering solution (see \code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}). When \code{method = NULL} as well, the optimization also includes all methods available in \code{\link[stats:hclust]{stats::hclust()}} in the grid search.

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{

#weekly covid prevalence in California
tsl <- tsl_initialize(
  x = covid_prevalence,
  name_column = "name",
  time_column = "time"
)

#subset 10 elements to accelerate example execution
tsl <- tsl_subset(
  tsl = tsl,
  names = 1:10
)

if(interactive()){
  #plotting first three time series
  tsl_plot(
    tsl = tsl[1:3],
    guide_columns = 3
  )
}

#dissimilarity analysis
distantia_df <- distantia(
  tsl = tsl,
  lock_step = TRUE
)

#hierarchical clustering
#automated number of clusters
#automated method selection
distantia_clust <- distantia_cluster_hclust(
  df = distantia_df,
  clusters = NULL,
  method = NULL
)

#names of the output object
names(distantia_clust)

#cluster object
distantia_clust$cluster_object

#distance matrix used for clustering
distantia_clust$d

#number of clusters
distantia_clust$clusters

#clustering data frame
#group label in column "cluster"
#negatives in column "silhouette_width" higlight anomalous cluster assignation
distantia_clust$df

#mean silhouette width of the clustering solution
distantia_clust$silhouette_width

#plot
if(interactive()){

  dev.off()

  clust <- distantia_clust$cluster_object
  k <- distantia_clust$clusters

  #tree plot
  plot(
    x = clust,
    hang = -1
  )

  #highlight groups
  stats::rect.hclust(
    tree = clust,
    k = k,
    cluster = stats::cutree(
      tree = clust,
      k = k
    )
  )

}

}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{importance_ls_cpp}
\alias{importance_ls_cpp}
\title{(C++) Contribution of Individual Variables to the Dissimilarity Between Two Aligned Time Series}
\usage{
importance_ls_cpp(x, y, distance = "euclidean")
}
\arguments{
\item{x}{(required, numeric matrix) multivariate time series.}

\item{y}{(required, numeric matrix) multivariate time series
with the same number of columns and rows as 'x'.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}
}
\value{
data frame
}
\description{
Computes the contribution of individual variables to the
similarity/dissimilarity between two aligned multivariate time series.
This function generates a data frame with the following columns:
\itemize{
\item variable: name of the individual variable for which the importance
is being computed, from the column names of the arguments \code{x} and \code{y}.
\item psi: global dissimilarity score \code{psi} of the two time series.
\item psi_only_with: dissimilarity between \code{x} and \code{y} computed from the given variable alone.
\item psi_without: dissimilarity between \code{x} and \code{y} computed from all other variables.
\item psi_difference: difference between \code{psi_only_with} and \code{psi_without}.
\item importance: contribution of the variable to the similarity/dissimilarity
between \code{x} and \code{y}, computed as \code{(psi_difference * 100) / psi_all}.
Positive scores represent contribution to dissimilarity,
while negative scores represent contribution to similarity.
}
}
\examples{
#simulate two regular time series
x <- zoo_simulate(
  seed = 1,
  irregular = FALSE
  )

y <- zoo_simulate(
  seed = 2,
  irregular = FALSE
  )

#same number of rows
nrow(x) == nrow(y)

#compute importance
df <- importance_ls_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)

df
}
\seealso{
Other Rcpp_importance:
\code{\link[=importance_dtw_cpp]{importance_dtw_cpp()}},
\code{\link[=importance_dtw_legacy_cpp]{importance_dtw_legacy_cpp()}}
}
\concept{Rcpp_importance}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_repair.R
\name{tsl_repair}
\alias{tsl_repair}
\title{Repair Issues in Time Series Lists}
\usage{
tsl_repair(tsl = NULL, full = TRUE)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{full}{(optional, logical) If TRUE, a full repair (structure and values) is triggered. Otherwise, only the data structure is repaired. Default: TRUE}
}
\value{
time series list
}
\description{
A Time Series List (\code{tsl} for short) is a list of zoo time series. This type of object, not defined as a class, is used throughout the \code{distantia} package to contain time series data ready for processing and analysis.

The structure and values of a \code{tsl} must fulfill several general conditions:

Structure:
\itemize{
\item The list names match the attributes "name" of the zoo time series.
\item All zoo time series must have at least one shared column name.
\item Data in univariate zoo time series (as extracted by \code{zoo::coredata(x)}) must be of the class "matrix". Univariate zoo time series are often represented as vectors, but this breaks several subsetting and transformation operations implemented in this package.
}

Values (optional, when \code{full = TRUE}):
\itemize{
\item All time series have at least one shared numeric column.
\item There are no NA, Inf, or NaN values in the time series.
}

This function analyzes a \code{tsl} and tries to fix all possible issues to make it comply with the conditions listed above without any user input. Use with care, as it might modify your data.
}
\examples{
#creating three zoo time series

#one with NA values
x <- zoo_simulate(
  name = "x",
  cols = 1,
  na_fraction = 0.1
  )

#with different number of columns
#with repeated name
y <- zoo_simulate(
  name = "x",
  cols = 2
  )

#with different time class
z <- zoo_simulate(
  name = "z",
  cols = 1,
  time_range = c(1, 100)
  )

#adding a few structural issues

#changing the column name of x
colnames(x) <- c("b")

#converting z to vector
z <- zoo::zoo(
  x = runif(nrow(z)),
  order.by = zoo::index(z)
)

#storing zoo objects in a list
#with mismatched names
tsl <- list(
  a = x,
  b = y,
  c = z
)

#running full diagnose
tsl_diagnose(
  tsl = tsl,
  full = TRUE
  )

tsl <- tsl_repair(tsl)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{albatross}
\alias{albatross}
\title{Flight Path Time Series of Albatrosses in the Pacific}
\format{
A data frame.
}
\usage{
data(albatross)
}
\description{
Daily mean flight path data of four individuals of Waved Albatross (Phoebastria irrorata) captured via GPS during the summer of 2008. An sf data frame with columns name, time, latitude, longitude, ground speed, heading, and (uncalibrated) temperature.

The full dataset at hourly resolution can be downloaded from \url{https://github.com/BlasBenito/distantia/blob/main/data_full/albatross.rda} (use the "Download raw file" button).
}
\examples{
#Load as TSL
#Scale all variables
#Aggregate to daily resolution
#Align all time series to same temporal span
tsl <- tsl_initialize(
  x = albatross,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_local
  ) |>
  tsl_aggregate(
    new_time = "days"
  )

if(interactive()){
  tsl_plot(
    tsl = tsl,
    guide_columns = 5
    )
}
}
\references{
\doi{10.5441/001/1.3hp3s250}
}
\seealso{
Other example_data:
\code{\link{cities_coordinates}},
\code{\link{cities_temperature}},
\code{\link{covid_counties}},
\code{\link{covid_prevalence}},
\code{\link{eemian_coordinates}},
\code{\link{eemian_pollen}},
\code{\link{fagus_coordinates}},
\code{\link{fagus_dynamics}},
\code{\link{honeycomb_climate}},
\code{\link{honeycomb_polygons}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_colnames.R
\name{tsl_colnames_clean}
\alias{tsl_colnames_clean}
\title{Clean Column Names in Time Series Lists}
\usage{
tsl_colnames_clean(
  tsl = NULL,
  lowercase = FALSE,
  separator = "_",
  capitalize_first = FALSE,
  capitalize_all = FALSE,
  length = NULL,
  suffix = NULL,
  prefix = NULL
)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{lowercase}{(optional, logical) If TRUE, all names are coerced to lowercase. Default: FALSE}

\item{separator}{(optional, character string) Separator when replacing spaces and dots. Also used to separate \code{suffix} and \code{prefix} from the main word. Default: "_".}

\item{capitalize_first}{(optional, logical) Indicates whether to capitalize the first letter of each name. Default: FALSE.}

\item{capitalize_all}{(optional, logical) Indicates whether to capitalize all letters of each name. Default: FALSE.}

\item{length}{(optional, integer) Minimum length of abbreviated names. Names are abbreviated via \code{\link[=abbreviate]{abbreviate()}}. Default: NULL.}

\item{suffix}{(optional, character string) String to append to the column names. Default: NULL.}

\item{prefix}{(optional, character string) String to prepend to the column names. Default: NULL.}
}
\value{
time series list
}
\description{
Uses the function \code{\link[=utils_clean_names]{utils_clean_names()}} to simplify and normalize messy column names in a time series list.

The cleanup operations are applied in the following order:
\itemize{
\item Remove leading and trailing whitespaces.
\item Generate syntactically valid names with \code{\link[base:make.names]{base::make.names()}}.
\item Replaces dots and spaces with the \code{separator}.
\item Coerces names to lowercase.
\item If \code{capitalize_first = TRUE}, the first letter is capitalized.
\item If \code{capitalize_all = TRUE}, all letters are capitalized.
\item If argument \code{length} is provided, \code{\link[base:abbreviate]{base::abbreviate()}} is used to abbreviate the new column names.
\item If \code{suffix} is provided, it is added at the end of the column name using the separator.
\item If \code{prefix} is provided, it is added at the beginning of the column name using the separator.
}
}
\examples{
#generate example data
tsl <- tsl_simulate(cols = 3)

#list all column names
tsl_colnames_get(
  tsl = tsl
)

#rename columns
tsl <- tsl_colnames_set(
  tsl = tsl,
  names = c(
  "New name 1",
  "new Name 2",
  "NEW NAME 3"
  )
)

#check new names
tsl_colnames_get(
  tsl = tsl,
  names = "all"
)

#clean names
tsl <- tsl_colnames_clean(
  tsl = tsl
)

tsl_colnames_get(
  tsl = tsl
)

#abbreviated
tsl <- tsl_colnames_clean(
  tsl = tsl,
  capitalize_first = TRUE,
  length = 6,
  suffix = "clean"
)

tsl_colnames_get(
  tsl = tsl
)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_check_args.R
\name{utils_check_args_path}
\alias{utils_check_args_path}
\title{Checks Least Cost Path}
\usage{
utils_check_args_path(path = NULL, arg_name = "path")
}
\arguments{
\item{path}{(required, data frame) least cost path generated with \code{\link[=psi_cost_path]{psi_cost_path()}}. This data frame must have the attribute \code{type == "cost_path"}, and must have been computed from the same time series used to compute the matrix \code{m}. Default: NULL.}

\item{arg_name}{(optional, character string) name of the argument being checked. Default: NULL.}
}
\value{
data frame.
}
\description{
Checks Least Cost Path
}
\seealso{
Other internal:
\code{\link[=utils_boxplot_common]{utils_boxplot_common()}},
\code{\link[=utils_check_args_distantia]{utils_check_args_distantia()}},
\code{\link[=utils_check_args_matrix]{utils_check_args_matrix()}},
\code{\link[=utils_check_args_momentum]{utils_check_args_momentum()}},
\code{\link[=utils_check_args_tsl]{utils_check_args_tsl()}},
\code{\link[=utils_check_args_zoo]{utils_check_args_zoo()}},
\code{\link[=utils_check_distance_args]{utils_check_distance_args()}},
\code{\link[=utils_check_list_class]{utils_check_list_class()}},
\code{\link[=utils_clean_names]{utils_clean_names()}},
\code{\link[=utils_digits]{utils_digits()}},
\code{\link[=utils_distantia_df_split]{utils_distantia_df_split()}},
\code{\link[=utils_prepare_df]{utils_prepare_df()}},
\code{\link[=utils_prepare_matrix]{utils_prepare_matrix()}},
\code{\link[=utils_prepare_matrix_list]{utils_prepare_matrix_list()}},
\code{\link[=utils_prepare_time]{utils_prepare_time()}},
\code{\link[=utils_prepare_vector_list]{utils_prepare_vector_list()}},
\code{\link[=utils_prepare_zoo_list]{utils_prepare_zoo_list()}},
\code{\link[=utils_tsl_pairs]{utils_tsl_pairs()}}
}
\concept{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cost_matrix_diagonal_weighted_cpp}
\alias{cost_matrix_diagonal_weighted_cpp}
\title{(C++) Compute Orthogonal and Weighted Diagonal Least Cost Matrix from a Distance Matrix}
\usage{
cost_matrix_diagonal_weighted_cpp(dist_matrix)
}
\arguments{
\item{dist_matrix}{(required, distance matrix). Distance matrix.}
}
\value{
Least cost matrix.
}
\description{
Computes the least cost matrix from a distance matrix.
Weights diagonals by a factor of 1.414214 (square root of 2) with respect to orthogonal paths.
}
\seealso{
Other Rcpp_matrix:
\code{\link[=cost_matrix_diagonal_cpp]{cost_matrix_diagonal_cpp()}},
\code{\link[=cost_matrix_orthogonal_cpp]{cost_matrix_orthogonal_cpp()}},
\code{\link[=distance_ls_cpp]{distance_ls_cpp()}},
\code{\link[=distance_matrix_cpp]{distance_matrix_cpp()}}
}
\concept{Rcpp_matrix}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_matrix.R
\name{distantia_matrix}
\alias{distantia_matrix}
\title{Convert Dissimilarity Analysis Data Frame to Distance Matrix}
\usage{
distantia_matrix(df = NULL)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=distantia]{distantia()}}, \code{\link[=distantia_ls]{distantia_ls()}}, \code{\link[=distantia_dtw]{distantia_dtw()}}, or \code{\link[=distantia_time_delay]{distantia_time_delay()}}. Default: NULL}
}
\value{
numeric matrix
}
\description{
Transforms a data frame resulting from \code{\link[=distantia]{distantia()}} into a dissimilarity matrix.
}
\examples{
#weekly covid prevalence in three California counties
#load as tsl
#subset 5 counties
#sum by month
tsl <- tsl_initialize(
  x = covid_prevalence,
  name_column = "name",
  time_column = "time"
) |>
  tsl_subset(
    names = 1:5
  ) |>
  tsl_aggregate(
    new_time = "months",
    method = sum
  )

if(interactive()){

  #plotting first three time series
  tsl_plot(
    tsl = tsl,
    guide_columns = 3
    )

  dev.off()

}

#dissimilarity analysis
#two combinations of arguments
distantia_df <- distantia(
  tsl = tsl,
  lock_step = c(TRUE, FALSE)
)

#to dissimilarity matrix
distantia_matrix <- distantia_matrix(
  df = distantia_df
)

#returns a list of matrices
lapply(
  X = distantia_matrix,
  FUN = class
  )

#these matrices have attributes tracing how they were generated
lapply(
  X = distantia_matrix,
  FUN = \(x) attributes(x)$distantia_args
)

#plot matrix
if(interactive()){

  #plot first matrix (default behavior of utils_matrix_plot())
  utils_matrix_plot(
    m = distantia_matrix
  )

  #plot second matrix
  utils_matrix_plot(
    m = distantia_matrix[[2]]
  )

}
}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_time_keywords_translate.R
\name{utils_time_keywords_translate}
\alias{utils_time_keywords_translate}
\title{Translate User Time Keywords Into Valid Ones}
\usage{
utils_time_keywords_translate(keyword = NULL)
}
\arguments{
\item{keyword}{(optional, character string) A time keyword such as "day". Default: NULL}
}
\value{
A valid time keyword.
}
\description{
Internal function to translate misnamed or abbreviated keywords into valid ones. Uses \code{\link[=utils_time_keywords_dictionary]{utils_time_keywords_dictionary()}} as reference dictionary.
}
\examples{
#millennia
utils_time_keywords_translate(
  keyword = "1000 years"
)

utils_time_keywords_translate(
  keyword = "1000 y"
)

utils_time_keywords_translate(
  keyword = "thousands"
)

#years
utils_time_keywords_translate(
  keyword = "year"
)

utils_time_keywords_translate(
  keyword = "y"
)

#days
utils_time_keywords_translate(
  keyword = "d"
)

utils_time_keywords_translate(
  keyword = "day"
)

#seconds
utils_time_keywords_translate(
  keyword = "s"
)

utils_time_keywords_translate(
  keyword = "sec"
)
}
\seealso{
Other internal_time_handling:
\code{\link[=utils_as_time]{utils_as_time()}},
\code{\link[=utils_coerce_time_class]{utils_coerce_time_class()}},
\code{\link[=utils_is_time]{utils_is_time()}},
\code{\link[=utils_new_time]{utils_new_time()}},
\code{\link[=utils_time_keywords]{utils_time_keywords()}},
\code{\link[=utils_time_keywords_dictionary]{utils_time_keywords_dictionary()}},
\code{\link[=utils_time_units]{utils_time_units()}}
}
\concept{internal_time_handling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_coerce_time_class.R
\name{utils_coerce_time_class}
\alias{utils_coerce_time_class}
\title{Coerces Vector to a Given Time Class}
\usage{
utils_coerce_time_class(x = NULL, to = "POSIXct")
}
\arguments{
\item{x}{(required, vector of class Date or POSIXct) time vector to convert. Default: NULL}

\item{to}{(required, class name) class to coerce \code{x} to. Either "Date", "POSIXct", "integer" or "numeric". Default: "POSIXct"}
}
\value{
time vector
}
\description{
Coerces Vector to a Given Time Class
}
\examples{
x <- utils_coerce_time_class(
  x = c("2024-01-01", "2024-02-01"),
  to = "Date"
)

x
class(x)

x <- utils_coerce_time_class(
  x = c("2024-01-01", "2024-02-01"),
  to = "POSIXct"
)

x
class(x)

x <- utils_coerce_time_class(
  x = c("2024-01-01", "2024-02-01"),
  to = "numeric"
)

x
class(x)
}
\seealso{
Other internal_time_handling:
\code{\link[=utils_as_time]{utils_as_time()}},
\code{\link[=utils_is_time]{utils_is_time()}},
\code{\link[=utils_new_time]{utils_new_time()}},
\code{\link[=utils_time_keywords]{utils_time_keywords()}},
\code{\link[=utils_time_keywords_dictionary]{utils_time_keywords_dictionary()}},
\code{\link[=utils_time_keywords_translate]{utils_time_keywords_translate()}},
\code{\link[=utils_time_units]{utils_time_units()}}
}
\concept{internal_time_handling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_spatial.R
\name{distantia_spatial}
\alias{distantia_spatial}
\title{Spatial Representation of \code{distantia()} Data Frames}
\usage{
distantia_spatial(df = NULL, sf = NULL, network = TRUE)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=distantia]{distantia()}} or \code{\link[=distantia_time_delay]{distantia_time_delay()}}. Default: NULL}

\item{sf}{(required, sf data frame) Points or polygons representing the location of the time series in argument 'df'. It must have a column with all time series names in \code{df$x} and \code{df$y}. Default: NULL}

\item{network}{(optional, logical) If TRUE, the resulting sf data frame is of type LINESTRING and represents network edges. Default: TRUE}
}
\value{
sf data frame (LINESTRING geometry)
}
\description{
Given an sf data frame with geometry types POLYGON, MULTIPOLYGON, or POINT representing time series locations, this function transforms the output of \code{\link[=distantia]{distantia()}}, \code{\link[=distantia_ls]{distantia_ls()}}, \code{\link[=distantia_dtw]{distantia_dtw()}} or \code{\link[=distantia_time_delay]{distantia_time_delay()}} to an sf data frame.

If \code{network = TRUE}, the sf data frame is of type LINESTRING, with edges connecting time series locations. This output is helpful to build many-to-many dissimilarity maps (see examples).

If \code{network = FALSE}, the sf data frame contains the geometry in the input \code{sf} argument. This output helps build one-to-many dissimilarity maps.
}
\examples{
tsl <- distantia::tsl_initialize(
  x = distantia::covid_prevalence,
  name_column = "name",
  time_column = "time"
) |>
distantia::tsl_subset(
  names = c(
    "Los_Angeles",
    "San_Francisco",
    "Fresno",
    "San_Joaquin"
    )
 )

df_psi <- distantia::distantia_ls(
  tsl = tsl
)

#network many to many
sf_psi <- distantia::distantia_spatial(
  df = df_psi,
  sf = distantia::covid_counties,
  network = TRUE
)

#network map
# mapview::mapview(
#   distantia::covid_counties,
#   col.regions = NA,
#   alpha.regions = 0,
#   color = "black",
#   label = "name",
#   legend = FALSE,
#   map.type = "OpenStreetMap"
# ) +
#   mapview::mapview(
#     sf_psi_subset,
#     layer.name = "Psi",
#     label = "edge_name",
#     zcol = "psi",
#     lwd = 3
#   ) |>
#   suppressWarnings()

}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{permute_free_cpp}
\alias{permute_free_cpp}
\title{(C++) Unrestricted Permutation of Cases}
\usage{
permute_free_cpp(x, block_size, seed = 1L)
}
\arguments{
\item{x}{(required, numeric matrix). Numeric matrix to permute.}

\item{block_size}{(optional, integer) this function ignores this argument and sets it to x.nrow().}

\item{seed}{(optional, integer) random seed to use.}
}
\value{
numeric matrix
}
\description{
Unrestricted shuffling of cases within the whole sequence.
}
\seealso{
Other Rcpp_permutation:
\code{\link[=permute_free_by_row_cpp]{permute_free_by_row_cpp()}},
\code{\link[=permute_restricted_by_row_cpp]{permute_restricted_by_row_cpp()}},
\code{\link[=permute_restricted_cpp]{permute_restricted_cpp()}}
}
\concept{Rcpp_permutation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_rescale_vector.R
\name{utils_rescale_vector}
\alias{utils_rescale_vector}
\title{Rescale Numeric Vector to a New Data Range}
\usage{
utils_rescale_vector(
  x = NULL,
  new_min = 0,
  new_max = 1,
  old_min = NULL,
  old_max = NULL
)
}
\arguments{
\item{x}{(required, numeric vector) Numeric vector. Default: \code{NULL}}

\item{new_min}{(optional, numeric) New minimum value. Default: \code{0}}

\item{new_max}{(optional, numeric) New maximum value. Default: \code{1}}

\item{old_min}{(optional, numeric) Old minimum value. Default: \code{NULL}}

\item{old_max}{(optional, numeric) Old maximum value. Default: \code{NULL}}
}
\value{
numeric vector
}
\description{
Rescale Numeric Vector to a New Data Range
}
\examples{

 out <- utils_rescale_vector(
   x = stats::rnorm(100),
   new_min = 0,
   new_max = 100,
   )

 out

}
\seealso{
Other tsl_processing_internal:
\code{\link[=utils_drop_geometry]{utils_drop_geometry()}},
\code{\link[=utils_global_scaling_params]{utils_global_scaling_params()}},
\code{\link[=utils_optimize_loess]{utils_optimize_loess()}},
\code{\link[=utils_optimize_spline]{utils_optimize_spline()}}
}
\concept{tsl_processing_internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{auto_sum_path_cpp}
\alias{auto_sum_path_cpp}
\title{(C++) Sum Distances Between All Consecutive Samples in the Least Cost Path Between Two Time Series}
\usage{
auto_sum_path_cpp(x, y, path, distance = "euclidean")
}
\arguments{
\item{x}{(required, numeric matrix) univariate or multivariate time series.}

\item{y}{(required, numeric matrix) univariate or multivariate time series
with the same number of columns as 'x'.}

\item{path}{(required, data frame) least-cost path produced by \code{\link[=cost_path_orthogonal_cpp]{cost_path_orthogonal_cpp()}}.
Default: NULL}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}
}
\value{
numeric
}
\description{
Computes the cumulative auto sum of auto-distances of two time series
for the coordinates of a trimmed least cost path. The output value is used
as normalization factor when computing dissimilarity scores.
}
\examples{
#simulate two time series
x <- zoo_simulate(seed = 1)
y <- zoo_simulate(seed = 2)

#distance matrix
dist_matrix <- distance_matrix_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)

#least cost matrix
cost_matrix <- cost_matrix_orthogonal_cpp(
  dist_matrix = dist_matrix
)

#least cost path
cost_path <- cost_path_orthogonal_cpp(
  dist_matrix = dist_matrix,
  cost_matrix = cost_matrix
)

nrow(cost_path)

#remove blocks from least-cost path
cost_path_trimmed <- cost_path_trim_cpp(
  path = cost_path
)

nrow(cost_path_trimmed)

#auto sum
auto_sum_path_cpp(
  x = x,
  y = y,
  path = cost_path_trimmed,
  distance = "euclidean"
)
}
\seealso{
Other Rcpp_auto_sum:
\code{\link[=auto_distance_cpp]{auto_distance_cpp()}},
\code{\link[=auto_sum_cpp]{auto_sum_cpp()}},
\code{\link[=auto_sum_full_cpp]{auto_sum_full_cpp()}},
\code{\link[=subset_matrix_by_rows_cpp]{subset_matrix_by_rows_cpp()}}
}
\concept{Rcpp_auto_sum}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psi_cost_matrix.R
\name{psi_cost_matrix}
\alias{psi_cost_matrix}
\title{Cost Matrix}
\usage{
psi_cost_matrix(dist_matrix = NULL, diagonal = TRUE)
}
\arguments{
\item{dist_matrix}{(required, numeric matrix). Distance matrix generated by \code{\link[=psi_distance_matrix]{psi_distance_matrix()}}. Default: NULL}

\item{diagonal}{(optional, logical vector). If TRUE, diagonals are included in the dynamic time warping computation. Default: TRUE}
}
\value{
numeric matrix
}
\description{
Demonstration function to compute a cost matrix from a distance matrix.
}
\examples{
#distance metric
d <- "euclidean"

#use diagonals in least cost computations
diagonal <- TRUE

#simulate two irregular time series
x <- zoo_simulate(
  name = "x",
  rows = 100,
  seasons = 2,
  seed = 1
)

y <- zoo_simulate(
  name = "y",
  rows = 80,
  seasons = 2,
  seed = 2
)

if(interactive()){
  zoo_plot(x = x)
  zoo_plot(x = y)
}

#distance matrix
dist_matrix <- psi_distance_matrix(
  x = x,
  y = y,
  distance = d
)

#cost matrix
cost_matrix <- psi_cost_matrix(
  dist_matrix = dist_matrix,
  diagonal = diagonal
)

if(interactive()){
  utils_matrix_plot(
    m = cost_matrix
    )
}
}
\seealso{
Other psi_demo:
\code{\link[=psi_auto_distance]{psi_auto_distance()}},
\code{\link[=psi_auto_sum]{psi_auto_sum()}},
\code{\link[=psi_cost_path]{psi_cost_path()}},
\code{\link[=psi_cost_path_sum]{psi_cost_path_sum()}},
\code{\link[=psi_distance_lock_step]{psi_distance_lock_step()}},
\code{\link[=psi_distance_matrix]{psi_distance_matrix()}},
\code{\link[=psi_equation]{psi_equation()}}
}
\concept{psi_demo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_time.R
\name{zoo_time}
\alias{zoo_time}
\title{Get Time Features from Zoo Objects}
\usage{
zoo_time(x = NULL, keywords = c("resample", "aggregate"))
}
\arguments{
\item{x}{(required, zoo object) Zoo time series to analyze. Default: NULL.}

\item{keywords}{(optional, character string or vector) Defines what keywords are returned. If "aggregate", returns valid keywords for \code{\link[=zoo_aggregate]{zoo_aggregate()}}. If "resample", returns valid keywords for \code{\link[=zoo_resample]{zoo_resample()}}. If both, returns all valid keywords. Default: c("aggregate", "resample").}
}
\value{
Data frame with the following columns:
\itemize{
\item \code{name} (string): time series name.
\item \code{rows} (integer): number of observations.
\item \code{class} (string): time class, one of "Date", "POSIXct", or "numeric."
\item \code{units} (string): units of the time series.
\item \code{length} (numeric): total length of the time series expressed in \code{units}.
\item \code{resolution} (numeric): average interval between observations expressed in \code{units}.
\item \code{begin} (date or numeric): begin time of the time series.
\item \code{end} (date or numeric): end time of the time series.
\item \code{keywords} (character vector): valid keywords for \code{\link[=tsl_aggregate]{tsl_aggregate()}} or \code{\link[=tsl_resample]{tsl_resample()}}, depending on the value of the argument \code{keywords}.
}
}
\description{
This function generates a data frame summarizing the time features (class, length, resolution, and others) of zoo time series.
}
\examples{
#simulate a zoo time series
x <- zoo_simulate(
  rows = 150,
  time_range = c(
    Sys.Date() - 365,
    Sys.Date()
  ),
  irregular = TRUE
)

#time data frame
zoo_time(
  x = x
)
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_ls.R
\name{distantia_ls}
\alias{distantia_ls}
\title{Lock-Step Dissimilarity Analysis of Time Series Lists}
\usage{
distantia_ls(tsl = NULL, distance = "euclidean")
}
\arguments{
\item{tsl}{(required, time series list) list of zoo time series. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}
}
\value{
data frame:
\itemize{
\item \code{x}: time series name.
\item \code{y}: time series name.
\item \code{distance}: name of the distance metric.
\item \code{lock_step}: always \code{TRUE}; flags this result as lock-step.
\item \code{psi}: psi dissimilarity of the sequences \code{x} and \code{y}.
}
}
\description{
A minimalistic but slightly faster version of \code{\link[=distantia]{distantia()}} that computes lock-step dissimilarity scores.
}
\examples{


#load fagus_dynamics as tsl
#global centering and scaling
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

if(interactive()){
  tsl_plot(
    tsl = tsl,
    guide_columns = 3
    )
}

#lock-step dissimilarity analysis
df_ls <- distantia_ls(
  tsl = tsl,
  distance = "euclidean"
)

df_ls

}
\seealso{
Other distantia:
\code{\link[=distantia]{distantia()}},
\code{\link[=distantia_dtw]{distantia_dtw()}},
\code{\link[=distantia_dtw_plot]{distantia_dtw_plot()}}
}
\concept{distantia}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_as_time.R
\name{utils_as_time}
\alias{utils_as_time}
\title{Ensure Correct Class for Time Arguments}
\usage{
utils_as_time(x = NULL, to_class = NULL)
}
\arguments{
\item{x}{(required, vector) Vectors of the classes 'numeric', 'Date', and 'POSIXct' are valid and returned without any transformation. Character vectors are analyzed to determine their more probable type, and are coerced to 'Date' or 'POSIXct' depending on their number of elements. Generally, any character vector representing an ISO 8601 standard, like "YYYY-MM-DD" or "YYYY-MM-DD HH:MM:SS" will be converted to a valid class. If a character vector cannot be coerced to date, it is returned as is. Default: NULL}

\item{to_class}{(optional, class) Options are: NULL, "numeric", "Date", and "POSIXct". If NULL, 'x' is returned as the most appropriate time class. Otherwise, 'x' is coerced to the given class. Default: NULL}
}
\value{
time vector
}
\description{
Guesses the class of a vector based on its elements. This function can handle numeric vectors, character vectors that can be coerced to either "Date" or "POSIXct" classes, and vectors already in "Date" or "POSIXct" classes.
}
\examples{
# numeric
utils_as_time(
  x = c(-123120, 1200)
  )

# character string to Date
utils_as_time(
  x = c("2022-03-17", "2024-02-05")
  )

# incomplete character strings to Date
utils_as_time(
  x = c("2022", "2024")
  )

utils_as_time(
  x = c("2022-02", "2024-03")
  )

# character string to POSIXct
utils_as_time(
  x = c("2022-03-17 12:30:45", "2024-02-05 11:15:45")
  )

# Date vector (returns the input)
utils_as_time(
  x = as.Date(c("2022-03-17", "2024-02-05"))
  )

# POSIXct vector (returns the input)
utils_as_time(
  x = as.POSIXct(c("2022-03-17 12:30:45", "2024-02-05 11:15:45"))
  )

}
\seealso{
Other internal_time_handling:
\code{\link[=utils_coerce_time_class]{utils_coerce_time_class()}},
\code{\link[=utils_is_time]{utils_is_time()}},
\code{\link[=utils_new_time]{utils_new_time()}},
\code{\link[=utils_time_keywords]{utils_time_keywords()}},
\code{\link[=utils_time_keywords_dictionary]{utils_time_keywords_dictionary()}},
\code{\link[=utils_time_keywords_translate]{utils_time_keywords_translate()}},
\code{\link[=utils_time_units]{utils_time_units()}}
}
\concept{internal_time_handling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psi_distance_lock_step.R
\name{psi_distance_lock_step}
\alias{psi_distance_lock_step}
\title{Lock-Step Distance}
\usage{
psi_distance_lock_step(x = NULL, y = NULL, distance = "euclidean")
}
\arguments{
\item{x}{(required, zoo object or numeric matrix) a time series with no NAs. Default: NULL}

\item{y}{(zoo object or numeric matrix) a time series with the same columns as \code{x} and no NAs. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}
}
\value{
numeric
}
\description{
Demonstration function to compute the lock-step distance between two univariate or multivariate time series.

This function does not accept NA data in the matrices \code{x} and \code{y}.
}
\examples{
#distance metric
d <- "euclidean"

#simulate two time series
#of the same length
x <- zoo_simulate(
  name = "x",
  rows = 100,
  seasons = 2,
  seed = 1
)

y <- zoo_simulate(
  name = "y",
  rows = 100,
  seasons = 2,
  seed = 2
)

if(interactive()){
  zoo_plot(x = x)
  zoo_plot(x = y)
}

#sum of distances
#between pairs of samples
psi_distance_lock_step(
  x = x,
  y = y,
  distance = d
)
}
\seealso{
Other psi_demo:
\code{\link[=psi_auto_distance]{psi_auto_distance()}},
\code{\link[=psi_auto_sum]{psi_auto_sum()}},
\code{\link[=psi_cost_matrix]{psi_cost_matrix()}},
\code{\link[=psi_cost_path]{psi_cost_path()}},
\code{\link[=psi_cost_path_sum]{psi_cost_path_sum()}},
\code{\link[=psi_distance_matrix]{psi_distance_matrix()}},
\code{\link[=psi_equation]{psi_equation()}}
}
\concept{psi_demo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psi_cost_path.R
\name{psi_cost_path}
\alias{psi_cost_path}
\title{Least Cost Path}
\usage{
psi_cost_path(
  dist_matrix = NULL,
  cost_matrix = NULL,
  diagonal = TRUE,
  bandwidth = 1
)
}
\arguments{
\item{dist_matrix}{(required, numeric matrix) Distance matrix generated by \code{\link[=psi_distance_matrix]{psi_distance_matrix()}}. Default: NULL}

\item{cost_matrix}{(required, numeric matrix) Cost matrix generated from the distance matrix with \code{\link[=psi_cost_matrix]{psi_cost_matrix()}}. Default: NULL}

\item{diagonal}{(optional, logical vector). If TRUE, diagonals are included in the dynamic time warping computation. Default: TRUE}

\item{bandwidth}{(optional, numeric) Proportion of space at each side of the cost matrix diagonal (aka \emph{Sakoe-Chiba band}) defining a valid region for dynamic time warping, used to control the flexibility of the warping path. This method prevents degenerate alignments due to differences in magnitude between time series when the data is not properly scaled. If \code{1} (default), DTW is unconstrained. If \code{0}, DTW is fully constrained and the warping path follows the matrix diagonal. Recommended values may vary depending on the nature of the data. Ignored if \code{lock_step = TRUE}. Default: 1.}
}
\value{
data frame
}
\description{
Demonstration function to compute the least cost path within a cost matrix
}
\examples{
#distance metric
d <- "euclidean"

#simulate two irregular time series
x <- zoo_simulate(
  name = "x",
  rows = 100,
  seasons = 2,
  seed = 1
)

y <- zoo_simulate(
  name = "y",
  rows = 80,
  seasons = 2,
  seed = 2
)

if(interactive()){
  zoo_plot(x = x)
  zoo_plot(x = y)
}

#distance matrix
dist_matrix <- psi_distance_matrix(
  x = x,
  y = y,
  distance = d
)

#diagonal least cost path
#------------------------

cost_matrix <- psi_cost_matrix(
  dist_matrix = dist_matrix,
  diagonal = TRUE
)

cost_path <- psi_cost_path(
  dist_matrix = dist_matrix,
  cost_matrix = cost_matrix,
  diagonal = TRUE
)

if(interactive()){
  utils_matrix_plot(
    m = cost_matrix,
    path = cost_path
    )
}


#orthogonal least cost path
#--------------------------
cost_matrix <- psi_cost_matrix(
  dist_matrix = dist_matrix,
  diagonal = FALSE
)

cost_path <- psi_cost_path(
  dist_matrix = dist_matrix,
  cost_matrix = cost_matrix,
  diagonal = FALSE
)

if(interactive()){
  utils_matrix_plot(
    m = cost_matrix,
    path = cost_path
  )
}
}
\seealso{
Other psi_demo:
\code{\link[=psi_auto_distance]{psi_auto_distance()}},
\code{\link[=psi_auto_sum]{psi_auto_sum()}},
\code{\link[=psi_cost_matrix]{psi_cost_matrix()}},
\code{\link[=psi_cost_path_sum]{psi_cost_path_sum()}},
\code{\link[=psi_distance_lock_step]{psi_distance_lock_step()}},
\code{\link[=psi_distance_matrix]{psi_distance_matrix()}},
\code{\link[=psi_equation]{psi_equation()}}
}
\concept{psi_demo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{honeycomb_polygons}
\alias{honeycomb_polygons}
\title{Hexagonal Grid}
\format{
An object of class \code{sf} (inherits from \code{data.frame}) with 72 rows and 2 columns.
}
\usage{
data(honeycomb_polygons)
}
\description{
Sf data frame with hexagonal grid of the dataset \link{honeycomb_climate}.
}
\seealso{
Other example_data:
\code{\link{albatross}},
\code{\link{cities_coordinates}},
\code{\link{cities_temperature}},
\code{\link{covid_counties}},
\code{\link{covid_prevalence}},
\code{\link{eemian_coordinates}},
\code{\link{eemian_pollen}},
\code{\link{fagus_coordinates}},
\code{\link{fagus_dynamics}},
\code{\link{honeycomb_climate}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/momentum_stats.R
\name{momentum_stats}
\alias{momentum_stats}
\title{Summary Statistics of Importance Data Frame}
\usage{
momentum_stats(df = NULL)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=momentum]{momentum()}}, \code{\link[=momentum_ls]{momentum_ls()}}, or \code{\link[=momentum_dtw]{momentum_dtw()}}. Default: NULL}
}
\value{
data frame
}
\description{
Computes summary statistics from the output of \code{\link[=momentum]{momentum()}}, returning a data frame with one row per variable and summary statistics of its importance scores across all time series pairs.
}
\examples{
tsl <- tsl_simulate(
  n = 5,
  irregular = FALSE
  )

df <- distantia(
  tsl = tsl,
  lock_step = TRUE
  )

df_stats <- distantia_stats(df = df)

df_stats

}
\seealso{
Other momentum_support:
\code{\link[=momentum_aggregate]{momentum_aggregate()}},
\code{\link[=momentum_boxplot]{momentum_boxplot()}},
\code{\link[=momentum_model_frame]{momentum_model_frame()}},
\code{\link[=momentum_spatial]{momentum_spatial()}},
\code{\link[=momentum_to_wide]{momentum_to_wide()}}
}
\concept{momentum_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_clean_names.R
\name{utils_clean_names}
\alias{utils_clean_names}
\title{Clean Character Vector of Names}
\usage{
utils_clean_names(
  x = NULL,
  lowercase = FALSE,
  separator = "_",
  capitalize_first = FALSE,
  capitalize_all = FALSE,
  length = NULL,
  suffix = NULL,
  prefix = NULL
)
}
\arguments{
\item{x}{(required, character vector) Names to be cleaned. Default: NULL.}

\item{lowercase}{(optional, logical) If TRUE, all names are coerced to lowercase. Default: FALSE.}

\item{separator}{(optional, character string) Separator when replacing spaces and dots and appending \code{suffix} and \code{prefix} to the main word. Default: "_".}

\item{capitalize_first}{(optional, logical) Capitalize the first letter of each name. Default: FALSE.}

\item{capitalize_all}{(optional, logical) Capitalize all letters of each name. Default: FALSE.}

\item{length}{(optional, integer) Minimum length of abbreviated names. Names are abbreviated via \code{\link[=abbreviate]{abbreviate()}}. Default: NULL.}

\item{suffix}{(optional, character string) String to append to the cleaned names. Default: NULL.}

\item{prefix}{(optional, character string) String to prepend to the cleaned names. Default: NULL.}
}
\value{
A character vector.
}
\description{
Clean and format character vectors for use as column names or variable names.
}
\details{
The cleanup operations are applied in the following order:
\itemize{
\item Remove leading and trailing whitespaces.
\item Generate syntactically valid names with \code{\link[base:make.names]{base::make.names()}}.
\item Replace dots and spaces with the \code{separator}.
\item Coerce names to lowercase.
\item If argument \code{length} is provided, \code{\link[base:abbreviate]{base::abbreviate()}} is used to abbreviate the new column names.
\item If \code{suffix} is provided, it is added at the end of the column name using the separator.
\item If \code{prefix} is provided, it is added at the beginning of the column name using the separator.
\item If \code{capitalize_first = TRUE}, the first letter is capitalized.
\item If \code{capitalize_all = TRUE}, all letters are capitalized.
}
}
\examples{
x <- c(
  "GerMany",
  "spain",
  "SWEDEN"
)

#abbreviate names
#---------------------------
#abbreviate to 4 characters
utils_clean_names(
  x = x,
  capitalize_all = TRUE,
  length = 4
)

#suffix and prefix
#---------------------------
utils_clean_names(
  x = x,
  capitalize_first = TRUE,
  separator = "_",
  prefix = "my_prefix",
  suffix = "my_suffix"
)
}
\seealso{
Other internal:
\code{\link[=utils_boxplot_common]{utils_boxplot_common()}},
\code{\link[=utils_check_args_distantia]{utils_check_args_distantia()}},
\code{\link[=utils_check_args_matrix]{utils_check_args_matrix()}},
\code{\link[=utils_check_args_momentum]{utils_check_args_momentum()}},
\code{\link[=utils_check_args_path]{utils_check_args_path()}},
\code{\link[=utils_check_args_tsl]{utils_check_args_tsl()}},
\code{\link[=utils_check_args_zoo]{utils_check_args_zoo()}},
\code{\link[=utils_check_distance_args]{utils_check_distance_args()}},
\code{\link[=utils_check_list_class]{utils_check_list_class()}},
\code{\link[=utils_digits]{utils_digits()}},
\code{\link[=utils_distantia_df_split]{utils_distantia_df_split()}},
\code{\link[=utils_prepare_df]{utils_prepare_df()}},
\code{\link[=utils_prepare_matrix]{utils_prepare_matrix()}},
\code{\link[=utils_prepare_matrix_list]{utils_prepare_matrix_list()}},
\code{\link[=utils_prepare_time]{utils_prepare_time()}},
\code{\link[=utils_prepare_vector_list]{utils_prepare_vector_list()}},
\code{\link[=utils_prepare_zoo_list]{utils_prepare_zoo_list()}},
\code{\link[=utils_tsl_pairs]{utils_tsl_pairs()}}
}
\concept{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_time_keywords_dictionary.R
\name{utils_time_keywords_dictionary}
\alias{utils_time_keywords_dictionary}
\title{Dictionary of Time Keywords}
\usage{
utils_time_keywords_dictionary()
}
\value{
data frame
}
\description{
Called by \code{\link[=utils_time_keywords_translate]{utils_time_keywords_translate()}} to generate a data frame that helps translate misnamed or abbreviated time keywords, like "day", "daily", or "d", into correct ones such as "days".
}
\examples{
df <- utils_time_keywords_dictionary()
}
\seealso{
Other internal_time_handling:
\code{\link[=utils_as_time]{utils_as_time()}},
\code{\link[=utils_coerce_time_class]{utils_coerce_time_class()}},
\code{\link[=utils_is_time]{utils_is_time()}},
\code{\link[=utils_new_time]{utils_new_time()}},
\code{\link[=utils_time_keywords]{utils_time_keywords()}},
\code{\link[=utils_time_keywords_translate]{utils_time_keywords_translate()}},
\code{\link[=utils_time_units]{utils_time_units()}}
}
\concept{internal_time_handling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_tsl_pairs.R
\name{utils_tsl_pairs}
\alias{utils_tsl_pairs}
\title{Data Frame with Pairs of Time Series in Time Series Lists}
\usage{
utils_tsl_pairs(tsl = NULL, args_list = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL.}

\item{args_list}{(required, list) List of arguments to combine with the pairs of time series. Default: NULL.}
}
\value{
A data frame.
}
\description{
Internal function used in \code{\link[=distantia]{distantia()}} and \code{\link[=momentum]{momentum()}} to generate a data frame with combinations of time series and function arguments.
}
\seealso{
Other internal:
\code{\link[=utils_boxplot_common]{utils_boxplot_common()}},
\code{\link[=utils_check_args_distantia]{utils_check_args_distantia()}},
\code{\link[=utils_check_args_matrix]{utils_check_args_matrix()}},
\code{\link[=utils_check_args_momentum]{utils_check_args_momentum()}},
\code{\link[=utils_check_args_path]{utils_check_args_path()}},
\code{\link[=utils_check_args_tsl]{utils_check_args_tsl()}},
\code{\link[=utils_check_args_zoo]{utils_check_args_zoo()}},
\code{\link[=utils_check_distance_args]{utils_check_distance_args()}},
\code{\link[=utils_check_list_class]{utils_check_list_class()}},
\code{\link[=utils_clean_names]{utils_clean_names()}},
\code{\link[=utils_digits]{utils_digits()}},
\code{\link[=utils_distantia_df_split]{utils_distantia_df_split()}},
\code{\link[=utils_prepare_df]{utils_prepare_df()}},
\code{\link[=utils_prepare_matrix]{utils_prepare_matrix()}},
\code{\link[=utils_prepare_matrix_list]{utils_prepare_matrix_list()}},
\code{\link[=utils_prepare_time]{utils_prepare_time()}},
\code{\link[=utils_prepare_vector_list]{utils_prepare_vector_list()}},
\code{\link[=utils_prepare_zoo_list]{utils_prepare_zoo_list()}}
}
\concept{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_subset.R
\name{tsl_subset}
\alias{tsl_subset}
\title{Subset Time Series Lists by Time Series Names, Time, and/or Column Names}
\usage{
tsl_subset(
  tsl = NULL,
  names = NULL,
  colnames = NULL,
  time = NULL,
  numeric_cols = TRUE,
  shared_cols = TRUE
)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{names}{(optional, character or numeric vector) Character vector of names or numeric vector with list indices. If NULL, all time series are kept. Default: NULL}

\item{colnames}{(optional, character vector) Column names of the zoo objects in \code{tsl}. If NULL, all columns are returned. Default: NULL}

\item{time}{(optional, numeric vector) time vector of length two used to subset rows by time. If NULL, all rows in \code{tsl} are preserved. Default: NULL}

\item{numeric_cols}{(optional, logical) If TRUE, only the numeric columns of the zoo objects are returned. Default: TRUE}

\item{shared_cols}{(optional, logical) If TRUE, only columns shared across all zoo objects are returned. Default: TRUE}
}
\value{
time series list
}
\description{
Subset Time Series Lists by Time Series Names, Time, and/or Column Names
}
\examples{
#initialize time series list
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
)

#checking available dimensions

#names
tsl_names_get(
  tsl = tsl
)

#colnames
tsl_colnames_get(
  tsl = tsl
)

#time
tsl_time(
  tsl = tsl
)[, c("name", "begin", "end")]

#subset
tsl_new <- tsl_subset(
  tsl = tsl,
  names = c("Sweden", "Germany"),
  colnames = c("rainfall", "temperature"),
  time = c("2010-01-01", "2015-01-01")
)

#check new dimensions

#names
tsl_names_get(
  tsl = tsl_new
)

#colnames
tsl_colnames_get(
  tsl = tsl_new
)

#time
tsl_time(
  tsl = tsl_new
)[, c("name", "begin", "end")]
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{permute_restricted_cpp}
\alias{permute_restricted_cpp}
\title{(C++) Restricted Permutation of Cases Within Blocks}
\usage{
permute_restricted_cpp(x, block_size, seed = 1L)
}
\arguments{
\item{x}{(required, numeric matrix). Numeric matrix to permute.}

\item{block_size}{(optional, integer) block size in number of rows.
Minimum value is 2, and maximum value is nrow(x).}

\item{seed}{(optional, integer) random seed to use.}
}
\value{
numeric matrix
}
\description{
Divides a sequence or time series in blocks and permutes cases
within these blocks. This function does not preserve rows, and should not be
used if the sequence has dependent columns.
Larger block sizes increasingly disrupt the data structure over time.
}
\seealso{
Other Rcpp_permutation:
\code{\link[=permute_free_by_row_cpp]{permute_free_by_row_cpp()}},
\code{\link[=permute_free_cpp]{permute_free_cpp()}},
\code{\link[=permute_restricted_by_row_cpp]{permute_restricted_by_row_cpp()}}
}
\concept{Rcpp_permutation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_ncol.R
\name{tsl_ncol}
\alias{tsl_ncol}
\title{Get Number of Columns in Time Series Lists}
\usage{
tsl_ncol(tsl = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}
}
\value{
list
}
\description{
Get Number of Columns in Time Series Lists
}
\examples{
#Initialize time series list
tsl <- tsl_simulate(
  n = 2,
  cols = 6
)

#Number of columns per zoo object
tsl_ncol(tsl = tsl)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_stats.R
\name{distantia_stats}
\alias{distantia_stats}
\title{Summary Statistics of Dissimilarity Data Frame}
\usage{
distantia_stats(df = NULL)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=distantia]{distantia()}}, \code{\link[=distantia_ls]{distantia_ls()}}, \code{\link[=distantia_dtw]{distantia_dtw()}}, or \code{\link[=distantia_time_delay]{distantia_time_delay()}}. Default: NULL}
}
\value{
data frame
}
\description{
Computes summary statistics from the output of \code{\link[=distantia]{distantia()}}, returning a data frame with one row per time series and summary statistics of its dissimilarity scores with all other time series.
}
\examples{
tsl <- tsl_simulate(
  n = 5,
  irregular = FALSE
  )

df <- distantia(
  tsl = tsl,
  lock_step = TRUE
  )

df_stats <- distantia_stats(df = df)

df_stats
}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_handle_NA.R
\name{tsl_handle_NA}
\alias{tsl_handle_NA}
\alias{tsl_Inf_to_NA}
\alias{tsl_NaN_to_NA}
\title{Handle NA Cases in Time Series Lists}
\usage{
tsl_handle_NA(tsl = NULL, na_action = c("impute", "omit"))

tsl_Inf_to_NA(tsl = NULL)

tsl_NaN_to_NA(tsl = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{na_action}{(required, character) NA handling action. Available options are:
\itemize{
\item "impute" (default): NA cases are interpolated from neighbors as a function of time (see \code{\link[zoo:na.approx]{zoo::na.approx()}} and \code{\link[zoo:na.spline]{zoo::na.spline()}}).
\item "omit": rows with NA cases are removed.
}}
}
\value{
time series list
}
\description{
Removes or imputes NA cases in time series lists. Imputation is done via interpolation against time via \code{\link[zoo:na.approx]{zoo::na.approx()}}, and if there are still leading or trailing NA cases after NA interpolation, then \code{\link[zoo:na.spline]{zoo::na.spline()}} is applied as well to fill these gaps. Interpolated values are forced to fall within the observed data range.

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{

#tsl with NA cases
tsl <- tsl_simulate(
  na_fraction = 0.25
)

tsl_count_NA(tsl = tsl)

if(interactive()){
  #issues warning
  tsl_plot(tsl = tsl)
}

#omit NA (default)
#--------------------------------------

#original row count
tsl_nrow(tsl = tsl)

#remove rows with NA
tsl_no_na <- tsl_handle_NA(
  tsl = tsl,
  na_action = "omit"
)

#count rows again
#large data loss in this case!
tsl_nrow(tsl = tsl_no_na)

#count NA again
tsl_count_NA(tsl = tsl_no_na)

if(interactive()){
  tsl_plot(tsl = tsl_no_na)
}


#impute NA with zoo::na.approx
#--------------------------------------

#impute NA cases
tsl_no_na <- tsl_handle_NA(
  tsl = tsl,
  na_action = "impute"
)

#count rows again
#large data loss in this case!
tsl_nrow(tsl = tsl_no_na)

if(interactive()){
  tsl_plot(tsl = tsl_no_na)
}
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_time_units.R
\name{utils_time_units}
\alias{utils_time_units}
\title{Data Frame with Supported Time Units}
\usage{
utils_time_units(all_columns = FALSE, class = NULL)
}
\arguments{
\item{all_columns}{(optional, logical) If TRUE, all columns are returned. Default: FALSE}

\item{class}{(optional, class name). Used to filter rows and columns. Accepted values are "numeric", "Date", and "POSIXct". Default: NULL}
}
\value{
data frame
}
\description{
Returns a data frame with the names of the supported time units, the classes that can handle each time unit, and the threshold used to identify which time units can be used when aggregating a time series.
}
\examples{

df <- utils_time_units()
head(df)
}
\seealso{
Other internal_time_handling:
\code{\link[=utils_as_time]{utils_as_time()}},
\code{\link[=utils_coerce_time_class]{utils_coerce_time_class()}},
\code{\link[=utils_is_time]{utils_is_time()}},
\code{\link[=utils_new_time]{utils_new_time()}},
\code{\link[=utils_time_keywords]{utils_time_keywords()}},
\code{\link[=utils_time_keywords_dictionary]{utils_time_keywords_dictionary()}},
\code{\link[=utils_time_keywords_translate]{utils_time_keywords_translate()}}
}
\concept{internal_time_handling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{distance_matrix_cpp}
\alias{distance_matrix_cpp}
\title{(C++) Distance Matrix of Two Time Series}
\usage{
distance_matrix_cpp(x, y, distance = "euclidean")
}
\arguments{
\item{x}{(required, numeric matrix) univariate or multivariate time series.}

\item{y}{(required, numeric matrix) univariate or multivariate time series
with the same number of columns as 'x'.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}
}
\value{
numeric matrix
}
\description{
Computes the distance matrix between the rows of two matrices
\code{y} and \code{x} representing regular or irregular time series with the same number of
columns. NA values should be removed before using this function. If the selected distance function is "chi" or "cosine", pairs of zeros should
be either removed or replaced with pseudo-zeros (i.e. 0.00001).
}
\examples{
#simulate two time series
x <- zoo_simulate(seed = 1)
y <- zoo_simulate(seed = 2)

#distance matrix
dist_matrix <- distance_matrix_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)
}
\seealso{
Other Rcpp_matrix:
\code{\link[=cost_matrix_diagonal_cpp]{cost_matrix_diagonal_cpp()}},
\code{\link[=cost_matrix_diagonal_weighted_cpp]{cost_matrix_diagonal_weighted_cpp()}},
\code{\link[=cost_matrix_orthogonal_cpp]{cost_matrix_orthogonal_cpp()}},
\code{\link[=distance_ls_cpp]{distance_ls_cpp()}}
}
\concept{Rcpp_matrix}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_resample.R
\name{tsl_resample}
\alias{tsl_resample}
\title{Resample Time Series Lists to a New Time}
\usage{
tsl_resample(
  tsl = NULL,
  new_time = NULL,
  method = "linear",
  max_complexity = FALSE
)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{new_time}{(required, zoo object, time series list, character string, time vector, numeric) New time to resample to. If a time vector is provided, it must be of a class compatible with the time of \code{tsl}.  If a zoo object or time series list is provided, its time is used as a template to resample \code{tsl}. Valid resampling keywords (see \code{\link[=tsl_time]{tsl_time()}}) are allowed. Numeric values are interpreted as interval widths in the time units of the time series. If NULL, irregular time series are predicted into a regular version of their own time. Default: NULL}

\item{method}{(optional, character string) Name of the method to resample the time series. One of "linear", "spline" or "loess". Default: "linear".}

\item{max_complexity}{(required, logical). Only relevant for methods "spline" and "loess". If TRUE, model optimization is ignored, and the a model of maximum complexity (an overfitted model) is used for resampling. Default: FALSE}
}
\value{
time series list
}
\description{
\strong{Objective}

Time series resampling interpolates new values for time steps not available in the original time series. This operation is useful to:
\itemize{
\item Transform irregular time series into regular.
\item Align time series with different temporal resolutions.
\item Increase (upsampling) or decrease (downsampling) the temporal resolution of a time series.
}

Time series resampling \strong{should not be used} to extrapolate new values outside of the original time range of the time series, or to increase the resolution of a time series by a factor of two or more. These operations are known to produce non-sensical results.

\strong{Warning}: This function resamples time series lists \strong{with overlapping times}. Please check such overlap by assessing the columns "begin" and "end" of the data frame resulting from \code{df <- tsl_time(tsl = tsl)}. Resampling will be limited by the shortest time series in your time series list. To resample non-overlapping time series, please subset the individual components of \code{tsl} one by one either using \code{\link[=tsl_subset]{tsl_subset()}} or the syntax \code{tsl = my_tsl[[i]]}.

\strong{Methods}

This function offers three methods for time series interpolation:

\itemize{
\item "linear" (default): interpolation via piecewise linear regression as implemented in \code{\link[zoo:na.approx]{zoo::na.approx()}}.
\item "spline": cubic smoothing spline regression as implemented in \code{\link[stats:smooth.spline]{stats::smooth.spline()}}.
\item "loess": local polynomial regression fitting as implemented in \code{\link[stats:loess]{stats::loess()}}.
}

These methods are used to fit models \code{y ~ x} where \code{y} represents the values of a univariate time series and \code{x} represents a numeric version of its time.

The functions \code{\link[=utils_optimize_spline]{utils_optimize_spline()}} and \code{\link[=utils_optimize_loess]{utils_optimize_loess()}} are used under the hood to optimize the complexity of the methods "spline" and "loess" by finding the configuration that minimizes the root mean squared error (RMSE) between  observed and predicted \code{y}. However, when the argument \code{max_complexity = TRUE}, the complexity optimization is ignored, and a maximum complexity model is used instead.

\strong{New time}

The argument \code{new_time} offers several alternatives to help define the new time of the resulting time series:

\itemize{
\item \code{NULL}: the target time series (\code{x}) is resampled to a regular time within its original time range and number of observations.
\item \verb{zoo object}: a zoo object to be used as template for resampling. Useful when the objective is equalizing the frequency of two separate zoo objects.
\item \verb{time series list}: a time series list to be used as template. The range of overlapping dates and the average resolution are used to generate the new resampling time. This method cannot be used to align two time series lists, unless the template is resampled beforehand.
\item \verb{time vector}: a time vector of a class compatible with the time in \code{x}.
\item \code{keyword}: character string defining a resampling keyword, obtained via \code{zoo_time(x, keywords = "resample")$keywords}.
\item \code{numeric}: a single number representing the desired interval between consecutive samples in the units of \code{x} (relevant units can be obtained via \code{zoo_time(x)$units}).
}

\strong{Step by Step}

The steps to resample a time series list are:

\enumerate{
\item The time interpolation range is computed from the intersection of all times in \code{tsl}. This step ensures that no extrapolation occurs during resampling, but it also makes resampling of non-overlapping time series impossible.
\item If \code{new_time} is provided, any values of \code{new_time} outside of the minimum and maximum interpolation times are removed to avoid extrapolation. If \code{new_time} is not provided, a regular time within the interpolation time range with the length of the shortest time series in \code{tsl} is generated.
\item For each univariate time series, a model \code{y ~ x}, where \code{y} is the time series and \code{x} is its own time coerced to numeric is fitted.
\itemize{
\item If \code{max_complexity == FALSE}, the model with the complexity that minimizes the root mean squared error between the observed and predicted \code{y} is returned.
\item If \code{max_complexity == TRUE} and \code{method = "spline"} or \code{method = "loess"}, the first valid model closest to a maximum complexity is returned.

}
\item The fitted model is predicted over \code{new_time} to generate the resampled time series.
}

\strong{Other Details}

Please use this operation with care, as there are limits to the amount of resampling that can be done without distorting the data. The safest option is to keep the distance between new time points within the same magnitude of the distance between the old time points.

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{

#generate irregular time series
tsl <- tsl_simulate(
  n = 2,
  rows = 100,
  irregular = TRUE
)

if(interactive()){
  tsl_plot(tsl)
}


#range of times between samples
tsl_time_summary(tsl)[
  c(
    "units",
    "resolution_min",
    "resolution_max"
    )
  ]

#resample to regular using linear interpolation
tsl_regular <- tsl_resample(
  tsl = tsl
)

if(interactive()){
  tsl_plot(tsl_regular)
}

#check new resolution
tsl_time_summary(tsl_regular)[
  c(
    "units",
    "resolution_min",
    "resolution_max"
  )
]

#resample using keywords

#valid resampling keywords
tsl_time_summary(
  tsl = tsl,
  keywords = "resample"
)$keywords

#by month
tsl_months <- tsl_resample(
  tsl = tsl,
  new_time = "months"
)

if(interactive()){
  tsl_plot(tsl_months)
}

#by week
tsl_weeks <- tsl_resample(
  tsl = tsl,
  new_time = "weeks"
)

if(interactive()){
  tsl_plot(tsl_weeks)
}

#resample using time interval

#get relevant units
tsl_time(tsl)$units

#resampling to 15 days intervals
tsl_15_days <- tsl_resample(
  tsl = tsl,
  new_time = 15 #days
)

tsl_time_summary(tsl_15_days)[
  c(
    "units",
    "resolution_min",
    "resolution_max"
  )
]

if(interactive()){
  tsl_plot(tsl_15_days)
}

#Aligning two time series lists

#two time series lists with different time ranges
tsl1 <- tsl_simulate(
  n = 2,
  rows = 80,
  time_range = c("2010-01-01", "2020-01-01"),
  irregular = TRUE
)

tsl2 <- tsl_simulate(
  n = 2,
  rows = 120,
  time_range = c("2005-01-01", "2024-01-01"),
  irregular = TRUE
)

#check time features
tsl_time_summary(tsl1)[
  c(
    "begin",
    "end",
    "resolution_min",
    "resolution_max"
  )
]

tsl_time_summary(tsl2)[
  c(
    "begin",
    "end",
    "resolution_min",
    "resolution_max"
  )
]

#tsl1 to regular
tsl1_regular <- tsl_resample(
  tsl = tsl1
)

#tsl2 resampled to time of tsl1_regular
tsl2_regular <- tsl_resample(
  tsl = tsl2,
  new_time = tsl1_regular
)

#check alignment
tsl_time_summary(tsl1_regular)[
  c(
    "begin",
    "end",
    "resolution_min",
    "resolution_max"
  )
]

tsl_time_summary(tsl2_regular)[
  c(
    "begin",
    "end",
    "resolution_min",
    "resolution_max"
  )
]

}
\seealso{
\code{\link[=zoo_resample]{zoo_resample()}}

Other tsl_processing:
\code{\link[=tsl_aggregate]{tsl_aggregate()}},
\code{\link[=tsl_smooth]{tsl_smooth()}},
\code{\link[=tsl_stats]{tsl_stats()}},
\code{\link[=tsl_transform]{tsl_transform()}}
}
\concept{tsl_processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{f_rescale_local}
\alias{f_rescale_local}
\title{Data Transformation: Local Rescaling to a New Range}
\usage{
f_rescale_local(
  x = NULL,
  new_min = 0,
  new_max = 1,
  old_min = NULL,
  old_max = NULL,
  ...
)
}
\arguments{
\item{x}{(required, zoo object) Zoo time series object to transform.}

\item{new_min}{(optional, numeric) New minimum value. Default: \code{0}}

\item{new_max}{(optional, numeric) New maximum value. Default: \code{1}}

\item{old_min}{(optional, numeric) Old minimum value. Default: \code{NULL}}

\item{old_max}{(optional, numeric) Old maximum value. Default: \code{NULL}}

\item{...}{(optional, additional arguments) Ignored in this function.}
}
\value{
zoo object
}
\description{
Data Transformation: Local Rescaling to a New Range
}
\examples{
x <- zoo_simulate(cols = 2)

y <- f_rescale_global(
  x = x,
  new_min = 0,
  new_max = 100
)

if(interactive()){
  zoo_plot(x)
  zoo_plot(y)
}
}
\seealso{
Other tsl_transformation:
\code{\link[=f_binary]{f_binary()}},
\code{\link[=f_clr]{f_clr()}},
\code{\link[=f_detrend_difference]{f_detrend_difference()}},
\code{\link[=f_detrend_linear]{f_detrend_linear()}},
\code{\link[=f_detrend_poly]{f_detrend_poly()}},
\code{\link[=f_hellinger]{f_hellinger()}},
\code{\link[=f_list]{f_list()}},
\code{\link[=f_log]{f_log()}},
\code{\link[=f_percent]{f_percent()}},
\code{\link[=f_proportion]{f_proportion()}},
\code{\link[=f_proportion_sqrt]{f_proportion_sqrt()}},
\code{\link[=f_rescale_global]{f_rescale_global()}},
\code{\link[=f_scale_global]{f_scale_global()}},
\code{\link[=f_scale_local]{f_scale_local()}},
\code{\link[=f_trend_linear]{f_trend_linear()}},
\code{\link[=f_trend_poly]{f_trend_poly()}}
}
\concept{tsl_transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_names.R
\name{zoo_name_clean}
\alias{zoo_name_clean}
\title{Clean Name of a Zoo Time Series}
\usage{
zoo_name_clean(
  x = NULL,
  lowercase = FALSE,
  separator = "_",
  capitalize_first = FALSE,
  capitalize_all = FALSE,
  length = NULL,
  suffix = NULL,
  prefix = NULL
)
}
\arguments{
\item{x}{(required, zoo object) Zoo time series to analyze. Default: NULL.}

\item{lowercase}{(optional, logical) If TRUE, all names are coerced to lowercase. Default: FALSE}

\item{separator}{(optional, character string) Separator when replacing spaces and dots. Also used to separate \code{suffix} and \code{prefix} from the main word. Default: "_".}

\item{capitalize_first}{(optional, logical) Indicates whether to capitalize the first letter of each name. Default: FALSE.}

\item{capitalize_all}{(optional, logical) Indicates whether to capitalize all letters of each name. Default: FALSE.}

\item{length}{(optional, integer) Minimum length of abbreviated names. Names are abbreviated via \code{\link[=abbreviate]{abbreviate()}}. Default: NULL.}

\item{suffix}{(optional, character string) Suffix for the clean names. Default: NULL.}

\item{prefix}{(optional, character string)  Prefix for the clean names. Default: NULL.}
}
\value{
zoo time series
}
\description{
Combines \code{\link[=utils_clean_names]{utils_clean_names()}} and \code{\link[=zoo_name_set]{zoo_name_set()}} to help clean, abbreviate, capitalize, and add a suffix or a prefix to the name of a zoo object.
}
\examples{
#simulate zoo time series
x <- zoo_simulate()

#get current name
zoo_name_get(x = x)

#change name
x <- zoo_name_set(
  x = x,
  name = "My.New.name"
)

zoo_name_get(x = x)

#clean name
x <- zoo_name_clean(
  x = x,
  lowercase = TRUE
)

zoo_name_get(x = x)
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_time.R
\name{tsl_time}
\alias{tsl_time}
\alias{tsl_time_summary}
\title{Time Features of Time Series Lists}
\usage{
tsl_time(tsl = NULL, keywords = c("resample", "aggregate"))

tsl_time_summary(tsl = NULL, keywords = c("resample", "aggregate"))
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{keywords}{(optional, character string or vector) Defines what keywords are returned. If "aggregate", returns valid keywords for \code{\link[=zoo_aggregate]{zoo_aggregate()}}. If "resample", returns valid keywords for \code{\link[=zoo_resample]{zoo_resample()}}. If both, returns all valid keywords. Default: c("aggregate", "resample").}
}
\value{
\itemize{
\item \code{\link[=tsl_time]{tsl_time()}}: data frame with the following columns:
\itemize{
\item \code{name} (string): time series name.
\item \code{rows} (integer): number of observations.
\item \code{class} (string): time class, one of "Date", "POSIXct", or "numeric"
\item \code{units} (string): units of the time series.
\item \code{length} (numeric): total length of the time series expressed in \code{units}.
\item \code{resolution} (numeric): average interval between observations expressed in \code{units}.
\item \code{begin} (date or numeric): begin time of the time series.
\item \code{end} (date or numeric): end time of the time series.
\item \code{keywords} (character vector): valid keywords for \code{\link[=tsl_aggregate]{tsl_aggregate()}} or \code{\link[=tsl_resample]{tsl_resample()}}, depending on the value of the argument \code{keywords}.
}
\item \code{\link[=tsl_time_summary]{tsl_time_summary()}}: list with the following objects:
\itemize{
\item \code{class} (string): time class, one of "Date", "POSIXct", or "numeric"
\item \code{units} (string): units of the time series.
\item \code{begin} (date or numeric): begin time of the time series.
\item \code{end} (date or numeric): end time of the time series.
\item \code{resolution_max} (numeric): longer time interval between consecutive samples expressed in \code{units}.
\item \code{resolution_min} (numeric): shorter time interval between consecutive samples expressed in \code{units}.
\item \code{keywords} (character vector): valid keywords for \code{\link[=tsl_aggregate]{tsl_aggregate()}} or \code{\link[=tsl_resample]{tsl_resample()}}, depending on the value of the argument \code{keywords}.
\item \code{units_df} (data frame) data frame for internal use within \code{\link[=tsl_aggregate]{tsl_aggregate()}} and \code{\link[=tsl_resample]{tsl_resample()}}.
}
}
}
\description{
The functions \code{\link[=tsl_time]{tsl_time()}} and \code{\link[=tsl_time_summary]{tsl_time_summary()}} summarize the time features of a time series list.
\itemize{
\item \code{\link[=tsl_time]{tsl_time()}} returns a data frame with one row per time series in the argument 'tsl'
\item \code{\link[=tsl_time_summary]{tsl_time_summary()}} returns a list with the features captured by \code{\link[=tsl_time]{tsl_time()}}, but aggregated across time series.
}

Both functions return keywords useful for the functions \code{\link[=tsl_aggregate]{tsl_aggregate()}} and \code{\link[=tsl_resample]{tsl_resample()}}, depending on the value of the argument \code{keywords}.
}
\examples{
#simulate a time series list
tsl <- tsl_simulate(
  n = 3,
  rows = 150,
  time_range = c(
    Sys.Date() - 365,
    Sys.Date()
  ),
  irregular = TRUE
)

#time data frame
tsl_time(
  tsl = tsl
)

#time summary
tsl_time_summary(
  tsl = tsl
)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{auto_sum_cpp}
\alias{auto_sum_cpp}
\title{(C++) Sum Distances Between Consecutive Samples in Two Time Series}
\usage{
auto_sum_cpp(x, y, path, distance = "euclidean", ignore_blocks = FALSE)
}
\arguments{
\item{x}{(required, numeric matrix) of same number of columns as 'y'.}

\item{y}{(required, numeric matrix) of same number of columns as 'x'.}

\item{path}{(required, data frame) output of \code{\link[=cost_path_orthogonal_cpp]{cost_path_orthogonal_cpp()}}.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean"}

\item{ignore_blocks}{(optional, logical). If TRUE, blocks of consecutive path
coordinates are trimmed to avoid inflating the psi distance. Default: FALSE.}
}
\value{
numeric
}
\description{
Sum of auto-distances of two time series.
This function switches between \code{\link[=auto_sum_full_cpp]{auto_sum_full_cpp()}} and \code{\link[=auto_sum_path_cpp]{auto_sum_path_cpp()}}
depending on the value of the argument \code{ignore_blocks}.
}
\examples{
#simulate two time series
x <- zoo_simulate(seed = 1)
y <- zoo_simulate(seed = 2)

#distance matrix
dist_matrix <- distance_matrix_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)

#least cost matrix
cost_matrix <- cost_matrix_orthogonal_cpp(
  dist_matrix = dist_matrix
)

#least cost path
cost_path <- cost_path_orthogonal_cpp(
  dist_matrix = dist_matrix,
  cost_matrix = cost_matrix
)

nrow(cost_path)

#remove blocks from least-cost path
cost_path_trimmed <- cost_path_trim_cpp(
  path = cost_path
)

nrow(cost_path_trimmed)

#auto sum
auto_sum_cpp(
  x = x,
  y = y,
  path = cost_path_trimmed,
  distance = "euclidean",
  ignore_blocks = FALSE
)
}
\seealso{
Other Rcpp_auto_sum:
\code{\link[=auto_distance_cpp]{auto_distance_cpp()}},
\code{\link[=auto_sum_full_cpp]{auto_sum_full_cpp()}},
\code{\link[=auto_sum_path_cpp]{auto_sum_path_cpp()}},
\code{\link[=subset_matrix_by_rows_cpp]{subset_matrix_by_rows_cpp()}}
}
\concept{Rcpp_auto_sum}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_colnames.R
\name{tsl_colnames_prefix}
\alias{tsl_colnames_prefix}
\title{Append Prefix to Column Names of Time Series List}
\usage{
tsl_colnames_prefix(tsl = NULL, prefix = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{prefix}{(optional, character string) String to prepend to the column names. Default: NULL.}
}
\value{
time series list
}
\description{
Append Prefix to Column Names of Time Series List
}
\examples{
tsl <- tsl_simulate()

tsl_colnames_get(tsl = tsl)

tsl <- tsl_colnames_prefix(
  tsl = tsl,
  prefix = "my_prefix_"
)

tsl_colnames_get(tsl = tsl)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_cluster_kmeans_optimizer.R
\name{utils_cluster_kmeans_optimizer}
\alias{utils_cluster_kmeans_optimizer}
\title{Optimize the Silhouette Width of K-Means Clustering Solutions}
\usage{
utils_cluster_kmeans_optimizer(d = NULL, seed = 1)
}
\arguments{
\item{d}{(required, matrix) distance matrix typically resulting from \code{\link[=distantia_matrix]{distantia_matrix()}}, but any other square matrix should work. Default: NULL}

\item{seed}{(optional, integer) Random seed to be used during the K-means computation. Default: 1}
}
\value{
data frame
}
\description{
Generates k-means solutions from 2 to \code{nrow(d) - 1} number of clusters and returns the number of clusters with the highest silhouette width median. See \code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}} for more details.

This function supports parallelization via \code{\link[future:plan]{future::plan()}} and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{

#weekly covid prevalence
#in 10 California counties
#aggregated by month
tsl <- tsl_initialize(
  x = covid_prevalence,
  name_column = "name",
  time_column = "time"
) |>
  tsl_subset(
    names = 1:10
  ) |>
  tsl_aggregate(
    new_time = "months",
    fun = max
  )

if(interactive()){
  #plotting first three time series
  tsl_plot(
    tsl = tsl_subset(
      tsl = tsl,
      names = 1:3
    ),
    guide_columns = 3
  )
}

#compute dissimilarity matrix
psi_matrix <- distantia(
  tsl = tsl,
  lock_step = TRUE
) |>
  distantia_matrix()

#optimize hierarchical clustering
kmeans_optimization <- utils_cluster_kmeans_optimizer(
  d = psi_matrix
)

#best solution in first row
head(kmeans_optimization)

}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_names.R
\name{tsl_names_set}
\alias{tsl_names_set}
\title{Set Time Series Names in a Time Series List}
\usage{
tsl_names_set(tsl = NULL, names = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{names}{(optional, character vector) names to set. Must be of the same length of \code{x}. If NULL, and the list \code{x} has names, then the names of the zoo objects inside of the list are taken from the names of the list elements.}
}
\value{
time series list
}
\description{
Sets the names of a time series list and the internal names of the zoo objects inside, stored in their attribute "name".
}
\examples{
#simulate time series list
tsl <- tsl_simulate(n = 3)

#assess validity
tsl_diagnose(
  tsl = tsl
)

#list and zoo names (default)
tsl_names_get(
  tsl = tsl
)

#list names
tsl_names_get(
  tsl = tsl,
  zoo = FALSE
)

#renaming list items and zoo objects
#------------------------------------
tsl <- tsl_names_set(
  tsl = tsl,
  names = c("X", "Y", "Z")
)

# check new names
tsl_names_get(
  tsl = tsl
)

#fixing naming issues
#------------------------------------

#creating a invalid time series list
names(tsl)[2] <- "B"

# check names
tsl_names_get(
  tsl = tsl
)

#validate tsl
#returns NOT VALID
#recommends a solution
tsl_diagnose(
  tsl = tsl
)

#fix issue with tsl_names_set()
#uses names of zoo objects for the list items
tsl <- tsl_names_set(
  tsl = tsl
)

#validate again
tsl_diagnose(
  tsl = tsl
)

#list names
tsl_names_get(
  tsl = tsl
)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/momentum_to_wide.R
\name{momentum_to_wide}
\alias{momentum_to_wide}
\title{Momentum Data Frame to Wide Format}
\usage{
momentum_to_wide(df = NULL, sep = "__")
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=momentum]{momentum()}}, \code{\link[=momentum_ls]{momentum_ls()}}, or \code{\link[=momentum_dtw]{momentum_dtw()}}. Default: NULL}

\item{sep}{(required, character string) Separator between the importance metric name and the time series variable name. Default: "__".}
}
\value{
data frame
}
\description{
Transforms a data frame returned by \code{\link[=momentum]{momentum()}} into wide format with the following columns:
\itemize{
\item \code{most_similar}: name of the variable with the highest contribution to similarity (most negative value in the \code{importance} column) for each pair of time series.
\item \code{most_dissimilar}: name of the variable with the highest contribution to dissimilarity (most positive value in the \code{importance} column) for each pair of time series.
\item \code{importance__variable_name}: contribution to similarity (negative values) or dissimilarity (positive values) of the given variable.
\item \code{psi_only_with__variable_name}: dissimilarity of the two time series when only using the given variable.
\item \code{psi_without__variable_name}: dissimilarity of the two time series when removing the given variable.
}
}
\examples{

tsl <- tsl_initialize(
  x = distantia::albatross,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

#importance data frame
df <- momentum(
  tsl = tsl
)

df

#to wide format
df_wide <- momentum_to_wide(
  df = df
)

df_wide

}
\seealso{
Other momentum_support:
\code{\link[=momentum_aggregate]{momentum_aggregate()}},
\code{\link[=momentum_boxplot]{momentum_boxplot()}},
\code{\link[=momentum_model_frame]{momentum_model_frame()}},
\code{\link[=momentum_spatial]{momentum_spatial()}},
\code{\link[=momentum_stats]{momentum_stats()}}
}
\concept{momentum_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia.R
\name{distantia}
\alias{distantia}
\title{Dissimilarity Analysis of Time Series Lists}
\usage{
distantia(
  tsl = NULL,
  distance = "euclidean",
  diagonal = TRUE,
  bandwidth = 1,
  lock_step = FALSE,
  permutation = "restricted_by_row",
  block_size = NULL,
  repetitions = 0,
  seed = 1
)
}
\arguments{
\item{tsl}{(required, time series list) list of zoo time series. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}

\item{diagonal}{(optional, logical vector). If TRUE, diagonals are included in the dynamic time warping computation. Default: TRUE}

\item{bandwidth}{(optional, numeric) Proportion of space at each side of the cost matrix diagonal (aka \emph{Sakoe-Chiba band}) defining a valid region for dynamic time warping, used to control the flexibility of the warping path. This method prevents degenerate alignments due to differences in magnitude between time series when the data is not properly scaled. If \code{1} (default), DTW is unconstrained. If \code{0}, DTW is fully constrained and the warping path follows the matrix diagonal. Recommended values may vary depending on the nature of the data. Ignored if \code{lock_step = TRUE}. Default: 1.}

\item{lock_step}{(optional, logical vector) If TRUE, time series captured at the same times are compared sample wise (with no dynamic time warping). Requires time series in argument \code{tsl} to be fully aligned, or it will return an error. Default: FALSE.}

\item{permutation}{(optional, character vector) permutation method, only relevant when \code{repetitions} is higher than zero. Valid values are: "restricted_by_row", "restricted", "free_by_row", and "free". Default: "restricted_by_row".}

\item{block_size}{(optional, integer) Size of the row blocks for the restricted permutation test. Only relevant when permutation methods are "restricted" or "restricted_by_row" and \code{repetitions} is higher than zero. A block of size \code{n} indicates that a row can only be permuted within a block of \code{n} adjacent rows. If NULL, defaults to the rounded one tenth of the shortest time series in \code{tsl}. Default: NULL.}

\item{repetitions}{(optional, integer vector) number of permutations to compute the p-value. If 0, p-values are not computed. Otherwise, the minimum is 2. The resolution of the p-values and the overall computation time depends on the number of permutations. Default: 0}

\item{seed}{(optional, integer) initial random seed to use for replicability when computing p-values. Default: 1}
}
\value{
data frame with columns:
\itemize{
\item \code{x}: time series name.
\item \code{y}: time series name.
\item \code{distance}: name of the distance metric.
\item \code{diagonal}: value of the argument \code{diagonal}.
\item \code{lock_step}: value of the argument \code{lock_step}.
\item \code{repetitions} (only if \code{repetitions > 0}): value of the argument \code{repetitions}.
\item \code{permutation} (only if \code{repetitions > 0}): name of the permutation method used to compute p-values.
\item \code{seed} (only if \code{repetitions > 0}): random seed used in the permutations.
\item \code{psi}: psi dissimilarity of the sequences \code{x} and \code{y}.
\item \code{null_mean} (only if \code{repetitions > 0}): mean of the null distribution of psi scores.
\item \code{null_sd} (only if \code{repetitions > 0}): standard deviation of the null distribution of psi values.
\item \code{p_value}  (only if \code{repetitions > 0}): proportion of scores smaller or equal than \code{psi} in the null distribution.
}
}
\description{
This function combines \emph{dynamic time warping} or \emph{lock-step comparison} with the \emph{psi dissimilarity score} and \emph{permutation methods} to assess dissimilarity between pairs of time series or any other sort of data composed of events ordered across a relevant dimension.

\strong{Dynamic Time Warping} (DTW) finds the optimal alignment between two time series by minimizing the cumulative distance between their samples. It applies dynamic programming to identify the least-cost path through a distance matrix between all pairs of samples. The resulting sum of distances along the least cost path is a metric of time series similarity. DTW disregards the exact timing of samples and focuses on their order and pattern similarity between time series, making it suitable for comparing both \emph{regular and irregular time series of the same or different lengths}, such as phenological data from different latitudes or elevations, time series from various years or periods, and movement trajectories like migration paths. Additionally, \code{distantia()} implements constrained DTW via Sakoe-Chiba bands with the \code{bandwidth} argument, which defines a region around the distance matrix diagonal to restrict the spread of the least cost path.

\strong{Lock-step} (LS) sums pairwise distances between samples in \emph{regular or irregular time series of the same length}, preferably captured at the same times. This method is an alternative to dynamic time warping when the goal is to assess the synchronicity of two time series.

The \strong{psi score} normalizes the cumulative sum of distances between two time series by the cumulative sum of distances between their consecutive samples to generate a comparable dissimilarity score. If for two time series \eqn{x} and \eqn{y} \eqn{D_xy} represents the cumulative sum of distances between them, either resulting from dynamic time warping or the lock-step method, and \eqn{S_xy} represents the cumulative sum of distances of their consecutive samples, then the psi score can be computed in two ways depending on the scenario:

\strong{Equation 1:} \eqn{\psi = \frac{D_{xy} - S_{xy}}{S_{xy}}}

\strong{Equation 2:} \eqn{\psi = \frac{D_{xy} - S_{xy}}{S_{xy}} + 1}

When $D_xy$ is computed via dynamic time warping \strong{ignoring the distance matrix diagonals} (\code{diagonal = FALSE}), then \emph{Equation 1} is used. On the other hand, if $D_xy$ results from the lock-step method (\code{lock_step = TRUE}), or from dynamic time warping considering diagonals (\code{diagonal = TRUE}), then \emph{Equation 2} is used instead:

In both equations, a psi score of zero indicates maximum similarity.

\strong{Permutation methods} are provided here to help assess the robustness of observed psi scores by direct comparison with a null distribution of psi scores resulting from randomized versions of the compared time series. The fraction of null scores smaller than the observed score is returned as a \emph{p_value} in the function output and interpreted as "the probability of finding a lower psi score (higher similarity) by chance".

In essence, restricted permutation is useful to answer the question "how robust is the similarity between two time series?"

Four different permutation methods are available:
\itemize{
\item \strong{"restricted"}: Separates the data into blocks of contiguous rows, and re-shuffles data points randomly within these blocks, independently by row and column. Applied when the data is structured in blocks that should be preserved during permutations (e.g., "seasons", "years", "decades", etc) and the columns represent independent variables.
\item \strong{"restricted_by_row"}: Separates the data into blocks of contiguous rows, and re-shuffles complete rows within these blocks. This method is suitable for cases where the data is organized into blocks as described above, but columns represent interdependent data (e.g., rows represent percentages or proportions), and maintaining the relationships between data within each row is important.
\item \strong{"free"}: Randomly re-shuffles data points across the entire time series, independently by row and column. This method is useful for loosely structured time series where data independence is assumed. When the data exhibits a strong temporal structure, this approach may lead to an overestimation of the robustness of dissimilarity scores.
\item \strong{"free_by_row"}: Randomly re-shuffles complete rows across the entire time series. This method is useful for loosely structured time series where dependency between columns is assumed (e.g., rows represent percentages or proportions). This method has the same drawbacks as the "free" method, when the data exhibits a strong temporal structure.
}

This function allows computing dissimilarity between pairs of time series using different combinations of arguments at once. For example, when the argument \code{distance} is set to \code{c("euclidean", "manhattan")}, the output data frame will show two dissimilarity scores for each pair of time series, one based on euclidean distances, and another based on manhattan distances. The same happens for most other parameters.

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}. However, due to the high performance of the C++ backend, parallelization might only result in efficiency gains when running permutation tests with large number of iterations, or working with very long time series.
}
\examples{

#parallelization setup
#not worth it for this data size
# future::plan(
#   strategy = future::multisession,
#   workers = 2
# )

#progress bar (does not work in R examples)
# progressr::handlers(global = TRUE)


#load fagus_dynamics as tsl
#global centering and scaling
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

if(interactive()){
  tsl_plot(
    tsl = tsl,
    guide_columns = 3
    )
}

#dynamic time warping dissimilarity analysis
#-------------------------------------------
#permutation restricted by row to preserve dependency of ndvi on temperature and rainfall
#block size is 3 months to permute within same season
df_dtw <- distantia(
  tsl = tsl,
  distance = "euclidean",
  permutation = "restricted_by_row",
  block_size = 3, #months
  repetitions = 10, #increase to 100 or more
  seed = 1
)

#focus on the important details
df_dtw[, c("x", "y", "psi", "p_value", "null_mean", "null_sd")]
#higher psi values indicate higher dissimilarity
#p-values indicate chance of finding a random permutation with a psi smaller than the observed

#visualize dynamic time warping
if(interactive()){

  distantia_dtw_plot(
    tsl = tsl[c("Spain", "Sweden")],
    distance = "euclidean"
  )

}

#recreating the null distribution
#direct call to C++ function
#use same args as distantia() call
psi_null <- psi_null_dtw_cpp(
  x = tsl[["Spain"]],
  y = tsl[["Sweden"]],
  distance = "euclidean",
  repetitions = 10, #increase to 100 or more

  permutation = "restricted_by_row",
  block_size = 3,
  seed = 1
)

#compare null mean with output of distantia()
mean(psi_null)
df_dtw$null_mean[3]

}
\seealso{
Other distantia:
\code{\link[=distantia_dtw]{distantia_dtw()}},
\code{\link[=distantia_dtw_plot]{distantia_dtw_plot()}},
\code{\link[=distantia_ls]{distantia_ls()}}
}
\concept{distantia}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_cluster_hclust_optimizer.R
\name{utils_cluster_hclust_optimizer}
\alias{utils_cluster_hclust_optimizer}
\title{Optimize the Silhouette Width of Hierarchical Clustering Solutions}
\usage{
utils_cluster_hclust_optimizer(d = NULL, method = NULL)
}
\arguments{
\item{d}{(required, matrix) distance matrix typically resulting from \code{\link[=distantia_matrix]{distantia_matrix()}}, but any other square matrix should work. Default: NULL}

\item{method}{(optional, character string) argument of \code{\link[stats:hclust]{stats::hclust()}} defining the agglomerative method. One of: "ward.D", "ward.D2", "single", "complete", "average" (= UPGMA), "mcquitty" (= WPGMA), "median" (= WPGMC), or "centroid" (= UPGMC). Unambiguous abbreviations are accepted as well.

This function supports parallelization via \code{\link[future:plan]{future::plan()}} and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.}
}
\value{
data frame
}
\description{
Performs a parallelized grid search to find the number of clusters maximizing the overall silhouette width of the clustering solution (see \code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}). When \code{method = NULL}, the optimization also includes all methods available in \code{\link[stats:hclust]{stats::hclust()}} in the grid search. This function supports parallelization via \code{\link[future:plan]{future::plan()}} and a progress bar generated by the \code{progressr} package (see Examples).
}
\examples{

#weekly covid prevalence
#in 10 California counties
#aggregated by month
tsl <- tsl_initialize(
  x = covid_prevalence,
  name_column = "name",
  time_column = "time"
) |>
  tsl_subset(
    names = 1:10
  ) |>
  tsl_aggregate(
    new_time = "months",
    fun = max
  )

if(interactive()){
  #plotting first three time series
  tsl_plot(
    tsl = tsl_subset(
      tsl = tsl,
      names = 1:3
    ),
    guide_columns = 3
  )
}

#compute dissimilarity matrix
psi_matrix <- distantia(
  tsl = tsl,
  lock_step = TRUE
) |>
  distantia_matrix()

#optimize hierarchical clustering
hclust_optimization <- utils_cluster_hclust_optimizer(
  d = psi_matrix
)

#best solution in first row
head(hclust_optimization)

}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_drop_geometry.R
\name{utils_drop_geometry}
\alias{utils_drop_geometry}
\title{Remove Geometry Column from SF Data Frames}
\usage{
utils_drop_geometry(df = NULL)
}
\arguments{
\item{df}{(required, data frame) Input data frame. Default: NULL.}
}
\value{
A data frame.
}
\description{
Replicates the functionality of \code{sf::st_drop_geometry()} without depending on the \code{sf} package.
}
\seealso{
Other tsl_processing_internal:
\code{\link[=utils_global_scaling_params]{utils_global_scaling_params()}},
\code{\link[=utils_optimize_loess]{utils_optimize_loess()}},
\code{\link[=utils_optimize_spline]{utils_optimize_spline()}},
\code{\link[=utils_rescale_vector]{utils_rescale_vector()}}
}
\concept{tsl_processing_internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_global_scaling_params.R
\name{utils_global_scaling_params}
\alias{utils_global_scaling_params}
\title{Global Centering and Scaling Parameters of Time Series Lists}
\usage{
utils_global_scaling_params(tsl = NULL, f = NULL, ...)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{f}{(required, function) function \code{\link[=f_scale_global]{f_scale_global()}}. Default: NULL}

\item{...}{(optional, arguments of \code{f}) Optional arguments for the transformation function.}
}
\value{
list
}
\description{
Internal function to compute global scaling parameters (mean and standard deviation) for time series lists. Used within \code{\link[=tsl_transform]{tsl_transform()}} when the scaling function \code{\link[=f_scale_global]{f_scale_global()}} is used as input for the argument \code{f}.

This function removes exclusive columns from the data. See function \code{\link[=tsl_subset]{tsl_subset()}}.
}
\seealso{
Other tsl_processing_internal:
\code{\link[=utils_drop_geometry]{utils_drop_geometry()}},
\code{\link[=utils_optimize_loess]{utils_optimize_loess()}},
\code{\link[=utils_optimize_spline]{utils_optimize_spline()}},
\code{\link[=utils_rescale_vector]{utils_rescale_vector()}}
}
\concept{tsl_processing_internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_names.R
\name{tsl_names_get}
\alias{tsl_names_get}
\title{Get Time Series Names from Time Series Lists}
\usage{
tsl_names_get(tsl = NULL, zoo = TRUE)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{zoo}{(optional, logical) If TRUE, the attributes "name" of the zoo objects are returned. Default: TRUE}
}
\value{
list
}
\description{
A time series list has two sets of names: the names of the list items (as returned by \code{names(tsl)}), and the names of the contained zoo objects, as stored in their attribute "name". These names should ideally be the same, for the sake of data consistency. This function extracts either set of names.
}
\examples{
#initialize a time series list
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
)


#get names of zoo objects
tsl_names_get(
  tsl = tsl,
  zoo = TRUE
)

#get list names only
tsl_names_get(
  tsl = tsl,
  zoo = FALSE
  )

#same as
names(tsl)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_names.R
\name{zoo_name_set}
\alias{zoo_name_set}
\title{Set Name of a Zoo Time Series}
\usage{
zoo_name_set(x = NULL, name = NULL)
}
\arguments{
\item{x}{(required, zoo object) Zoo time series to analyze. Default: NULL.}

\item{name}{(required, character string) Name or new name of the zoo object. If NULL, \code{x} is returned as is. Default: NULL.}
}
\value{
zoo time series
}
\description{
Zoo time series do not have an attribute 'name'. However, within \code{distantia}, to keep data consistency in several plotting and analysis operations, an attribute 'name' is used for these objects. This function is a convenient wrapper of \code{attr(x = x, which = "name") <- "xxx"}.
}
\examples{
#simulate zoo time series
x <- zoo_simulate()

#get current name
zoo_name_get(x = x)

#change name
x <- zoo_name_set(
  x = x,
  name = "My.New.name"
)

zoo_name_get(x = x)

#clean name
x <- zoo_name_clean(
  x = x,
  lowercase = TRUE
)

zoo_name_get(x = x)
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_permute.R
\name{zoo_permute}
\alias{zoo_permute}
\title{Random or Restricted Permutation of Zoo Time Series}
\usage{
zoo_permute(
  x = NULL,
  repetitions = 1L,
  permutation = "restricted_by_row",
  block_size = NULL,
  seed = 1L
)
}
\arguments{
\item{x}{(required, zoo object) zoo time series. Default: NULL}

\item{repetitions}{(optional, integer) number of permutations to compute. Large numbers may compromise your R session. Default: 1}

\item{permutation}{(optional, character string) permutation method. Valid values are listed below from higher to lower induced randomness:
\itemize{
\item "free": unrestricted re-shuffling of individual cases across rows and columns. Ignores block_size.
\item "free_by_row": unrestricted re-shuffling of complete rows. Ignores block size.
\item "restricted": restricted shuffling across rows and columns within blocks of rows.
\item "restricted_by_row": restricted re-shuffling of rows within blocks.
}}

\item{block_size}{(optional, integer) Block size in number of rows for restricted permutations. Only relevant when permutation methods are "restricted" or "restricted_by_row". A block of size \code{n} indicates that the original data is pre-divided into blocks of such size, and a given row can only be permuted within their original block. If NULL, defaults to the rounded one tenth of the number of rows in \code{x}. Default: NULL.}

\item{seed}{(optional, integer) initial random seed to use during permutations. Default: 1}
}
\value{
Time Series List
}
\description{
Fast permutation of zoo time series for null model testing using a fast and efficient C++ implementations of different restricted and free permutation methods.

The available permutation methods are:

\itemize{
\item "free" (see \code{\link[=permute_free_cpp]{permute_free_cpp()}}): Unrestricted and independent re-shuffling of individual cases across rows and columns. Individual values are relocated to a new row and column within the dimensions of the original matrix.
\item "free_by_row" (see \code{\link[=permute_free_by_row_cpp]{permute_free_by_row_cpp()}}): Unrestricted re-shuffling of complete rows. Each individual row is given a new random row number, and the data  matrix is re-ordered accordingly.
\item "restricted" (see \code{\link[=permute_restricted_cpp]{permute_restricted_cpp()}}): Data re-shuffling across rows and columns is restricted to blocks of contiguous rows. The algorithm divides the data matrix into a set of blocks of contiguous rows, and individual cases are then assigned to a new row and column within their original block.
\item "restricted_by_row" (see \code{\link[=permute_restricted_by_row_cpp]{permute_restricted_by_row_cpp()}}): Re-shuffling of complete rows is restricted to blocks of contiguous rows. The algorithm divides the data matrix into a set of blocks of contiguous rows, each individual row is given a new random row number within its original block, and the block is reordered accordingly to generate the permuted output.
}

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{

#simulate zoo time series
x <- zoo_simulate(cols = 2)

if(interactive()){
  zoo_plot(x)
}

#free
x_free <- zoo_permute(
  x = x,
  permutation = "free",
  repetitions = 2
)

if(interactive()){
  tsl_plot(
    tsl = x_free,
    guide = FALSE
    )
}

#free by row
x_free_by_row <- zoo_permute(
  x = x,
  permutation = "free_by_row",
  repetitions = 2
)

if(interactive()){
  tsl_plot(
    tsl = x_free_by_row,
    guide = FALSE
  )
}

#restricted
x_restricted <- zoo_permute(
  x = x,
  permutation = "restricted",
  repetitions = 2
)

if(interactive()){
  tsl_plot(
    tsl = x_restricted,
    guide = FALSE
  )
}

#restricted by row
x_restricted_by_row <- zoo_permute(
  x = x,
  permutation = "restricted_by_row",
  repetitions = 2
)

if(interactive()){
  tsl_plot(
    tsl = x_restricted_by_row,
    guide = FALSE
  )
}

}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_aggregate.R
\name{tsl_aggregate}
\alias{tsl_aggregate}
\title{Aggregate Time Series List Over Time Periods}
\usage{
tsl_aggregate(tsl = NULL, new_time = NULL, f = mean, ...)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{new_time}{(required, numeric, numeric vector, Date vector, POSIXct vector, or keyword) Definition of the aggregation pattern. The available options are:
\itemize{
\item numeric vector: only for the "numeric" time class, defines the breakpoints for time series aggregation.
\item "Date" or "POSIXct" vector: as above, but for the time classes "Date" and "POSIXct." In any case, the input vector is coerced to the time class of the \code{tsl} argument.
\item numeric: defines fixed time intervals in the units of \code{tsl} for time series aggregation. Used as is when the time class is "numeric", and coerced to integer and interpreted as days for the time classes "Date" and "POSIXct".
\item keyword (see \code{\link[=utils_time_units]{utils_time_units()}}): the common options for the time classes "Date" and "POSIXct" are: "millennia", "centuries", "decades", "years", "quarters", "months", and "weeks". Exclusive keywords for the "POSIXct" time class are: "days", "hours", "minutes", and "seconds". The time class "numeric" accepts keywords coded as scientific numbers, from "1e8" to "1e-8".
}}

\item{f}{(required, function name) Name of function taking a vector as input and returning a single value as output. Typical examples are \code{mean}, \code{max},\code{min}, \code{median}, and \code{quantile}. Default: \code{mean}.}

\item{...}{(optional) further arguments for \code{f}.}
}
\value{
time series list
}
\description{
Time series aggregation involves grouping observations and summarizing group values with a statistical function. This operation is useful to:
\itemize{
\item Decrease (downsampling) the temporal resolution of a time series.
\item Highlight particular states of a time series over time. For example, a daily temperature series can be aggregated by month using \code{max} to represent the highest temperatures each month.
\item Transform irregular time series into regular.
}

This function aggregates time series lists \strong{with overlapping times}. Please check such overlap by assessing the columns "begin" and "end " of the data frame resulting from \code{df <- tsl_time(tsl = tsl)}. Aggregation will be limited by the shortest time series in your time series list. To aggregate non-overlapping time series, please subset the individual components of \code{tsl} one by one either using \code{\link[=tsl_subset]{tsl_subset()}} or the syntax \code{tsl = my_tsl[[i]]}.

\strong{Methods}

Any function returning a single number from a numeric vector can be used to aggregate a time series list. Quoted and unquoted function names can be used. Additional arguments to these functions can be passed via the argument \code{...}. Typical examples are:

\itemize{
\item \code{mean} or \code{"mean"}: see \code{\link[=mean]{mean()}}.
\item \code{median} or \code{"median"}: see \code{\link[stats:median]{stats::median()}}.
\item \code{quantile} or "quantile": see \code{\link[stats:quantile]{stats::quantile()}}.
\item \code{min} or \code{"min"}: see \code{\link[=min]{min()}}.
\item \code{max} or \code{"max"}: see \code{\link[=max]{max()}}.
\item \code{sd} or \code{"sd"}: to compute standard deviation, see \code{\link[stats:sd]{stats::sd()}}.
\item \code{var} or \code{"var"}: to compute the group variance, see \code{\link[stats:var]{stats::var()}}.
\item \code{length} or \code{"length"}: to compute group length.
\item \code{sum} or \code{"sum"}: see \code{\link[=sum]{sum()}}.
}

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{

# yearly aggregation
#----------------------------------
#long-term monthly temperature of 20 cities
tsl <- tsl_initialize(
  x = cities_temperature,
  name_column = "name",
  time_column = "time"
)

#plot time series
if(interactive()){
  tsl_plot(
    tsl = tsl[1:4],
    guide_columns = 4
  )
}

#check time features
tsl_time(tsl)[, c("name", "resolution", "units")]

#aggregation: mean yearly values
tsl_year <- tsl_aggregate(
  tsl = tsl,
  new_time = "year",
  f = mean
)

#check time features
tsl_time(tsl_year)[, c("name", "resolution", "units")]

if(interactive()){
  tsl_plot(
    tsl = tsl_year[1:4],
    guide_columns = 4
  )
}


# other supported keywords
#----------------------------------

#simulate full range of calendar dates
tsl <- tsl_simulate(
  n = 2,
  rows = 1000,
  time_range = c(
    "0000-01-01",
    as.character(Sys.Date())
  )
)

#mean value by millennia (extreme case!!!)
tsl_millennia <- tsl_aggregate(
  tsl = tsl,
  new_time = "millennia",
  f = mean
)

if(interactive()){
  tsl_plot(tsl_millennia)
}

#max value by centuries
tsl_century <- tsl_aggregate(
  tsl = tsl,
  new_time = "century",
  f = max
)

if(interactive()){
  tsl_plot(tsl_century)
}

#quantile 0.75 value by centuries
tsl_centuries <- tsl_aggregate(
  tsl = tsl,
  new_time = "centuries",
  f = stats::quantile,
  probs = 0.75 #argument of stats::quantile()
)

}
\seealso{
\code{\link[=zoo_aggregate]{zoo_aggregate()}}

Other tsl_processing:
\code{\link[=tsl_resample]{tsl_resample()}},
\code{\link[=tsl_smooth]{tsl_smooth()}},
\code{\link[=tsl_stats]{tsl_stats()}},
\code{\link[=tsl_transform]{tsl_transform()}}
}
\concept{tsl_processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_nrow.R
\name{tsl_nrow}
\alias{tsl_nrow}
\title{Get Number of Rows in Time Series Lists}
\usage{
tsl_nrow(tsl = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}
}
\value{
list
}
\description{
Get Number of Rows in Time Series Lists
}
\examples{
#Simulate zoo time series
tsl <- tsl_simulate(
  rows = 150
  )

#Count rows
tsl_nrow(
  tsl = tsl
)
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_names.R
\name{zoo_name_get}
\alias{zoo_name_get}
\title{Get Name of a Zoo Time Series}
\usage{
zoo_name_get(x = NULL)
}
\arguments{
\item{x}{(required, zoo object) Zoo time series to analyze. Default: NULL.}
}
\value{
character string
}
\description{
Just a convenient wrapper of \code{attributes(x)$name}.
}
\examples{
#simulate zoo time series
x <- zoo_simulate()

#get current name
zoo_name_get(x = x)

#change name
x <- zoo_name_set(
  x = x,
  name = "My.New.name"
)

zoo_name_get(x = x)

#clean name
x <- zoo_name_clean(
  x = x,
  lowercase = TRUE
)

zoo_name_get(x = x)
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_dtw_plot.R
\name{distantia_dtw_plot}
\alias{distantia_dtw_plot}
\title{Two-Way Dissimilarity Plots of Time Series Lists}
\usage{
distantia_dtw_plot(
  tsl = NULL,
  distance = "euclidean",
  diagonal = TRUE,
  bandwidth = 1,
  matrix_type = "cost",
  matrix_color = NULL,
  path_width = 1,
  path_color = "black",
  diagonal_width = 1,
  diagonal_color = "white",
  line_color = NULL,
  line_width = 1,
  text_cex = 1
)
}
\arguments{
\item{tsl}{(required, time series list) list of zoo time series. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}

\item{diagonal}{(optional, logical vector). If TRUE, diagonals are included in the dynamic time warping computation. Default: TRUE}

\item{bandwidth}{(optional, numeric) Proportion of space at each side of the cost matrix diagonal (aka \emph{Sakoe-Chiba band}) defining a valid region for dynamic time warping, used to control the flexibility of the warping path. This method prevents degenerate alignments due to differences in magnitude between time series when the data is not properly scaled. If \code{1} (default), DTW is unconstrained. If \code{0}, DTW is fully constrained and the warping path follows the matrix diagonal. Recommended values may vary depending on the nature of the data. Ignored if \code{lock_step = TRUE}. Default: 1.}

\item{matrix_type}{(optional, character string): one of "cost" or "distance" (the abbreviation "dist" is accepted as well). Default: "cost".}

\item{matrix_color}{(optional, character vector) Vector of colors for the distance or cost matrix. If NULL, uses the palette "Zissou 1" provided by \code{\link[grDevices:hcl.colors]{grDevices::hcl.colors()}}. Default: NULL}

\item{path_width}{(optional, numeric) width of the least cost path. Default: 1}

\item{path_color}{(optional, character string) color of the least-cost path. Default: "black"}

\item{diagonal_width}{(optional, numeric) width of the diagonal. Set to 0 to remove the diagonal line. Default: 0.5}

\item{diagonal_color}{(optional, character string) color of the diagonal. Default: "white"}

\item{line_color}{(optional, character vector) Vector of colors for the time series plot. If NULL, defaults to a subset of \code{matrix_color}.}

\item{line_width}{(optional, numeric vector) Width of the time series plot. Default: 1}

\item{text_cex}{(optional, numeric) Multiplier of the text size. Default: 1}
}
\value{
multipanel plot
}
\description{
Plots two sequences, their distance or cost matrix, their least cost path, and all relevant values used to compute dissimilarity.

Unlike \code{\link[=distantia]{distantia()}}, this function does not accept vectors as inputs for the arguments to compute dissimilarity (\code{distance}, \code{diagonal},  and \code{weighted}), and only plots a pair of sequences at once.

The argument \code{lock_step} is not available because this plot does not make sense in such a case.
}
\examples{
#three time series
#climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
#convert to time series list
#scale and center to neutralize effect of different scales in temperature, rainfall, and ndvi
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global #see help(f_scale_global)
  )

if(interactive()){
  tsl_plot(
    tsl = tsl,
    guide_columns = 3
    )
}

#visualize dynamic time warping
if(interactive()){

  #plot pair with cost matrix (default)
  distantia_dtw_plot(
    tsl = tsl[c("Spain", "Sweden")] #only two time series!
  )

  #plot pair with distance matrix
  distantia_dtw_plot(
    tsl = tsl[c("Spain", "Sweden")],
    matrix_type = "distance"
  )

  #plot pair with different distance
  distantia_dtw_plot(
    tsl = tsl[c("Spain", "Sweden")],
    distance = "manhattan", #sed data(distances)
    matrix_type = "distance"
  )


  #with different colors
  distantia_dtw_plot(
    tsl = tsl[c("Spain", "Sweden")],
    matrix_type = "distance",
    matrix_color = grDevices::hcl.colors(
      n = 100,
      palette = "Inferno"
    ),
    path_color = "white",
    path_width = 2,
    line_color = grDevices::hcl.colors(
      n = 3, #same as variables in tsl
      palette = "Inferno"
    )
  )

}
}
\seealso{
Other distantia:
\code{\link[=distantia]{distantia()}},
\code{\link[=distantia_dtw]{distantia_dtw()}},
\code{\link[=distantia_ls]{distantia_ls()}}
}
\concept{distantia}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_to_tsl.R
\name{zoo_to_tsl}
\alias{zoo_to_tsl}
\title{Convert Individual Zoo Objects to Time Series List}
\usage{
zoo_to_tsl(x = NULL)
}
\arguments{
\item{x}{(required, zoo object) Time series. Default: NULL}
}
\value{
time series list of length one.
}
\description{
Internal function to wrap a zoo object into a time series list.
}
\examples{
#create zoo object
x <- zoo_simulate()
class(x)

#to time series list
tsl <- zoo_to_tsl(
  x = x
)

class(tsl)
class(tsl[[1]])
names(tsl)
attributes(tsl[[1]])$name
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_boxplot.R
\name{distantia_boxplot}
\alias{distantia_boxplot}
\title{Distantia Boxplot}
\usage{
distantia_boxplot(df = NULL, fill_color = NULL, f = median, text_cex = 1)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=distantia]{distantia()}}, \code{\link[=distantia_ls]{distantia_ls()}}, \code{\link[=distantia_dtw]{distantia_dtw()}}, or \code{\link[=distantia_time_delay]{distantia_time_delay()}}. Default: NULL}

\item{fill_color}{(optional, character vector) boxplot fill color. Default: NULL}

\item{f}{(optional, function) function used to aggregate the input data frame and arrange the boxes. One of \code{mean} or \code{median}. Default: \code{median}.}

\item{text_cex}{(optional, numeric) Multiplier of the text size. Default: 1}
}
\value{
boxplot
}
\description{
Boxplot of a data frame returned by \code{\link[=distantia]{distantia()}} showing the summary statistics of psi scores for each time series against all others.
}
\examples{
tsl <- tsl_initialize(
  x = distantia::albatross,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

df <- distantia(
  tsl = tsl,
  lock_step = TRUE
  )

distantia_boxplot(
  df = df,
  text_cex = 1.5
  )
}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_diagnose.R
\name{tsl_diagnose}
\alias{tsl_diagnose}
\title{Diagnose Issues in Time Series Lists}
\usage{
tsl_diagnose(tsl = NULL, full = TRUE)
}
\arguments{
\item{tsl}{(required, list of zoo time series) Time series list to diagnose. Default: NULL}

\item{full}{(optional, logical) If TRUE, a full diagnostic is triggered. Otherwise, only the data structure is tested. Default: TRUE}
}
\value{
invisible
}
\description{
A Time Series List (\code{tsl} for short) is a named list of zoo time series. This type of object, not defined as a class, is used throughout the \code{distantia} package to contain time series data ready for processing and analysis.

The structure and values of a \code{tsl} must fulfill several general conditions:

Structure:
\itemize{
\item List names match the attributes "name" of the zoo time series.
\item Zoo time series must have at least one shared column name.
\item The index (as extracted by \code{\link[zoo:index]{zoo::index()}}) of all zoo objects must be of the same class (either "Date", "POSIXct", "numeric", or "integer").
\item The "core data" (as extracted by \code{\link[zoo:coredata]{zoo::coredata()}}) of univariate zoo time series must be of class "matrix".
}

Values (optional, when \code{full = TRUE}):
\itemize{
\item All time series have at least one shared numeric column.
\item There are no NA, Inf, or NaN values in the time series.
}

This function analyzes a \code{tsl} without modifying it and returns messages describing what conditions are not met, and provides hints on how to fix most issues.
}
\examples{
#creating three zoo time series

#one with NA values
x <- zoo_simulate(
  name = "x",
  cols = 1,
  na_fraction = 0.1
  )

#with different number of columns
#with repeated name
y <- zoo_simulate(
  name = "x",
  cols = 2
  )

#with different time class
z <- zoo_simulate(
  name = "z",
  cols = 1,
  time_range = c(1, 100)
  )

#adding a few structural issues

#changing the column name of x
colnames(x) <- c("b")

#converting z to vector
z <- zoo::zoo(
  x = runif(nrow(z)),
  order.by = zoo::index(z)
)

#storing zoo objects in a list
#with mismatched names
tsl <- list(
  a = x,
  b = y,
  c = z
)

#running full diagnose
tsl_diagnose(
  tsl = tsl,
  full = TRUE
  )
}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{psi_ls_cpp}
\alias{psi_ls_cpp}
\title{(C++) Psi Dissimilarity Score of Two Aligned Time Series}
\usage{
psi_ls_cpp(x, y, distance = "euclidean")
}
\arguments{
\item{x}{(required, numeric matrix) of same number of columns as 'y'.}

\item{y}{(required, numeric matrix) of same number of columns as 'x'.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}
}
\value{
numeric
}
\description{
Computes the psi dissimilarity score between two time series
observed at the same times. Time series \code{y} and \code{x} with the same
number of columns and rows. NA values should be removed before using this function.
If the selected distance function is "chi" or "cosine", pairs of zeros should
be either removed or replaced with pseudo-zeros (i.e. 0.00001).
}
\seealso{
Other Rcpp_dissimilarity_analysis:
\code{\link[=psi_dtw_cpp]{psi_dtw_cpp()}},
\code{\link[=psi_equation_cpp]{psi_equation_cpp()}},
\code{\link[=psi_null_dtw_cpp]{psi_null_dtw_cpp()}},
\code{\link[=psi_null_ls_cpp]{psi_null_ls_cpp()}}
}
\concept{Rcpp_dissimilarity_analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{distance_bray_curtis_cpp}
\alias{distance_bray_curtis_cpp}
\title{(C++) Bray-Curtis Distance Between Two Vectors}
\usage{
distance_bray_curtis_cpp(x, y)
}
\arguments{
\item{x}{(required, numeric vector).}

\item{y}{(required, numeric vector) of same length as \code{x}.}
}
\value{
numeric
}
\description{
Computes the Bray-Curtis distance, suitable for species abundance data.
}
\examples{
distance_bray_curtis_cpp(x = runif(100), y = runif(100))
}
\seealso{
Other Rcpp_distance_methods:
\code{\link[=distance_canberra_cpp]{distance_canberra_cpp()}},
\code{\link[=distance_chebyshev_cpp]{distance_chebyshev_cpp()}},
\code{\link[=distance_chi_cpp]{distance_chi_cpp()}},
\code{\link[=distance_cosine_cpp]{distance_cosine_cpp()}},
\code{\link[=distance_euclidean_cpp]{distance_euclidean_cpp()}},
\code{\link[=distance_hamming_cpp]{distance_hamming_cpp()}},
\code{\link[=distance_hellinger_cpp]{distance_hellinger_cpp()}},
\code{\link[=distance_jaccard_cpp]{distance_jaccard_cpp()}},
\code{\link[=distance_manhattan_cpp]{distance_manhattan_cpp()}},
\code{\link[=distance_russelrao_cpp]{distance_russelrao_cpp()}},
\code{\link[=distance_sorensen_cpp]{distance_sorensen_cpp()}}
}
\concept{Rcpp_distance_methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{auto_sum_full_cpp}
\alias{auto_sum_full_cpp}
\title{(C++) Sum Distances Between All Consecutive Samples in Two Time Series}
\usage{
auto_sum_full_cpp(x, y, distance = "euclidean")
}
\arguments{
\item{x}{(required, numeric matrix) univariate or multivariate time series.}

\item{y}{(required, numeric matrix) univariate or multivariate time series
with the same number of columns as 'x'.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean"}
}
\value{
numeric
}
\description{
Computes the cumulative auto sum of autodistances of two time series.
The output value is used as normalization factor when computing dissimilarity scores.
}
\examples{
#simulate two time series
x <- zoo_simulate(seed = 1)
y <- zoo_simulate(seed = 2)

#auto sum
auto_sum_full_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)
}
\seealso{
Other Rcpp_auto_sum:
\code{\link[=auto_distance_cpp]{auto_distance_cpp()}},
\code{\link[=auto_sum_cpp]{auto_sum_cpp()}},
\code{\link[=auto_sum_path_cpp]{auto_sum_path_cpp()}},
\code{\link[=subset_matrix_by_rows_cpp]{subset_matrix_by_rows_cpp()}}
}
\concept{Rcpp_auto_sum}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{fagus_coordinates}
\alias{fagus_coordinates}
\title{Site Coordinates of Fagus sylvatica Stands}
\format{
An sf data frame with 3 rows and 4 columns.
}
\usage{
data(fagus_coordinates)
}
\description{
Site Coordinates of Fagus sylvatica Stands
}
\seealso{
Other example_data:
\code{\link{albatross}},
\code{\link{cities_coordinates}},
\code{\link{cities_temperature}},
\code{\link{covid_counties}},
\code{\link{covid_prevalence}},
\code{\link{eemian_coordinates}},
\code{\link{eemian_pollen}},
\code{\link{fagus_dynamics}},
\code{\link{honeycomb_climate}},
\code{\link{honeycomb_polygons}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_boxplot_common.R
\name{utils_boxplot_common}
\alias{utils_boxplot_common}
\title{Common Boxplot Component of \code{distantia_boxplot()} and \code{momentum_boxplot()}}
\usage{
utils_boxplot_common(
  variable = NULL,
  value = NULL,
  fill_color = NULL,
  f = median,
  main = NULL,
  xlab = NULL,
  ylab = NULL,
  text_cex = 1
)
}
\arguments{
\item{variable}{(required, character vector) vector of variable or time series names. Default: NULL}

\item{value}{(required, numeric vector) numeric values to compute the boxplot. Must have the same length as \code{variable}. Default: NULL}

\item{fill_color}{(optional, character vector) boxplot fill color. Default: NULL}

\item{f}{(optional, function) function used to aggregate the input data frame and arrange the boxes. One of \code{mean} or \code{median}. Default: \code{median}.}

\item{main}{(optional, string) boxplot title. Default: NULL}

\item{xlab}{(optional, string) x axis label. Default: NULL}

\item{ylab}{(optional, string) y axis label. Default: NULL}

\item{text_cex}{(optional, numeric) Multiplier of the text size. Default: 1}
}
\value{
boxplot
}
\description{
Common Boxplot Component of \code{distantia_boxplot()} and \code{momentum_boxplot()}
}
\examples{
utils_boxplot_common(
  variable = rep(x = c("a", "b"), times = 50),
  value = stats::runif(100)
)
}
\seealso{
Other internal:
\code{\link[=utils_check_args_distantia]{utils_check_args_distantia()}},
\code{\link[=utils_check_args_matrix]{utils_check_args_matrix()}},
\code{\link[=utils_check_args_momentum]{utils_check_args_momentum()}},
\code{\link[=utils_check_args_path]{utils_check_args_path()}},
\code{\link[=utils_check_args_tsl]{utils_check_args_tsl()}},
\code{\link[=utils_check_args_zoo]{utils_check_args_zoo()}},
\code{\link[=utils_check_distance_args]{utils_check_distance_args()}},
\code{\link[=utils_check_list_class]{utils_check_list_class()}},
\code{\link[=utils_clean_names]{utils_clean_names()}},
\code{\link[=utils_digits]{utils_digits()}},
\code{\link[=utils_distantia_df_split]{utils_distantia_df_split()}},
\code{\link[=utils_prepare_df]{utils_prepare_df()}},
\code{\link[=utils_prepare_matrix]{utils_prepare_matrix()}},
\code{\link[=utils_prepare_matrix_list]{utils_prepare_matrix_list()}},
\code{\link[=utils_prepare_time]{utils_prepare_time()}},
\code{\link[=utils_prepare_vector_list]{utils_prepare_vector_list()}},
\code{\link[=utils_prepare_zoo_list]{utils_prepare_zoo_list()}},
\code{\link[=utils_tsl_pairs]{utils_tsl_pairs()}}
}
\concept{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_plot.R
\name{zoo_plot}
\alias{zoo_plot}
\title{Plot Zoo Time Series}
\usage{
zoo_plot(
  x = NULL,
  line_color = NULL,
  line_width = 1,
  xlim = NULL,
  ylim = NULL,
  title = NULL,
  xlab = NULL,
  ylab = NULL,
  text_cex = 1,
  guide = TRUE,
  guide_position = "topright",
  guide_cex = 0.8,
  vertical = FALSE,
  subpanel = FALSE
)
}
\arguments{
\item{x}{(required, zoo object) zoo time series. Default: NULL}

\item{line_color}{(optional, character vector) vector of colors for the distance or cost matrix. If NULL, uses an appropriate palette generated with \code{\link[grDevices:palette.colors]{grDevices::palette.colors()}}. Default: NULL}

\item{line_width}{(optional, numeric vector) Width of the time series lines. Default: 1}

\item{xlim}{(optional, numeric vector) Numeric vector with the limits of the x axis. Default: NULL}

\item{ylim}{(optional, numeric vector) Numeric vector with the limits of the y axis. Default: NULL}

\item{title}{(optional, character string) Main title of the plot. If NULL, it's set to the name of the time series. Default: NULL}

\item{xlab}{(optional, character string) Title of the x axis. Disabled if \code{subpanel} or \code{vertical} are TRUE. If NULL, the word "Time" is used. Default: NULL}

\item{ylab}{(optional, character string) Title of the y axis. Disabled if \code{subpanel} or \code{vertical} are TRUE. If NULL, it is left empty. Default: NULL}

\item{text_cex}{(optional, numeric) Multiplicator of the text size. Default: 1}

\item{guide}{(optional, logical) If TRUE, plots a legend. Default: TRUE}

\item{guide_position}{(optional, vector of xy coordinates or character string). This is a condensed version of the \code{x} and \code{y} arguments of the \code{\link[graphics:legend]{graphics::legend()}} function. Coordinates (in the range 0 1) or keyword to position the legend. Accepted keywords are: "bottomright", "bottom", "bottomleft", "left", "topleft", "top", "topright", "right" and "center". Default: "topright".}

\item{guide_cex}{(optional, numeric) Size of the guide's text and separation between the guide's rows. Default: 0.7.}

\item{vertical}{(optional, logical) For internal use within the package in multipanel plots. Switches the plot axes. Disabled if \code{subpanel = FALSE}. Default: FALSE}

\item{subpanel}{(optional, logical) For internal use within the package in multipanel plots. Strips down the plot for a sub-panel. Default: FALSE}
}
\value{
A plot.
}
\description{
Plot Zoo Time Series
}
\examples{

#simulate zoo time series
x <- zoo_simulate()

if(interactive()){

  zoo_plot(
    x = x,
    xlab = "Date",
    ylab = "Value",
    title = "My time series"
  )

}
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}
\keyword{plotting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{f_trend_linear}
\alias{f_trend_linear}
\title{Data Transformation: Linear Trend of Zoo Time Series}
\usage{
f_trend_linear(x = NULL, center = TRUE, ...)
}
\arguments{
\item{x}{(required, zoo object) Zoo time series object to transform.}

\item{center}{(required, logical) If TRUE, the output is centered at zero. If FALSE, it is centered at the data mean. Default: TRUE}

\item{...}{(optional, additional arguments) Ignored in this function.}
}
\value{
zoo object
}
\description{
Fits a linear model on each column of a zoo object using time as a predictor, and predicts the outcome.
}
\examples{
x <- zoo_simulate(cols = 2)

y <- f_trend_linear(
  x = x
)

if(interactive()){
  zoo_plot(x)
  zoo_plot(y)
}
}
\seealso{
Other tsl_transformation:
\code{\link[=f_binary]{f_binary()}},
\code{\link[=f_clr]{f_clr()}},
\code{\link[=f_detrend_difference]{f_detrend_difference()}},
\code{\link[=f_detrend_linear]{f_detrend_linear()}},
\code{\link[=f_detrend_poly]{f_detrend_poly()}},
\code{\link[=f_hellinger]{f_hellinger()}},
\code{\link[=f_list]{f_list()}},
\code{\link[=f_log]{f_log()}},
\code{\link[=f_percent]{f_percent()}},
\code{\link[=f_proportion]{f_proportion()}},
\code{\link[=f_proportion_sqrt]{f_proportion_sqrt()}},
\code{\link[=f_rescale_global]{f_rescale_global()}},
\code{\link[=f_rescale_local]{f_rescale_local()}},
\code{\link[=f_scale_global]{f_scale_global()}},
\code{\link[=f_scale_local]{f_scale_local()}},
\code{\link[=f_trend_poly]{f_trend_poly()}}
}
\concept{tsl_transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{honeycomb_climate}
\alias{honeycomb_climate}
\title{Rainfall and Temperature in The Americas}
\format{
An object of class \code{tbl_df} (inherits from \code{tbl}, \code{data.frame}) with 9432 rows and 4 columns.
}
\usage{
data(honeycomb_climate)
}
\description{
Monthly temperature and rainfall between 2009 and 2019 in 72 hexagonal cells covering The Americas.
}
\seealso{
Other example_data:
\code{\link{albatross}},
\code{\link{cities_coordinates}},
\code{\link{cities_temperature}},
\code{\link{covid_counties}},
\code{\link{covid_prevalence}},
\code{\link{eemian_coordinates}},
\code{\link{eemian_pollen}},
\code{\link{fagus_coordinates}},
\code{\link{fagus_dynamics}},
\code{\link{honeycomb_polygons}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{subset_matrix_by_rows_cpp}
\alias{subset_matrix_by_rows_cpp}
\title{(C++) Subset Matrix by Rows}
\usage{
subset_matrix_by_rows_cpp(m, rows)
}
\arguments{
\item{m}{(required, numeric matrix) a univariate or multivariate time series.}

\item{rows}{(required, integer vector) vector of rows to subset from a
least-cost path data frame.}
}
\value{
numeric matrix
}
\description{
Subsets a time series matrix to the coordinates of a trimmed
least-cost path when blocks are ignored during a dissimilarity analysis.
}
\examples{
#simulate a time series
m <- zoo_simulate(seed = 1)

#sample some rows
rows <- sample(
  x = nrow(m),
  size = 10
  ) |>
  sort()

#subset by rows
m_subset <- subset_matrix_by_rows_cpp(
  m = m,
  rows = rows
  )

#compare with original
m[rows, ]

}
\seealso{
Other Rcpp_auto_sum:
\code{\link[=auto_distance_cpp]{auto_distance_cpp()}},
\code{\link[=auto_sum_cpp]{auto_sum_cpp()}},
\code{\link[=auto_sum_full_cpp]{auto_sum_full_cpp()}},
\code{\link[=auto_sum_path_cpp]{auto_sum_path_cpp()}}
}
\concept{Rcpp_auto_sum}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_join.R
\name{tsl_join}
\alias{tsl_join}
\title{Join Time Series Lists}
\usage{
tsl_join(..., na_action = "impute")
}
\arguments{
\item{...}{(required, time series lists) names of the time series lists to merge.}

\item{na_action}{(required, character) NA handling action. Available options are:
\itemize{
\item "impute" (default): NA cases are interpolated from neighbors as a function of time (see \code{\link[zoo:na.approx]{zoo::na.approx()}} and \code{\link[zoo:na.spline]{zoo::na.spline()}}).
\item "omit": rows with NA cases are removed.
}}
}
\value{
time series list
}
\description{
Joins an arbitrary number of time series lists by name and time. Pairs of zoo objects are joined with \code{\link[zoo:merge.zoo]{zoo::merge.zoo()}}. Names that are not shared across all input TSLs are ignored, and observations with no matching time are filled with NA and then managed via \code{\link[=tsl_handle_NA]{tsl_handle_NA()}} depending on the value of the argument \code{na_action}.
}
\examples{

#generate two time series list to join
tsl_a <- tsl_simulate(
  n = 2,
  cols = 2,
  irregular = TRUE,
  seed = 1
)

#needs renaming
tsl_b <- tsl_simulate(
  n = 3,
  cols = 2,
  irregular = TRUE,
  seed = 2
) |>
  tsl_colnames_set(
    names = c("c", "d")
  )

#join
tsl <- tsl_join(
  tsl_a,
  tsl_b
)

#plot result
if(interactive()){
  tsl_plot(
    tsl = tsl
  )
}

}
\seealso{
Other tsl_management:
\code{\link[=tsl_burst]{tsl_burst()}},
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{f_detrend_difference}
\alias{f_detrend_difference}
\title{Data Transformation: Detrending and Differencing}
\usage{
f_detrend_difference(x = NULL, lag = 1, center = TRUE, ...)
}
\arguments{
\item{x}{(required, zoo object) Zoo time series object to transform.}

\item{lag}{(optional, integer) Lag value for differencing. Default: 1}

\item{center}{(required, logical) If TRUE, the output is centered at zero. If FALSE, it is centered at the data mean. Default: TRUE}

\item{...}{(optional, additional arguments) Ignored in this function.}
}
\value{
zoo object
}
\description{
Performs differencing to remove trends from a zoo time series, isolating short-term fluctuations by subtracting values at specified lags. The function preserves the original index and metadata, with an option to center the output around the mean of the original series. Suitable for preprocessing time series data to focus on random fluctuations unrelated to overall trends.
}
\examples{
x <- zoo_simulate(cols = 2)

y_lag1 <- f_detrend_difference(
  x = x,
  lag = 1
)

y_lag5 <- f_detrend_difference(
  x = x,
  lag = 5
)

if(interactive()){
  zoo_plot(x)
  zoo_plot(y_lag1)
  zoo_plot(y_lag5)
}
}
\seealso{
Other tsl_transformation:
\code{\link[=f_binary]{f_binary()}},
\code{\link[=f_clr]{f_clr()}},
\code{\link[=f_detrend_linear]{f_detrend_linear()}},
\code{\link[=f_detrend_poly]{f_detrend_poly()}},
\code{\link[=f_hellinger]{f_hellinger()}},
\code{\link[=f_list]{f_list()}},
\code{\link[=f_log]{f_log()}},
\code{\link[=f_percent]{f_percent()}},
\code{\link[=f_proportion]{f_proportion()}},
\code{\link[=f_proportion_sqrt]{f_proportion_sqrt()}},
\code{\link[=f_rescale_global]{f_rescale_global()}},
\code{\link[=f_rescale_local]{f_rescale_local()}},
\code{\link[=f_scale_global]{f_scale_global()}},
\code{\link[=f_scale_local]{f_scale_local()}},
\code{\link[=f_trend_linear]{f_trend_linear()}},
\code{\link[=f_trend_poly]{f_trend_poly()}}
}
\concept{tsl_transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_smooth_exponential.R
\name{zoo_smooth_exponential}
\alias{zoo_smooth_exponential}
\title{Exponential Smoothing of Zoo Time Series}
\usage{
zoo_smooth_exponential(x = NULL, alpha = 0.2)
}
\arguments{
\item{x}{(required, zoo object) Time series to smooth. Default: NULL}

\item{alpha}{(required, numeric) Smoothing factor in the range (0, 1]. Determines the weight of the current value relative to past values. A higher value gives more weight to recent observations, while a lower value gives more weight to past observations. Default: 0.2}
}
\value{
zoo object
}
\description{
Applies exponential smoothing to a zoo time series object, where each value is a weighted average of the current value and past smoothed values. This method is useful for reducing noise in time series data while preserving the general trend.
}
\examples{
x <- zoo_simulate()

x_smooth <- zoo_smooth_exponential(
  x = x,
  alpha = 0.2
)

if(interactive()){
  zoo_plot(x)
  zoo_plot(x_smooth)
}
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{f_clr}
\alias{f_clr}
\title{Data Transformation: Rowwise Centered Log-Ratio}
\usage{
f_clr(x = NULL, ...)
}
\arguments{
\item{x}{(required, zoo object) Zoo time series object to transform.}

\item{...}{(optional, additional arguments) Ignored in this function.}
}
\value{
zoo object
}
\description{
Centers log-transformed proportions by subtracting the geometric mean of the row.
}
\examples{
x <- zoo_simulate(
  cols = 5,
  data_range = c(0, 500)
  )

y <- f_clr(
  x = x
)

if(interactive()){
  zoo_plot(x)
  zoo_plot(y)
}
}
\seealso{
Other tsl_transformation:
\code{\link[=f_binary]{f_binary()}},
\code{\link[=f_detrend_difference]{f_detrend_difference()}},
\code{\link[=f_detrend_linear]{f_detrend_linear()}},
\code{\link[=f_detrend_poly]{f_detrend_poly()}},
\code{\link[=f_hellinger]{f_hellinger()}},
\code{\link[=f_list]{f_list()}},
\code{\link[=f_log]{f_log()}},
\code{\link[=f_percent]{f_percent()}},
\code{\link[=f_proportion]{f_proportion()}},
\code{\link[=f_proportion_sqrt]{f_proportion_sqrt()}},
\code{\link[=f_rescale_global]{f_rescale_global()}},
\code{\link[=f_rescale_local]{f_rescale_local()}},
\code{\link[=f_scale_global]{f_scale_global()}},
\code{\link[=f_scale_local]{f_scale_local()}},
\code{\link[=f_trend_linear]{f_trend_linear()}},
\code{\link[=f_trend_poly]{f_trend_poly()}}
}
\concept{tsl_transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cost_path_trim_cpp}
\alias{cost_path_trim_cpp}
\title{(C++) Remove Blocks from a Least Cost Path}
\usage{
cost_path_trim_cpp(path)
}
\arguments{
\item{path}{(required, data frame) least-cost path produced by \code{\link[=cost_path_orthogonal_cpp]{cost_path_orthogonal_cpp()}}.}
}
\value{
data frame
}
\description{
(C++) Remove Blocks from a Least Cost Path
}
\examples{
#simulate two time series
x <- zoo_simulate(seed = 1)
y <- zoo_simulate(seed = 2)

#distance matrix
dist_matrix <- distance_matrix_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)

#least cost matrix
cost_matrix <- cost_matrix_orthogonal_cpp(
  dist_matrix = dist_matrix
)

#least cost path
cost_path <- cost_path_slotting_cpp(
  dist_matrix = dist_matrix,
  cost_matrix = cost_matrix
)

nrow(cost_path)

#remove blocks from least-cost path
cost_path_trimmed <- cost_path_trim_cpp(
  path = cost_path
)

nrow(cost_path_trimmed)
}
\seealso{
Other Rcpp_cost_path:
\code{\link[=cost_path_cpp]{cost_path_cpp()}},
\code{\link[=cost_path_diagonal_bandwidth_cpp]{cost_path_diagonal_bandwidth_cpp()}},
\code{\link[=cost_path_diagonal_cpp]{cost_path_diagonal_cpp()}},
\code{\link[=cost_path_orthogonal_bandwidth_cpp]{cost_path_orthogonal_bandwidth_cpp()}},
\code{\link[=cost_path_orthogonal_cpp]{cost_path_orthogonal_cpp()}},
\code{\link[=cost_path_slotting_cpp]{cost_path_slotting_cpp()}},
\code{\link[=cost_path_sum_cpp]{cost_path_sum_cpp()}}
}
\concept{Rcpp_cost_path}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psi_auto_distance.R
\name{psi_auto_distance}
\alias{psi_auto_distance}
\title{Cumulative Sum of Distances Between Consecutive Cases in a Time Series}
\usage{
psi_auto_distance(x = NULL, distance = "euclidean")
}
\arguments{
\item{x}{(required, zoo object or matrix) univariate or multivariate time series with no NAs. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}
}
\value{
numeric value
}
\description{
Demonstration function to compute the sum of distances between consecutive cases in a time series
}
\examples{
#distance metric
d <- "euclidean"

#simulate zoo time series
x <- zoo_simulate(
  name = "x",
  rows = 100,
  seasons = 2,
  seed = 1
)

#sum distance between consecutive samples
psi_auto_distance(
  x = x,
  distance = d
)
}
\seealso{
Other psi_demo:
\code{\link[=psi_auto_sum]{psi_auto_sum()}},
\code{\link[=psi_cost_matrix]{psi_cost_matrix()}},
\code{\link[=psi_cost_path]{psi_cost_path()}},
\code{\link[=psi_cost_path_sum]{psi_cost_path_sum()}},
\code{\link[=psi_distance_lock_step]{psi_distance_lock_step()}},
\code{\link[=psi_distance_matrix]{psi_distance_matrix()}},
\code{\link[=psi_equation]{psi_equation()}}
}
\concept{psi_demo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{psi_equation_cpp}
\alias{psi_equation_cpp}
\title{(C++) Equation of the Psi Dissimilarity Score}
\usage{
psi_equation_cpp(a, b, diagonal = TRUE)
}
\arguments{
\item{a}{(required, numeric) output of \code{\link[=cost_path_sum_cpp]{cost_path_sum_cpp()}} on a least cost path.}

\item{b}{(required, numeric) auto sum of both sequences,
result of \code{\link[=auto_sum_cpp]{auto_sum_cpp()}}.}

\item{diagonal}{(optional, logical). Must be TRUE when diagonals are used in
dynamic time warping and for lock-step distances. Default: FALSE.}
}
\value{
numeric
}
\description{
Equation to compute the \code{psi} dissimilarity score
(Birks and Gordon 1985). Psi is computed as \eqn{\psi = (2a / b) - 1},
where \eqn{a} is the sum of distances between the relevant samples of two
time series, and \eqn{b} is the cumulative sum of distances between
consecutive samples in the two time series.
If \code{a} is computed with dynamic time warping, and diagonals are used in the
computation of the least cost path, then one is added to the result of the equation above.
}
\seealso{
Other Rcpp_dissimilarity_analysis:
\code{\link[=psi_dtw_cpp]{psi_dtw_cpp()}},
\code{\link[=psi_ls_cpp]{psi_ls_cpp()}},
\code{\link[=psi_null_dtw_cpp]{psi_null_dtw_cpp()}},
\code{\link[=psi_null_ls_cpp]{psi_null_ls_cpp()}}
}
\concept{Rcpp_dissimilarity_analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{distance}
\alias{distance}
\title{Distance Between Two Numeric Vectors}
\usage{
distance(x = NULL, y = NULL, distance = "euclidean")
}
\arguments{
\item{x}{(required, numeric vector).}

\item{y}{(required, numeric vector) of same length as \code{x}.}

\item{distance}{(optional, character string) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \code{distances}. Default: "euclidean".}
}
\value{
numeric value
}
\description{
Computes the distance between two numeric vectors using a distance metric included in the data frame \code{distantia::distances}.
}
\examples{

distance(
  x = runif(100),
  y = runif(100),
  distance = "euclidean"
)

}
\seealso{
Other distances:
\code{\link[=distance_matrix]{distance_matrix()}},
\code{\link{distances}}
}
\concept{distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_matrix_guide.R
\name{utils_matrix_guide}
\alias{utils_matrix_guide}
\title{Color Guide for Matrix Plot}
\usage{
utils_matrix_guide(
  m = NULL,
  matrix_color = NULL,
  breaks = NULL,
  title = NULL,
  text_cex = 1
)
}
\arguments{
\item{m}{(required, numeric matrix) distance or cost matrix generated by \code{\link[=psi_distance_matrix]{psi_distance_matrix()}} or \code{\link[=psi_cost_matrix]{psi_cost_matrix()}}, but any numeric matrix will work. Default: NULL}

\item{matrix_color}{(optional, character vector) vector of colors. Default: NULL}

\item{breaks}{(optional, numeric vector) vector of breaks for the color guide. Default: NULL}

\item{title}{(optional, character string) guide title. Default: NULL}

\item{text_cex}{(optional, numeric) multiplier for the text size. Default: 1}
}
\value{
Plot
}
\description{
Plot a color legend for a distance or cost matrix for multi-panel plots or external image editors.
}
\examples{
#prepare time series list
tsl <- tsl_simulate(
  n = 2,
  independent = TRUE
)

#distance matrix between time series
dm <- psi_distance_matrix(
  x = tsl[[1]],
  y = tsl[[2]]
)

if(interactive()){
  utils_matrix_guide(m = dm)
}
}
\seealso{
Other internal_plotting:
\code{\link[=color_continuous]{color_continuous()}},
\code{\link[=color_discrete]{color_discrete()}},
\code{\link[=utils_color_breaks]{utils_color_breaks()}},
\code{\link[=utils_line_color]{utils_line_color()}},
\code{\link[=utils_line_guide]{utils_line_guide()}},
\code{\link[=utils_matrix_plot]{utils_matrix_plot()}}
}
\concept{internal_plotting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psi_equation.R
\name{psi_equation}
\alias{psi_equation}
\title{Normalized Dissimilarity Score}
\usage{
psi_equation(a = NULL, b = NULL, diagonal = TRUE)
}
\arguments{
\item{a}{(required, numeric) Result of \code{\link[=psi_cost_path_sum]{psi_cost_path_sum()}}, the sum of distances of the least cost path between two time series. Default: NULL}

\item{b}{(required, numeric) Result of \code{\link[=psi_auto_sum]{psi_auto_sum()}}, the cumulative sum of the consecutive cases of two time series. Default: NULL}

\item{diagonal}{(optional, logical) Used to correct \code{psi} when diagonals are used during the computation of the least cost path. If the cost matrix and least cost path were computed using \code{diagonal = TRUE}, this argument should be \code{TRUE} as well. Default: TRUE}
}
\value{
numeric value
}
\description{
Demonstration function to compute the \code{psi} dissimilarity score (Birks and Gordon 1985). Psi is computed as \eqn{\psi = (2a / b) - 1}, where \eqn{a} is the sum of distances between the relevant samples of two time series, and \eqn{b} is the cumulative sum of distances between consecutive samples in the two time series.

When \code{diagonal = TRUE}, one is added to the result of the equation above.
}
\details{
When \code{b} is zero (both series are perfectly flat with no variation between
consecutive samples), the function returns \code{NA} because the normalisation
denominator is undefined.
}
\examples{
#distance metric
d <- "euclidean"

#use diagonals in least cost computations
diagonal <- TRUE

#simulate two irregular time series
x <- zoo_simulate(
  name = "x",
  rows = 100,
  seasons = 2,
  seed = 1
  )

y <- zoo_simulate(
  name = "y",
  rows = 80,
  seasons = 2,
  seed = 2
  )

if(interactive()){
  zoo_plot(x = x)
  zoo_plot(x = y)
}

#dynamic time warping

#distance matrix
dist_matrix <- psi_distance_matrix(
  x = x,
  y = y,
  distance = d
)

#cost matrix
cost_matrix <- psi_cost_matrix(
  dist_matrix = dist_matrix,
  diagonal = diagonal
)

#least cost path
cost_path <- psi_cost_path(
  dist_matrix = dist_matrix,
  cost_matrix = cost_matrix,
  diagonal = diagonal
)

if(interactive()){
  utils_matrix_plot(
    m = cost_matrix,
    path = cost_path
    )
}


#computation of psi score

#sum of distances in least cost path
a <- psi_cost_path_sum(
  path = cost_path
  )

#auto sum of both time series
b <- psi_auto_sum(
  x = x,
  y = y,
  distance = d
)

#dissimilarity score
psi_equation(
  a = a,
  b = b,
  diagonal = diagonal
)

#full computation with distantia()
tsl <- list(
  x = x,
  y = y
)

distantia(
  tsl = tsl,
  distance = d,
  diagonal = diagonal
)$psi

if(interactive()){
  distantia_dtw_plot(
    tsl = tsl,
    distance = d,
    diagonal = diagonal
  )
}
}
\seealso{
Other psi_demo:
\code{\link[=psi_auto_distance]{psi_auto_distance()}},
\code{\link[=psi_auto_sum]{psi_auto_sum()}},
\code{\link[=psi_cost_matrix]{psi_cost_matrix()}},
\code{\link[=psi_cost_path]{psi_cost_path()}},
\code{\link[=psi_cost_path_sum]{psi_cost_path_sum()}},
\code{\link[=psi_distance_lock_step]{psi_distance_lock_step()}},
\code{\link[=psi_distance_matrix]{psi_distance_matrix()}}
}
\concept{psi_demo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_line_guide.R
\name{utils_line_guide}
\alias{utils_line_guide}
\title{Guide for Time Series Plots}
\usage{
utils_line_guide(
  x = NULL,
  position = "topright",
  line_color = NULL,
  line_width = 1,
  length = 1,
  text_cex = 0.7,
  guide_columns = 1,
  subpanel = FALSE
)
}
\arguments{
\item{x}{(required, sequence) a zoo object or a time series list. Default: NULL}

\item{position}{(optional, vector of xy coordinates or character string). This is a condensed version of the \code{x} and \code{y} arguments of the \code{\link[graphics:legend]{graphics::legend()}} function. Coordinates (in the range 0 1) or keyword to position the legend. Accepted keywords are: "bottomright", "bottom", "bottomleft", "left", "topleft", "top", "topright", "right" and "center". Default: "topright".}

\item{line_color}{(optional, character vector) vector of colors for the time series columns. If NULL, uses the palette "Zissou 1" provided by the function \code{\link[grDevices:hcl.colors]{grDevices::hcl.colors()}}. Default: NULL}

\item{line_width}{(optional, numeric vector) Widths of the time series lines. Default: 1}

\item{length}{(optional, numeric) maps to the argument \code{seg.len} of \code{\link[graphics:legend]{graphics::legend()}}. Length of the lines drawn in the legend. Default: 1}

\item{text_cex}{(optional, numeric) Multiplier of the text size. Default: 0.7}

\item{guide_columns}{(optional, integer) number of columns for the legend items. Default: 1.}

\item{subpanel}{(optional, logical) internal argument used when generating the multipanel plot produced by \code{\link[=distantia_dtw_plot]{distantia_dtw_plot()}}.}
}
\value{
plot
}
\description{
Guide for Time Series Plots
}
\examples{
x <- zoo_simulate()

if(interactive()){

  zoo_plot(x, guide = FALSE)

  utils_line_guide(
    x = x,
    position = "right"
  )

}

}
\seealso{
Other internal_plotting:
\code{\link[=color_continuous]{color_continuous()}},
\code{\link[=color_discrete]{color_discrete()}},
\code{\link[=utils_color_breaks]{utils_color_breaks()}},
\code{\link[=utils_line_color]{utils_line_color()}},
\code{\link[=utils_matrix_guide]{utils_matrix_guide()}},
\code{\link[=utils_matrix_plot]{utils_matrix_plot()}}
}
\concept{internal_plotting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_matrix_plot.R
\name{utils_matrix_plot}
\alias{utils_matrix_plot}
\title{Plot Distance or Cost Matrix and Least Cost Path}
\usage{
utils_matrix_plot(
  m = NULL,
  matrix_color = NULL,
  title = NULL,
  subtitle = NULL,
  xlab = NULL,
  ylab = NULL,
  text_cex = 1,
  path = NULL,
  path_width = 1,
  path_color = "black",
  diagonal_width = 1,
  diagonal_color = "white",
  guide = TRUE,
  subpanel = FALSE
)
}
\arguments{
\item{m}{(required, numeric matrix) distance or cost matrix generated by \code{\link[=psi_distance_matrix]{psi_distance_matrix()}} or \code{\link[=psi_cost_matrix]{psi_cost_matrix()}}, but any numeric matrix will work. Default: NULL}

\item{matrix_color}{(optional, character vector) vector of colors. Uses the palette "Zissou 1" by default. Default: NULL}

\item{title}{(optional, character string) plot title. By default, names of the sequences used to compute the matrix \code{m}. Default: NULL}

\item{subtitle}{(optional, character string) plot subtitle. Default: NULL}

\item{xlab}{(optional, character string) title of the x axis (matrix columns). By default, the name of one of the sequences used to compute the matrix \code{m}. Default: NULL}

\item{ylab}{(optional, character string) title of the y axis (matrix rows). By default, the name of one of the sequences used to compute the matrix \code{m}. Default: NULL}

\item{text_cex}{(optional, numeric) multiplier of the text size for the plot labels and titles. Default: 1}

\item{path}{(optional, data frame) least cost path generated with \code{\link[=psi_cost_path]{psi_cost_path()}}. This data frame must have the attribute \verb{type == "cost_path}, and must have been computed from the same sequences used to compute the matrix \code{m}. Default: NULL.}

\item{path_width}{(optional, numeric) width of the least cost path. Default: 1}

\item{path_color}{(optional, character string) color of the least-cost path. Default: "black"}

\item{diagonal_width}{(optional, numeric) width of the diagonal. Set to 0 to remove the diagonal line. Default: 0.5}

\item{diagonal_color}{(optional, character string) color of the diagonal. Default: "white"}

\item{guide}{(optional, logical) if TRUE, a color guide for the matrix \code{m} is added by \code{\link[=utils_matrix_guide]{utils_matrix_guide()}}.}

\item{subpanel}{(optional, logical) internal argument used when generating the multi-panel plot produced by \code{\link[=distantia_dtw_plot]{distantia_dtw_plot()}}.}
}
\value{
plot
}
\description{
This function is a simplified version of \code{\link[fields:imagePlot]{fields::imagePlot()}}, by \href{https://dnychka.github.io/}{Douglas Nychka}. The original version is recommended in case more customization than the provided here is needed.
}
\examples{
#prepare time series list
tsl <- tsl_simulate(
  n = 2,
  independent = TRUE
)

#distance matrix between time series
dm <- psi_distance_matrix(
  x = tsl[[1]],
  y = tsl[[2]]
)

#cost matrix
cm <- psi_cost_matrix(
  dist_matrix = dm
  )

#least cost path
cp <- psi_cost_path(
  dist_matrix = dm,
  cost_matrix = cm
)

#plot cost matrix and least cost path
if(interactive()){
  utils_matrix_plot(
    m = cm,
    path = cp,
    guide = TRUE
  )
}
}
\seealso{
Other internal_plotting:
\code{\link[=color_continuous]{color_continuous()}},
\code{\link[=color_discrete]{color_discrete()}},
\code{\link[=utils_color_breaks]{utils_color_breaks()}},
\code{\link[=utils_line_color]{utils_line_color()}},
\code{\link[=utils_line_guide]{utils_line_guide()}},
\code{\link[=utils_matrix_guide]{utils_matrix_guide()}}
}
\concept{internal_plotting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_stats.R
\name{tsl_stats}
\alias{tsl_stats}
\title{Summary Statistics of Time Series Lists}
\usage{
tsl_stats(tsl = NULL, lags = 1L)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{lags}{(optional, integer) An integer specifying the number of autocorrelation lags to compute. If NULL, autocorrelation computation is disabled. Default: 1.}
}
\value{
data frame:
\itemize{
\item name: name of the zoo object.
\item rows: rows of the zoo object.
\item columns: columns of the zoo object.
\item time_units: time units of the zoo time series (see \code{\link[=zoo_time]{zoo_time()}}).
\item time_begin: beginning time of the time series.
\item time_end: end time of the time series.
\item time_length: total length of the time series, expressed in time units.
\item time_resolution: average distance between consecutive observations
\item variable: name of the variable, a column of the zoo object.
\item min: minimum value of the zoo column.
\item q1: first quartile (25th percentile).
\item median: 50th percentile.
\item q3: third quartile (75th percentile).
\item max: maximum value.
\item mean: average value.
\item sd: standard deviation.
\item range: range of the variable, computed as max - min.
\item iq_range: interquartile range of the variable, computed as q3 - q1.
\item skewness: asymmetry of the variable distribution.
\item kurtosis: "tailedness" of the variable distribution.
\item ac_lag_1, ac_lag_2, ...: autocorrelation values for the specified lags.
}
}
\description{
This function computes a variety of summary statistics for each time series and numeric column within a time series list. The statistics include common metrics  such as minimum, maximum, quartiles, mean, standard deviation, range, interquartile range, skewness, kurtosis, and autocorrelation for specified lags.

For irregular time series, autocorrelation computation is performed after regularizing the time series via interpolation with \code{\link[=zoo_resample]{zoo_resample()}}. This regularization does not affect the computation of all other stats.

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{


#three time series
#climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
)


#stats computation
df <- tsl_stats(
  tsl = tsl,
  lags = 3
  )

df
}
\seealso{
Other tsl_processing:
\code{\link[=tsl_aggregate]{tsl_aggregate()}},
\code{\link[=tsl_resample]{tsl_resample()}},
\code{\link[=tsl_smooth]{tsl_smooth()}},
\code{\link[=tsl_transform]{tsl_transform()}}
}
\concept{tsl_processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_time_delay.R
\name{distantia_time_delay}
\alias{distantia_time_delay}
\title{Time Delay Between Time Series}
\usage{
distantia_time_delay(
  tsl = NULL,
  distance = "euclidean",
  bandwidth = 1,
  directional = FALSE
)
}
\arguments{
\item{tsl}{(required, time series list) list of zoo time series. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}

\item{bandwidth}{(optional, numeric) Proportion of space at each side of the cost matrix diagonal (aka \emph{Sakoe-Chiba band}) defining a valid region for dynamic time warping, used to control the flexibility of the warping path. This method prevents degenerate alignments due to differences in magnitude between time series when the data is not properly scaled. If \code{1} (default), DTW is unconstrained. If \code{0}, DTW is fully constrained and the warping path follows the matrix diagonal. Recommended values may vary depending on the nature of the data. Ignored if \code{lock_step = TRUE}. Default: 1.}

\item{directional}{(optional, logical) If TRUE, a directional time delay is computed as \verb{x to y} and \verb{y to x}, resulting in two rows per pair of time series. Otherwise, the absolute magnitude of the delay between \code{x} and \code{y} is returned as a single row per pair. Default: TRUE}
}
\value{
data frame
}
\description{
This function computes an approximation to the time-delay between pairs of time series as the difference between observations connected by the dynamic time warping path.

Given a pair of time series \code{x} and \code{y}, and the time of their samples in the dynamic time warping \code{time(x)} and \code{time(y)}, when the argument \code{directional} is \code{TRUE}, the time delay is computed as follows:
\itemize{
\item Time delay from \code{x} to \code{y}: \code{time(y) - time(x)}.
\item Time delay from \code{y} to \code{x}: \code{time(x) - time(y)}
}
In such case, two rows per pair of time series are returned. Otherwise, the time delay is computed as \code{abs(time(y) - time(x))}, and only one row per pair of time series is returned.

If the time series have more than 30 observations, 5\% of cases are omitted at each extreme of the warping path to avoid overestimating time delays due to early misalignments.

The function returns a data frame with the names of the time series in columns \emph{x} and \emph{y}, and summary statistics of the time delay. The mode and median are generally the most accurate time-delay metrics.

This function requires scaled and detrended time series. It may yield non-sensical results in case of degenerate warping paths. Plotting dubious results with \code{\link[=distantia_dtw_plot]{distantia_dtw_plot()}} is a good approach to identify these cases.
}
\examples{
#load two long-term temperature time series
#local scaling to focus on shape rather than values
#polynomial detrending to make them stationary
tsl <- tsl_init(
  x = cities_temperature[
    cities_temperature$name \%in\% c("London", "Kinshasa"),
    ],
  name = "name",
  time = "time"
) |>
  tsl_transform(
    f = f_scale_local
  ) |>
  tsl_transform(
    f = f_detrend_poly,
    degree = 35 #data years
  )


if(interactive()){
  tsl_plot(
    tsl = tsl,
    guide = FALSE
  )
}

#compute shifts
df_shift <- distantia_time_delay(
  tsl = tsl,
  directional = TRUE
)

df_shift
}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{distance_ls_cpp}
\alias{distance_ls_cpp}
\title{(C++) Sum of Pairwise Distances Between Cases in Two Aligned Time Series}
\usage{
distance_ls_cpp(x, y, distance = "euclidean")
}
\arguments{
\item{x}{(required, numeric matrix) univariate or multivariate time series.}

\item{y}{(required, numeric matrix) univariate or multivariate time series
with the same number of columns and rows as 'x'.}

\item{distance}{(optional, character string) distance name from the "names"
column of the dataset \code{distances} (see \code{distances$name}). Default: "euclidean".}
}
\value{
numeric
}
\description{
Computes the lock-step sum of distances between two regular
and aligned time series. NA values should be removed before using this function.
If the selected distance function is "chi" or "cosine", pairs of zeros should
be either removed or replaced with pseudo-zeros (i.e. 0.00001).
}
\examples{
#simulate two regular time series
x <- zoo_simulate(
  seed = 1,
  irregular = FALSE
  )
y <- zoo_simulate(
  seed = 2,
  irregular = FALSE
  )

#distance matrix
dist_matrix <- distance_ls_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)
}
\seealso{
Other Rcpp_matrix:
\code{\link[=cost_matrix_diagonal_cpp]{cost_matrix_diagonal_cpp()}},
\code{\link[=cost_matrix_diagonal_weighted_cpp]{cost_matrix_diagonal_weighted_cpp()}},
\code{\link[=cost_matrix_orthogonal_cpp]{cost_matrix_orthogonal_cpp()}},
\code{\link[=distance_matrix_cpp]{distance_matrix_cpp()}}
}
\concept{Rcpp_matrix}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_transform.R
\name{tsl_transform}
\alias{tsl_transform}
\title{Transform Values in Time Series Lists}
\usage{
tsl_transform(tsl = NULL, f = NULL, ...)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{f}{(required, transformation function) name of a function taking a matrix as input. Currently, the following options are implemented, but any other function taking a matrix as input (for example, \code{\link[=scale]{scale()}}) should work as well:
\itemize{
\item f_proportion: proportion computed by row.
\item f_percent: percentage computed by row.
\item f_hellinger: Hellinger transformation computed by row
\item f_scale_local: Local centering and/or scaling based on the variable mean and standard deviation in each time series within \code{tsl}.
\item f_scale_global: Global centering and/or scaling based on the variable mean and standard deviation across all time series within \code{tsl}.
\item f_smooth: Time series smoothing with a user defined rolling window.
\item f_detrend_difference: Differencing detrending of time series via \code{\link[=diff]{diff()}}.
\item f_detrend_linear: Detrending of seasonal time series via linear modeling.
\item f_detrend_gam: Detrending of seasonal time series via Generalized Additive Models.
}}

\item{...}{(optional, additional arguments of \code{f}) Optional arguments for the transformation function.}
}
\value{
time series list
}
\description{
Function for time series transformations without changes in data dimensions. Generally, functions introduced via the argument \code{f} should not change the dimensions of the output time series list. See \code{\link[=tsl_resample]{tsl_resample()}} and \code{\link[=tsl_aggregate]{tsl_aggregate()}} for transformations requiring changes in time series dimensions.

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}, and progress bars provided by the package \href{https://CRAN.R-project.org/package=progressr}{progressr}.
}
\examples{

#two time series
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
) |>
  tsl_subset(
    names = c("Spain", "Sweden"),
    colnames = c("rainfall", "temperature")
  )

if(interactive()){
  tsl_plot(
    tsl = tsl
  )
}

#centering and scaling
#-----------------------------------------
#same mean and standard deviation are used to scale each variable across all time series
tsl_scale <- tsl_transform(
  tsl = tsl,
  f = f_scale_local
)

if(interactive()){
  tsl_plot(
    tsl = tsl_scale,
    guide_columns = 3
  )
}


#rescaling to a new range
#-----------------------------------------

#rescale between -100 and 100
tsl_rescaled <- tsl_transform(
  tsl = tsl,
  f = f_rescale_local,
  new_min = -100,
  new_max = 100
)

#old range
sapply(X = tsl, FUN = range)

#new range
sapply(X = tsl_rescaled, FUN = range)



#numeric transformations
#-----------------------------------------
#eemian pollen counts
tsl <- tsl_initialize(
  x = distantia::eemian_pollen,
  name_column = "name",
  time_column = "time"
)

if(interactive()){
  tsl_plot(
    tsl = tsl
  )
}

#percentages
tsl_percentage <- tsl_transform(
  tsl = tsl,
  f = f_percent
)

if(interactive()){
  tsl_plot(
    tsl = tsl_percentage
  )
}

#hellinger transformation
tsl_hellinger <- tsl_transform(
  tsl = tsl,
  f = f_hellinger
)

if(interactive()){
  tsl_plot(
    tsl = tsl_hellinger
  )
}

}
\seealso{
Other tsl_processing:
\code{\link[=tsl_aggregate]{tsl_aggregate()}},
\code{\link[=tsl_resample]{tsl_resample()}},
\code{\link[=tsl_smooth]{tsl_smooth()}},
\code{\link[=tsl_stats]{tsl_stats()}}
}
\concept{tsl_processing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_new_time.R
\name{utils_new_time}
\alias{utils_new_time}
\alias{utils_new_time_type}
\title{New Time for Time Series Aggregation}
\usage{
utils_new_time(tsl = NULL, new_time = NULL, keywords = "aggregate")

utils_new_time_type(
  tsl = NULL,
  new_time = NULL,
  keywords = c("resample", "aggregate")
)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{new_time}{(required, zoo object, numeric, numeric vector, Date vector, POSIXct vector, or keyword) breakpoints defining aggregation groups. Options are:
\itemize{
\item numeric vector: only for the "numeric" time class, defines the breakpoints for time series aggregation.
\item "Date" or "POSIXct" vector: as above, but for the time classes "Date" and "POSIXct." In any case, the input vector is coerced to the time class of the \code{tsl} argument.
\item numeric: defines fixed with time intervals for time series aggregation. Used as is when the time class is "numeric", and coerced to integer and interpreted as days for the time classes "Date" and "POSIXct".
\item keyword (see \code{\link[=utils_time_units]{utils_time_units()}} and \code{\link[=tsl_time_summary]{tsl_time_summary()}}): the common options for the time classes "Date" and "POSIXct" are: "millennia", "centuries", "decades", "years", "quarters", "months", and "weeks". Exclusive keywords for the "POSIXct" time class are: "days", "hours", "minutes", and "seconds". The time class "numeric" accepts keywords coded as scientific numbers, from "1e8" to "1e-8".
}}

\item{keywords}{(optional, character string or vector) Defines what keywords are returned. If "aggregate", returns valid keywords for \code{\link[=zoo_aggregate]{zoo_aggregate()}}. If "resample", returns valid keywords for \code{\link[=zoo_resample]{zoo_resample()}}. Default: "aggregate".}
}
\value{
Vector of class numeric, Date, or POSIXct
}
\description{
Internal function called by \code{\link[=tsl_aggregate]{tsl_aggregate()}} and \code{\link[=tsl_resample]{tsl_resample()}} to help transform the input argument \code{new_time} into the proper format for time series aggregation or resampling.
}
\examples{
#three time series
#climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
)

# new time for aggregation using keywords
#-----------------------------------

#get valid keywords for aggregation
tsl_time_summary(
  tsl = tsl,
  keywords = "aggregate"
)$keywords

#if no keyword is used, for aggregation the highest resolution keyword is selected automatically
new_time <- utils_new_time(
  tsl = tsl,
  new_time = NULL,
  keywords = "aggregate"
)

new_time

#if no keyword is used
#for resampling a regular version
#of the original time based on the
#average resolution is used instead
new_time <- utils_new_time(
  tsl = tsl,
  new_time = NULL,
  keywords = "resample"
)

new_time

#aggregation time vector form keyword "years"
new_time <- utils_new_time(
  tsl = tsl,
  new_time = "years",
  keywords = "aggregate"
)

new_time

#same from shortened keyword
#see utils_time_keywords_dictionary()
utils_new_time(
  tsl = tsl,
  new_time = "year",
  keywords = "aggregate"
)

#same for abbreviated keyword
utils_new_time(
  tsl = tsl,
  new_time = "y",
  keywords = "aggregate"
)

#from a integer defining a time interval in days
utils_new_time(
  tsl = tsl,
  new_time = 365,
  keywords = "aggregate"
)

#using this vector as input for aggregation
tsl_aggregated <- tsl_aggregate(
  tsl = tsl,
  new_time = new_time
)
}
\seealso{
Other internal_time_handling:
\code{\link[=utils_as_time]{utils_as_time()}},
\code{\link[=utils_coerce_time_class]{utils_coerce_time_class()}},
\code{\link[=utils_is_time]{utils_is_time()}},
\code{\link[=utils_time_keywords]{utils_time_keywords()}},
\code{\link[=utils_time_keywords_dictionary]{utils_time_keywords_dictionary()}},
\code{\link[=utils_time_keywords_translate]{utils_time_keywords_translate()}},
\code{\link[=utils_time_units]{utils_time_units()}}
}
\concept{internal_time_handling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{f_binary}
\alias{f_binary}
\title{Data Transformation: Convert Zoo Object to Binary}
\usage{
f_binary(x = NULL, threshold = NULL)
}
\arguments{
\item{x}{(required, zoo object) Zoo time series object to transform.}

\item{threshold}{(required, numeric) Values greater than this number become 1, others become 0. Default: NULL (set to the mean of the time series)}
}
\value{
zoo object
}
\description{
Converts a zoo object to binary (1 and 0) based on a given threshold.
}
\examples{
x <- zoo_simulate(
  data_range = c(0, 1)
  )

y <- f_binary(
  x = x,
  threshold = 0.5
)

if(interactive()){
  zoo_plot(x)
  zoo_plot(y)
}
}
\seealso{
Other tsl_transformation:
\code{\link[=f_clr]{f_clr()}},
\code{\link[=f_detrend_difference]{f_detrend_difference()}},
\code{\link[=f_detrend_linear]{f_detrend_linear()}},
\code{\link[=f_detrend_poly]{f_detrend_poly()}},
\code{\link[=f_hellinger]{f_hellinger()}},
\code{\link[=f_list]{f_list()}},
\code{\link[=f_log]{f_log()}},
\code{\link[=f_percent]{f_percent()}},
\code{\link[=f_proportion]{f_proportion()}},
\code{\link[=f_proportion_sqrt]{f_proportion_sqrt()}},
\code{\link[=f_rescale_global]{f_rescale_global()}},
\code{\link[=f_rescale_local]{f_rescale_local()}},
\code{\link[=f_scale_global]{f_scale_global()}},
\code{\link[=f_scale_local]{f_scale_local()}},
\code{\link[=f_trend_linear]{f_trend_linear()}},
\code{\link[=f_trend_poly]{f_trend_poly()}}
}
\concept{tsl_transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_plot.R
\name{tsl_plot}
\alias{tsl_plot}
\title{Plot Time Series List}
\usage{
tsl_plot(
  tsl = NULL,
  columns = 1,
  xlim = NULL,
  ylim = "absolute",
  line_color = NULL,
  line_width = 1,
  text_cex = 1,
  guide = TRUE,
  guide_columns = 1,
  guide_cex = 0.8
)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{columns}{(optional, integer) Number of columns of the multipanel plot. Default: 1}

\item{xlim}{(optional, numeric vector) Numeric vector with the limits of the x axis. Applies to all sequences. Default: NULL}

\item{ylim}{(optional, numeric vector or character string) Numeric vector of length two with the limits of the vertical axis or a keyword. Accepted keywords are:
\itemize{
\item "absolute" (default): all time series are plotted using the overall data range. When this option is used, horizontal lines indicating the overall mean, minimum, and maximum are shown as reference.
\item "relative": each time series is plotted using its own range. Equivalent result can be achieved using \code{ylim = NULL}.
}}

\item{line_color}{(optional, character vector) vector of colors for the distance or cost matrix. If NULL, uses an appropriate palette generated with \code{\link[grDevices:palette.colors]{grDevices::palette.colors()}}. Default: NULL}

\item{line_width}{(optional, numeric vector) Width of the time series plot. Default: 1}

\item{text_cex}{(optional, numeric) Multiplicator of the text size. Default: 1}

\item{guide}{(optional, logical) If TRUE, plots a legend. Default: TRUE}

\item{guide_columns}{(optional, integer) Columns of the line guide. Default: 1}

\item{guide_cex}{(optional, numeric) Size of the guide's text and separation between the guide's rows. Default: 0.7.}
}
\value{
plot
}
\description{
Plot Time Series List
}
\examples{
#simulate zoo time series
tsl <- tsl_simulate(
  cols = 3
  )

if(interactive()){

  #default plot
  tsl_plot(
    tsl = tsl
    )

  #relative vertical limits
  tsl_plot(
    tsl = tsl,
    ylim = "relative"
  )

  #changing layout
  tsl_plot(
    tsl = tsl,
    columns = 2,
    guide_columns = 2
  )

  #no legend
  tsl_plot(
    tsl = tsl,
    guide = FALSE
  )

  #changing color
  tsl_plot(
    tsl = tsl,
    line_color = c("red", "green", "blue"))

}
}
\concept{tsl_visualization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_aggregate.R
\name{zoo_aggregate}
\alias{zoo_aggregate}
\title{Aggregate Cases in Zoo Time Series}
\usage{
zoo_aggregate(x = NULL, new_time = NULL, f = mean, ...)
}
\arguments{
\item{x}{(required, zoo object) Time series to aggregate. Default: NULL}

\item{new_time}{(optional, zoo object, keyword, or time vector) New time to aggregate \code{x} to. The available options are:
\itemize{
\item NULL: the highest resolution keyword returned by \code{zoo_time(x)$keywords} is used to generate a new time vector to aggregate \code{x}.
\item zoo object: the index of the given zoo object is used as template to aggregate \code{x}.
\item time vector: a vector with new times to resample \code{x} to. If time in \code{x} is of class "numeric", this vector must be numeric as well. Otherwise, vectors of classes "Date" and "POSIXct" can be used indistinctly.
\item keyword: a valid keyword returned by \code{zoo_time(x)$keywords}, used to generate a time vector with the relevant units.
\item numeric of length 1: interpreted as new time interval, in the highest resolution units returned by \code{zoo_time(x)$units}.
}}

\item{f}{(optional, quoted or unquoted function name) Name of a function to aggregate numeric vectors. Typical examples are \code{mean}, \code{max},\code{min}, \code{median}, and \code{quantile}. Default: \code{mean}.}

\item{...}{(optional, additional arguments) additional arguments to \code{f}.}
}
\value{
zoo object
}
\description{
Aggregate Cases in Zoo Time Series
}
\examples{
#full range of calendar dates
x <- zoo_simulate(
  rows = 1000,
  time_range = c(
    "0000-01-01",
    as.character(Sys.Date())
    )
)

#plot time series
if(interactive()){
  zoo_plot(x)
}


#find valid aggregation keywords
x_time <- zoo_time(x)
x_time$keywords

#mean value by millennia (extreme case!!!)
x_millennia <- zoo_aggregate(
  x = x,
  new_time = "millennia",
  f = mean
)

if(interactive()){
  zoo_plot(x_millennia)
}

#max value by centuries
x_centuries <- zoo_aggregate(
  x = x,
  new_time = "centuries",
  f = max
)

if(interactive()){
  zoo_plot(x_centuries)
}

#quantile 0.75 value by centuries
x_centuries <- zoo_aggregate(
  x = x,
  new_time = "centuries",
  f = stats::quantile,
  probs = 0.75 #argument of stats::quantile()
)

if(interactive()){
  zoo_plot(x_centuries)
}
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distantia_aggregate.R
\name{distantia_aggregate}
\alias{distantia_aggregate}
\title{Aggregate \code{distantia()} Data Frames Across Parameter Combinations}
\usage{
distantia_aggregate(df = NULL, f = mean, ...)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=distantia]{distantia()}}, \code{\link[=distantia_ls]{distantia_ls()}}, \code{\link[=distantia_dtw]{distantia_dtw()}}, or \code{\link[=distantia_time_delay]{distantia_time_delay()}}. Default: NULL}

\item{f}{(optional, function) Function to summarize psi scores (for example, \code{mean}) when there are several combinations of parameters in \code{df}. Ignored when there is a single combination of arguments in the input. Default: \code{mean}}

\item{...}{(optional, arguments of \code{f}) Further arguments to pass to the function \code{f}.}
}
\value{
data frame
}
\description{
The function \code{\link[=distantia]{distantia()}} allows dissimilarity assessments based on several combinations of arguments at once. For example, when the argument \code{distance} is set to \code{c("euclidean", "Manhattan")}, the output data frame will show two dissimilarity scores for each pair of compared time series, one based on Euclidean distances, and another based on Manhattan distances.

This function computes dissimilarity stats across combinations of parameters.

If psi scores smaller than zero occur in the aggregated output, then the smaller psi value is added to the column \code{psi} to start dissimilarity scores at zero.

If there is only one combination of arguments in the input data frame, no aggregation occurs and all parameter columns are removed.
}
\examples{
#three time series
#climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

if(interactive()){
  tsl_plot(
    tsl = tsl,
    guide_columns = 3
    )
}

#distantia with multiple parameter combinations
#-------------------------------------
df <- distantia(
  tsl = tsl,
  distance = c("euclidean", "manhattan"),
  lock_step = TRUE
)

df[, c(
  "x",
  "y",
  "distance",
  "psi"
)]

#aggregation using means
df <- distantia_aggregate(
  df = df,
  f = mean
)

df
}
\seealso{
Other distantia_support:
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_block_size]{utils_block_size()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_block_size.R
\name{utils_block_size}
\alias{utils_block_size}
\title{Default Block Size for Restricted Permutation in Dissimilarity Analyses}
\usage{
utils_block_size(tsl = NULL, block_size = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL.}

\item{block_size}{(optional, integer vector) Row block sizes for restricted permutation tests. Used only when the permutation method is "restricted" or "restricted_by_row". A block of size \code{n} indicates that a row can only be permuted within a block of \code{n} adjacent rows. If NULL, defaults to 20 percent of the rows of the shortest time series in \code{tsl}. Minimum value is 2; maximum value is 50 percent of the rows of the shortest time series in \code{tsl}. Default: NULL.}
}
\value{
An integer.
}
\description{
Default Block Size for Restricted Permutation in Dissimilarity Analyses
}
\seealso{
Other distantia_support:
\code{\link[=distantia_aggregate]{distantia_aggregate()}},
\code{\link[=distantia_boxplot]{distantia_boxplot()}},
\code{\link[=distantia_cluster_hclust]{distantia_cluster_hclust()}},
\code{\link[=distantia_cluster_kmeans]{distantia_cluster_kmeans()}},
\code{\link[=distantia_matrix]{distantia_matrix()}},
\code{\link[=distantia_model_frame]{distantia_model_frame()}},
\code{\link[=distantia_spatial]{distantia_spatial()}},
\code{\link[=distantia_stats]{distantia_stats()}},
\code{\link[=distantia_time_delay]{distantia_time_delay()}},
\code{\link[=utils_cluster_hclust_optimizer]{utils_cluster_hclust_optimizer()}},
\code{\link[=utils_cluster_kmeans_optimizer]{utils_cluster_kmeans_optimizer()}},
\code{\link[=utils_cluster_silhouette]{utils_cluster_silhouette()}}
}
\concept{distantia_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance_matrix.R
\name{distance_matrix}
\alias{distance_matrix}
\title{Data Frame to Distance Matrix}
\usage{
distance_matrix(df = NULL, name_column = NULL, distance = "euclidean")
}
\arguments{
\item{df}{(required, data frame) Data frame with numeric columns to transform into a distance matrix. Default: NULL}

\item{name_column}{(optional, column name) Column naming individual time series. Numeric names are converted to character with the prefix "X". Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}
}
\value{
square matrix
}
\description{
Data Frame to Distance Matrix
}
\examples{
#compute distance matrix
m <- distance_matrix(
  df = cities_coordinates,
  name_column = "name",
  distance = "euclidean"
)

#get data used to compute the matrix
attributes(m)$df

#check matrix
m

}
\seealso{
Other distances:
\code{\link[=distance]{distance()}},
\code{\link{distances}}
}
\concept{distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/momentum_dtw.R
\name{momentum_dtw}
\alias{momentum_dtw}
\title{Dynamic Time Warping Variable Importance Analysis of Multivariate Time Series Lists}
\usage{
momentum_dtw(tsl = NULL, distance = "euclidean")
}
\arguments{
\item{tsl}{(required, time series list) list of zoo time series. Default: NULL}

\item{distance}{(optional, character vector) name or abbreviation of the distance method. Valid values are in the columns "names" and "abbreviation" of the dataset \link{distances}. Default: "euclidean".}
}
\value{
data frame:
\itemize{
\item \code{x}: name of the time series \code{x}.
\item \code{y}: name of the time series \code{y}.
\item \code{psi}: psi score of \code{x} and \code{y}.
\item \code{variable}: name of the individual variable.
\item \code{importance}: importance score of the variable.
\item \code{effect}: interpretation of the "importance" column, with the values "increases similarity" and "decreases similarity".
}
}
\description{
A minimalistic but slightly faster version of \code{\link[=momentum]{momentum()}} that computes dynamic time warping importance analysis with the "robust" setup in multivariate time series lists.
}
\examples{

tsl <- tsl_initialize(
  x = distantia::albatross,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

df <- momentum_dtw(
  tsl = tsl,
  distance = "euclidean"
  )

#focus on important columns
df[, c(
  "x",
  "y",
  "variable",
  "importance",
  "effect"
  )]

}
\seealso{
Other momentum:
\code{\link[=momentum]{momentum()}},
\code{\link[=momentum_ls]{momentum_ls()}}
}
\concept{momentum}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_optimize_loess.R
\name{utils_optimize_loess}
\alias{utils_optimize_loess}
\title{Optimize Loess Models for Time Series Resampling}
\usage{
utils_optimize_loess(x = NULL, y = NULL, max_complexity = FALSE)
}
\arguments{
\item{x}{(required, numeric vector) predictor, a time vector coerced to numeric. Default: NULL}

\item{y}{(required, numeric vector) response, a column of a zoo object. Default: NULL}

\item{max_complexity}{(required, logical). If TRUE, RMSE optimization is ignored, and the model of maximum complexity is returned. Default: FALSE}
}
\value{
Loess model.
}
\description{
Internal function used in \code{\link[=zoo_resample]{zoo_resample()}}. It finds the \code{span} parameter of a univariate Loess (Locally Estimated Scatterplot Smoothing) model \code{y ~ x} fitted with \code{\link[stats:loess]{stats::loess()}} that minimizes the root mean squared error (RMSE) between observations and predictions, and returns a model fitted with such \code{span}.
}
\examples{

#zoo time series
xy <- zoo_simulate(
  cols = 1,
  rows = 30
)

#optimize Loess model
m <- utils_optimize_loess(
  x = as.numeric(zoo::index(xy)), #predictor
  y = xy[, 1] #response
)

print(m)

#plot observation
plot(
  x = zoo::index(xy),
  y = xy[, 1],
  col = "forestgreen",
  type = "l",
  lwd = 2
  )

#plot prediction
points(
  x = zoo::index(xy),
  y = stats::predict(
    object = m,
    newdata = as.numeric(zoo::index(xy))
    ),
  col = "red4"
  )

}
\seealso{
Other tsl_processing_internal:
\code{\link[=utils_drop_geometry]{utils_drop_geometry()}},
\code{\link[=utils_global_scaling_params]{utils_global_scaling_params()}},
\code{\link[=utils_optimize_spline]{utils_optimize_spline()}},
\code{\link[=utils_rescale_vector]{utils_rescale_vector()}}
}
\concept{tsl_processing_internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_simulate.R
\name{tsl_simulate}
\alias{tsl_simulate}
\title{Simulate a Time Series List}
\usage{
tsl_simulate(
  n = 2,
  cols = 5,
  rows = 100,
  time_range = c("2010-01-01", "2020-01-01"),
  data_range = c(0, 1),
  seasons = 0,
  na_fraction = 0,
  independent = FALSE,
  irregular = TRUE,
  seed = NULL
)
}
\arguments{
\item{n}{(optional, integer) Number of time series to simulate. Default: 2.}

\item{cols}{(optional, integer) Number of columns of each time series. Default: 5}

\item{rows}{(optional, integer) Length of each time series. Minimum is 10, but maximum is not limited. Very large numbers might crash the R session. Default: 100}

\item{time_range}{(optional character or numeric vector) Time interval of the time series. Either a character vector with dates in format YYYY-MM-DD or a numeric vector. If there is a mismatch between \code{time_range} and \code{rows} (for example, the number of days in \code{time_range} is smaller than \code{rows}), the upper value in \code{time_range} is adapted to \code{rows}. Default: c("2010-01-01", "2020-01-01")}

\item{data_range}{(optional, numeric vector of length 2) Extremes of the time series values. Default: c(0, 1)}

\item{seasons}{(optional, integer) Number of seasons in the resulting time series. The maximum number of seasons is computed as \code{floor(rows/3)}. Default: 0}

\item{na_fraction}{(optional, numeric) Value between 0 and 0.5 indicating the approximate fraction of NA data in the simulated time series. Default: 0.}

\item{independent}{(optional, logical) If TRUE, each new column in a simulated time series is averaged with the previous column to generate dependency across columns, and each new simulated time series is weighted-averaged with a time series template to generate dependency across time series. Irrelevant when \code{cols < 2} or \code{n < 2}, and hard to perceive in the output when \code{seasons > 0}. Default: FALSE}

\item{irregular}{(optional, logical) If TRUE, the time intervals between consecutive samples and the number of rows are irregular. Default: TRUE}

\item{seed}{(optional, integer) Random seed used to simulate the zoo object. If NULL (default), a seed is selected at random. Default: NULL}
}
\value{
time series list
}
\description{
Generates simulated time series lists for testing and learning.

This function supports progress bars generated by the \code{progressr} package, and accepts a parallelization setup via \code{\link[future:plan]{future::plan()}} (see examples).
}
\examples{

# generates a different time series list on each iteration when seed = NULL
tsl <- tsl_simulate(
  n = 2,
  seasons = 4
)

if(interactive()){
  tsl_plot(
    tsl = tsl
  )
}

# generate 3 independent time series
tsl_independent <- tsl_simulate(
  n = 3,
  cols = 3,
  independent = TRUE
)

if(interactive()){
  tsl_plot(
    tsl = tsl_independent
  )
}

# generate 3 independent time series
tsl_dependent <- tsl_simulate(
  n = 3,
  cols = 3,
  independent = FALSE
)

if(interactive()){
  tsl_plot(
    tsl = tsl_dependent
  )
}

# with seasons
tsl_seasons <- tsl_simulate(
  n = 3,
  cols = 3,
  seasons = 4,
  independent = FALSE
)

if(interactive()){
  tsl_plot(
    tsl = tsl_seasons
  )
}

}
\seealso{
Other simulate_time_series:
\code{\link[=zoo_simulate]{zoo_simulate()}}
}
\concept{simulate_time_series}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_simulate.R
\name{zoo_simulate}
\alias{zoo_simulate}
\title{Simulate a Zoo Time Series}
\usage{
zoo_simulate(
  name = "A",
  cols = 5,
  rows = 100,
  time_range = c("2010-01-01", "2020-01-01"),
  data_range = c(0, 1),
  seasons = 0,
  na_fraction = 0,
  independent = FALSE,
  irregular = TRUE,
  seed = NULL
)
}
\arguments{
\item{name}{(optional, character string) Name of the zoo object, to be stored in the attribute "name". Default: "A"}

\item{cols}{(optional, integer) Number of time series. Default: 5}

\item{rows}{(optional, integer) Length of the time series. Minimum is 10, but maximum is not limited. Very large numbers might crash the R session. Default: 100}

\item{time_range}{(optional character or numeric vector) Interval of the time series. Either a character vector with dates in format YYYY-MM-DD or or a numeric vector. If there is a mismatch between \code{time_range} and \code{rows} (for example, the number of days in \code{time_range} is smaller than \code{rows}), the upper value in \code{time_range} is adapted to \code{rows}. Default: c("2010-01-01", "2020-01-01")}

\item{data_range}{(optional, numeric vector of length 2) Extremes of the simulated time series values. The simulated time series are independently adjusted to random values within the provided range. Default: c(0, 1)}

\item{seasons}{(optional, integer) Number of seasons in the resulting time series. The maximum number of seasons is computed as \code{floor(rows/3)}. Default: 0}

\item{na_fraction}{(optional, numeric) Value between 0 and 0.5 indicating the approximate fraction of NA data in the simulated time series. Default: 0.}

\item{independent}{(optional, logical) If TRUE, each new column in a simulated time series is averaged with the previous column. Irrelevant when \code{cols <= 2}, and hard to perceive in the output when \code{seasons > 0}. Default: FALSE}

\item{irregular}{(optional, logical) If TRUE, the time series is created with 20 percent more rows, and a random 20 percent of rows are removed at random. Default: TRUE}

\item{seed}{(optional, integer) Random seed used to simulate the zoo object. Default: NULL}
}
\value{
zoo object
}
\description{
Generates simulated zoo time series.
}
\examples{

#generates a different time series on each execution when 'seed = NULL'
x <- zoo_simulate()

#returns a zoo object
class(x)

#time series names are uppercase letters
#this attribute is not defined in the zoo class and might be lost during data transformations
attributes(x)$name

#column names are lowercase letters
names(x)

#plotting methods
if(interactive()){

  #plot time series with default zoo method
  plot(x)

  #plot time series with distantia
  zoo_plot(
    x = x,
    xlab = "Date",
    ylab = "Value",
    title = "My time series"
  )

}
}
\seealso{
Other simulate_time_series:
\code{\link[=tsl_simulate]{tsl_simulate()}}
}
\concept{simulate_time_series}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_prepare_zoo_list.R
\name{utils_prepare_zoo_list}
\alias{utils_prepare_zoo_list}
\title{Convert List of Data Frames to List of Zoo Objects}
\usage{
utils_prepare_zoo_list(x = NULL, time_column = NULL)
}
\arguments{
\item{x}{(required, list of data frames). Default: NULL.}

\item{time_column}{(required, column name) Name of the column representing time, if any. Default: NULL.}
}
\value{
A named list of zoo objects.
}
\description{
Convert List of Data Frames to List of Zoo Objects
}
\examples{
x <- utils_prepare_zoo_list(
  x = list(
    spain = fagus_dynamics[fagus_dynamics$name == "Spain", ],
    sweden = fagus_dynamics[fagus_dynamics$name == "Sweden", ]
  ),
  time_column = "time"
)
}
\seealso{
Other internal:
\code{\link[=utils_boxplot_common]{utils_boxplot_common()}},
\code{\link[=utils_check_args_distantia]{utils_check_args_distantia()}},
\code{\link[=utils_check_args_matrix]{utils_check_args_matrix()}},
\code{\link[=utils_check_args_momentum]{utils_check_args_momentum()}},
\code{\link[=utils_check_args_path]{utils_check_args_path()}},
\code{\link[=utils_check_args_tsl]{utils_check_args_tsl()}},
\code{\link[=utils_check_args_zoo]{utils_check_args_zoo()}},
\code{\link[=utils_check_distance_args]{utils_check_distance_args()}},
\code{\link[=utils_check_list_class]{utils_check_list_class()}},
\code{\link[=utils_clean_names]{utils_clean_names()}},
\code{\link[=utils_digits]{utils_digits()}},
\code{\link[=utils_distantia_df_split]{utils_distantia_df_split()}},
\code{\link[=utils_prepare_df]{utils_prepare_df()}},
\code{\link[=utils_prepare_matrix]{utils_prepare_matrix()}},
\code{\link[=utils_prepare_matrix_list]{utils_prepare_matrix_list()}},
\code{\link[=utils_prepare_time]{utils_prepare_time()}},
\code{\link[=utils_prepare_vector_list]{utils_prepare_vector_list()}},
\code{\link[=utils_tsl_pairs]{utils_tsl_pairs()}}
}
\concept{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformations.R
\name{f_list}
\alias{f_list}
\title{Lists Available Transformation Functions}
\usage{
f_list()
}
\value{
character vector
}
\description{
Lists Available Transformation Functions
}
\examples{
f_list()
}
\seealso{
Other tsl_transformation:
\code{\link[=f_binary]{f_binary()}},
\code{\link[=f_clr]{f_clr()}},
\code{\link[=f_detrend_difference]{f_detrend_difference()}},
\code{\link[=f_detrend_linear]{f_detrend_linear()}},
\code{\link[=f_detrend_poly]{f_detrend_poly()}},
\code{\link[=f_hellinger]{f_hellinger()}},
\code{\link[=f_log]{f_log()}},
\code{\link[=f_percent]{f_percent()}},
\code{\link[=f_proportion]{f_proportion()}},
\code{\link[=f_proportion_sqrt]{f_proportion_sqrt()}},
\code{\link[=f_rescale_global]{f_rescale_global()}},
\code{\link[=f_rescale_local]{f_rescale_local()}},
\code{\link[=f_scale_global]{f_scale_global()}},
\code{\link[=f_scale_local]{f_scale_local()}},
\code{\link[=f_trend_linear]{f_trend_linear()}},
\code{\link[=f_trend_poly]{f_trend_poly()}}
}
\concept{tsl_transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{eemian_pollen}
\alias{eemian_pollen}
\title{Pollen Counts of Nine Interglacial Sites in Central Europe}
\format{
A data frame with 24 columns and 376 rows.
}
\usage{
data(eemian_pollen)
}
\description{
Pollen counts of nine interglacial sites in central Europe.

Site coordinates for this dataset are in \link{eemian_coordinates}.
}
\examples{
data("eemian_pollen")

#Convert to time series list
tsl <- tsl_initialize(
  x = eemian_pollen,
  name_column = "name",
  time_column = "time"
)

#time series plot
if(interactive()){

 tsl_plot(
  tsl = tsl_subset(
    tsl = tsl,
    names = 1:3
    ),
  columns = 2,
  guide_columns = 2
  )

}

}
\seealso{
Other example_data:
\code{\link{albatross}},
\code{\link{cities_coordinates}},
\code{\link{cities_temperature}},
\code{\link{covid_counties}},
\code{\link{covid_prevalence}},
\code{\link{eemian_coordinates}},
\code{\link{fagus_coordinates}},
\code{\link{fagus_dynamics}},
\code{\link{honeycomb_climate}},
\code{\link{honeycomb_polygons}}
}
\concept{example_data}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{cost_path_orthogonal_cpp}
\alias{cost_path_orthogonal_cpp}
\title{(C++) Orthogonal Least Cost Path}
\usage{
cost_path_orthogonal_cpp(dist_matrix, cost_matrix)
}
\arguments{
\item{dist_matrix}{(required, numeric matrix). Distance matrix between two
time series.}

\item{cost_matrix}{(required, numeric matrix). Cost matrix generated from
\code{dist_matrix}.}
}
\value{
data frame
}
\description{
Computes an orthogonal least-cost path within a cost matrix. Each steps within
the least-cost path either moves in the x or the y direction, but never diagonally.
}
\examples{
#simulate two time series
x <- zoo_simulate(seed = 1)
y <- zoo_simulate(seed = 2)

#distance matrix
dist_matrix <- distance_matrix_cpp(
  x = x,
  y = y,
  distance = "euclidean"
)

#least cost matrix
cost_matrix <- cost_matrix_orthogonal_cpp(
  dist_matrix = dist_matrix
)

#least cost path
cost_path <- cost_path_orthogonal_cpp(
  dist_matrix = dist_matrix,
  cost_matrix = cost_matrix
)

cost_path
}
\seealso{
Other Rcpp_cost_path:
\code{\link[=cost_path_cpp]{cost_path_cpp()}},
\code{\link[=cost_path_diagonal_bandwidth_cpp]{cost_path_diagonal_bandwidth_cpp()}},
\code{\link[=cost_path_diagonal_cpp]{cost_path_diagonal_cpp()}},
\code{\link[=cost_path_orthogonal_bandwidth_cpp]{cost_path_orthogonal_bandwidth_cpp()}},
\code{\link[=cost_path_slotting_cpp]{cost_path_slotting_cpp()}},
\code{\link[=cost_path_sum_cpp]{cost_path_sum_cpp()}},
\code{\link[=cost_path_trim_cpp]{cost_path_trim_cpp()}}
}
\concept{Rcpp_cost_path}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/momentum_aggregate.R
\name{momentum_aggregate}
\alias{momentum_aggregate}
\title{Aggregate \code{momentum()} Data Frames Across Parameter Combinations}
\usage{
momentum_aggregate(df = NULL, f = mean, ...)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=momentum]{momentum()}}, \code{\link[=momentum_ls]{momentum_ls()}}, or \code{\link[=momentum_dtw]{momentum_dtw()}}. Default: NULL}

\item{f}{(optional, function) Function to summarize importance scores (for example, \code{mean}) when there are several combinations of parameters in \code{df}. Ignored when there is a single combination of arguments in the input. Default: \code{mean}}

\item{...}{(optional, arguments of \code{f}) Further arguments to pass to the function \code{f}.}
}
\value{
data frame
}
\description{
The function \code{\link[=momentum]{momentum()}} allows variable importance assessments based on several combinations of arguments at once. For example, when the argument \code{distance} is set to \code{c("euclidean", "manhattan")}, the output data frame will show two importance scores for each pair of compared time series and variable, one based on euclidean distances, and another based on manhattan distances.

This function computes importance statistics across combinations of parameters.

If there are no different combinations of arguments in the input data frame, no aggregation happens, but all parameter columns are removed.
}
\examples{
#three time series
#climate and ndvi in Fagus sylvatica stands in Spain, Germany, and Sweden
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

if(interactive()){
  tsl_plot(
    tsl = tsl,
    guide_columns = 3
    )
}

#momentum with multiple parameter combinations
#-------------------------------------
df <- momentum(
  tsl = tsl,
  distance = c("euclidean", "manhattan"),
  lock_step = TRUE
)

df[, c(
  "x",
  "y",
  "distance",
  "importance"
)]

#aggregation using means
df <- momentum_aggregate(
  df = df,
  f = mean
)

df
}
\seealso{
Other momentum_support:
\code{\link[=momentum_boxplot]{momentum_boxplot()}},
\code{\link[=momentum_model_frame]{momentum_model_frame()}},
\code{\link[=momentum_spatial]{momentum_spatial()}},
\code{\link[=momentum_stats]{momentum_stats()}},
\code{\link[=momentum_to_wide]{momentum_to_wide()}}
}
\concept{momentum_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/momentum_boxplot.R
\name{momentum_boxplot}
\alias{momentum_boxplot}
\title{Momentum Boxplot}
\usage{
momentum_boxplot(df = NULL, fill_color = NULL, f = median, text_cex = 1)
}
\arguments{
\item{df}{(required, data frame) Output of \code{\link[=momentum]{momentum()}}, \code{\link[=momentum_ls]{momentum_ls()}}, or \code{\link[=momentum_dtw]{momentum_dtw()}}. Default: NULL}

\item{fill_color}{(optional, character vector) boxplot fill color. Default: NULL}

\item{f}{(optional, function) Function to summarize importance scores (for example, \code{mean}) when there are several combinations of parameters in \code{df}. Ignored when there is a single combination of arguments in the input. Default: \code{mean}}

\item{text_cex}{(optional, numeric) Multiplier of the text size. Default: 1}
}
\value{
boxplot
}
\description{
Creates a boxplot of a data frame returned by \code{\link[=momentum]{momentum()}}, summarizing the contribution to similarity (negative) and/or dissimilarity (positive) of individual variables across all time series.
}
\examples{
tsl <- tsl_initialize(
  x = distantia::albatross,
  name_column = "name",
  time_column = "time"
) |>
  tsl_transform(
    f = f_scale_global
  )

df <- momentum(
  tsl = tsl,
  lock_step = TRUE
  )

momentum_boxplot(
  df = df
  )
}
\seealso{
Other momentum_support:
\code{\link[=momentum_aggregate]{momentum_aggregate()}},
\code{\link[=momentum_model_frame]{momentum_model_frame()}},
\code{\link[=momentum_spatial]{momentum_spatial()}},
\code{\link[=momentum_stats]{momentum_stats()}},
\code{\link[=momentum_to_wide]{momentum_to_wide()}}
}
\concept{momentum_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/momentum_model_frame.R
\name{momentum_model_frame}
\alias{momentum_model_frame}
\title{Importance Model Frame}
\usage{
momentum_model_frame(
  response_df = NULL,
  predictors_df = NULL,
  composite_predictors = NULL,
  scale = TRUE,
  distance = "euclidean"
)
}
\arguments{
\item{response_df}{(required, data frame) Output of \code{\link[=momentum]{momentum()}}, \code{\link[=momentum_ls]{momentum_ls()}}, or \code{\link[=momentum_dtw]{momentum_dtw()}}. Default: NULL}

\item{predictors_df}{(required, data frame or sf data frame) Data frame with numeric predictors for the model frame. Must have a column with the time series names in \code{response_df$x} and \code{response_df$y}. If an \code{sf} data frame, the column "geographic_distance" with distances between pairs of time series is added to the model frame. Default: NULL}

\item{composite_predictors}{(optional, list) List defining composite predictors. For example, \code{composite_predictors = list(a = c("b", "c"))} uses the columns \code{"b"} and \code{"c"} from \code{predictors_df} to generate the predictor \code{a} as the multivariate distance between \code{"b"} and \code{"c"} for each pair of time series in \code{response_df}. Default: NULL}

\item{scale}{(optional, logical) If TRUE, all predictors are scaled and centered with \code{\link[=scale]{scale()}}. Default: TRUE}

\item{distance}{(optional, string) Method to compute the distance between predictor values for all pairs of time series in \code{response_df}. Default: "euclidean".}
}
\value{
A data frame with the attribute "predictors".
}
\description{
This function generates a model frame for statistical or machine learning analysis from these objects:

\itemize{
\item An importance data frame generated by \code{\link[=momentum]{momentum()}}, \code{\link[=momentum_ls]{momentum_ls()}}, or \code{\link[=momentum_dtw]{momentum_dtw()}}. The output model frame will have as many rows as this data frame.
\item A data frame with static descriptors of the time series. These descriptors are converted to distances between pairs of time series via \code{\link[=distance_matrix]{distance_matrix()}}.
\item A list defining composite predictors. This feature allows grouping together predictors that have a common meaning. For example, \code{composite_predictors = list(temperature = c("temperature_mean", "temperature_min", "temperature_max"))} generates a new predictor named "temperature", which results from computing the multivariate distances between the vectors of temperature variables of each pair of time series. Predictors in one such group will be scaled before distance computation if their maximum standard deviations differ by a factor of 10 or more.
}

The resulting data frame contains the following columns:
\itemize{
\item \code{x} and \code{y}: names of the pair of time series represented in the row.
\item response columns.
\item predictors columns: representing the distance between the values of the given static predictor between \code{x} and \code{y}.
\item (optional) \code{geographic_distance}: If \code{predictors_df} is an \code{sf} data frame, then geographic distances are computed via \code{\link[sf:st_distance]{sf::st_distance()}}.
}

This function supports a parallelization setup via \code{\link[future:plan]{future::plan()}}.
}
\examples{

#Fagus sylvatica dynamics in Europe
tsl <- tsl_initialize(
  x = fagus_dynamics,
  name_column = "name",
  time_column = "time"
)

#importance analysis
df <- momentum_ls(tsl = tsl)

#generate model frame
model_frame <- momentum_model_frame(
  response_df = df,
  predictors_df = fagus_coordinates,
  scale = TRUE
)

head(model_frame)

#names of response and predictors
#and an additive formula
#are stored as attributes
attributes(model_frame)$predictors

}
\seealso{
Other momentum_support:
\code{\link[=momentum_aggregate]{momentum_aggregate()}},
\code{\link[=momentum_boxplot]{momentum_boxplot()}},
\code{\link[=momentum_spatial]{momentum_spatial()}},
\code{\link[=momentum_stats]{momentum_stats()}},
\code{\link[=momentum_to_wide]{momentum_to_wide()}}
}
\concept{momentum_support}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_time_keywords.R
\name{utils_time_keywords}
\alias{utils_time_keywords}
\title{Valid Aggregation Keywords}
\usage{
utils_time_keywords(tsl = NULL)
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}
}
\value{
Character string representing a valid aggregation keyword, or "none".
}
\description{
Internal function to obtain valid aggregation keywords from a zoo object or a time series list.
}
\examples{
#one minute time series
#-----------------------------------
tsl <- tsl_simulate(
  time_range = c(
    Sys.time() - 60,
    Sys.time()
  )
)

#valid keywords for aggregation and/or resampling
utils_time_keywords(
  tsl = tsl
)

#10 minutes time series
#-----------------------------------
tsl <- tsl_simulate(
  time_range = c(
    Sys.time() - 600,
    Sys.time()
  )
)

utils_time_keywords(
  tsl = tsl
)

#10 hours time series
#-----------------------------------
tsl <- tsl_simulate(
  time_range = c(
    Sys.time() - 6000,
    Sys.time()
  )
)

utils_time_keywords(
  tsl = tsl
)

#10 days time series
#-----------------------------------
tsl <- tsl_simulate(
  time_range = c(
    Sys.Date() - 10,
    Sys.Date()
  )
)

utils_time_keywords(
  tsl = tsl
)

#10 years time series
#-----------------------------------
tsl <- tsl_simulate(
  time_range = c(
    Sys.Date() - 3650,
    Sys.Date()
  )
)

utils_time_keywords(
  tsl = tsl
)
}
\seealso{
Other internal_time_handling:
\code{\link[=utils_as_time]{utils_as_time()}},
\code{\link[=utils_coerce_time_class]{utils_coerce_time_class()}},
\code{\link[=utils_is_time]{utils_is_time()}},
\code{\link[=utils_new_time]{utils_new_time()}},
\code{\link[=utils_time_keywords_dictionary]{utils_time_keywords_dictionary()}},
\code{\link[=utils_time_keywords_translate]{utils_time_keywords_translate()}},
\code{\link[=utils_time_units]{utils_time_units()}}
}
\concept{internal_time_handling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_vector_to_matrix.R
\name{zoo_vector_to_matrix}
\alias{zoo_vector_to_matrix}
\title{Coerce Coredata of Univariate Zoo Time Series to Matrix}
\usage{
zoo_vector_to_matrix(x = NULL, name = NULL)
}
\arguments{
\item{x}{(required, zoo object) zoo time series. Default: NULL}

\item{name}{(required, character string) name of the matrix column. Default: NULL}
}
\value{
zoo time series
}
\description{
Transforms vector coredata of univariate zoo time series to class matrix. If the input zoo time series has the attribute "name", the output inherits the value of such attribute.

Multivariate zoo objects (those with multiple columns) are returned unchanged.
}
\examples{
#create zoo object from vector
x <- zoo::zoo(
  x = runif(100)
)

#coredata is not a matrix
is.matrix(zoo::coredata(x))

#convert to matrix
y <- zoo_vector_to_matrix(
  x = x
)

#coredata is now a matrix
is.matrix(zoo::coredata(y))
}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_resample]{zoo_resample()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}}
}
\concept{zoo_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsl_burst.R
\name{tsl_burst}
\alias{tsl_burst}
\title{Multivariate TSL to Univariate TSL}
\usage{
tsl_burst(tsl = NULL, sep = "__")
}
\arguments{
\item{tsl}{(required, list) Time series list. Default: NULL}

\item{sep}{(required, character string) separator between the time series name and the column name. Default: "__"}
}
\value{
time series list: list of univariate zoo objects with a column named "x".
}
\description{
Takes a time series list with multivariate zoo objects to generate a new one with one univariate zoo object per variable. A time series list with the zoo objects "A" and "B", each with the columns "a", "b", and "c", becomes a time series list with the zoo objects "A__a", "A__b", "A__c", "B__a", "B__b", and "B__c". The only column of each new zoo object is named "x".
}
\examples{

tsl <- tsl_simulate(
  n = 2,
  time_range = c(
    "2010-01-01",
    "2024-12-31"
  ),
  cols = 3
)

tsl_names_get(tsl)
tsl_colnames_get(tsl)

if(interactive()){
  tsl_plot(tsl)
}

tsl <- tsl_burst(tsl)

tsl_names_get(tsl)
tsl_colnames_get(tsl)

if(interactive()){
  tsl_plot(tsl)
}
}
\seealso{
Other tsl_management:
\code{\link[=tsl_colnames_clean]{tsl_colnames_clean()}},
\code{\link[=tsl_colnames_get]{tsl_colnames_get()}},
\code{\link[=tsl_colnames_prefix]{tsl_colnames_prefix()}},
\code{\link[=tsl_colnames_set]{tsl_colnames_set()}},
\code{\link[=tsl_colnames_suffix]{tsl_colnames_suffix()}},
\code{\link[=tsl_count_NA]{tsl_count_NA()}},
\code{\link[=tsl_diagnose]{tsl_diagnose()}},
\code{\link[=tsl_handle_NA]{tsl_handle_NA()}},
\code{\link[=tsl_join]{tsl_join()}},
\code{\link[=tsl_names_clean]{tsl_names_clean()}},
\code{\link[=tsl_names_get]{tsl_names_get()}},
\code{\link[=tsl_names_set]{tsl_names_set()}},
\code{\link[=tsl_names_test]{tsl_names_test()}},
\code{\link[=tsl_ncol]{tsl_ncol()}},
\code{\link[=tsl_nrow]{tsl_nrow()}},
\code{\link[=tsl_repair]{tsl_repair()}},
\code{\link[=tsl_subset]{tsl_subset()}},
\code{\link[=tsl_time]{tsl_time()}},
\code{\link[=tsl_to_df]{tsl_to_df()}}
}
\concept{tsl_management}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_optimize_spline.R
\name{utils_optimize_spline}
\alias{utils_optimize_spline}
\title{Optimize Spline Models for Time Series Resampling}
\usage{
utils_optimize_spline(x = NULL, y = NULL, max_complexity = FALSE)
}
\arguments{
\item{x}{(required, numeric vector) predictor, a time vector coerced to numeric. Default: NULL}

\item{y}{(required, numeric vector) response, a column of a zoo object. Default: NULL}

\item{max_complexity}{(required, logical). If TRUE, RMSE optimization is ignored, and the model of maximum complexity is returned. Default: FALSE}
}
\value{
Object of class "smooth.spline".
}
\description{
Internal function used in \code{\link[=zoo_resample]{zoo_resample()}}. It finds the optimal \code{df} parameter of a smoothing spline model \code{y ~ x} fitted with \code{\link[stats:smooth.spline]{stats::smooth.spline()}} that minimizes the root mean squared error (RMSE) between observations and predictions, and returns a model fitted with such \code{df}.
}
\examples{

#zoo time series
xy <- zoo_simulate(
  cols = 1,
  rows = 30
)

#optimize spline model
m <- utils_optimize_spline(
  x = as.numeric(zoo::index(xy)), #predictor
  y = xy[, 1] #response
)

print(m)

#plot observation
plot(
  x = zoo::index(xy),
  y = xy[, 1],
  col = "forestgreen",
  type = "l",
  lwd = 2
  )

#plot prediction
points(
  x = zoo::index(xy),
  y = stats::predict(
    object = m,
    x = as.numeric(zoo::index(xy))
  )$y,
  col = "red"
)

}
\seealso{
Other tsl_processing_internal:
\code{\link[=utils_drop_geometry]{utils_drop_geometry()}},
\code{\link[=utils_global_scaling_params]{utils_global_scaling_params()}},
\code{\link[=utils_optimize_loess]{utils_optimize_loess()}},
\code{\link[=utils_rescale_vector]{utils_rescale_vector()}}
}
\concept{tsl_processing_internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zoo_resample.R
\name{zoo_resample}
\alias{zoo_resample}
\title{Resample Zoo Objects to a New Time}
\usage{
zoo_resample(
  x = NULL,
  new_time = NULL,
  method = "linear",
  max_complexity = FALSE
)
}
\arguments{
\item{x}{(required, zoo object) Time series to resample. Default: NULL}

\item{new_time}{(optional, zoo object, keyword, or time vector) New time to resample \code{x} to. The available options are:
\itemize{
\item NULL: a regular version of the time in \code{x} is generated and used for resampling.
\item zoo object: the index of the given zoo object is used as template to resample \code{x}.
\item time vector: a vector with new times to resample \code{x} to. If time in \code{x} is of class "numeric", this vector must be numeric as well. Otherwise, vectors of classes "Date" and "POSIXct" can be used indistinctly.
\item keyword: a valid keyword returned by \code{zoo_time(x)$keywords}, used to generate a time vector with the relevant units.
\item numeric of length 1: interpreted as new time interval, in the highest resolution units returned by \code{zoo_time(x)$units}.
}}

\item{method}{(optional, character string) Name of the method to resample the time series. One of "linear", "spline" or "loess". Default: "linear".}

\item{max_complexity}{(required, logical). Only relevant for methods "spline" and "loess". If TRUE, model optimization is ignored, and the a model of maximum complexity (an overfitted model) is used for resampling. Default: FALSE}
}
\value{
zoo object
}
\description{
\strong{Objective}

Time series resampling involves interpolating new values for time steps not available in the original time series. This operation is useful to:
\itemize{
\item Transform irregular time series into regular.
\item Align time series with different temporal resolutions.
\item Increase (upsampling) or decrease (downsampling) the temporal resolution of a time series.
}

On the other hand, time series resampling \strong{should not be used} to extrapolate new values outside of the original time range of the time series, or to increase the resolution of a time series by a factor of two or more. These operations are known to produce non-sensical results.

\strong{Methods}
This function offers three methods for time series interpolation:

\itemize{
\item "linear" (default): interpolation via piecewise linear regression as implemented in \code{\link[zoo:na.approx]{zoo::na.approx()}}.
\item "spline": cubic smoothing spline regression as implemented in \code{\link[stats:smooth.spline]{stats::smooth.spline()}}.
\item "loess": local polynomial regression fitting as implemented in \code{\link[stats:loess]{stats::loess()}}.
}

These methods are used to fit models \code{y ~ x} where \code{y} represents the values of a univariate time series and \code{x} represents a numeric version of its time.

The functions \code{\link[=utils_optimize_spline]{utils_optimize_spline()}} and \code{\link[=utils_optimize_loess]{utils_optimize_loess()}} are used under the hood to optimize the complexity of the methods "spline" and "loess" by finding the configuration that minimizes the root mean squared error (RMSE) between  observed and predicted \code{y}. However, when the argument \code{max_complexity = TRUE}, the complexity optimization is ignored, and a maximum complexity model is used instead.

\strong{New time}

The argument \code{new_time} offers several alternatives to help define the new time of the resulting time series:

\itemize{
\item \code{NULL}: the target time series (\code{x}) is resampled to a regular time within its original time range and number of observations.
\item \verb{zoo object}: a zoo object to be used as template for resampling. Useful when the objective is equalizing the frequency of two separate zoo objects.
\item \verb{time vector}: a time vector of a class compatible with the time in \code{x}.
\item \code{keyword}: character string defining a resampling keyword, obtained via \code{zoo_time(x, keywords = "resample")$keywords}.
\item \code{numeric}: a single number representing the desired interval between consecutive samples in the units of \code{x} (relevant units can be obtained via \code{zoo_time(x)$units}).
}

\strong{Step by Step}

The steps to resample a time series list are:

\enumerate{
\item The time interpolation range taken from the index of the zoo object. This step ensures that no extrapolation occurs during resampling.
\item If \code{new_time} is provided, any values of \code{new_time} outside of the minimum and maximum interpolation times are removed to avoid extrapolation. If \code{new_time} is not provided, a regular time within the interpolation time range of the zoo object is generated.
\item For each univariate time time series, a model \code{y ~ x}, where \code{y} is the time series and \code{x} is its own time coerced to numeric is fitted.
\itemize{
\item If \code{max_complexity == FALSE} and \code{method = "spline"} or \code{method = "loess"}, the model with the complexity that minimizes the root mean squared error between the observed and predicted \code{y} is returned.
\item If \code{max_complexity == TRUE} and \code{method = "spline"} or \code{method = "loess"}, the first valid model closest to a maximum complexity is returned.
}
\item The fitted model is predicted over \code{new_time} to generate the resampled time series.
}

\strong{Other Details}

Please use this operation with care, as there are limits to the amount of resampling that can be done without distorting the data. The safest option is to keep the distance between new time points within the same magnitude of the distance between the old time points.
}
\examples{
#simulate irregular time series
x <- zoo_simulate(
  cols = 2,
  rows = 50,
  time_range = c("2010-01-01", "2020-01-01"),
  irregular = TRUE
  )

#plot time series
if(interactive()){
  zoo_plot(x)
}

#intervals between samples
x_intervals <- diff(zoo::index(x))
x_intervals

#create regular time from the minimum of the observed intervals
new_time <- seq.Date(
  from = min(zoo::index(x)),
  to = max(zoo::index(x)),
  by = floor(min(x_intervals))
)

new_time
diff(new_time)

#resample using piecewise linear regression
x_linear <- zoo_resample(
  x = x,
  new_time = new_time,
  method = "linear"
)

#resample using max complexity splines
x_spline <- zoo_resample(
  x = x,
  new_time = new_time,
  method = "spline",
  max_complexity = TRUE
)

#resample using max complexity loess
x_loess <- zoo_resample(
  x = x,
  new_time = new_time,
  method = "loess",
  max_complexity = TRUE
)


#intervals between new samples
diff(zoo::index(x_linear))
diff(zoo::index(x_spline))
diff(zoo::index(x_loess))

#plotting results
if(interactive()){

  par(mfrow = c(4, 1), mar = c(3,3,2,2))

  zoo_plot(
    x,
    guide = FALSE,
    title = "Original"
    )

  zoo_plot(
    x_linear,
    guide = FALSE,
    title = "Method: linear"
  )

  zoo_plot(
    x_spline,
    guide = FALSE,
    title = "Method: spline"
    )

  zoo_plot(
    x_loess,
    guide = FALSE,
    title = "Method: loess"
  )

}

}
\seealso{
Other zoo_functions:
\code{\link[=zoo_aggregate]{zoo_aggregate()}},
\code{\link[=zoo_name_clean]{zoo_name_clean()}},
\code{\link[=zoo_name_get]{zoo_name_get()}},
\code{\link[=zoo_name_set]{zoo_name_set()}},
\code{\link[=zoo_permute]{zoo_permute()}},
\code{\link[=zoo_plot]{zoo_plot()}},
\code{\link[=zoo_smooth_exponential]{zoo_smooth_exponential()}},
\code{\link[=zoo_smooth_window]{zoo_smooth_window()}},
\code{\link[=zoo_time]{zoo_time()}},
\code{\link[=zoo_to_tsl]{zoo_to_tsl()}},
\code{\link[=zoo_vector_to_matrix]{zoo_vector_to_matrix()}}
}
\concept{zoo_functions}
